/********************************************************************
 *		File			:		bheap.c
 *		Description		:		heap implementation
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "metautil.h"

/********************************************************************
 *		binary heap
 *********************************************************************/
#define bheapGetCellLoc(cell)		(((BinHeapCell)cell)->loc)
#define bheapGetKey(cell)			(((Jint*)cell->key)[0])
#define bheapSetKey(cell,k)			(((Jint*)cell->key)[0] = k)

static void
bheapUpdateMin( BinHeap h ) {
	while((h->fMin <= h->fHigh)
		  && (h->fBuckets[h->fMin-h->fLow] == NULL) )
		h->fMin ++;
	if( h->fMin > h->fHigh )
		h->fMax = h->fLow - 1;
	}

static void
bheapUpdateMax( BinHeap h ) {
	while( (h->fMax >= h->fLow)
		   && (h->fBuckets[h->fMax-h->fLow] == NULL) )
		h->fMax--;
	if( h->fMax < h->fLow )
		h->fMin = h->fHigh + 1;
	}

extern Jerror
bheapConstruct( BinHeap h, IMem m, Jint a, Jint b ) {
	Jint		i;

	h->fMem = m;
	h->fLow = a;
	h->fHigh = b;
	h->fMin = h->fHigh + 1;
	h->fMax = h->fLow - 1;
	h->fBuckets = comAlloc( m, sizeof(HeapCell)*(b-a+1) );
	if( h->fBuckets == NULL )
		return COM_E_OUTOFMEMORY;
	h->fNumElts = 0;
	for( i = a; i <= b; i++ ) 
		h->fBuckets[i-a] = NULL;
	return 0;
	}

extern Jerror
bheapDestruct( BinHeap h ) {
	comFree( h->fMem, h->fBuckets );
	return 0;
	}

extern Jerror
bheapInsertElement( BinHeap h, HeapCell it ) {
	Jint		key = bheapGetKey( it );
		
	assert( !( key < h->fLow || key > h->fHigh ) );
	if( key < h->fMin )
		h->fMin = key;
	if( key > h->fMax )
		h->fMax = key;
	it->next = h->fBuckets[key-h->fLow];
	h->fBuckets[key-h->fLow] = it;
	h->fNumElts ++;
	return 0;
	}

extern Jerror
bheapRemoveAll( BinHeap h ) {
	Jint	i;
	
	for( i = h->fLow; i < h->fHigh; i ++ )
		h->fBuckets[i-h->fLow] = NULL;
	h->fMin = h->fHigh + 1;
	h->fMax = h->fLow - 1;
	h->fNumElts = 0;
	return 0;
	}

extern Jerror
bheapDelAll( BinHeap h ) {
	Jint		i;
	HeapCell	it;
	
	for( i = h->fLow; i < h->fHigh; i ++ ) {
		for( it = h->fBuckets[i-h->fLow]; it != NULL; it = it->next )
			comFree( h->fMem, it );
		}
	h->fMin = h->fHigh + 1;
	h->fMax = h->fLow - 1;
	h->fNumElts = 0;
	return 0;
	}

extern Jerror
bheapRemoveElement( BinHeap h, HeapCell it ) {
	HeapCell	prev, head;
	
	assert( it != NULL );
	head = h->fBuckets[bheapGetKey(it)-h->fLow];
	assert( head != NULL );
	if( it == head ) {
		h->fBuckets[bheapGetKey(it)-h->fLow] = it->next;
		}
	else {
		assert( head->next != NULL );
		for( prev = head; prev->next != it; prev = prev->next );
		prev->next = it->next;
		}
	bheapUpdateMin( h );
	bheapUpdateMax( h );
	h->fNumElts --;
	return 0;
	}

extern HeapCell
bheapNewCell( BinHeap h, Jint key, IGeneric data ) {
	IMem		mem = h->fMem;
	HeapCell	it;

	it = (HeapCell)comAlloc( mem, sizeof(HeapCellDecl) );
	if( it == NULL )
		return NULL;
	it->val = data;
	bheapSetKey( it, key );
	return it;
	}

extern HeapCell
bheapFindMin( BinHeap h ) {
	if( h->fMin > h->fHigh )
		return NULL;
	return h->fBuckets[ h->fMin-h->fLow ];
	}

extern HeapCell
bheapFindMax( BinHeap h ) {
	if( h->fMax < h->fLow )
		return NULL;
	return h->fBuckets[ h->fMax-h->fLow ];
	}

extern HeapCell
bheapDelMin( BinHeap h ) {
	HeapCell	cell;
	
	assert( h->fMin <= h->fHigh );
	cell = h->fBuckets[h->fMin-h->fLow];
	if( cell == NULL )
		return NULL;
	
	h->fBuckets[h->fMin-h->fLow] = cell->next;
	bheapUpdateMin( h );
	h->fNumElts --;
	return cell;
	}

extern HeapCell
bheapDelMax( BinHeap h ) {
	HeapCell	cell;
	
	cell = h->fBuckets[h->fMax-h->fLow];
	if( cell == NULL )
		return NULL;
	h->fBuckets[h->fMax-h->fLow] = cell->next;
	bheapUpdateMax( h );
	h->fNumElts --;
	return cell;
	}

extern HeapCell
bheapDecKey( BinHeap h, HeapCell it, Jint k ) {
	assert( it != NULL );
	assert( !( k < h->fLow || k > h->fHigh) );

	bheapRemoveElement( h, it );
	bheapSetKey( it, k );
	bheapInsertElement( h, it );
	return it;
	}

extern HeapCell
bheapIncKey( BinHeap h, HeapCell it, Jint k ) {
	assert( it != NULL );
	assert( !( k < h->fLow || k > h->fHigh ) );
	
	bheapRemoveElement( h, it );
	bheapSetKey( it, k );
	bheapInsertElement( h, it );
	return it;
	}

/********************************************************************
 *		binary tree heap
 *********************************************************************/
#define heapSetElement(h,i,it)	\
    (it)->hash = (i);	\
    vecSetElement( heapGetTree(h), (i), (it) )

#define heapGetElement(h,i)	\
    ((HeapCell)vecGetElement( heapGetTree(h), (i) ))

#define heapMakeBtm(it)		((it)->next = it)
#define heapIsBtm(it)		((it)->next == it)

#define heapGreater(h,c1,c2) (				\
    heapIsBtm(c1) ? FALSE :					\
    heapIsBtm(c2) ? TRUE :					\
    (COM_ICALL( compare, (h)->scheme )( 	\
        (h)->scheme, (h)->keysize, 			\
        (h)->weights, 						\
        (c1)->key, (c2)->key ) == 1))

static HeapCell
heapNewCell0( Heap h, IGeneric data ) {
	HeapCell	cell;

	cell = comAlloc(
		heapMem(h), sizeof(*cell)+(h->keysize-1)*sizeof(IGeneric)
		);
	if( cell == NULL )
		return NULL;
	cell->next = NULL;
	cell->val = data;
	return cell;
	}

extern Jerror
heapConstruct( Heap h, IMem mem, Jint size ) {
	return heapConstruct0( h, mem, size, 1, NULL, orderIntScheme );
	}

extern Jerror
heapConstruct0(
	Heap h, IMem mem, Jint size,
	Jint keysize, IGeneric weights, IOrderScheme scheme
	) {
	HeapCell	it;
	
	vecConstruct( heapGetTree(h), mem, size );
	h->keysize = keysize;
	h->weights = weights;
	h->scheme = scheme;
	
	it = heapNewCell0( h, NULL );
	heapMakeBtm( it );
	heapSetElement( h, 0, it );
	
	return 0;
	}

extern Jerror
heapInsertElement( Heap h, HeapCell it ) {
	Jint	i;
	
	i = vecGetSize( heapGetTree(h) );
	while( heapGreater( h, heapGetElement( h, i >> 1 ), it ) ) {
		heapSetElement( h, i, heapGetElement( h, i>>1 ) );
		i >>= 1;
		}
	heapSetElement( h, i, it );
	return 0;
	}

extern HeapCell
heapNewCell1( Heap h, IGeneric key, IGeneric data ) {
	HeapCell	cell;

	cell = (HeapCell)comAlloc( heapMem(h), sizeof(*cell) );
	if( cell == NULL )
		return NULL;
	cell->key[0] = key;
	cell->val = data;
	cell->next = NULL;
	return cell;
	}

extern HeapCell
heapNewCell2(
	Heap h, IGeneric key1, IGeneric key2, IGeneric data
	) {
	HeapCell	cell = heapNewCell0( h, data );

	if( cell ) {
		cell->key[0] = key1;
		cell->key[1] = key2;
		}
	return cell;
	}

extern HeapCell
heapNewCell3(
	Heap h, IGeneric key1, IGeneric key2, IGeneric key3, IGeneric data
	) {
	HeapCell	cell = heapNewCell0( h, data );

	if( cell ) {
		cell->key[0] = key1;
		cell->key[1] = key2;
		cell->key[2] = key3;
		}
	return cell;
	}

extern HeapCell
heapNewCellN(
	Heap h, IGeneric* keys, IGeneric data
	) {
	HeapCell	cell = heapNewCell0( h, data );
	Jint		i;
	
	if( cell ) {
		for( i = 0; i < h->keysize; i ++ )
			cell->key[i] = keys[i];
		}
	return cell;
	}
	
extern HeapCell
heapFindMin( Heap h ) {
	return heapGetElement( h, 1 );
	}

extern HeapCell
heapDelMin( Heap h ) {
	Jint		child, i;
	HeapCell	min, last;
	Jint		size;
	HeapCell	left, right;
	
	assert( !heapIsEmpty(h) );
	min = heapGetElement( h, 1 );
	last = heapGetElement( h, --vecGetSize(heapGetTree(h)) );
	if( min == last )
		return min;
	size = vecGetSize(heapGetTree(h))-1;
	for( i = 1; (i<<1) <= size; i = child ) {
		child = i << 1;
		if( child != size ) {
			left = heapGetElement( h, child );
			right = heapGetElement( h, child+1 );
			if( heapGreater( h, left, right ) ) 
				child ++;
			}

		/* percolate one level */
		left = heapGetElement( h, child );
		if( heapGreater( h, last, left ) ) {
			/* swap */
			heapSetElement( h, i, left );
			}
		else
			break;
		}
	heapSetElement( h, i, last );
	return min;
	}

static void
heapSiftUp( Heap h, HeapCell it ) {
	Jint		i = it->hash;
	HeapCell	parent = heapGetElement( h, i >> 1 );

	while( heapGreater( h, parent, it ) ) {
		/* swap parent and it */
		heapSetElement( h, i, parent );

		/* move one level up */
		i >>= 1;
		parent = heapGetElement( h, i >> 1 );
		}
	
	/* settle it here */
	heapSetElement( h, i, it );
	}

extern Jerror
heapRemoveElement( Heap h, HeapCell it ) {
	heapMakeBtm( it );
	heapSiftUp( h, it );
	heapDelMin( h );
	return 0;
	}

extern HeapCell
heapDecKey( Heap h, HeapCell it, Jint k ) {
	assert( h->keysize == 1 );
	assert( h->scheme = orderIntScheme );
	((Jint *)it->key)[0] -= k;
	heapSiftUp( h, it );
	return it;
	}

extern Jerror
heapDestruct( Heap h ) {
	return vecDestruct( heapGetTree(h) );
	}

extern Jerror
heapDelAll( Heap h ) {
	Jint		size = vecGetSize( heapGetTree(h) );
	Jint		i;
	HeapCell	cell;
	
	for( i = 0; i < size; i ++ ) {
		cell = heapGetElement( h, i );
		comFree( heapMem(h), cell );
		}
	heapDestruct( h );
	return 0;
	}




