/********************************************************************
 *		File			:		error.c
 *		Description		:		error handler implementation
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>

#include <ierror.h>
#include "metautil.h"

static char*	curFile;
static Jint		curLine, curColumn;
static Jint		warningLevel = 1;
static Jint		numOfErrors = 0;
static Jint		numOfWarnings = 0;

extern char*			comFormatedPrint(
	FILE* f, char* bp, const char* fmt, va_list ap
	);

static COMDECL
query( IErrorHandler ei, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, ErrorHandler ) ) {
		*i = ei;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
addref( IErrorHandler ei ) {
	return 1;
	}

static COMDECLU
release( IErrorHandler ei ) {
	return 1;
	}

static COMDECL
begin( IErrorHandler ei ) {
	curFile = NULL;
	curLine = curColumn = 0;
	warningLevel = 1;
	numOfErrors = numOfWarnings = 0;
	return 0;
	}

static COMDECL
end( IErrorHandler ei ) {
	return 0;
	}

static COMDECL
getCoordinate( IErrorHandler ei, char** file, Jint* line, Jint* column ) {
	*file = curFile;
	*line = curLine;
	*column = curColumn;
	return 0;
	}

static COMDECL
setCoordinate( IErrorHandler ei, char* file, Jint line, Jint column ) {
	curFile = file;
	curLine = line;
	curColumn = column;
	return 0;
	}

extern char*	metaFormatedPrint(
	FILE* f, char* bp, const char* fmt, va_list ap
	);

static COMDECL
syntaxError( IErrorHandler ei, char* format, ... ) {
	va_list ap;

	va_start( ap, format );
	if( curFile != NULL )
		comPrintf(
			stderr, NULL, "%s:%d: (syntax error) ", curFile, curLine
			);
	comFormatedPrint( stderr, NULL, format, ap );
	va_end(ap);
	numOfErrors ++;

	return 0;
	}

static COMDECL
semanticError( IErrorHandler ei, char* format, ... ) {
	va_list ap;

	va_start( ap, format );
	if( curFile != NULL )
		comPrintf(
			stderr, NULL, "%s:%d: (semantic error) ", curFile, curLine
			);
	comFormatedPrint( stderr, NULL, format, ap );
	va_end(ap);
	numOfErrors ++;

	return 0;
	}

static COMDECL
internalError( IErrorHandler ei, char* format, ... ) {
	va_list ap;

	va_start( ap, format );
	comFormatedPrint( stderr, NULL, "(internal error) ", ap );
	comFormatedPrint( stderr, NULL, format, ap );
	va_end(ap);
	numOfErrors ++;

	return 0;
	}

static COMDECL
fatalError( IErrorHandler ei, char* format, ... ) {
	va_list ap;

	va_start( ap, format );
	comFormatedPrint( stderr, NULL, "(fatal error) ", ap );
	comFormatedPrint( stderr, NULL, format, ap );
	va_end(ap);
	numOfErrors ++;

	exit( -1 );
	}

static COMDECL
warning( IErrorHandler ei, Jint level, char* format, ... ) {
	va_list ap;

	va_start( ap, format );
	if( level <= warningLevel ) {
		if( curFile != NULL )
			comPrintf(
				stderr, NULL, "%s:%d: (warning) ", curFile, curLine
				);
		else
			comPrintf( stderr, NULL, "(warning) " );
			
		comFormatedPrint( stderr, NULL, format, ap );
		numOfWarnings ++;
		}
	va_end(ap);
	return 0;
	}

static COMDECL
setWarningLevel( IErrorHandler ei, int level ) {
	warningLevel = level;
	return 0;
	}

static COMDECL
getNumOfErrors( IErrorHandler ei, Jint *pnum ) {
	*pnum = numOfErrors;
	return 0;
	}

static COMDECL
getNumOfWarnings( IErrorHandler ei, Jint *pnum ) {
	*pnum = numOfWarnings;
	return 0;
	}

struct _DErrorHandler	dstderrhdl = {
	query,
	addref,
	release,
	begin,
	end,
	getCoordinate,
	setCoordinate,
	syntaxError, 
	semanticError,
	internalError,
	fatalError,
	warning,
	setWarningLevel,
	getNumOfErrors,
	getNumOfWarnings
    };

