/********************************************************************
 *		File			:		hash.c
 *		Description		:		hash table
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "metautil.h"

extern Jerror
hashConstruct(
	HashTable h, IMem mem, IHashScheme scheme,
	Jshort level1Size, Jshort level2Size,
	float loadFactor
	) {
	Jint	  	capacity;
	HashCell**	table;
	Jshort**	sizes;
	Jint		j;
	
	assert( h != NULL && mem != NULL );

	table = comAlloc( mem, (1 << level1Size)*sizeof(HashCell*) );
	if( table == NULL )
		return COM_E_OUTOFMEMORY;
	sizes = comAlloc( mem, (1 << level1Size)*sizeof(Jshort*) );
	if( sizes == NULL )
		return COM_E_OUTOFMEMORY;
	table[0] = comAlloc( mem, (1 << level2Size)*sizeof(HashCell) );
	if( table[0] == NULL ) {
		comFree( mem, table );
		return COM_E_OUTOFMEMORY;
		}
	sizes[0] = comAlloc( mem, (1 << level2Size)*sizeof(Jshort) );
	if( sizes[0] == NULL ) {
		comFree( mem, table );
		comFree( mem, table[0] );
		comFree( mem, sizes );
		return COM_E_OUTOFMEMORY;
		}
	for( j = 0; j < (1 << level2Size); j ++ ) {
		table[0][j] = NULL;
		sizes[0][j] = 0;
		}
	h->mem = mem;
	h->scheme = scheme;
	h->level1MaxSize = level1Size;
	h->level1Size = 0;
	h->level2Size = level2Size;
	h->count = 0;
	h->loadFactor = loadFactor;
	capacity =  1 << (level2Size + h->level1Size);
	h->threshold = (Jint)(capacity * loadFactor);
	h->table = table;
	h->bucketSize = sizes;
	h->keySize = 0;
	h->assoc = -1;
	h->shift = sizeof(Jint) * 8 - (h->level1MaxSize + h->level2Size);
	return 0;
	}

extern Jerror
hashDefaultConstruct( HashTable h, IMem mem ) {
	return hashConstruct( h, mem, hashPtrScheme, 0, 8, 0.75 );
	}

extern Jerror
hashDestruct( HashTable h ) {
	IMem	mem;
	Jint	i, j;
	Jint	l1Capacity, l2Capacity;
	
	l2Capacity = (1 << h->level2Size);
	l1Capacity = (1 << h->level1Size);
	
	mem = h->mem;
	for( i = 0; i < l1Capacity; i ++ ) {
		for( j = 0; j < l2Capacity; j ++ ) {
			HashCell	cell;
			
			for( cell = h->table[i][j]; cell != NULL; ) {
				HashCell	e = cell;

				cell = cell->next;
				comFree( mem, e );
				}
			}
		comFree( mem, h->table[i] );
		comFree( mem, h->bucketSize[i] );
		}
	comFree( mem, h->table );
	comFree( mem, h->bucketSize );
	return 0;
	}

extern HashCell
hashGetCell( HashTable h, void* key ) {
	Juint		hash;
	Jint		level2Index, level1Index, level1Size, level2Size;
	HashCell	cell;
	IHashScheme	scheme;

	scheme = h->scheme;
	hash = COM_CALL( hashCode, scheme )( scheme, h, key );
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);
	for( cell = h->table[level1Index][level2Index];
		 cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			COM_CALL( equals, scheme )( scheme, h, key, cell->key[0] ) )
			return cell;
		}
	return NULL;
	}

extern void*
hashGet( HashTable h, void* key ) {
	Juint		hash;
	Jint		level2Index, level1Index, level1Size, level2Size;
	HashCell	cell;
	IHashScheme	scheme;

	scheme = h->scheme;
	hash = COM_CALL( hashCode, scheme )( scheme, h, key );
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);
	for( cell = h->table[level1Index][level2Index];
		 cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			COM_CALL( equals, scheme )( scheme, h, key, cell->key[0] ) )
			return cell->val;
		}
	return NULL;
	}

static void
_rehash( HashTable h ) {
	Jint		oldCapacity, newCapacity, l2Capacity;
	HashCell*	newTable;
	Jshort*		newSize;
	Jint		i, j;
	Jint		level1Size, level2Size;
	HashCell	cell, prev;
	HashCell	e;
	Jint		l1Index;

	assert( h != NULL );
	
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	l2Capacity = (1 << level2Size);
	oldCapacity = (1 << level1Size);
	newCapacity = (oldCapacity << 1);
	
	for( i = 0; i < oldCapacity; i ++ ) {
		newTable = comAlloc(
			h->mem, l2Capacity*sizeof(HashCell)
			);
		newSize = comAlloc(
			h->mem, l2Capacity*sizeof(Jshort)
			);
		if( newTable == NULL || newSize == NULL ) {
			for( j = 0; j < i; i ++ ) {
				comFree( h->mem, h->table[oldCapacity+i] );
				comFree( h->mem, h->bucketSize[oldCapacity+i] );
				h->table[oldCapacity+i] = NULL;
				h->bucketSize[oldCapacity+i] = NULL;
				}
			return;
			}
		for( j = 0; j < l2Capacity; j ++ ) {
			newTable[j] = NULL;
			newSize[j] = 0;
			}
		h->table[oldCapacity+i] = newTable;
		h->bucketSize[oldCapacity+i] = newSize;
		}
	h->threshold = newCapacity * h->loadFactor;
	h->level1Size ++;
	level1Size ++;
	
	for( i = 0; i < oldCapacity; i ++ ) {
		for( j = 0; j < l2Capacity; j ++ ) {
			prev = NULL;
			for( cell = h->table[i][j]; cell != NULL; ) {
				e = cell;
				cell = e->next;
				l1Index = (e->hash >> level2Size) & ((1 << level1Size) - 1);
				if( l1Index != i ) {
					/* remove it from old table */
					if( prev == NULL ) {
						h->table[i][j] = cell;
						}
					else {
						prev->next = cell;
						}
					h->bucketSize[i][j] --;
					/* add it to new table */
					e->next = h->table[l1Index][j];
					h->table[l1Index][j] = e;
					h->bucketSize[l1Index][j] ++;
					}
				else
					prev = e;
				}
			}
		}
	}

extern Jerror
hashPut( HashTable h, void* key, void* v ) {
	Juint		hash;
	Jint		level2Index, level1Index, level1Size, level2Size;
	HashCell	cell;
	IHashScheme	scheme;

	assert( h != NULL );
	assert( v != NULL );
	
	scheme = h->scheme;
	if( h->count >= h->threshold && h->level1Size < h->level1MaxSize ) {
		_rehash( h );
		return hashPut( h, key, v );
		}
	
	hash = COM_CALL( hashCode, scheme )( scheme, h, key );
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);

	cell = comAlloc( h->mem, sizeof(*cell) );
	if( cell == NULL )
		return COM_E_OUTOFMEMORY;
	cell->hash = hash;
	cell->key[0] = key;
	cell->val = v;
	cell->next = h->table[level1Index][level2Index];
	h->table[level1Index][level2Index] = cell;
	h->bucketSize[level1Index][level2Index] ++;
	h->count ++;
	return 0;
	}

extern Jerror
hashPutCell( HashTable h, HashCell cell ) {
	Juint		hash;
	Jint		level2Index, level1Index, level1Size, level2Size;
	IHashScheme	scheme;
	void*		key = h->keySize == 0 ? cell->key[0] : (IGeneric)cell->key;
	void*		v = cell->val;
	
	assert( h != NULL );
	assert( v != NULL );
	
	scheme = h->scheme;
	if( h->count >= h->threshold && h->level1Size < h->level1MaxSize ) {
		_rehash( h );
		return hashPutCell( h, cell );
		}
	
	hash = COM_CALL( hashCode, scheme )( scheme, h, key );
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);

	cell->hash = hash;
	cell->next = h->table[level1Index][level2Index];
	h->table[level1Index][level2Index] = cell;
	h->bucketSize[level1Index][level2Index] ++;
	h->count ++;
	return 0;
	}

extern Jerror
hashRemoveAllCells( HashTable h, HashCell* freehead, HashCell *freetail ) {
	Jint		i, j;
	HashCell	free = NULL;
	HashCell	head, tail;
	Jboolean	first = TRUE;

	*freetail = NULL;
    for( i = 0; i < (1 << h->level1Size); i ++ ) 		
		for( j = 0; j < (1 << h->level2Size); j ++ ) {
			head = h->table[i][j];
			if( head != NULL ) {
				for( tail = head; tail->next != NULL; tail = tail->next );
				tail->next = free;
				free = head;
				if( first ) {
					*freetail = tail;
					first = FALSE;
					}
				h->table[i][j] = NULL;
				h->bucketSize[i][j] = 0;
				}
			}
	h->count = 0;
	*freehead = free;
	return 0;
	}

extern HashCell
hashRemoveCell( HashTable h, HashCell c ) {
	Juint		hash;
	Jint		level2Index, level1Index, level1Size, level2Size;
	HashCell	cell, prev;
	
	hash = c->hash;
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);

	for( prev = NULL, cell = h->table[level1Index][level2Index];
		 cell != NULL; prev = cell, cell = cell->next ) {
		if( cell == c ) {
			if( prev == NULL )
				h->table[level1Index][level2Index] = cell->next;
			else
				prev->next = cell->next;
			h->count --;
			h->bucketSize[level1Index][level2Index] --;
			return cell;
			}
		}
	return NULL;
	}

extern HashCell
hashRemove( HashTable h, void* key ) {
	Juint		hash;
	Jint		level2Index, level1Index, level1Size, level2Size;
	IHashScheme	scheme;
	HashCell	cell, prev;
	
	assert( h != NULL );
	
	scheme = h->scheme;
	hash = COM_CALL( hashCode, scheme )( scheme, h, key );
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);

	for( prev = NULL, cell = h->table[level1Index][level2Index];
		 cell != NULL; prev = cell, cell = cell->next ) {
		if( cell->hash == hash &&
			COM_CALL( equals, scheme )( scheme, h, key, cell->key[0] ) ) {
			if( prev == NULL )
				h->table[level1Index][level2Index] = cell->next;
			else
				prev->next = cell->next;
			h->count --;
			h->bucketSize[level1Index][level2Index] --;
			return cell;
			}
		}
	return NULL;
	}

extern Jerror
hashDump( HashTable h ) {
	Jint		l1Capacity, l2Capacity;
	Jint		i, j;
	Jint		level1Size, level2Size;
	HashCell	cell;

	assert( h != NULL );
	
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	l2Capacity = (1 << level2Size);
	l1Capacity = (1 << level1Size);

	comPrintf( stderr, NULL, "dumping hash table\n" );
	for( i = 0; i < l1Capacity; i ++ ) {
		for( j = 0; j < l2Capacity; j ++ ) {
			for( cell = h->table[i][j]; cell != NULL; cell = cell->next ) {
				comPrintf( stderr, NULL, "key: %x  val: %x next: %x \n",
						(Jint)cell->key[0],
						(Jint)cell->val,
						(Jint)(cell->next ? cell->next->key[0] : 0)
						);
				}
			}
		}
	comPrintf( stderr, NULL, "\n" );
	return 0;
	}


struct guid comIHashSchemeIID = COM_IHASHSCHEME_IID;

static COMDECL
_hashQuery(
	IHashScheme si, const ComIID iid, IGeneric *i
	) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, HashScheme ) ) {
		*i = si;
		return 0;
		}
	*i = NULL;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
_hashAddref( IHashScheme si ) {
	return 1;
	}

static COMDECLU
_hashRelease( IHashScheme si ) {
	return 1;
	}

static Juint COMCALL
_hashHashCodePtr( IHashScheme si, HashTable h, void* key ) {
	return (*(Juint *)&key) >> 2;
	}

static Jboolean COMCALL
_hashEqualsPtr( IHashScheme si, HashTable h, void* key, void *another ) {
	return key == another;
	}

static Juint COMCALL
_hashHashCodeString( IHashScheme si, HashTable t, void* key ) {
	char*	s = (char *)key;
	char*	end;
	Juint	h;
	Jint	i;
	Jint	len = strlen( s );
	extern int hashScatter[];
	
	for( h = 0, i = len, end = s; i > 0; i -- )
		h = (h<<1) + hashScatter[*(unsigned char *)end++];
	return h;
	}

static Jboolean COMCALL
_hashEqualsString( IHashScheme si, HashTable t, void* key, void *another ) {
	return !strcmp( (char *)key, (char *)another );
	}

static Juint COMCALL
_hashHashCodeMD( IHashScheme si, HashTable t, void* key ) {
	MDID	md = (MDID)key;

	return md->digest[0];
	}

static Jboolean COMCALL
_hashEqualsMD( IHashScheme si, HashTable t, void* key, void *another ) {
	MDID	md1 = (MDID)key;
	MDID	md2 = (MDID)another;

	return (
		md1->digest[0] == md2->digest[0] && 
		md1->digest[1] == md2->digest[1] && 
		md1->digest[2] == md2->digest[2] && 
		md1->digest[3] == md2->digest[3] && 
		md1->digest[4] == md2->digest[4]
		);
	}

/*
 * Multiple pointer hashing function
 */
/* Primes for cache hash functions. */
#define PRIME1			12582917
#define PRIME2			4256249
#define PRIME3			741457
#define PRIME4			1618033999
#define SIZEOFP			4
#define SIZEOFINT		4

#if SIZEOFP == 8 && SIZEOFINT == 4
#define phash2(f,g,shift) \
((((unsigned)(unsigned long)(f) * PRIME1 + \
   (unsigned)(unsigned long)(g)) * PRIME2) >> (shift))
#else
#define phash2(f,g,shift) \
((((unsigned)(f) * PRIME1 + (unsigned)(g)) * PRIME2) >> (shift))
#endif

#if SIZEOF_VOID_P == 8 && SIZEOF_INT == 4
#define phash3(o,f,g,s) \
(((((unsigned)(unsigned long)(f) + (unsigned)(unsigned long)(o)) * PRIME1 + \
   (unsigned)(unsigned long)(g)) * PRIME2) >> (s))
#else
#define phash3(o,f,g,s) \
(((((unsigned)(f) + (unsigned)(o)) * PRIME1 + (unsigned)(g)) * PRIME2) >> (s))
#endif

static Juint COMCALL
_hashHashCodeCache( IHashScheme si, HashTable h, void* key ) {
	IGeneric*	key1 = (IGeneric*)key;
    unsigned int val = (unsigned int) key1[0];
    unsigned int i;

	switch( h->keySize ) {
	case 1 : return phash2( key1[0], key1[0], h->shift );
	case 2 : return phash2( key1[0], key1[1], h->shift );
	case 3 : return phash3( key1[0], key1[1], key1[2], h->shift );
	default :	
		for( i = 1; i < h->keySize; i ++ ) {
			val = val * PRIME1 + (int) (unsigned int)key1[i];
			}
		return (val >> h->shift);
		}
	}

static Jboolean COMCALL
_hashEqualsCache( IHashScheme si, HashTable h, void* key, void *another ) {
	Jint		i;
	IGeneric*	key1 = (IGeneric*)key;
	IGeneric*	another1 = (IGeneric*)another;
	
	switch( h->keySize ) {
	case 1 : return key1[0] == another1[0];
	case 2 : return key1[0] == another1[0] && key1[1] == another1[2];
	case 3 : return key1[0] == another1[0] && key1[1] == another1[2] &&
				 key1[3] == another1[3];
	default :
		for( i = 0; i < h->keySize; i ++ )
			if( key1[i] != another1[i] )
				return FALSE;
		return TRUE;
		}
	}

static struct  _DHashScheme	_hashPtrSchemeVtbl = {
	_hashQuery, _hashAddref, _hashRelease,
	_hashHashCodePtr, _hashEqualsPtr
    };

static struct  _DHashScheme	_hashStringSchemeVtbl = {
	_hashQuery, _hashAddref, _hashRelease,
	_hashHashCodeString, _hashEqualsString
    };

static struct  _DHashScheme	_hashMDSchemeVtbl = {
	_hashQuery, _hashAddref, _hashRelease,
	_hashHashCodeMD, _hashEqualsMD
    };

static struct  _DHashScheme	_hashCacheSchemeVtbl = {
	_hashQuery, _hashAddref, _hashRelease,
	_hashHashCodeCache, _hashEqualsCache
    };


static struct _IHashScheme	_hashPtrScheme = { &_hashPtrSchemeVtbl };
IHashScheme			hashPtrScheme = &_hashPtrScheme;
static struct _IHashScheme	_hashStringScheme = { &_hashStringSchemeVtbl };
IHashScheme			hashStringScheme = &_hashStringScheme;
static struct _IHashScheme	_hashMDScheme = { &_hashMDSchemeVtbl };
IHashScheme			hashMDScheme = &_hashMDScheme;
static struct _IHashScheme	_hashCacheScheme = { &_hashCacheSchemeVtbl };
IHashScheme			hashCacheScheme = &_hashCacheScheme;

/*******************************************************************
 *	iterator
 *******************************************************************/
typedef struct {
	struct _IMap		intf;
	HashTable			tbl;
	Jint				i, j;
	HashCell			cell;
	} *HashIter;

static COMDECL
_hashIterQuery( IMap mi, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, Iterator ) 
		|| COM_SAME_IIDNAME( iid, Map ) ) {
		*i = mi;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
_hashIterAddRef( IMap ci ) {
	return 1;
	}

static COMDECLU
_hashIterRelease( IMap ci ) {
	return 1;
	}

static COMDECL
_hashIterReset( IMap iti ) {
	HashIter	it = (HashIter)iti;

    for( it->i = 0; it->i < (1 << it->tbl->level1Size); it->i ++ ) 		
		for( it->j = 0; it->j < (1 << it->tbl->level2Size); it->j ++ ) 	
			if( (it->cell = it->tbl->table[it->i][it->j]) != NULL )
				return 0;
	return 0;
	}

static Jboolean	COMCALL
_hashIterHasMoreElements( IMap iti ) {
	HashIter	it = (HashIter)iti;

	if( it->i >= (1 << it->tbl->level1Size) )
		return FALSE;
	return TRUE;
	}

static IGeneric COMCALL
_hashIterNextElement( IMap iti ) {
	HashIter	it = (HashIter)iti;
	IGeneric	rtn = hashData( IGeneric, it->cell );

	it->cell = it->cell->next;
	if( it->cell == NULL ) {
		for( ; it->i < (1 << it->tbl->level1Size); it->i ++ ) 		
			for( it->j = it->j+1; it->j < (1 << it->tbl->level2Size);
				 it->j ++ )
				if( (it->cell = it->tbl->table[it->i][it->j]) != NULL )
					break;
		}
	return rtn;
	}

static COMDECL
_hashIterPut( IMap iti, IGeneric key, IGeneric data ) {
	HashIter	it = (HashIter)iti;
	
	return hashPut( it->tbl, key, data );
	}

static IGeneric COMCALL
_hashIterGet( IMap iti, IGeneric key ) {
	HashIter	it = (HashIter)iti;
	
	return hashGet( it->tbl, key );
	}

static IGeneric COMCALL
_hashIterKey( IMap iti ) {
	HashIter	it = (HashIter)iti;
	IGeneric	rtn = hashKey( IGeneric, it->cell );
	
	return rtn;
	}

static struct _DMap	iterVtbl = {
	_hashIterQuery, _hashIterAddRef, _hashIterRelease,
	_hashIterReset, _hashIterHasMoreElements, _hashIterNextElement,
	_hashIterPut, _hashIterGet, _hashIterKey
    };


extern Jerror
hashConstructMap( HashTable h, IMap mem ) {
	HashIter	it = (HashIter)mem;

	it->intf.vtbl = &iterVtbl;
	it->tbl = h;
	_hashIterReset( mem );
	
	return 0;
	}

/*******************************************************************
 *	cache management
 *******************************************************************/
extern Jerror
cacheConstruct(
	Cache h, IMem mem, Jshort level1Size, Jshort level2Size,
	Jint keySize, Jint assoc
	) {
	Jerror	rc;
	
	rc = hashConstruct( h, mem, hashCacheScheme, level1Size, level2Size, 0.8 );
	if( COM_FAILED(rc) )
		return rc;
	h->keySize = keySize;
	h->shift = sizeof(Jint) * 8 - (h->level1MaxSize + h->level2Size);
	h->assoc = assoc;
	return rc;
	}

extern Jerror
cacheDefaultConstruct( Cache h, IMem mem, Jint keySize ) {
	Jerror	rc;
	
	rc = hashDefaultConstruct( h, mem );
	if( COM_FAILED(rc) )
		return rc;
	h->scheme = hashCacheScheme;
	h->keySize = keySize;
	h->assoc = 0;
	return rc;
	}

extern Jerror
cacheDestruct( Cache h ) {
	return hashDestruct( h );
	}

static HashCell
cachePut0( Cache h, Juint hash, IGeneric v ) {
	Jint		level2Index, level1Index, level1Size, level2Size;
	HashCell	cell;
	Jint		extra;
	
	assert( v != NULL );
	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);

	if( h->assoc > 0 &&
		h->bucketSize[level1Index][level2Index] == h->assoc ) {
		/* need to do cache replacement */
		cell = h->table[level1Index][level2Index];
		}
	else {
		extra = h->keySize > 2 ? h->keySize -1 : 1;
		cell = comAlloc(
			h->mem, sizeof(*cell) + extra*sizeof(IGeneric)
			);
		if( cell == NULL )
			return NULL;
		cell->hash = hash;
		cell->next = h->table[level1Index][level2Index];
		h->table[level1Index][level2Index] = cell;
		h->bucketSize[level1Index][level2Index] ++;
		h->count ++;
		}
	cell->val = v;
	
	return cell;
	}

extern Jerror
cachePut( Cache h, IGeneric key, IGeneric v ) {
	HashCell	cell;
	Juint		hash;
	
	hash = phash2( key, key, h->shift );
	cell = cachePut0( h, hash, v );
	if( cell != NULL ) {
		cell->key[0] = key;
		cell->key[1] = key;
		return 0;
		}
	return COM_E_OUTOFMEMORY;
	}

extern Jerror
cachePut2( Cache h, IGeneric key1, IGeneric key2, IGeneric v ) {
	HashCell	cell;
	Juint		hash;
	
	hash = phash2( key1, key2, h->shift );
	cell = cachePut0( h, hash, v );		
	if( cell != NULL ) {
		cell->key[0] = key1;
		cell->key[1] = key2;
		return 0;
		}
	return COM_E_OUTOFMEMORY;
	}

extern Jerror
cachePut3( Cache h, IGeneric key1, IGeneric key2, IGeneric key3, IGeneric v ) {
	HashCell	cell;
	Juint		hash;
	
	hash = phash3( key1, key2, key3, h->shift );
	cell = cachePut0( h, hash, v );		
	if( cell != NULL ) {
		cell->key[0] = key1;
		cell->key[1] = key2;
		cell->key[2] = key3;
		return 0;
		}
	return COM_E_OUTOFMEMORY;
	}

extern Jerror
cachePutN( Cache h, IGeneric* key, IGeneric v ) {
	HashCell		cell;
	Juint			hash;
	IHashScheme		scheme = h->scheme;
	Jint			i;
	
	hash = COM_ICALL( hashCode, scheme )( scheme, h, key );
	cell = cachePut0( h, hash, v );
	if( cell != NULL ) {
		for( i = 0; i < h->keySize; i ++ ) 
			cell->key[i] = key[i];
		return 0;
		}
	return COM_E_OUTOFMEMORY;
	}

static IGeneric
cacheGet0( Cache h, Juint hash ) {
	Jint		level2Index, level1Index, level1Size, level2Size;

	level1Size = h->level1Size;
	level2Size = h->level2Size;
	level2Index = hash & ((1 << level2Size) - 1);
	level1Index = (hash >> level2Size) & ((1 << level1Size) - 1);
	return h->table[level1Index][level2Index];
	}

extern IGeneric
cacheGet( Cache h, IGeneric key ) {
	Juint		hash;
	HashCell	cell;

	hash = phash2( key, key, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash && cell->key[0] == key )
			return cell->val;
		}
	return NULL;
	}

extern IGeneric
cacheGet2( Cache h, IGeneric key1, IGeneric key2 ) {
	Juint		hash;
	HashCell	cell;

	hash = phash2( key1, key2, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			cell->key[0] == key1 && cell->key[1] == key2 )
			return cell->val;
		}
	return NULL;
	}

extern IGeneric
cacheGet3( Cache h, IGeneric key1, IGeneric key2, IGeneric key3 ) {
	Juint		hash;
	HashCell	cell;

	hash = phash3( key1, key2, key3, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			cell->key[0] == key1 && cell->key[1] == key2 &&
			cell->key[2] == key3 )
			return cell->val;
		}
	return NULL;
	}

extern IGeneric
cacheGetN( Cache h, IGeneric* key, IGeneric v ) {
	Juint			hash;
	IHashScheme		scheme = h->scheme;
	HashCell	cell;

	hash = phash2( key, key, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			COM_ICALL( equals, scheme )( scheme, h, cell->key, key ) )
			return cell->val;
		}
	return NULL;
	}


extern HashCell
cacheGetCell( Cache h, IGeneric key ) {
	Juint		hash;
	HashCell	cell;

	hash = phash2( key, key, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash && cell->key[0] == key )
			return cell;
		}
	return NULL;
	}

extern HashCell
cacheGetCell2( Cache h, IGeneric key1, IGeneric key2 ) {
	Juint		hash;
	HashCell	cell;

	hash = phash2( key1, key2, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			cell->key[0] == key1 && cell->key[1] == key2 )
			return cell;
		}
	return NULL;
	}

extern HashCell
cacheGetCell3( Cache h, IGeneric key1, IGeneric key2, IGeneric key3 ) {
	Juint		hash;
	HashCell	cell;

	hash = phash3( key1, key2, key3, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			cell->key[0] == key1 && cell->key[1] == key2 &&
			cell->key[2] == key3 )
			return cell;
		}
	return NULL;
	}

extern HashCell
cacheGetCellN( Cache h, IGeneric* key, IGeneric v ) {
	Juint			hash;
	IHashScheme		scheme = h->scheme;
	HashCell	cell;

	hash = phash2( key, key, h->shift );
	for( cell = cacheGet0( h, hash ); cell != NULL; cell = cell->next ) {
		if( cell->hash == hash &&
			COM_ICALL( equals, scheme )( scheme, h, cell->key, key ) )
			return cell;
		}
	return NULL;
	}
