/********************************************************************
 *		File			:		list.c
 *		Description		:		linked list
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "metautil.h"

extern Jerror
listConstruct( List l ) {
	l->head = l->tail = NULL;
	l->size = 0;
	l->mem = NULL;
	return 0;
	}

extern Jerror
listConstruct0( List l, IMem mem ) {
	l->head = l->tail = NULL;
	l->size = 0;
	l->mem = mem;
	return 0;
	}

extern Jerror
listDestruct( List l ) {
	l->head = l->tail = NULL;
	l->size = 0;
	return 0;
	}

extern Jerror
listDelAll( List l ) {
	ListCell 	cell, next;
	IMem		mem = l->mem;

	if( mem ) {
		cell = listGetHead( l );
		while( cell != NULL ){
			next = cell->next;
			listDelCell( mem, cell );
			cell = next;
			}
		}
	listDestruct( l );
	return 0;
	}

extern Jint
listDelDataAll( List list, void* data ) {
	ListCell 	cell, next;
	IMem		mem = list->mem;
	Jint 		rvl = 0;

	cell = listGetHead( list );
	while( cell != NULL ){
		next = cell->next;
		if( cell->data == data ){
			listRemoveElement( list, cell );
			if( mem )
				listDelCell( mem, cell );
			rvl++;
			}
		cell = next;
		}
	return rvl;
	}

extern Jboolean
listDelData( List list, IGeneric data ) {
	ListCell	cell;
	IMem		mem = list->mem;

	cell = listSearch( list, data );
	if( cell ) {
		listRemoveElement( list, cell );
		if( mem )
			listDelCell( mem, cell );
		return TRUE;
		}
	return FALSE;
	}


extern ListCell
listInsertElement( List l, ListCell e, ListCell after ) {
	ListCell  	el;

	if( l->head == NULL && l->tail == NULL ) {
		assert( after == NULL );
		l->head = e;
		l->tail = e;
		e->next = e->prev = NULL;
		l->size = 1;

		return e;
		}

	assert( l->head != NULL && l->tail != NULL );
	if( after == NULL ) {
		el = l->head;
		l->head = e;
		e->next = el;
		el->prev = e;
		e->prev = NULL;
		l->size ++;
		}
	else {
		el = after->next;
		after->next = e;
		e->prev = after;
		e->next = el;
		if( el != NULL )
			el->prev = e;
		else
			l->tail = e;
		l->size ++;
		}
	return e;
	}

extern ListCell
listRemoveElement( List l, ListCell e ) {
	assert( e != NULL );
	
	if( e->prev != NULL )
		e->prev->next = e->next;
	else
		l->head = e->next;
	if( e->next != NULL )
		e->next->prev = e->prev;
	else
		l->tail = e->prev;
	l->size --;
	
	e->prev = e->next = NULL;
	return e;
	}


extern List
listConcat( List l, List l2 ) {
	ListCell  e;

	assert( l->mem == l2->mem );
	if( l == NULL && l2 == NULL ) return NULL;
	if( l2 == NULL ) return l;
	if( l == NULL ) return l2;

	if( l2->size == 0 )
		return l;
	
	if( l->size == 0 ) 
		l->head = l2->head;
	else {
		e = l->tail;
		e->next = l2->head;
		l2->head->prev = e;
		}
	
	l->tail = l2->tail;               /* l1's new tail is l2.tail */
	l->size = l->size + l2->size;
	
	l2->size = 0;
	l2->head = l2->tail = NULL;
	return l;
	}

extern ListCell
listSearch( List l, void* data ) {
	ListCell	c;
	
	listForeach( l, c ) {
		if( c->data == data )
			return c;
		}
	return NULL;
	}

extern Jerror
listClone( List from, List to ) {
	ListCell cell, it;

	assert( to->mem != NULL );
	listForeach( from, it ){
		listNewCell( to->mem, cell, listData(IGeneric, it) );
		listAppendElement( to, cell );
		}
	return 0;
	}

typedef struct {
	struct _ISequence	intf;
	List				list;
	ListCell			cur;
	} *ListIter;

static COMDECL
_listIterQuery( ISequence mi, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, Iterator )
		|| COM_SAME_IIDNAME( iid, Sequence ) ) {
		*i = mi;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
_listIterAddRef( ISequence ci ) {
	return 1;
	}

static COMDECLU
_listIterRelease( ISequence ci ) {
	return 1;
	}

static COMDECL
_listIterReset( ISequence iti ) {
	ListIter	it = (ListIter)iti;

	it->cur = listGetHead( it->list );
	return 0;
	}

static Jboolean	COMCALL
_listIterHasMoreElements( ISequence iti ) {
	ListIter	it = (ListIter)iti;

	return it->cur != NULL;
	}

static IGeneric COMCALL
_listIterNextElement( ISequence iti ) {
	ListIter	it = (ListIter)iti;
	IGeneric	rtn = listData( IGeneric, it->cur );
	
	it->cur = it->cur->next;
	return rtn;
	}

static COMDECL
_listIterPushElement( ISequence iti, IMem mem, IGeneric data ) {
	ListIter	it = (ListIter)iti;
	ListCell	cell;

	listNewCell( mem, cell, data );
	listPush( it->list, cell );
	
	return 0;
	}

static IGeneric COMCALL
_listIterPopElement( ISequence iti ) {
	ListIter	it = (ListIter)iti;
	ListCell	cell;

	cell = listPop( it->list );
	if( cell != NULL )
		return cell->data;
	return NULL;
	}

static COMDECL
_listIterAppendElement( ISequence iti, IMem mem, IGeneric data ) {
	ListIter	it = (ListIter)iti;
	ListCell	cell;

	listNewCell( mem, cell, data );
	listAppendElement( it->list, cell );

	return 0;
	}

static COMDECL
_listIterRemoveElement( ISequence iti, IGeneric data ) {
	ListIter	it = (ListIter)iti;
	ListCell	cell;

	cell = listSearch( it->list, data );
	if( cell != NULL )
		listRemoveElement( it->list, cell );
	return 0;
	}

static COMDECL
_listIterRemoveAllElements( ISequence iti ) {
	ListIter	it = (ListIter)iti;
	
	listConstruct( it->list );
	return 0;
	}


static struct _DSequence	iterVtbl = {
	_listIterQuery, _listIterAddRef, _listIterRelease,
	_listIterReset, _listIterHasMoreElements, _listIterNextElement,
	_listIterPushElement, _listIterPopElement,
	_listIterAppendElement, _listIterRemoveElement, _listIterRemoveAllElements
    };


extern Jerror
listConstructSequence( List l, ISequence mem ) {
	ListIter	it = (ListIter)mem;

	it->list = l;
	it->cur = listGetHead( it->list );
	it->intf.vtbl = &iterVtbl;
	
	return 0;
	}

