/********************************************************************
 *		File			:		map.c
 *		Description		:		a sparse integer map with lazy allocation
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "metautil.h"

/*******************************************************************
 *	map implementation
 *******************************************************************/
static Jerror
mapGrow( Map m, Jint level1Log ) {
	Jint		i;
	IGeneric**	table;
	
	assert( level1Log > m->level1Log );
	table = comAlloc( m->mem, (1 << level1Log)*sizeof(IGeneric*) );
	if( table == NULL )
		return COM_E_OUTOFMEMORY;
	for( i = 0; i < (1 << m->level1Log); i ++ )
		table[i] = m->table[i];
	for( ; i < (1 << level1Log); i ++ )
		table[i] = NULL;
	comFree( m->mem, m->table );
	m->table = table;
	m->level1Log = level1Log;
	return 0;
	}

extern Jerror
mapConstruct( Map m, IMem mem, Jint level1Log, Jint level2Size ) {
	Jint		i, level2Log;
	IGeneric**	table;
	
	table = comAlloc( mem, (1 << level1Log)*sizeof(IGeneric*) );
	if( table == NULL )
		return COM_E_OUTOFMEMORY;
	for( i = 0; i < (1<<level1Log); i ++ )
		table[i] = NULL;
	for( level2Log = 0; (1<<level2Log) < level2Size; level2Log ++ );
	m->mem = mem;
	m->level1Log = level1Log;
	m->level2Size = level2Size;
	m->level2Log = level2Log;
	m->table = table;
	return 0;
	}

extern Jerror
mapDestruct( Map m ) {
	Jint		i;
	IGeneric*	bucket;
	
	for( i = 0; i < (1<<m->level1Log); i ++ ) {
		bucket = m->table[i];
		if( bucket )
			comFree( m->mem, bucket );
		}
	comFree( m->mem, m->table );
	return 0;
	}

extern IGeneric
mapGet( Map m, IGeneric key ) {
	Jint		hash = (Jint)key;
	Jint		level2Index, level1Index;
	IGeneric*	bucket;
	
	level1Index = hash >> m->level2Log;
	bucket = m->table[level1Index];
	if( bucket == NULL )
		return NULL;
	level2Index = hash & ((1<<m->level2Log) - 1);
	assert( level2Index < m->level2Size );
	return bucket[level2Index];
	}

extern Jint
mapGetI( Map m, IGeneric key ) {
	IGeneric	rtn;

	rtn = mapGet( m, key );
	return (Jint)rtn;
	}

extern Jfloat
mapGetF( Map m, IGeneric key ) {
	IGeneric	rtn;

	rtn = mapGet( m, key );
	return *(Jfloat *)&rtn;
	}

extern Jerror
mapPutI( Map m, IGeneric key, Jint data ) {
	return mapPut( m, key, (IGeneric)data );
	}

extern Jerror
mapPutF( Map m, IGeneric key, Jfloat data ) {
	IGeneric 	v;

	v = *(IGeneric*)&data;
	return mapPut( m, key, v );
	}

extern Jerror
mapPut( Map m, IGeneric key, IGeneric data ) {
	Jint		hash = (Jint)key;
	Jint		level2Index, level1Index, level1Log;
	IGeneric*	bucket;
	Jerror		rc;
	Jint		j;
	
	level1Index = hash >> m->level2Log;
	if( level1Index >= (1 << m->level1Log) ) {
		for( level1Log = 0; level1Index >= (1<<level1Log); level1Log ++ );
		rc = mapGrow( m, level1Log );
		if( COM_FAILED(rc) )
			return rc;
		}
	bucket = m->table[level1Index];
	if( bucket == NULL ) {
		bucket = comAlloc( m->mem, m->level2Size*sizeof(IGeneric) );
		if( bucket == NULL )
			return COM_E_OUTOFMEMORY;
		for( j = 0; j < m->level2Size; j ++ )
			bucket[j] = NULL;
		m->table[level1Index] = bucket;
		}
	level2Index = hash & ((1<<m->level2Log) - 1);
	assert( level2Index < m->level2Size );
	bucket[level2Index] = data;
	return 0;
	}

/*******************************************************************
 *	mapset implementation
 *******************************************************************/
static void
makeBucketEmpty( Jint* bucket, Jint size ) {
	Jint*		end = bucket + size;

	while( bucket < end ) {
		*bucket ++ = 0;
		}
	}

static void
copyBucket( Jint* bucket1, Jint* bucket2, Jint size ) {
	Jint*		end = bucket1 + size;

	while( bucket1 < end ) {
		*bucket2 ++ = *bucket1 ++;
		}
	}

static void
orBucket( Jint* bucket1, Jint* bucket2, Jint* bucket3, Jint size ) {
	Jint*		end = bucket1 + size;

	while( bucket1 < end ) {
		*bucket3 ++ = *bucket1++ | *bucket2 ++;
		}
	}

static void
andBucket( Jint* bucket1, Jint* bucket2, Jint* bucket3, Jint size ) {
	Jint*		end = bucket1 + size;

	while( bucket1 < end ) {
		*bucket3 ++ = *bucket1++ & *bucket2 ++;
		}
	}

static Jboolean
isBucketEmpty( Jint* bucket, Jint size ) {
	Jint*		end = bucket + size;

	while( bucket < end ) {
		if( *bucket ++ != 0 )
			return FALSE;
		}
	return TRUE;
	}

typedef void	(*BucketCallback)(
	int* bucket1, Jint* bucket2, Jint* bucket3, Jint size
	);

extern Jerror
msetCopy( Map s, Map another ) {
	Jint		level1Index;
	Jint*		bucket;
	Jint*		bucket1;
	Jint		size, i;
	Jerror		rc;
	
	assert( s->level2Size == another->level2Size );
	size = s->level2Size;
	
	/* make sure another is at least as big as s */
	if( s->level1Log > another->level1Log ) {
		rc = mapGrow( s, s->level1Log );
		if( COM_FAILED(rc) )
			return rc;
		}
	
	/* copy */
	for( level1Index = 0; level1Index < (1<<s->level1Log); level1Index ++ ) {
		bucket = (Jint*)s->table[level1Index];
		bucket1 = (Jint*)another->table[level1Index];
		if( bucket == NULL ) {
			if( bucket1 == NULL )
				continue;
			makeBucketEmpty( bucket1, size );
			}
		else {
			if( bucket1 == NULL ) {
				bucket1 = comAlloc( another->mem, size*sizeof(IGeneric) );
				if( bucket == NULL )
					return COM_E_OUTOFMEMORY;
				another->table[level1Index] = (IGeneric *)bucket1;
				}
			for( i = 0; i < size; i ++ )
				bucket1[i] = bucket[i];
			}
		}
	for( ; level1Index < (1<<another->level1Log); level1Index ++ ) {
		bucket1 = (Jint*)another->table[level1Index];
		makeBucketEmpty( bucket1, size );
		}
	return 0;
	}

static Jerror
msetApply( Map s, Map another, Map result, BucketCallback cb ) {
	Jint		level1Index;
	Jint*		bucket;
	Jint*		bucket1;
	Jint*		bucket2;
	Jint		size;
	Map			tmp;
	Jerror		rc;
	
	assert( s->level2Size == another->level2Size );
	assert( s->level2Size == result->level2Size );
	size = s->level2Size;
	
	/* make sure s is less than another */
	if( s->level1Log > another->level2Log ) {
		tmp = s;
		s = another;
		another = tmp;
		}
	/* make sure result is at least as big as another */
	if( another->level1Log > result->level1Log ) {
		rc = mapGrow( result, another->level1Log );
		if( COM_FAILED(rc) )
			return rc;
		}
	
	/* unite */
	for( level1Index = 0; level1Index < (1<<s->level1Log); level1Index ++ ) {
		bucket = (Jint*)s->table[level1Index];
		bucket1 = (Jint*)another->table[level1Index];
		bucket2 = (Jint*)result->table[level1Index];
		if( bucket == NULL ) {
			if( bucket1 == NULL ) {
				if( bucket2 == NULL )
					continue;
				makeBucketEmpty( bucket2, size );
				}
			else
				copyBucket( bucket1, bucket2, size );
			}
		else {
			if( bucket2 == NULL ) {
				bucket2 = comAlloc( result->mem, size*sizeof(IGeneric) );
				if( bucket2 == NULL )
					return COM_E_OUTOFMEMORY;
				result->table[level1Index] = (IGeneric *)bucket2;
				}
			if( bucket1 == NULL ) {
				copyBucket( bucket, bucket2, size );
				}
			else
				(*cb)( bucket, bucket1, bucket2, size );
			}
		}
	/* copy the rest of another */
	for( ; level1Index < (1 << another->level1Log); level1Index ++ ) {
		bucket1 = (Jint*)another->table[level1Index];
		bucket2 = (Jint*)result->table[level1Index];
		if( bucket1 == NULL ) {
			if( bucket2 == NULL )
				continue;
			else
				makeBucketEmpty( bucket2, size );
			}	
		else {
			if( bucket1 == NULL ) {
				bucket2 = comAlloc( result->mem, size*sizeof(IGeneric) );
				if( bucket2 == NULL )
					return COM_E_OUTOFMEMORY;
				result->table[level1Index] = (IGeneric *)bucket2;
				}
			copyBucket( bucket1, bucket2, size );
			}
		}
    /* empty the rest of result */
	for( ; level1Index < (1 << result->level1Log); level1Index ++ ) {
		bucket2 = (Jint*)result->table[level1Index];
		if( bucket2 == NULL )
			continue;
		makeBucketEmpty( bucket2, size );
		}
	return 0;
	}

extern Jerror
msetUnite( Map s, Map another, Map result ) {
	return msetApply( s, another, result, orBucket );
	}

extern Jerror
msetIntersect( Map s, Map another, Map result ) {
	return msetApply( s, another, result, andBucket );
	}


extern Jerror
msetComplement( Map s, Map result ) {
	assert( 0 );
	return COM_E_NOTIMPL;
	}

extern Jerror
msetIsMember( Map s, Jint m ) {
	Jint		hash = m >> 5;
	Jint		level2Index, level1Index;
	Jint*		bucket;
	
	level1Index = hash >> s->level2Log;
	level2Index = hash & ((1<<s->level2Log) - 1);
	if( level1Index >= (1 << s->level1Log) )
		return FALSE;
	bucket = (Jint*)s->table[level1Index];
	if( bucket == NULL )
		return FALSE;
	return ( bucket[level2Index] &
			 (0x01 << ( m & 31 )) ) != 0;
	}

extern Jerror
msetInclude( Map s, Jint m ) {
	Jint		hash = m >> 5;
	Jint		level2Index, level1Index, level1Log;
	Jint*		bucket;
	Jint		j;
	
	level1Index = hash >> s->level2Log;
	level2Index = hash & ((1<<s->level2Log) - 1);
	
	if( level1Index >= (1 << s->level1Log) ) {
		for( level1Log = 0; level1Index >= (1<<level1Log); level1Log ++ );
		mapGrow( s, level1Log );
		}
	
	bucket = (Jint*)s->table[level1Index];
	if( bucket == NULL ) {
		bucket = comAlloc( s->mem, s->level2Size*sizeof(IGeneric) );
		if( bucket == NULL )
			return COM_E_OUTOFMEMORY;
		for( j = 0; j < s->level2Size; j ++ )
			bucket[j] = NULL;
		s->table[level1Index] = (IGeneric *)bucket;
		}
	bucket[ level2Index ] |= (0x01 << ( m & 31 ) );
	return 0;
	}

extern Jerror
msetExclude( Map s, Jint m ) {
	Jint		hash = m >> 5;
	Jint		level2Index, level1Index;
	Jint*		bucket;
	
	level1Index = hash >> s->level2Log;
	level2Index = hash & ((1<<s->level2Log) - 1);
	if( level1Index >= (1 << s->level1Log) )
		return 0;
	bucket = (Jint*)s->table[level1Index];
	if( bucket == NULL )
		return 0;
	bucket[ level2Index ] &= ~(0x01 << ( m & 31 ));
	return 0;
	}

extern Jerror
msetIsEmpty( Map s ) {
	Jint		level1Index;
	Jint*		bucket;
	Jint		size;
	
	size = s->level2Size;
	for( level1Index = 0; level1Index < (1<<s->level1Log); level1Index ++ ) {
		bucket = (Jint*)s->table[level1Index];
		if( bucket != NULL )
			if( !isBucketEmpty( bucket, size ) )
				return FALSE;
		}
	return TRUE;
	}

extern Jerror
msetEquals( Map s, Map another ) {
	Jint		level1Index;
	Jint*		bucket;
	Jint*		bucket1;
	Jint		size, i;
	Map			tmp;

	assert( s->level2Size == another->level2Size );
	size = s->level2Size;
	
	/* make sure level1Log of s is less than another */
	if( s->level1Log > another->level2Log ) {
		tmp = s;
		s = another;
		another = tmp;
		}
	/* compare */
	for( level1Index = 0; level1Index < (1<<s->level1Log); level1Index ++ ) {
		bucket = (Jint*)s->table[level1Index];
		bucket1 = (Jint*)another->table[level1Index];
		if( bucket == NULL ) {
			if( bucket1 == NULL )
				continue;
			if( !isBucketEmpty( bucket1, size ) )
				return FALSE;
			}
		else {
			if( bucket == NULL && !isBucketEmpty( bucket, size ) )
				return FALSE;
			for( i = 0; i < size; i ++ )
				if( bucket[i] != bucket1[i] )
					return FALSE;
			}
		}
	/* see if the extra part of another is empty */
	for( ; level1Index < (1<<another->level1Log); level1Index ++ ) {
		bucket1 = (Jint*)another->table[level1Index];
		if( bucket1 != NULL && !isBucketEmpty( bucket1, size ) )
			return FALSE;
		}

	return TRUE;
	}

extern Jerror
msetMakeEmpty( Map s ) {
	Jint		level1Index;
	Jint*		bucket;
	Jint		size, i;
	
	size = s->level2Size;
	
	for( level1Index = 0; level1Index < (1<<s->level1Log); level1Index ++ ) {
		bucket = (Jint*)s->table[level1Index];
		if( bucket != NULL ) {
			for( i = 0; i < size; i ++ )
				bucket[i] = 0;
			}
		}
	return 0;
	}

extern Jerror
msetMakeUniv( Map s ) {
	assert( 0 );
	return COM_E_NOTIMPL;
	}

extern Jerror
msetDump( Map s ) {
	Jint	i, j;
	Jint*	bucket;
	
	for( i = 0; i < (1 << s->level1Log); i ++ ) {
		bucket = (Jint *)s->table[i];
		if( bucket == NULL )
			continue;
		for( j = 0; j < (s->level2Size << 5); j ++ )
			if( (bucket[j >> 5] & (0x01 << ( j & 31 ) )) != 0 ) 
				fprintf( stderr, " %d", (i<<s->level2Log<<5) + j );
		}
	fprintf( stderr, "\n" );
	return 0;
	}

extern Jint
msetChoose( Map s ) {
	Jint		level1Index;
	Jint*		bucket;
	Jint		size, i, j, d;
	
	size = s->level2Size;
	
	for( level1Index = 0; level1Index < (1<<s->level1Log); level1Index ++ ) {
		bucket = (Jint*)s->table[level1Index];
		if( bucket != NULL ) {
			for( i = 0; i < size; i ++ )
				if( bucket[i] != 0 )
					break;
			if( i < size ) {
				d = bucket[i];
				for( j = 0; j < 32; j ++ ) {
					if( (d & 0x01) != 0 ) break;
					d >>= 1;
					}
				j += (i << 5) + (level1Index << s->level2Log);
				}
			}
		}
	return -1;
	}

extern Jerror
msetConstruct( Map m, IMem mem, Jint level1Log, Jint level2Size ) {
	return mapConstruct( m, mem, level1Log, (level2Size+31)>>5 );
	}
 
extern Jerror
msetDestruct( Map m ) {
	return mapDestruct( m );
	}
 
/*******************************************************************
 *	iterator
 *******************************************************************/
typedef struct {
	struct _IMap		intf;
	Map					tbl;
	Jint				i, j;
	} *MapIter;

static COMDECL
_mapIterQuery( IMap mi, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, Iterator ) 
		|| COM_SAME_IIDNAME( iid, Map ) ) {
		*i = mi;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
_mapIterAddRef( IMap ci ) {
	return 1;
	}

static COMDECLU
_mapIterRelease( IMap ci ) {
	return 1;
	}

static COMDECL
_mapIterReset( IMap iti ) {
	MapIter		it = (MapIter)iti;
	IGeneric*	bucket;
	
    for( it->i = 0; it->i < (1 << it->tbl->level1Log); it->i ++ ) {
		bucket = it->tbl->table[it->i];
		if( bucket == NULL )
			continue;
		for( it->j = 0; it->j < it->tbl->level2Size; it->j ++ ) 	
			if( bucket[it->j] != NULL )
				return 0;
		}
	return 0;
	}

static Jboolean	COMCALL
_mapIterHasMoreElements( IMap iti ) {
	MapIter	it = (MapIter)iti;

	if( it->i >= (1 << it->tbl->level1Log) )
		return FALSE;
	return TRUE;
	}

static IGeneric COMCALL
_mapIterNextElement( IMap iti ) {
	MapIter		it = (MapIter)iti;
	IGeneric*	bucket;
	IGeneric	rtn = it->tbl->table[it->i][it->j ++];

	for( ; it->i < (1 << it->tbl->level1Log); it->i ++, it->j = 0 ) {
		bucket = it->tbl->table[it->i];
		if( bucket == NULL )
			continue;
		for( ; it->j < it->tbl->level2Size; it->j ++ ) 	
			if( bucket[it->j] != NULL )
				return rtn;
		}
	return rtn;
	}

static COMDECL
_mapIterPut( IMap iti, IGeneric key, IGeneric data ) {
	MapIter	it = (MapIter)iti;
	
	return mapPut( it->tbl, key, data );
	}

static IGeneric COMCALL
_mapIterGet( IMap iti, IGeneric key ) {
	MapIter	it = (MapIter)iti;
	
	return mapGet( it->tbl, key );
	}

static IGeneric COMCALL
_mapIterKey( IMap iti ) {
	MapIter		it = (MapIter)iti;

	return (IGeneric)((it->i<<it->tbl->level2Log) + it->j);
	}


static struct _DMap	iterVtbl = {
	_mapIterQuery, _mapIterAddRef, _mapIterRelease,
	_mapIterReset, _mapIterHasMoreElements, _mapIterNextElement,
	_mapIterPut, _mapIterGet, _mapIterKey
    };


extern Jerror
mapConstructMap( Map h, IMap mem ) {
	MapIter	it = (MapIter)mem;

	it->intf.vtbl = &iterVtbl;
	it->tbl = h;
	_mapIterReset( mem );
	
	return 0;
	}


