/********************************************************************
 *		File			:		mem.c
 *		Description		:		memory managers
 *		Author			:		Jianwen Zhu
 *		Reference		:		lcc
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <malloc.h>

#include "metautil.h"

struct _DMem dheap;
static struct _DMem darena;

COM_DECLARE_CLASS( HeapMem );
COM_DECLARE_CLASS( ArenaMem );
COM_DECLARE_CLASS( Block );

struct _Block {
	Block		next;
	char*		limit;
	char*		avail;
	};

union align {
	long 		l;
	char*		p;
	double 		d;
	int 		(*f)(void);
	};

union header {
	struct _Block 	b;
	union align 	a;
	};

struct _ArenaMem {
	struct _IMem	intf;
	Block			first;		/* the first memory block */
	Block			last;		/* the last memory block */
	struct _Block	sentinal;
	};

static	Block		freeblocks = NULL;

struct _HeapMem {
	struct _IMem	intf;
	};

static COMDECL
query( IMem o, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, Component )
		|| COM_SAME_IIDNAME( iid, Mem ) ) {
		*i = o;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
addref( IMem m ) {
	return 1;
	}

static COMDECLU
release( IMem m ) {
	return 1;
	}

static Jint	COMCALL
avail( IMem m ) {
	return COM_E_NOTIMPL;
	}

static char*
dump( IMem m, FILE* fp, char* bp ) {
	return bp;
	}

static COMDECL
_heapDestruct( IMem m ) {
	return 0;
	}

static COMDECL
_heapConstruct( IMem mi ) {
	HeapMem		heap = (HeapMem)mi;

	heap->intf.vtbl = &dheap;
	return 0;
	}

static void* COMCALL
_heapAlloc( IMem m, Juint size ) {
	return malloc( size );
	}

static void* COMCALL
_heapAllocArray( IMem m, Juint size, Juint length ) {
	return malloc( size * length );
	}

static void* COMCALL
_heapAllocMDArray( IMem m, Juint size, Jlong dim ) {
	return NULL;
	}

static void* COMCALL
_heapAllocAligned( IMem m, Juint size, Juint align ) {
	return memalign( align, size );
	}

static COMDECL
_heapGrow( IMem mi, Juint size ) {
	return COM_E_FAIL;
	}

static COMDECLV
_heapFree( IMem m, void *ptr ) {
	free( ptr );
	}

#define PURIFY
#ifdef PURIFY
static void* COMCALL
arenaAlloc( IMem mi, Juint size ) {
	ArenaMem		m = (ArenaMem)mi;
	union header *b = malloc( sizeof *b + size );

	if( b == NULL ) {
		return NULL;
        }
	b->b.next = m->first;
	m->first = (Block)b;
	return b + 1;
	}

static COMDECL
arenaGrow( IMem mi, Juint size ) {
	return COM_E_FAIL;
	}

static Jerror
arenaDestruct( IMem mi ) {
	ArenaMem			m = (ArenaMem)mi;
	union header 	*p, *q;

	for( p = (union header *)m->first; p != (union header *)m->last; p = q ) {
		q = (void *)p->b.next;
		free(p);
        }
	m->first = m->last;
	return 0;
	}

#else
#define ROUNDUP(x,n) (((x)+((n)-1))&(~((n)-1)))

static void* COMCALL
arenaAlloc( IMem mi, Juint size ) {
	ArenaMem		m = (ArenaMem)mi;
	Block		ap = m->last;
	long		n = size;
	
	assert( n > 0 );
	n = ROUNDUP( n, sizeof (union align) );
	
	while( n > ap->limit - ap->avail ) {
		Block		b;

		ap->next = b = freeblocks;
		if( b != NULL ) {
			freeblocks = freeblocks->next;
			ap = b;
			}
		else {
			unsigned m = sizeof (union header) + n + 1024;
			ap->next = malloc( m );
			ap = ap->next;
			if( ap == NULL ) {
				return NULL;
				}
			ap->limit = (char *)ap + m;
			}
		ap->avail = (char *)((union header *)ap + 1);
		ap->next = NULL;
		m->last = ap;
        }
	ap->avail += n;
	return ap->avail - n;
	}

static COMDECL
arenaGrow( IMem mi, Juint size ) {
	ArenaMem		m = (ArenaMem)mi;
	Block		ap = m->last;

	if( ap->avail + size < ap->limit ) {
		ap->avail += size;
		return 0;
		}
	return COM_E_FAIL;
	}

static Jerror
arenaDestruct( IMem mi ) {
	ArenaMem		m = (ArenaMem)mi;

	m->last->next = freeblocks;
	freeblocks = m->first->next;
	m->first->next = NULL;
	m->last = m->first;
	return 0;
	}

#endif

static COMDECL
arenaConstruct( IMem mi ) {
	ArenaMem		arena = (ArenaMem)mi;

	arena->intf.vtbl = &darena;
	arena->first = &arena->sentinal;
	arena->last = arena->first;
	arena->sentinal.next = NULL;
	arena->sentinal.avail = arena->sentinal.limit = NULL;
	return 0;
	}


static void* COMCALL
arenaAllocArray( IMem mi, Juint size, Juint length ) {
	return arenaAlloc( mi, size*length );
	}

static void* COMCALL
arenaAllocMDArray( IMem mi, Juint size, Jlong dim ) {
	return NULL;
	}

static void* COMCALL
arenaAllocAligned( IMem m, Juint size, Juint align ) {
	/* xxxx fix me */
	return arenaAlloc( m, size );
	}

static COMDECLV
arenaFree( IMem m, void *ptr ) {
	return;
	}

extern void
arenaDumpFreeBlocks( void ) {
	Block		ap;

	for( ap = freeblocks; ap != NULL; ap = ap->next )
		fprintf( stderr, "%x ", (Jint)ap );
	fprintf( stderr, "\n" );
	}

extern void
arenaDumpBlocks( IMem mi ) {
	Block		ap;
	ArenaMem	m = (ArenaMem)mi;
	
	for( ap = m->first; ap != NULL; ap = ap->next )
		fprintf( stderr, "%x ", (Jint)ap );
	fprintf( stderr, "\n" );
	}

extern IMem
comNewHeap( IMem mem ) {
	HeapMem		heap = NULL;
	
	assert( mem == NULL );
	heap = malloc( sizeof(*heap) );
	if( heap != NULL ) {
		_heapConstruct( &heap->intf );
		return &heap->intf;
		}
	return NULL;
	}

extern IMem
comNewGC( IMem mem ) {
	return 0;
	}

extern IMem
comNewArena( IMem mem ) {
	ArenaMem		arena = NULL;

	if( mem == NULL ) 
		mem = comHeap();
	arena = comAlloc( mem, sizeof(*arena) );
	if( arena == NULL )
		return NULL;
	arenaConstruct( &arena->intf );
	return &arena->intf;
	}

struct _DMem dheap = {
	query,
	addref,
	release,
	dump,
	_heapDestruct,
	_heapAlloc,
	_heapAllocArray,
	_heapAllocMDArray,
	_heapAllocAligned,
	_heapGrow,
	_heapFree,
	avail
	};

static struct _DMem darena = {
	query,
	addref,
	release,
	dump,
	arenaDestruct,
	arenaAlloc,
	arenaAllocArray,
	arenaAllocMDArray,
	arenaAllocAligned,
	arenaGrow,
	arenaFree,
	avail
    };


