/********************************************************************
 *		File			:		metautil.h
 *		Description		:		utilities
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#ifndef METAUTIL_H
#define METAUTIL_H

#include <stdio.h>

#include <com.h>
#include <imd.h>
#include <iservice.h>
#include <ilexicon.h>
#include <ierror.h>

/***********************************************************************
 *	classes exported
 ***********************************************************************/
COM_DECLARE_CLASS( List );
COM_DECLARE_CLASS( Vector );
COM_DECLARE_CLASS( HashTable );
COM_DECLARE_CLASS( BitSet );
COM_DECLARE_CLASS( SparseSet );
COM_DECLARE_CLASS( Map );
COM_DECLARE_CLASS( Graph );
COM_DECLARE_CLASS( BinHeap );
COM_DECLARE_CLASS( Heap );
COM_DECLARE_CLASS( Spreadsheet );
COM_DECLARE_CLASS( RandomProcess );

COM_DECLARE_CLASS( ListCell );
COM_DECLARE_CLASS( KeyCell );
COM_DECLARE_CLASS( GraphNode );
COM_DECLARE_CLASS( GraphEdge );

COM_ALIAS_CLASS( Map, MapSet );
COM_ALIAS_CLASS( HashTable, Cache );
COM_ALIAS_CLASS( KeyCell, HashCell );
COM_ALIAS_CLASS( KeyCell, HeapCell );

COM_DECLARE_INTERFACE( HashScheme );
COM_DECLARE_INTERFACE( OrderScheme );

#define COM_IHASHSCHEME_IID COM_DEFINE_GUID(0x00000000, 0x0000, 0x0000, \
                0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46)
#define COM_IORDERSCHEME_IID COM_DEFINE_GUID(0x00000000, 0x0000, 0x0000, \
                0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x47)
	
struct _ListCell {
	IGeneric		data;	
	ListCell		next;
	ListCell		prev;
	};

struct _KeyCell {
	IGeneric		val;
	KeyCell			next;
	Juint			hash;
	IGeneric*	  	key[1];
	};

struct _List {
	IMem			mem;
	Jint			size;
	ListCell		head;
	ListCell		tail;
	};

struct _Vector {
	IMem			mem;
	Jint			count;
	Jint			maxCount;
	IGeneric*		data;
	};

struct _HashTable {
	IMem			mem;
	Jint			count;
	IHashScheme		scheme;
	Jshort			level1MaxSize;	/* size = 1 << level1Size */
	Jshort			level1Size;		/* size = 1 << level1Size */
	Jshort			level2Size;		/* size = 1 << level2Size */
	Jfloat			loadFactor;
	Jint			threshold;
	HashCell**		table;			/* buckets */
	Jshort**		bucketSize;		/* size of buckets */

	/* the following is specific for cache implementation */
	Jshort			keySize;		/* size of key */
	Jshort			shift;			/* used for hash calculation */
	Jshort			assoc;			/* associativity (-1 if unlimited) */
	};

struct _DHashScheme {
	COM_INHERIT_UNKNOWN( HashScheme );
	
	Juint COMCALL	(*hashCode)(
		IHashScheme si, HashTable h, void* key
		);
	Jboolean COMCALL(*equals)(
		IHashScheme si, HashTable h, void* key, void* another
		);
	};

struct _DOrderScheme {
	COM_INHERIT_UNKNOWN( OrderScheme );

	/* returns 0 if a==b, 1 if a>b, -1 if a<b
	 * keysize > 1 for multiple keys
	 * weight != NULL if weighted sum of keys
	 */
	Jint COMCALL	(*compare)(
		IOrderScheme si, Jint keysize, void* weights,
		void* a, void* b
		);
	};

struct _BitSet {
	IMem			fMem;
	Juint*			fData;
	Jint			fSize;
	};

struct _SparseSet {
	IMem			fMem;
	Jint			fUnivSize;
	Jint*			fArray;
	Jint*			fStack;
	Jint			fMemCount;
	};

struct _Map {
	IMem			mem;
	Jshort			level1Log;		/* level1Size = 1 << level1Size */
	Jshort			level2Size;		/* level2Size = 1 << level2Log */
	Jshort			level2Log;		
	IGeneric**		table;
	};

typedef void (*GraphMonitor)( GraphNode n );

struct _GraphNode {
	Graph			fGraph;
	Jint			fID;
	ListDecl		fIncoming;
	ListDecl		fOutgoing;
	Jint			fIncomingDegree;
	Jint			fOutgoingDegree;

	/* for annotating cost */
	Jint			fWeight;

	/* for weighted degree */
	Jint			fDegreeWeight;
	Jint			fIncomingWDegree;
	Jint			fOutgoingWDegree;
	
	/* for hide/show/cluster operation */
	short			fState; 	/* 0 shown; 1 hidden; 3 clusterred */

	/* for union-find clustering */
	GraphNode		fParent;
	int				fRank;
	ListDecl		fCluster;
	ListCell		fClusterLink;
	};

struct _GraphEdge {
	GraphNode		fSrc, fSink;
	Jint			fID;
	Jint			fWeight;
	Jboolean		fIsHidden;
	ListCell		fLinkIncoming;
	ListCell		fLinkOutgoing;
	HashCellDecl	fLinkMap;
	};

struct _Graph {
	IMem			fMem;
	Jint			fNodeNum, fEdgeNum;
	MapDecl			fNodes;
	MapDecl			fEdges;
	CacheDecl		fEdgeMap;
	GraphMonitor	fHideMonitor;
	GraphMonitor	fShowMonitor;
	GraphMonitor	fClusterMonitor;
	Jint			fMaxID, fMaxNodeID;
	};

struct _BinHeap {
	IMem			fMem;
	Jint			fMin;
	Jint			fMax;
	Jint			fLow;
	Jint			fHigh;
	Jint			fNumElts;
	HeapCell*		fBuckets;
	};

struct _Heap {
	VectorDecl		tree;
	Jint			keysize;
	IGeneric		weights;
	IOrderScheme	scheme;
	};

#define SS_MAX_COUNTER		64

struct _Spreadsheet {
	IMem		mem;
	Jint		nrows, ncols;
	Jint		maxrow;
	char**		headers;
	char*		types;
	char**		formats;
	
	union {
		Jint	i;
		char*	s;
		Jfloat	f;
		}*	data;

	Jlong		time[SS_MAX_COUNTER];
	Jint		timeTop;
	};

struct _RandomProcess {
	char		state[128];
	};
	
/***********************************************************************
 *	APIs
 ***********************************************************************/
#ifdef __cplusplus
extern "C" {
#endif



/***********************************************************************
 *	utilities components
 ***********************************************************************/
#define 				comBigEndian()	!comLittleEndian()
#define					comEndianSwap(in)					\
	((in << 24) & 0xff000000) | ((in <<  8) & 0x00ff0000) |	\
	((in >>  8) & 0x0000ff00) | ((in >> 24) & 0x000000ff)	
#define 				comHeap()		(iheap)
#define					comMin(a,b)		((a) < (b) ? (a) : (b))
#define					comMax(a,b)		((a) > (b) ? (a) : (b))
	
#define newHeap			comNewHeap		/* deprecated */
#define newArena		comNewArena		/* deprecated */
#define newLexicon		comNewLexicon	/* deprecated */
#define newShaDigest	comNewShaDigest	/* deprecated */
#define newService		comNewService	/* deprecated */
	
	
	extern Jerror			comBegin( void );
	extern Jerror			comEnd( void );
	extern char*			comPrintf(
		FILE* fp, char* bp, const char* fmt, ...
		);
	extern void				comVprintf0( Jint level, const char* fmt, ... );
	extern void				comVprintf( const char* fmt, ... );
	extern void				comIndent( void );
	extern void				comUnindent( void );
	
	extern Jboolean			comLittleEndian( void );
	extern Jint				comLog2( Jlong n );
	
	extern IMem				comNewHeap( IMem mi );
	extern IMem				comNewArena( IMem mi );
	extern ILexicon			comNewLexicon( IMem mem, int size );
	extern IMsgDigest		comNewShaDigest( IMem mem );
	extern IService			comNewService( IMem mem );

/***********************************************************************
 *	double linked list
 ***********************************************************************/
#define	listNewCell( mem, cell, d )	\
                cell = (ListCell)COM_CALL( alloc, mem )( mem, sizeof(*cell) );\
                cell->prev = cell->next = NULL;	\
                cell->data = d;
#define	listDelCell( mem, cell ) 	COM_CALL( free, mem )( mem, cell )
	
#define	listIsEmpty( l )			((l)->size == 0)
#define	listGetSize( l )			((l)->size)
#define	listGetHead( l )			((l)->head)
#define	listGetTail( l )			((l)->tail)
#define	listData( type, c )			((type)(c)->data)
#define listPrev( c )				((c)->prev)
#define listNext( c )				((c)->next)
	
#define	listAddElement( l, c )		listInsertElement( (l), (c), NULL )
#define	listAppendElement( l, c )	listInsertElement( (l), (c), (l)->tail )
#define listDelElement(l, c) { 		\
    listRemoveElement( l, c ); 		\
    listDelCell( (l)->mem,c ); }
	
#define listAppend( list, data )							\
	{ ListCell cell; listNewCell( (list)->mem, cell, (void*)data ); 	\
      listAppendElement( (list), cell ); }
	
#define	listMoveElement( l, c, l2 )	listRemoveElement( (l), (c) ); \
				                    listInsertElement( (l2), (c), (l2)->tail)
#define	listPop( l )		  		listRemoveElement( (l), (l)->head )
#define	listPush( l, c )			listAppendElement( (l), (c) )
	
#define	listForeach( l, e )			for((e) = (l)->head; (e); (e) = (e)->next)
#define	listForeachBackward( l, e )	for((e) = (l)->tail; (e); (e) = (e)->prev)
	
#define listConstructIterator( l, it )	\
    listConstructSequence( l, *(ISequence*)&it )
	
	extern Jerror		listConstruct( List l );
	extern Jerror		listConstruct0( List l, IMem mem );
	extern Jerror		listDestruct( List l );
	extern ListCell		listRemoveElement( List l, ListCell c );
	extern ListCell		listInsertElement( List l, ListCell c, ListCell at );
	extern List			listConcat( List l, List l2 );
	extern ListCell		listSearch( List l, void* data );
	extern Jerror		listClone( List from, List to );
	extern Jerror		listDelAll( List list );
	extern Jboolean		listDelData( List list, IGeneric data );
	extern Jint			listDelDataAll( List list, IGeneric data );
	extern Jerror		listConstructSequence( List l, ISequence it );
	extern Jerror		listSort(
		List l, Jint keysize, IGeneric weights, IOrderScheme scheme
		);
	
/***********************************************************************
 *	Vector
 ***********************************************************************/
#define	vecGetSize( v )				((v)->count)
#define	vecIsEmpty( v )				((v)->count == 0)
#define	vecGetElement( v, i )		((v)->data[i])
#define vecGetData( v )				((v)->data)
#define vecReset( v )				((v)->count = 0)
#define vecConstructIterator( l, it )		\
	(vecConstructSequence( l, *(ISequence*)&it ))

	extern Jerror		vecConstruct( Vector m, IMem mem, Jint count );
	extern Jerror		vecDestruct( Vector m );
	extern Jerror		vecExpand( Vector m, Jint count );
	extern Jerror		vecSetElement( Vector v, Jint i, IGeneric d );
	extern Jerror		vecAddElement( Vector v, IGeneric d );
	extern Jerror		vecRemoveElement( Vector v, Jint i );
	extern Jerror		vecPush( Vector v, IGeneric d );
	extern IGeneric		vecPop( Vector v );
	extern Jerror		vecConstructSequence( Vector v, ISequence seq );
	extern Jerror		vecSort(
		Vector v, Jint keysize, IGeneric weights, IOrderScheme scheme
		);
	
/***********************************************************************
 *	Hash table
 ***********************************************************************/
#define			hashGetSize(h)			(h->count)	
#define			hashData( type, h )		((type)h->val)
#define			hashKey( type, h )		((type)h->key[0])
#define			hashForeach( h, i, j, cell )	\
    for( (i) = 0; (i) < (1 << (h)->level1Size); (i) ++ ) 		\
		for( (j) = 0; (j) < (1 << (h)->level2Size); (j) ++ ) 	\
			for( (cell) = (h)->table[(i)][(j)]; (cell) != NULL; 	\
					 (cell) = (cell)->next ) 					\
																	
#define			hashNewCell( h, cell, k, d ) (		\
	(cell) = comAlloc( (h)->mem, sizeof(*cell) ), 	\
	(cell)->key[0] = (k), \
	(cell)->val = (d) )

#define hashDel(h,key)				\
	{ HashCell hc = hashRemove((h),(key)); comFree((h)->mem, hc); }	
																	
#define			hashConstructIterator( h, it )	\
	(hashConstructMap(h,*(IMap*)&it))

	extern IHashScheme	hashPtrScheme;
	extern IHashScheme	hashStringScheme;
	extern IHashScheme	hashMDScheme;
	extern IHashScheme	hashCacheScheme;
	
	extern Jerror	hashConstruct(
		HashTable h, IMem mem, IHashScheme scheme,
		Jshort level1Size, Jshort level2Size,
		float loadFactor
		);
	extern Jerror	hashDefaultConstruct( HashTable h, IMem mem );
	extern Jerror	hashDestruct( HashTable h );
	extern void*	hashGet( HashTable h, void* key );
	extern Jerror	hashPut( HashTable h, void* key, void* v );
	extern HashCell	hashGetCell( HashTable h, void* key );
	extern Jerror	hashPutCell( HashTable h, HashCell cell );
	extern HashCell	hashRemoveCell( HashTable h, HashCell cell );
	extern HashCell	hashRemove( HashTable h, void* key );
	extern Jerror	hashRemoveAllCells(
		HashTable h, HashCell* freehead, HashCell *freetail
		);
	extern Jerror	hashConstructMap( HashTable h, IMap it );

/***********************************************************************
 *	Cache management
 ***********************************************************************/
#define cachePutCell		hashPutCell
#define cacheRemoveCell		hashRemoveCell
	
	extern Jerror	cacheConstruct(
		Cache h, IMem mem, Jshort level1Size, Jshort level2Size,
		Jint keySize, Jint assoc
		);
	extern Jerror	cacheDefaultConstruct( Cache h, IMem mem, Jint keySize );
	extern Jerror	cacheDestruct( Cache h );
	extern Jerror	cachePut( Cache h, IGeneric key, IGeneric v );
	extern Jerror	cachePut2(
		Cache h, IGeneric key1, IGeneric key2, IGeneric v
		);
	extern Jerror	cachePut3(
		Cache h, IGeneric key1, IGeneric key2, IGeneric key3, IGeneric v
		);
	extern Jerror	cachePutN( Cache h, IGeneric* key, IGeneric v );
	extern IGeneric	cacheGet( Cache h, IGeneric key );
	extern IGeneric	cacheGet2( Cache h, IGeneric key1, IGeneric key2 );
	extern IGeneric	cacheGet3(
		Cache h, IGeneric key1, IGeneric key2, IGeneric key3
		);
	extern IGeneric	cacheGetN( Cache h, IGeneric* key, IGeneric v );
	extern HashCell	cacheGetCell( Cache h, IGeneric key );
	extern HashCell	cacheGetCell2( Cache h, IGeneric key1, IGeneric key2 );
	extern HashCell	cacheGetCell3(
		Cache h, IGeneric key1, IGeneric key2, IGeneric key3
		);
	extern HashCell	cacheGetCellN( Cache h, IGeneric* key, IGeneric v );
	
/***********************************************************************
 *	Bit vector based set
 ***********************************************************************/
#define bsetForeach( s, i )					\
    for( i = 0; i < (s)->fSize; i ++ )		\
	if( bsetIsMember((s), i) )

#define bsetConstructIterator(s,it)	(bsetConstructSet(s,*(ISet*)&it))	
	
	extern Jerror		bsetConstruct( BitSet s, IMem mem, Jint univSize );
	extern Jerror		bsetDestruct( BitSet s );
	extern Jerror		bsetCopy( BitSet s, BitSet to );
	extern Jerror		bsetUnite( BitSet s, BitSet another, BitSet result );
	extern Jerror		bsetIntersect(
		BitSet s, BitSet another, BitSet result
		);
	extern Jerror		bsetComplement( BitSet s, BitSet result );
	extern Jerror		bsetIsMember( BitSet s, Jint m );
	extern Jerror		bsetInclude( BitSet s, Jint m );
	extern Jerror		bsetExclude( BitSet s, Jint m );
	extern Jerror		bsetIsEmpty( BitSet s );
	extern Jerror		bsetEquals( BitSet s, BitSet another );
	extern Jerror		bsetMakeEmpty( BitSet s );
	extern Jerror		bsetMakeUniv( BitSet s );
	extern Jerror		bsetDump( BitSet s );
	extern Jint			bsetChoose( BitSet s );
	extern Jint 		bsetSize( BitSet s );
	extern Jerror		bsetConstructSet( BitSet s, ISet mem );
	
/***********************************************************************
 *	Brigg's sparse set
 ***********************************************************************/
#define			spsetIsMember( s, m )	\
    ((s)->fArray[m] >= 0 && (s)->fArray[m] < (s)->fMemCount )

#define 		spsetIsEmpty( s )	((s)->fMemCount == 0 )
#define			spsetSize( s )		((s)->fMemCount )
#define			spsetForeach( s, i, m )			\
	for( i = 0, m = (s)->fStack[i]; i < (s)->fMemCount; m = (s)->fStack[++i] )
#define			spsetConstructIterator(s,it)	\
    (spsetConstructSet(s,*(ISet*)&it))
	
	extern Jerror	spsetConstruct( SparseSet s, IMem mem, Jint univSize );
	extern Jerror	spsetDestruct( SparseSet s );
	extern Jerror	spsetMakeEmpty( SparseSet s );
	extern Jerror	spsetMakeUniverse( SparseSet s );
	extern Jerror	spsetInclude( SparseSet s, Jint m );
	extern Jerror	spsetExclude( SparseSet s, Jint m );
	extern Jerror	spsetCopy( SparseSet s, SparseSet to );
	extern Jerror	spsetUnite(
		SparseSet s, SparseSet another, SparseSet result
		);
	extern Jerror	spsetIntersect(
		SparseSet s, SparseSet another, SparseSet result
		);
	extern Jerror	spsetComplement( SparseSet s, SparseSet result );
	extern Jerror	spsetDiff(
		SparseSet s, SparseSet another, SparseSet result
		);
	extern Jerror	spsetEquals( SparseSet s, SparseSet another );
	extern Jerror	spsetDump( SparseSet s );
	extern Jint		spsetChoose( SparseSet s );
	extern Jerror	spsetConstructSet( SparseSet s, ISet it );

/***********************************************************************
 *	Sparse integer Map
 ***********************************************************************/
	extern Jerror	mapConstruct(
		Map m, IMem mem, Jint level1Log, Jint level2Size
		);
	extern Jerror	mapDestruct( Map m );
	extern IGeneric	mapGet( Map m, IGeneric key );
	extern Jint		mapGetI( Map m, IGeneric key );
	extern Jfloat	mapGetF( Map m, IGeneric key );
	extern Jerror	mapPut( Map m, IGeneric key, IGeneric v );
	extern Jerror	mapPutI( Map m, IGeneric key, Jint data );
	extern Jerror	mapPutF( Map m, IGeneric key, Jfloat data );
	extern Jerror	mapConstructMap( Map m, IMap it );

/***********************************************************************
 *	MapSet (bitset implementation using Map)
 ***********************************************************************/
	extern Jerror	msetConstruct(
		MapSet s, IMem mem, Jint level1Log, Jint level2Size
		);
	extern Jerror	msetDestruct( MapSet s );
	extern Jerror	msetMakeEmpty( MapSet s );
	extern Jerror	msetIsEmpty( MapSet s );
	extern Jerror	msetMakeEmpty( MapSet s );
	extern Jerror	msetMakeUniv( MapSet s );
	extern Jerror	msetIsMember( MapSet s, Jint m );
	extern Jerror	msetInclude( MapSet s, Jint m );
	extern Jerror	msetExclude( MapSet s, Jint m );
	extern Jerror	msetCopy( MapSet s, MapSet to );
	extern Jerror	msetUnite( MapSet s, MapSet another, MapSet result );
	extern Jerror	msetIntersect(
		MapSet s, MapSet another, MapSet result
		);
	extern Jerror	msetComplement( MapSet s, MapSet result );
	extern Jerror	msetEquals( MapSet s, MapSet another );
	extern Jerror	msetDump( MapSet s );
	extern Jint		msetChoose( MapSet s );
	extern Jerror	msetConstructSet( MapSet s, ISet it );

	
/***********************************************************************
 *	Graph
 ***********************************************************************/
#define grGetNodes( g )							((g)->fNodes)
#define grGetEdges( g )							((g)->fEdges)
#define grGetNode( g, id )						\
    ((GraphNode)mapGet(grGetNodes(g),(IGeneric)(id)))
#define grGetEdge( g, id )						\
    ((GraphEdge)mapGet(grGetEdges(g),(IGeneric)(id)))
#define grGetMaxNodeID( g )						((g)->fMaxNodeID)
#define grGetNumOfEdges( g )					((g)->fEdgeNum)
#define grGetNumOfNodes( g )					((g)->fNodeNum)
#define grGetMaxID( g )							((g)->fMaxID)
#define grSetHideMonitor( g, m )				((g)->fHideMonitor = (m))
#define grSetShowMonitor( g, m )				((g)->fShowMonitor = (m))
#define grIsEmpty( g )							(grGetNumOfNodes(g) == 0)
#define grGetNodeParent( n )			   		((n)->fParent)
#define grGetNodeWeight( n )			   		((n)->fWeight)
#define grSetNodeWeight( n, v )					((n)->fWeight = (v))
#define grGetNodeID( n )						((n)->fID)
#define grSetNodeID( n, v )						((n)->fID = (v))
#define grGetNodeIncoming( n )					((n)->fIncoming)
#define grGetNodeOutgoing( n )					((n)->fOutgoing)
#define grGetNodeCluster( n )					((n)->fCluster)
#define grGetNodeIncomingDegree( n )			((n)->fIncomingDegree)
#define grGetNodeOutgoingDegree( n )			((n)->fOutgoingDegree)
#define grGetNodeIncomingWeightedDegree( n )	((n)->fIncomingWDegree)
#define grGetNodeOutgoingWeightedDegree( n )	((n)->fOutgoingWDegree)
#define grGetNodeDegree( n )					\
    ((n)->fIncomingDegree + (n)->fOutgoingDegree)
#define grGetNodeWeightedDegree( n )			\
    ((n)->fIncomingWDegree + (n)->fOutgoingWDegree)
#define grGetDegreeWeight( n )					((n)->fDegreeWeight)
#define grSetDegreeWeight( n, v )				((n)->fDegreeWeight = (v))
#define grIsNodeHidden( n )						(((n)->fState & 0x01) != 0)
#define	grIsNodeClustered( n )					((n)->fState == 0x11)
#define grGetEdgeID( e )						((e)->fID)
#define grGetEdgeSrc( e )						((e)->fSrc)
#define grGetEdgeSnk( e )						((e)->fSink)
#define grGetEdgeIncomingLink( e )				((e)->fLinkIncoming)
#define grGetEdgeOutgoingLink( e )				((e)->fLinkOutgoing)
#define grIsEdgeHidden( e )						((e)->fIsHidden)
#define grGetEdgeWeight( e )					((e)->fWeight)
#define grSetEdgeWeight( e, w )					((e)->fWeight = (w))
#define grIsNodePred( n, p )					\
    (grFindEdge((n)->fGraph, (p), (n)) != NULL )
#define grIsNodeSucc( n, s )					\
    (grFindEdge((n)->fGraph, (n), (s)) != NULL )
#define grIsNodeAdj( n, adj )					\
    (grIsNodePred( n, adj ) || grIsNodeSucc( n, adj ) )

#define grNodeMapPut( m, n, v )		mapPut( (m), (IGeneric)grGetNodeID(n), v )
#define grNodeMapPutI( m, n, v )	mapPutI( (m), (IGeneric)grGetNodeID(n), v )
#define grNodeMapPutF( m, n, v )	mapPutF( (m), (IGeneric)grGetNodeID(n), v )
#define grNodeMapGet( m, n )		mapGet( (m), (IGeneric)grGetNodeID(n) )
#define grNodeMapGetI( m, n )		mapGetI( (m), (IGeneric)grGetNodeID(n) )
#define grNodeMapGetF( m, n )		mapGetF( (m), (IGeneric)grGetNodeID(n) )
	
	
	extern Jerror		grConstruct( Graph g, IMem mem, Jint nodeSize );
	extern Jerror		grConstructDimacs( Graph g, IMem mem, char* name );
	extern Jerror		grConstructLarge(
		Graph g, IMem mem,
		Jint nodeLevel1Log, Jint nodeLevel2Log,
		Jint edgeLevel1Log, Jint edgeLevel2Log
		);
	extern Jerror		grDestruct( Graph g );
	extern GraphNode	grNewNode( Graph g, Jint id, Jint weight );
	extern GraphEdge	grNewEdge(
		Graph g, GraphNode src, GraphNode snk, Jint weight
		);
	extern GraphEdge	grFindEdge(
		Graph g, GraphNode src, GraphNode snk
		);
	extern Jerror		grIsAcyclicDG( Graph g );
	extern Jerror		grHideNode( GraphNode n );
	extern Jerror		grShowNode( GraphNode n );
	extern Jerror		grHideEdge( Graph f, GraphEdge e );
	extern Jerror		grShowEdge( Graph f, GraphEdge e );
	extern Jerror		grShowAllEdges( Graph f );
	extern Jerror		grAddEdge( Graph f, GraphEdge e );
	extern Jerror		grRemoveEdge( Graph f, GraphEdge e );
	extern Jerror		grOrientEdge(
		GraphEdge e, GraphNode src, GraphNode snk
		);
	extern Jerror		grUnionDG( GraphNode n, GraphNode into );
	extern Jerror		grUnionUG( GraphNode n, GraphNode into );
	extern GraphNode	grFindNode( GraphNode n );
	extern Jerror		grDumpDot( Graph g, char* name );
	extern Jerror		grDumpDotWithWeight( Graph g, char* name );
	extern Jerror		grDump( Graph g );

	extern Jerror		grConstructNodeIter( Graph g, IIterator it );
	extern Jerror		grConstructEdgeIter( Graph g, IIterator it );
	extern Jerror		grConstructIncomingIter( GraphNode n, IIterator it );
	extern Jerror		grConstructOutgoingIter( GraphNode n, IIterator it );
	extern Jerror		grConstructAdjIter( GraphNode n, IIterator it );
	extern Jerror		grConstructPredIter( GraphNode n, IIterator it );
	extern Jerror		grConstructSuccIter( GraphNode n, IIterator it );
	extern Jerror		grConstructIncidentIter( GraphNode n, IIterator it );
	extern Jerror		grConstructRandomIter( GraphNode n, IIterator it );
	extern Jerror		grConstructNodeMap( Map map, Graph g );

/***********************************************************************
 *	Binary Heap
 ***********************************************************************/
#define bheapData( type, cell )	((type)(cell)->val)
#define bheapKey( cell )		(((Jint*)(cell)->key)[0])
#define bheapGetSize( h )		((h)->fNumElts)
#define bheapIsEmpty( h )		((h)->fNumElts == 0)
	

	extern Jerror 		bheapConstruct( BinHeap h, IMem m, Jint a, Jint b );
	extern Jerror 		bheapDestruct( BinHeap h );
	extern Jerror 		bheapInsertElement( BinHeap h, HeapCell it );
	extern Jerror 		bheapRemoveAll( BinHeap h );
	extern Jerror 		bheapRemoveElement( BinHeap h, HeapCell it );
	extern HeapCell 	bheapNewCell( BinHeap h, Jint key, IGeneric data );
	extern HeapCell		bheapFindMin( BinHeap h );
	extern HeapCell		bheapFindMax( BinHeap h );
	extern HeapCell		bheapDelMin( BinHeap h );
	extern HeapCell		bheapDelMax( BinHeap h );
	extern HeapCell		bheapDecKey( BinHeap h, HeapCell it, Jint k );
	extern HeapCell		bheapIncKey( BinHeap h, HeapCell it, Jint k );
	extern Jerror		bheapDelAll( BinHeap h );
	
/***********************************************************************
 *	Heap
 ***********************************************************************/
#define heapData( type, cell )	((type)(cell)->val)
#define heapKey( cell )			(((Jint*)(cell)->key)[0])
	
#define heapGetTree(h)			((h)->tree)
#define heapIsEmpty(h)			(vecGetSize(heapGetTree(h)) == 1)
#define heapMem(h)				(heapGetTree(h)->mem)

#define heapNewCell(h,k,d)		(heapNewCell1((h),(IGeneric)(k), (d)))	
#define heapInsert(h,k1,d) {								\
    HeapCell cell = heapNewCell1((h),(IGeneric)(k),(d));	\
    if(cell) heapInsertElement( h, cell );					\
    }
					   
#define heapInsert2(h,k1,k2,d) {							\
    HeapCell cell = heapNewCell2((h),(IGeneric)(k1),(IGeneric)(k2),(d));	\
    if(cell) heapInsertElement( h, cell );					\
    }

#define heapInsert3(h,k1,k2,k3,d) {							\
    HeapCell cell = heapNewCell3((h),						\
        (IGeneric)(k1),(IGeneric)(k2),(IGeneric)(k3),(d));	\
    if(cell) heapInsertElement( h, cell );					\
    }
								 
	extern Jerror		heapConstruct( Heap h, IMem mem, Jint size );
	extern Jerror		heapConstruct0(
		Heap h, IMem mem, Jint size,
		Jint keysize, IGeneric weights, IOrderScheme scheme
		);
	extern Jerror		heapInsertElement( Heap h, HeapCell it );
	extern Jerror		heapRemoveElement( Heap h, HeapCell it );
	extern HeapCell		heapFindMin( Heap h );
	extern HeapCell		heapDelMin( Heap h );
	extern HeapCell		heapDecKey( Heap h, HeapCell it, Jint k );
	extern Jerror		heapDelAll( Heap h );
	
	extern HeapCell		heapNewCell1(
		Heap h, IGeneric key, IGeneric data
		);
	extern HeapCell		heapNewCell2(
		Heap h, IGeneric key1, IGeneric key2, IGeneric data
		);
	extern HeapCell		heapNewCell3(
		Heap h, IGeneric key1, IGeneric key2, IGeneric key3, IGeneric data
		);
	extern HeapCell		heapNewCellN(
		Heap h, IGeneric* keys, IGeneric data
		);
	
	extern IOrderScheme	orderIntScheme, orderFloatScheme;
	extern IOrderScheme orderIntDictScheme, orderFloatDictScheme;
	extern IOrderScheme orderIntWeightedScheme, orderFloatWeightedScheme;
	extern IOrderScheme	orderPtrScheme, orderStringScheme;
	
/***********************************************************************
 *	Postscript graphics driver
 ***********************************************************************/
	extern Jerror	psBegin( char* file, Jint x1, Jint y1, Jint x2, Jint y2 );
	extern Jerror	psEnd( void );
	extern Jerror	psDrawLine( Jint x1, Jint y1, Jint x2, Jint y2 );
	extern Jerror	psDrawArrow( Jint x1, Jint y1, Jint x2, Jint y2 );
	extern Jerror	psDrawRect(
		Jint x1, Jint y1, Jint x2, Jint y2, Jboolean fill
		);
	extern Jerror	psDrawArc(
		Jint xc, Jint yc, Jint rad,
		Jint angstart, Jint angextent, Jboolean fill
		);
	extern Jerror	psDrawText( Jint x, Jint y, char *text );
	extern Jerror	psSetColor( char* color );
	extern Jerror	psSetLineStyle( char* style );
	extern Jerror	psSetLineWidth( Jint linewidth );
	extern Jerror	psSetFontSize( Jint pointsize );

	
/***********************************************************************
 *	Simple spreadsheet
 ***********************************************************************/
#define ssEnsure(m,row)				\
		((m)->maxrow = row > (m)->maxrow ? row : (m)->maxrow)	
#define ssGetI(m,row,col)			((m)->data[col*(m)->nrows+row].i)	
#define ssSetI(m,row,col,v)			\
        (ssEnsure(m,row), (m)->data[col*(m)->nrows+row].i=(v))
#define ssGetS(m,row,col)			((m)->data[col*(m)->nrows+row].s)	
#define ssSetS(m,row,col,v)			\
        (ssEnsure(m,row), (m)->data[col*(m)->nrows+row].s=(v))
#define ssGetF(m,row,col)			((m)->data[col*(m)->nrows+row].f)	
#define ssSetF(m,row,col,v)			\
        (ssEnsure(m,row), (m)->data[col*(m)->nrows+row].f=(v))
#define ssGetColI(m,col)			((Jint*)&(m)->data[col*(m)->nrows])	
#define ssGetColS(m,col)			((char**)&(m)->data[col*(m)->nrows])	
#define ssGetColF(m,col)			((Jfloat*)&(m)->data[col*(m)->nrows])	

	extern Jerror	ssConstruct(
		Spreadsheet h, IMem mem, Jint nrows, char** descriptor
		);
	extern Jerror	ssDestruct( Spreadsheet h );
	extern Jerror	ssSetColByIndex(
		Spreadsheet h, Jint col, Jint base, Jint step
		);
	extern Jerror	ssSetColByIndexLog(
		Spreadsheet h, Jint col, Jint base, Jint step
		);
	extern Jerror	ssSum( Spreadsheet h, Jint row, Jint col );
	extern Jerror	ssAverage( Spreadsheet h, Jint row, Jint col );
	extern Jerror	ssBeginTime( Spreadsheet h );
	extern Jerror	ssEndTime( Spreadsheet h, Jint row, Jint col );
	
	extern Jerror	ssDumpCSV(
		Spreadsheet h, char* file, Jboolean append
		);
	extern Jerror	ssDumpGrace( Spreadsheet h, char* file, Jboolean append );
	extern Jerror	ssDumpLatex( Spreadsheet h, char* file, Jboolean append );


/***********************************************************************
 *	random number generator
 ***********************************************************************/

	extern Jerror	rpConstruct( RandomProcess rp );
	extern Jdouble	rpGenGaussian(
		RandomProcess rp, Jdouble mean, Jdouble dev
		);
	extern Jint		rpGenUniformInt( RandomProcess rp, Jint nbits );
	
#ifdef __cplusplus
}
#endif

#endif

