/********************************************************************
 *		File			:		order.c
 *		Description		:		order scheme
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "metautil.h"

struct guid comIOrderSchemeIID = COM_IORDERSCHEME_IID;

static IOrderScheme 	_scheme;
static Jint				_keysize;
static IGeneric			_weights;

static Jint
_compare( const void* p1, const void* p2 ) {
	KeyCell	c1 = (KeyCell)p1;
	KeyCell	c2 = (KeyCell)p2;
	
	return COM_ICALL( compare, _scheme )(
		_scheme, _keysize, _weights, c1->key, c2->key
		);
	}
	
extern Jerror
vecSort( Vector v, Jint keysize, IGeneric weights, IOrderScheme scheme ) {
	_scheme = scheme;
	_keysize = keysize;
	_weights = weights;
	qsort(
		(void *)v->data, vecGetSize(v), sizeof(KeyCell), _compare
		);
	return 0;
	}

extern Jerror
listSort( List l, Jint keysize, IGeneric weights, IOrderScheme scheme ) {
	return -1;
	}

static COMDECL
_orderQuery(
	IOrderScheme si, const ComIID iid, IGeneric *i
	) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, OrderScheme ) ) {
		*i = si;
		return 0;
		}
	*i = NULL;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
_orderAddref( IOrderScheme si ) {
	return 1;
	}

static COMDECLU
_orderRelease( IOrderScheme si ) {
	return 1;
	}

static Jint COMCALL
_orderIntCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	Jint	nkey1 = *(Jint*)key1;
	Jint	nkey2 = *(Jint*)key2;

	if( nkey1 > nkey2 )
		return 1;
	if( nkey1 == nkey2 )
		return 0;
	return -1;
	}

static Jint COMCALL
_orderFloatCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	Jfloat	fkey1 = *(Jfloat*)key1;
	Jfloat	fkey2 = *(Jfloat*)key2;

	if( fkey1 > fkey2 )
		return 1;
	if( fkey1 == fkey2 )
		return 0;
	return -1;
	}

static Jint COMCALL
_orderPtrCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	Jint	nkey1 = *(Jint*)key1;
	Jint	nkey2 = *(Jint*)key2;

	nkey1 >>= 2;
	nkey2 >>= 2;

	if( nkey1 > nkey2 )
		return 1;
	if( nkey1 == nkey2 )
		return 0;
	return -1;
	}

static Jint COMCALL
_orderStringCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	char*	skey1 = *(char**)key1;
	char*	skey2 = *(char**)key2;

	return strcmp( skey1, skey2 );
	}

static Jint COMCALL
_orderIntDictCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	Jint*	pkey1 = (Jint*)key1;
	Jint*	pkey2 = (Jint*)key2;
	Jint	i;
	
	for( i = 0; i < keysize; i ++ ) {
		if( pkey1[i] > pkey2[i] )
			return 1;
		else if( pkey1[i] < pkey2[i] )
			return -1;
		}
	return 0;
	}

static Jint COMCALL
_orderFloatDictCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	Jfloat*	pkey1 = (Jfloat*)key1;
	Jfloat*	pkey2 = (Jfloat*)key2;
	Jint	i;

	for( i = 0; i < keysize; i ++ ) {
		if( pkey1[i] > pkey2[i] )
			return 1;
		else if( pkey1[i] < pkey2[i] )
			return -1;
		}
	return 0;
	}

static Jint COMCALL
_orderIntWeightedCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	Jint*	pkey1 = (Jint*)key1;
	Jint*	pkey2 = (Jint*)key2;
	Jint*	pweights = (Jint*)weights;
	Jint	i;
	Jint	sum1 = 0, sum2 = 0;
	
	for( i = 0; i < keysize; i ++ ) {
		sum1 += pweights[i] * pkey1[i];
		sum2 += pweights[i] * pkey2[i];
		}
		
	if( sum1 > sum2 )
		return 1;
	else if( sum1 < sum2 )
		return -1;
	return 0;
	}


static Jint COMCALL
_orderFloatWeightedCompare(
	IOrderScheme si, Jint keysize, IGeneric weights,
	IGeneric key1, IGeneric key2
	) {
	Jfloat*	pkey1 = (Jfloat*)key1;
	Jfloat*	pkey2 = (Jfloat*)key2;
	Jfloat*	pweights = (Jfloat*)weights;
	Jint	i;
	Jfloat	sum1 = 0, sum2 = 0;
	
	for( i = 0; i < keysize; i ++ ) {
		sum1 += pweights[i] * pkey1[i];
		sum2 += pweights[i] * pkey2[i];
		}
		
	if( sum1 > sum2 )
		return 1;
	else if( sum1 < sum2 )
		return -1;
	return 0;
	}


static struct  _DOrderScheme	_orderIntSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderIntCompare
    };

static struct  _DOrderScheme	_orderFloatSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderFloatCompare
    };

static struct  _DOrderScheme	_orderPtrSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderPtrCompare
    };

static struct  _DOrderScheme	_orderStringSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderStringCompare
    };

static struct  _DOrderScheme	_orderIntDictSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderIntDictCompare
    };

static struct  _DOrderScheme	_orderFloatDictSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderFloatDictCompare
    };

static struct  _DOrderScheme	_orderIntWeightedSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderIntWeightedCompare
    };

static struct  _DOrderScheme	_orderFloatWeightedSchemeVtbl = {
	_orderQuery, _orderAddref, _orderRelease,
	_orderFloatWeightedCompare
    };

static struct _IOrderScheme	_orderIntScheme = { &_orderIntSchemeVtbl };
IOrderScheme			orderIntScheme = &_orderIntScheme;

static struct _IOrderScheme	_orderFloatScheme = { &_orderFloatSchemeVtbl };
IOrderScheme			orderFloatScheme = &_orderFloatScheme;

static struct _IOrderScheme	_orderPtrScheme = { &_orderPtrSchemeVtbl };
IOrderScheme			orderPtrScheme = &_orderPtrScheme;

static struct _IOrderScheme	_orderStringScheme = { &_orderStringSchemeVtbl };
IOrderScheme			orderStringScheme = &_orderStringScheme;

static struct _IOrderScheme	_orderIntDictScheme = { &_orderIntDictSchemeVtbl };
IOrderScheme			orderIntDictScheme = &_orderIntDictScheme;

static struct _IOrderScheme	_orderFloatDictScheme =
    { &_orderFloatDictSchemeVtbl };
IOrderScheme			orderFloatDictScheme = &_orderFloatDictScheme;

static struct _IOrderScheme	_orderIntWeightedScheme = {
	&_orderIntWeightedSchemeVtbl
	};
IOrderScheme			orderIntWeightedScheme = &_orderIntWeightedScheme;

static struct _IOrderScheme	_orderFloatWeightedScheme =
    { &_orderFloatWeightedSchemeVtbl };
IOrderScheme			orderFloatWeightedScheme = &_orderFloatWeightedScheme;

