/********************************************************************
 *		File			:		print.c
 *		Description		:		formatted printing
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <com.h>

static Jint _indent = 0;


static char*
comPrintString( const char* str, FILE* fp, char* bp ) {
	if( fp != NULL )
		fputs( str, fp );
	else
		while( (*bp = *str++) )
			bp++;
	return bp;
	}

static char*
comPrintDecimal( Jlong n, FILE* f, char* bp ) {
	Julong 	m;
	char buf[25], *s = buf + sizeof buf;

	*--s = '\0';
	if( n < 0 )
		m = -n;
	else
		m = n;
	do
		*--s = m%10 + '0';
	while ((m /= 10) != 0);
	if (n < 0)
		*--s = '-';
	return comPrintString( s, f, bp );
	}

static char*
comPrintUnsigned( Julong n, Jint base, FILE* fp, char* bp ) {
	char buf[25], *s = buf + sizeof buf;

	*--s = '\0';
	do
		*--s = "0123456789abcdef"[n%base];
	while( (n /= base) != 0 );
	return comPrintString( s, fp, bp );
	}

extern char*
comFormatedPrint( FILE* f, char* bp, const char* fmt, va_list ap ) {
	for( ; *fmt; fmt++ ) {
		if( *fmt == '%' ) {
			switch (*++fmt) {
			case 'd':
				bp = comPrintDecimal( va_arg(ap, int), f, bp );
				break;
			case 'D':
				bp = comPrintDecimal(va_arg(ap, Jlong), f, bp);
				break;
			case 'U':
				bp = comPrintUnsigned(va_arg(ap, Julong), 10, f, bp);
				break;
			case 'u':
				bp = comPrintUnsigned(va_arg(ap, Juint), 10, f, bp);
				break;
			case 'o':
				bp = comPrintUnsigned(va_arg(ap, Juint), 8, f, bp);
				break;
			case 'X':
				bp = comPrintUnsigned(va_arg(ap, Julong), 16, f, bp);
				break;
			case 'x':
				bp = comPrintUnsigned(va_arg(ap, Juint), 16, f, bp);
				break;
			case 'f': case 'e':
			case 'g': {
				static char format[] = "%f";
				char buf[128];
				format[1] = *fmt;
				sprintf( buf, format, va_arg(ap, Jdouble) );
				bp = comPrintString( buf, f, bp );
				break;
				}
			case 's':
				bp = comPrintString( va_arg(ap, char *), f, bp );
				break;
			case 'p': {
				void *p = va_arg(ap, void *);
				if( p )
					bp = comPrintString( "0x", f, bp );
				bp = comPrintUnsigned( (unsigned long)p, 16, f, bp );
				break;
				}
			case 'c':
				if( f )
					fputc(va_arg(ap, int), f);
				else
					*bp++ = va_arg(ap, int);
				break;
			case 'S': {
				char *s = va_arg(ap, char *);
				int n = va_arg(ap, int);
				if( s ) {
					for( ; n-- > 0; s++ )
						if (f) putc(*s, f); else *bp++ = *s;
					}
				break;
				}	
			case 'm': {
				IComponent m = va_arg( ap, IComponent );
				
				bp = COM_CALL( dump, m )( m, f, bp );
				break;
				}
			case 'i': {
				/* indent */
				int n = _indent * 4;
				while (--n >= 0)
					if( f )
						putc( ' ', f );
					else
						*bp++ = ' ';
				break;				
				} 
			case 'I': {
				/* indent with header */
				int id = va_arg(ap, int);
				int n = _indent * 4;
				char idstr[80];

				sprintf( idstr, "(%d)", id );
				bp = comPrintString( idstr, f, bp );
				n -= strlen( idstr );
				while (--n >= 0)
					if( f )
						putc( ' ', f );
					else
						*bp++ = ' ';
				break;				
				} 
			default:
				if( f )
					putc( *fmt, f );
				else
					*bp++ = *fmt;
				break;
				}
			}
		else if( f )
			putc(*fmt, f);
		else
			*bp++ = *fmt;
		}
	if( !f )
		*bp = '\0';
	return bp;
	}

extern char*
comPrintf( FILE* fp, char* bp, const char* fmt, ... ) {
	va_list ap;

	va_start( ap, fmt );
	bp = comFormatedPrint( fp, bp, fmt, ap );
	va_end(ap);
	return bp;
	}

extern void
comIndent( void ) {
	_indent ++;
	}

extern void
comUnindent( void ) {
	_indent --;
	}
