/********************************************************************
 *		File			:		com.c
 *		Description		:		Microsoft COM registeration
 *		Author			:		Jianwen Zhu
 *		Reference		:		http://www.microsoft.com/com
 *		                 		oskit
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "metautil.h"

COM_DECLARE_CLASS( ObjNode );
COM_DECLARE_CLASS( IidNode );
COM_DECLARE_CLASS( CService );


struct _DService dservice;

struct _ObjNode {
	IUnKnown	intf;
	ObjNode		next;
	};

struct _IidNode {
	struct guid		iid;
	ObjNode			objs;
	int				count;
	IidNode			next;
	};

struct _CService {
	struct _IService	intf;
	IidNode				iids;
	int					count;		/* reference count */
	IMem				mem;		/* memory manager */
	};

static COMDECL
svQuery( IService s, const ComIID iid, IGeneric *handle ) {
	CService		db = (CService)s;

	if( COM_SAME_IIDNAME( iid, UnKnown ) ||
		COM_SAME_IIDNAME( iid, Component ) || 
		COM_SAME_IIDNAME( iid, Service ) ) {
		*handle = &db->intf;
		db->count ++;
		return 0;
		}
	*handle = NULL;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
svAddref( IService si ) {
	CService s = (CService) si;

	assert(s->count);
	return ++s->count;
	}

static COMDECLU
svRelease( IService si ) {
	CService s = (CService) si;

	assert( s->count );
	if( --s->count == 0 ) {
		IMem	mem = s->mem;
		IidNode		in, nin;
		ObjNode		on, non;

        /*
		 * Must free up all the registered interface objects.
		 */
		in = s->iids;
		while (in) {
			on = in->objs;
			while (on) {
				comRelease( on->intf );
				non = on->next;
				comFree( mem, (void *)on );
				on  = non;
				}
			nin = in->next;
			comFree( mem, (void *)in );
			in  = nin;
			}

		comFree( mem, (void *)s );
		comRelease( mem );
		return 0;
        }
	return s->count;
	}

char* COMCALL
svDump( IService si, FILE* fp, char* bp ) {
	return bp;
	}

static COMDECLP
svCreate( IService si, IMem mem ) {
	return newService( mem );
	}

static COMDECL
svConstructWithMem( IService si, IMem mem ) {
	CService 	s = (CService)si;
	
	s->count     = 1;
	s->iids      = 0;
	s->mem		 = comRef( mem );
	s->intf.vtbl = &dservice;
	return 0;
	}

static COMDECL
svConstruct( IService si ) {
	return svConstructWithMem( si, iheap );
	}

static COMDECL
svDestruct( IService si ) {
	return 0;
	}

static COMDECL
svAdd( IService si, const ComIID iid, IGeneric intf ) {
	CService 		s = (CService) si;
	IidNode			in;
	ObjNode			on, *onp;

	for( in = s->iids; ; in = in->next ) {
		if (in == NULL) {
			in = comAlloc( s->mem, sizeof(*in) );
			if( in == NULL )
				return COM_E_OUTOFMEMORY;
			in->iid = *iid;
			in->objs = NULL;
			in->count = 0;
			in->next = s->iids;
			s->iids = in;
			break;
			}
		if( memcmp(&in->iid, iid, sizeof(*iid)) == 0 )
			break;
		}

	for( onp = &in->objs; *onp; onp = &(*onp)->next ) {
		if( (*onp)->intf == intf )
			return 0;
		}

	on = comAlloc( s->mem, sizeof(*on) );
	if( on == NULL )
		return COM_E_OUTOFMEMORY;
	on->next = NULL;
	on->intf = comRef( intf );
	*onp = on;
	in->count++;

	return 0;
	}

static COMDECL
svRemove( IService si, const ComIID iid, IGeneric intf ) {
	CService	   s = (CService)si;
	IidNode			in;
	ObjNode			on, *onp;

	/* Find the appropriate iidnode */
	for( in = s->iids; ; in = in->next ) {
		if( in == NULL )
			return COM_E_INVALIDARG;
		if( COM_SAME_IID( &in->iid, iid ) )
			break;
		}

	/* Find and remove the objnode */
	for( onp = &in->objs; ; onp = &on->next ) {
		on = *onp;
		if( on == NULL )
			return COM_E_INVALIDARG;
		if( on->intf == intf )
			break;
		}
	*onp = on->next;
	comRelease( on->intf );
	comFree( s->mem, (void *) on );
	in->count--;

	return 0;
	}

static COMDECL
svLookup( IService si, const ComIID iid, IGeneric **iarray, IMem mem ) {
	CService		s = (CService)si;
	IidNode			in;
	ObjNode			on;
	IGeneric*		arr;
	int 			i;

	/* Find the appropriate iidnode */
	for( in = s->iids; ; in = in->next ) {
		if (in == NULL) {
			*iarray = NULL;
			return 0;
		}
		if( COM_SAME_IID( &in->iid, iid ) )
			break;
		}
	if( in->count == 0 ) {
		*iarray = NULL;
		return 0;
		}

	/*
	 * Note that we use memory manager provided by the client 
	 * for freeing up the array. 
	 */
	arr = comAlloc( mem, sizeof(*arr)*in->count );
	if( arr == NULL )
		return COM_E_OUTOFMEMORY;

	for( i = 0, on = in->objs; i < in->count; i++, on = on->next ) {
		assert(on != NULL);
		arr[i] = comRef( on->intf );
		}
	assert(on == NULL);

	*iarray = arr;
	return in->count;
	}

static COMDECL
svLookupFirst( IService si, const ComIID iid, IGeneric *intf ) {
	CService s = (CService) si;
	IidNode	in;
	IUnKnown result;

	/* Find the appropriate iidnode */
	for( in = s->iids; ; in = in->next ) {
		if( in == NULL ) {
			*intf = NULL;
			return 0;
			}
		if( COM_SAME_IID( &in->iid, iid ) )
			break;
		}
	if( in->count == 0 ) {
		*intf = NULL;
		return 0;
		}
	result = in->objs->intf;
	*intf = comRef( result );
	return 0;
	}

static COMDECL
svClone(  IService si, IService *intf ) {
	CService 		s = (CService) si;
	CService		ns;
	IidNode			in;
	ObjNode			on;
	Jerror			rc;

	ns = comAlloc( s->mem, sizeof(*ns) );
	if( ns == NULL )
		return COM_E_OUTOFMEMORY;

	ns->count     = 1;
	ns->mem       = comRef( s->mem );
	ns->intf.vtbl = &dservice;
	ns->iids      = 0;

	in = s->iids;
	while (in) {
		on = in->objs;
		while (on) {
			if ((rc = COM_CALL( add, &ns->intf )( &ns->intf, 
						      &in->iid, on->intf) )
			    != 0) {
				COM_CALL( fatalError, ierror )(
					ierror, "service::clone"
					);
				}
			on = on->next;
			}
		in = in->next;
		}

	*intf = &ns->intf;
	return 0;
	}

struct _DService dservice = {
	svQuery,
	svAddref,
	svRelease,
	svDump,
	svCreate,
	svDestruct,
	svConstruct,
	svConstructWithMem,
	svAdd,
	svRemove,
	svLookup,
	svLookupFirst,
	svClone
    };

extern IService
comNewService( IMem mem ) {
	CService 	s;
	Jerror		rc;
	
	if( !mem ) {
		mem = comHeap();

		if( !mem )
			COM_CALL( fatalError, ierror )(
				ierror, "constructService: NULL memory object!"
				);
		}

	s = comAlloc( mem, sizeof(*s) );
	if( s == NULL )
		return NULL;
	rc = svConstructWithMem( &s->intf, mem );
	if( COM_SUCCEEDED(rc) )
		return &s->intf;
	else {
		comFree( mem, s );
		return NULL;
		}
	}
