/********************************************************************
 *		File			:		set.c
 *		Description		:		set implementation
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "metautil.h"

extern Jerror
bsetConstruct( BitSet s, IMem mem, Jint univSize ) {
	s->fSize = univSize;
	s->fData = comAlloc( mem, ((univSize + 31) >> 5) * sizeof(Jint) );
	s->fMem = mem;
	bsetMakeEmpty( s );

	return 0;
	}

extern Jerror
bsetDestruct( BitSet s ) {
	comFree( s->fMem, s->fData );
	return 0;
	}

extern Jerror
bsetCopy( BitSet s, BitSet another ) {
	Jint		i;
	Jint		size;

	assert( another->fSize == s->fSize );
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		another->fData[i] = s->fData[i];
	return 0;
	}

extern Jerror
bsetUnite( BitSet s, BitSet another, BitSet result ) {
	Jint	i;
	Jint	size;
	
	assert( s->fSize == result->fSize );
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		result->fData[i] = s->fData[i] | another->fData[i];
	return 0;
	}

extern Jerror
bsetIntersect( BitSet s, BitSet another, BitSet result ) {
	Jint	i;
	Jint	size;
	
	assert( s->fSize == another->fSize );
	assert( s->fSize == result->fSize );
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		result->fData[i] = s->fData[i] & another->fData[i];
	return 0;
	}

extern Jerror
bsetComplement( BitSet s, BitSet result ) {
	Jint	i;
	Jint	size;
	
	assert( s->fSize == result->fSize );
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		result->fData[i] = ~s->fData[i];
	return 0;
	}

extern Jerror
bsetIsMember( BitSet s, Jint m ) {
	assert( m >= 0 && m < s->fSize );
	return ( s->fData[ m >> 5 ] & (0x01 << ( m & 31 )) )  != 0;
	}

extern Jerror
bsetInclude( BitSet s, Jint m ) {
	assert( m >= 0 && m < s->fSize );
	s->fData[ m >> 5 ] |= (0x01 << ( m & 31 ));
	return 0;
	}

extern Jerror
bsetExclude( BitSet s, Jint m ) {
	assert( m >= 0 && m < s->fSize );
	s->fData[ m >> 5 ] &= ~(0x01 << ( m & 31 ));
	return 0;
	}

extern Jerror
bsetIsEmpty( BitSet s ) {
	Jint	i;
	Jint	size;
	
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		if( s->fData[i] != 0 )
			return FALSE;
	return TRUE;
	}

extern Jint 
bsetSize( BitSet s ) {
	int 	i;
	int 	memberCount = 0;
	
	bsetForeach( s, i ){
		memberCount++;
		}  
	return memberCount;
	}


extern Jerror
bsetEquals( BitSet s, BitSet another ) {
	Jint	i;
	Jint	size;
	
	assert( s->fSize == another->fSize );
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		if( s->fData[i] != another->fData[i] )
			return FALSE;
	return TRUE;
	}

extern Jerror
bsetMakeEmpty( BitSet s ) {
	Jint	i;
	Jint	size;
	
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		s->fData[i] = 0;
	return 0;
	}

extern Jerror
bsetMakeUniv( BitSet s ) {
	Jint	i;
	Jint	size;
	
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		s->fData[i] = 0xffffffff;
	s->fData[size-1] &= 0xffffffff >> (32 - (s->fSize & 31));
	return 0;
	}

extern Jerror
bsetDump( BitSet s ) {
	Jint	i;
	
	for( i = 0; i < s->fSize; i ++ ) {
		if( bsetIsMember( s, i ) ) {
			fprintf( stderr, " %d", i );
			}
		}
	fprintf( stderr, "\n" );
	return 0;
	}

extern Jint
bsetChoose( BitSet s ) {
	Jint		i, j;
	Jint		size;
	Juint		d;
	
	size = (s->fSize + 31) >> 5;
	for( i = 0; i < size; i ++ )
		if( s->fData[i] != 0 )
			break;
	if( i < size ) {
		d = s->fData[i];
			
		for( j = 0; j < 32; j ++ ) {
			if( (d & 0x01) != 0 ) break;
			d >>= 1;
			}
		j += (i << 5);
		if( j < s->fSize )
			return j;
		}
	return -1;
	}


extern Jerror
spsetConstruct( SparseSet s, IMem mem, Jint univSize ) {
	Jint	i;
	
	assert( univSize > 0 );
	s->fUnivSize = univSize;
	s->fMemCount = 0;
	s->fMem = mem;
	s->fArray = comAlloc( mem, sizeof(Jint)*univSize );
	s->fStack = comAlloc( mem, sizeof(Jint)*(univSize+1) );

	for( i = 0; i < univSize; i ++ )
		s->fArray[i] = -1;
	return 0;
	}

extern Jerror
spsetDestruct( SparseSet s ) {
	comFree( s->fMem, s->fArray );
	comFree( s->fMem, s->fStack);
	return 0;
	}

extern Jerror
spsetMakeEmpty( SparseSet s ) {
	Jint		i;
	
	s->fMemCount = 0;
	for( i = 0; i < s->fUnivSize; i ++ )
		s->fArray[i] = -1;
	return 0;
	}

extern Jerror
spsetMakeUniv( SparseSet s ) {
	Jint		i;
	
	for( i = 0; i < s->fUnivSize; i ++ )
		spsetInclude( s, i );
	return 0;
	}

extern Jerror
spsetInclude( SparseSet s, Jint a ) {
	assert( !spsetIsMember( s, a ) );
	s->fArray[a] = s->fMemCount;
	s->fStack[s->fMemCount] = a;
	s->fMemCount ++;
	return 0;
	}

extern Jerror
spsetExclude( SparseSet s, Jint a ) {
	assert( spsetIsMember( s, a ) );
	s->fStack[ s->fArray[a] ] = s->fStack[ s->fMemCount-1 ];
	s->fArray[ s->fStack[ s->fMemCount-1 ] ] = s->fArray[a];
	s->fArray[a] = -1;
	s->fMemCount --;
	return 0;
	}

extern Jerror
spsetUnite( SparseSet s, SparseSet another, SparseSet result ) {
	int		cur;

	spsetMakeEmpty( result );
	for( cur = 0; cur < s->fMemCount; cur ++ )
		spsetInclude( result, s->fStack[cur] );
	for( cur = 0; cur < another->fMemCount; cur ++ )
		if( !spsetIsMember( s, another->fStack[cur] ) )
			spsetInclude( result, another->fStack[cur] );
	return 0;
	}

extern Jerror
spsetIntersect(
	SparseSet s, SparseSet another, SparseSet result
	) {
	int		cur;
	
	spsetMakeEmpty( result );
	for( cur = 0; cur < s->fMemCount; cur ++ )
		if( spsetIsMember( another, s->fStack[cur] ) )
			spsetInclude( result, s->fStack[cur] );
	return 0;
	}

extern Jerror
spsetComplement( SparseSet s, SparseSet result ) {
	assert( 0 );
	return 0;
	}

extern Jerror
spsetDiff( SparseSet s, SparseSet another, SparseSet result ) {
	int		cur;

	spsetMakeEmpty( result );
	for( cur = 0; cur < s->fMemCount; cur ++ )
		if( !spsetIsMember( another, s->fStack[cur] ) )
			spsetInclude( result, s->fStack[cur] );
	return 0;
	}

extern Jerror
spsetCopy( SparseSet s, SparseSet another ) {
	Jint	cur;
	
	spsetMakeEmpty( another );
	for( cur = 0; cur < s->fMemCount; cur ++ )
		spsetInclude( another, s->fStack[cur] );
	return 0;
	}

extern Jerror
spsetEquals( SparseSet s, SparseSet another ) {
	int		cur;

	if( s->fMemCount != another->fMemCount )
		return FALSE;
	for( cur = 0; cur < s->fMemCount; cur ++ )
		if( !spsetIsMember( another, s->fStack[cur] ) )
			return FALSE;
	return TRUE;
	}

extern Jerror
spsetDump( SparseSet s ) {
	Jint	i;
	
	for( i = 0; i < s->fMemCount; i ++ ) {
		fprintf( stderr, " %d", s->fStack[i] );
		}
	fprintf( stderr, "\n" );
	return 0;
	}

extern Jint
spsetChoose( SparseSet s ) {
	if( s->fMemCount == 0 ) return -1;
	return s->fStack[0];
	}


/*******************************************************/
static COMDECL
query( ISet mi, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, Set ) ) {
		*i = mi;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
addref( ISet ci ) {
	return 1;
	}

static COMDECLU
release( ISet ci ) {
	return 1;
	}

#define DECL_VTBL( iter )									\
static struct _DSet	iter##Vtbl = {							\
	query, addref, release,									\
	iter##Reset, iter##HasMoreElements, iter##NextElement,	\
	iter##MakeEmpty,							\
    iter##MakeUniv,								\
	iter##Copy,									\
	iter##Unite,								\
	iter##Intersect,							\
	iter##Complement,							\
	iter##Equals,								\
	iter##IsEmpty,								\
	iter##IsMember,								\
	iter##IncludeMember,						\
	iter##ExcludeMember							\
    }

typedef struct {
	struct _ISet	intf;
	BitSet			set;
	Jint			cur;
	} * BitSetIter;

typedef struct {
	struct _ISet	intf;
	SparseSet		set;
	Jint			cur;
	} * SparseSetIter;

typedef struct {
	struct _ISet	intf;
	Map				set;
	Jint			level1, level2;
	} * MapSetIter;

#define GET_BSET(it)		(((BitSetIter)it)->set)
#define GET_SPSET(it)		(((SparseSetIter)it)->set)
#define GET_CUR(it)			(((BitSetIter)it)->cur)
#define GET_MSET(it)		(((MapSetIter)it)->set)
#define GET_LEVEL1(it)		(((MapSetIter)it)->level1)
#define GET_LEVEL2(it)		(((MapSetIter)it)->level2)

static COMDECL
_bsetReset( ISet iti ) {
	GET_CUR(iti) = 0;
	while( GET_CUR(iti) < GET_BSET(iti)->fSize &&
		   !bsetIsMember( GET_BSET(iti), GET_CUR(iti) ) )
		GET_CUR(iti) ++;
	return 0;
	}

static Jboolean COMCALL
_bsetHasMoreElements( ISet iti ) {
	return GET_CUR(iti) < GET_BSET( iti )->fSize;
	}

static IGeneric COMCALL
_bsetNextElement( ISet iti ) {
	Jint		rtn = GET_CUR( iti );
	
	while( ++GET_CUR(iti) < GET_BSET(iti)->fSize &&
		   !bsetIsMember( GET_BSET(iti), GET_CUR(iti) ) );
	return *(IGeneric *)&rtn;
	}

static COMDECL
_bsetMakeEmpty( ISet si ) {
	return bsetMakeEmpty( GET_BSET(si) );
	}

static COMDECL
_bsetMakeUniv( ISet si ) {
	return bsetMakeUniv( GET_BSET(si) );
	}

static COMDECL
_bsetCopy( ISet si, ISet result ) {	
	return bsetCopy( GET_BSET(si), GET_BSET(result) );
	}

static COMDECL
_bsetUnite( ISet si, ISet another, ISet result ) {
	return bsetUnite( GET_BSET(si), GET_BSET(another), GET_BSET(result) );
	}

static COMDECL
_bsetIntersect( ISet si, ISet another, ISet result ) {
	return bsetIntersect( GET_BSET(si), GET_BSET(another), GET_BSET(result) );
	}

static COMDECL
_bsetComplement( ISet si, ISet result ) {
	return bsetComplement( GET_BSET(si), GET_BSET(result) );
	}

static COMDECL
_bsetEquals( ISet si, ISet another ) {
	return bsetEquals( GET_BSET(si), GET_BSET(another) );
    }

static COMDECL
_bsetIsEmpty( ISet si ) {
	return bsetIsEmpty( GET_BSET(si) );
	}

static COMDECL
_bsetIsMember( ISet si, Jint m ) {
	return bsetIsMember( GET_BSET(si), m );
	}

static COMDECL
_bsetIncludeMember( ISet si, Jint m ) {
	return bsetInclude( GET_BSET(si), m );
	}

static COMDECL
_bsetExcludeMember( ISet si, Jint m ) {
	return bsetExclude( GET_BSET(si), m );
	}

DECL_VTBL( _bset );

extern Jerror
bsetConstructSet( BitSet s, ISet mem ) {
	BitSetIter	it = (BitSetIter)mem;
	
	it->set = s;
	it->intf.vtbl = &_bsetVtbl;
	_bsetReset( mem );
	
	return 0;
	}


static COMDECL
_spsetReset( ISet iti ) {
	GET_CUR(iti) = 0;
	return 0;
	}

static Jboolean COMCALL
_spsetHasMoreElements( ISet iti ) {
	return GET_CUR(iti) < GET_SPSET(iti)->fMemCount;
	}

static IGeneric COMCALL
_spsetNextElement( ISet iti ) {
	return *(IGeneric *)&GET_SPSET(iti)->fStack[GET_CUR(iti) ++];
	}

static COMDECL
_spsetMakeEmpty( ISet si ) {
	return spsetMakeEmpty( GET_SPSET(si) );
	}

static COMDECL
_spsetMakeUniv( ISet si ) {
	return spsetMakeUniv( GET_SPSET(si) );
	}

static COMDECL
_spsetCopy( ISet si, ISet result ) {	
	return spsetCopy( GET_SPSET(si), GET_SPSET(result) );
	}

static COMDECL
_spsetUnite( ISet si, ISet another, ISet result ) {
	return spsetUnite( GET_SPSET(si), GET_SPSET(another), GET_SPSET(result) );
	}

static COMDECL
_spsetIntersect( ISet si, ISet another, ISet result ) {
	return spsetIntersect( GET_SPSET(si), GET_SPSET(another), GET_SPSET(result) );
	}

static COMDECL
_spsetComplement( ISet si, ISet result ) {
	return spsetComplement( GET_SPSET(si), GET_SPSET(result) );
	}

static COMDECL
_spsetEquals( ISet si, ISet another ) {
	return spsetEquals( GET_SPSET(si), GET_SPSET(another) );
    }

static COMDECL
_spsetIsEmpty( ISet si ) {
	return spsetIsEmpty( GET_SPSET(si) );
	}

static COMDECL
_spsetIsMember( ISet si, Jint m ) {
	return spsetIsMember( GET_SPSET(si), m );
	}

static COMDECL
_spsetIncludeMember( ISet si, Jint m ) {
	return spsetInclude( GET_SPSET(si), m );
	}

static COMDECL
_spsetExcludeMember( ISet si, Jint m ) {
	return spsetExclude( GET_SPSET(si), m );
	}

DECL_VTBL( _spset );

extern Jerror
spsetConstructSet( SparseSet s, ISet mem ) {
	SparseSetIter	it = (SparseSetIter)mem;

	it->set = s;
	it->intf.vtbl = &_spsetVtbl;
	_spsetReset( mem );
	
	return 0;
	}


/*******************************************************************
 *	MapSet iterator
 *******************************************************************/
static COMDECL
_msetReset( ISet iti ) {
	Jint	size = 	(GET_MSET(iti)->level2Size) << 5;
	Jint	d;
	Jint*	bucket;
	
	for( GET_LEVEL1(iti) = 0;
		 GET_LEVEL1(iti) < (1 << GET_MSET(iti)->level1Log);
		 GET_LEVEL1(iti) ++
		 ) {
		bucket = (Jint *)GET_MSET(iti)->table[GET_LEVEL1(iti)];
		if( bucket == NULL )
			continue;
		for( GET_LEVEL2(iti) = 0;
			 GET_LEVEL2(iti) < size &&
				 (d = bucket[GET_LEVEL2(iti) >> 5]) &&
				 (d & (0x1 << (GET_LEVEL2(iti) & 31) ));
			 GET_LEVEL2(iti) ++
			 );
		}
	return 0;
	}

static Jboolean COMCALL
_msetHasMoreElements( ISet iti ) {
	return GET_LEVEL1(iti) < (1 << GET_MSET(iti)->level1Log);
	}

static IGeneric COMCALL
_msetNextElement( ISet iti ) {
	Jint	size = 	(GET_MSET(iti)->level2Size) << 5;
	Jint	d;
	Jint	rtn;
	Jint*	bucket;

	rtn = GET_LEVEL2(iti) + 
		+ (GET_LEVEL1( iti ) << GET_MSET(iti)->level2Log);

	for( ;
		 GET_LEVEL1(iti) < (1 << GET_MSET(iti)->level1Log);
		 GET_LEVEL1(iti) ++
		 ) {
		bucket = (Jint *)GET_MSET(iti)->table[GET_LEVEL1(iti)];
		if( bucket == NULL )
			continue;
		for( ;
			 GET_LEVEL2(iti) < size &&
				 (d = bucket[GET_LEVEL2(iti) >> 5]) &&
				 (d & (0x1 << (GET_LEVEL2(iti) & 31) ));
			 GET_LEVEL2(iti) ++
			 );
		}
	return *(IGeneric *)&rtn;
	}

static COMDECL
_msetMakeEmpty( ISet si ) {
	return msetMakeEmpty( GET_MSET(si) );
	}

static COMDECL
_msetMakeUniv( ISet si ) {
	return msetMakeUniv( GET_MSET(si) );
	}

static COMDECL
_msetCopy( ISet si, ISet result ) {	
	return msetCopy( GET_MSET(si), GET_MSET(result) );
	}

static COMDECL
_msetUnite( ISet si, ISet another, ISet result ) {
	return msetUnite( GET_MSET(si), GET_MSET(another), GET_MSET(result) );
	}

static COMDECL
_msetIntersect( ISet si, ISet another, ISet result ) {
	return msetIntersect( GET_MSET(si), GET_MSET(another), GET_MSET(result) );
	}

static COMDECL
_msetComplement( ISet si, ISet result ) {
	return msetComplement( GET_MSET(si), GET_MSET(result) );
	}

static COMDECL
_msetEquals( ISet si, ISet another ) {
	return msetEquals( GET_MSET(si), GET_MSET(another) );
    }

static COMDECL
_msetIsEmpty( ISet si ) {
	return msetIsEmpty( GET_MSET(si) );
	}

static COMDECL
_msetIsMember( ISet si, Jint m ) {
	return msetIsMember( GET_MSET(si), m );
	}

static COMDECL
_msetIncludeMember( ISet si, Jint m ) {
	return msetInclude( GET_MSET(si), m );
	}

static COMDECL
_msetExcludeMember( ISet si, Jint m ) {
	return msetExclude( GET_MSET(si), m );
	}

DECL_VTBL( _mset );

extern Jerror
msetConstructSet( Map s, ISet mem ) {
	MapSetIter	it = (MapSetIter)mem;

	it->set = s;
	it->intf.vtbl = &_msetVtbl;
	_msetReset( mem );
	
	return 0;
	}
