/********************************************************************
 *		File			:		sha.c
 *		Description		:		Message Digest using SHA algorithm
 *		Author			:		Jianwen Zhu
 *		Reference		:		Uwe Hollerbach <uh@alumni.caltech.edu>'s
 *								public domain implementation of SHA
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "metautil.h"

COM_DECLARE_CLASS( Sha );

struct _Sha {
	struct _IMsgDigest	intf;
	Juint		digest[5];				/* result of digest */
	Juint		countLo, countHi;		/* bit count */
	Juint		data[16];				/* data buffer */
	Jint		local;					/* data to be processed */
	};

struct _DMsgDigest dsha;

#define SHA_BLOCKSIZE           64
#define SHA_DIGESTSIZE          20
/* UNRAVEL should be fastest & biggest */
/* UNROLL_LOOPS should be just as big, but slightly slower */
/* both undefined should be smallest and slowest */

#define UNRAVEL
/* #define UNROLL_LOOPS */

/* by default, compile for little-endian machines (Intel, Vax) */
/* change for big-endian machines; for machines which are neither, */
/* you will need to change the definition of maybe_byte_reverse */
/* #define LITTLE_ENDIAN */

/* NIST's proposed modification to SHA of 7/11/94 may be */
/* activated by defining USE_MODIFIED_SHA; leave it off for now */
#undef USE_MODIFIED_SHA

/* SHA f()-functions */

#define f1(x,y,z)       ((x & y) | (~x & z))
#define f2(x,y,z)       (x ^ y ^ z)
#define f3(x,y,z)       ((x & y) | (x & z) | (y & z))
#define f4(x,y,z)       (x ^ y ^ z)

/* SHA constants */

#define CONST1          0x5a827999L
#define CONST2          0x6ed9eba1L
#define CONST3          0x8f1bbcdcL
#define CONST4          0xca62c1d6L

/* 32-bit rotate */

#define ROT32(x,n)      ((x << n) | (x >> (32 - n)))

/* the generic case, for when the overall rotation is not unraveled */

#define FG(n)   \
    T = ROT32(A,5) + f##n(B,C,D) + E + *WP++ + CONST##n;        \
    E = D; D = C; C = ROT32(B,30); B = A; A = T

/* specific cases, for when the overall rotation is unraveled */

#define FA(n)   \
    T = ROT32(A,5) + f##n(B,C,D) + E + *WP++ + CONST##n; B = ROT32(B,30)

#define FB(n)   \
    E = ROT32(T,5) + f##n(A,B,C) + D + *WP++ + CONST##n; A = ROT32(A,30)

#define FC(n)   \
    D = ROT32(E,5) + f##n(T,A,B) + C + *WP++ + CONST##n; T = ROT32(T,30)

#define FD(n)   \
    C = ROT32(D,5) + f##n(E,T,A) + B + *WP++ + CONST##n; E = ROT32(E,30)

#define FE(n)   \
    B = ROT32(C,5) + f##n(D,E,T) + A + *WP++ + CONST##n; D = ROT32(D,30)

#define FT(n)   \
    A = ROT32(B,5) + f##n(C,D,E) + T + *WP++ + CONST##n; C = ROT32(C,30)

/* do SHA transformation */

static COMDECL
query( IMsgDigest o, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, Component )
		|| COM_SAME_IIDNAME( iid, MsgDigest ) ) {
		*i = o;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
addref( IMsgDigest m ) {
	return 1;
	}

static COMDECLU
release( IMsgDigest m ) {
	return 1;
	}

static void
shaTransform( Sha m ) {
    int 		i;
    Jint		T, A, B, C, D, E, W[80], *WP;

    for (i = 0; i < 16; ++i) {
        W[i] = m->data[i];
		}
    for( i = 16; i < 80; ++i ) {
        W[i] = W[i-3] ^ W[i-8] ^ W[i-14] ^ W[i-16];
#ifdef USE_MODIFIED_SHA
        W[i] = ROT32(W[i], 1);
#endif /* USE_MODIFIED_SHA */
		}
    A = m->digest[0];
    B = m->digest[1];
    C = m->digest[2];
    D = m->digest[3];
    E = m->digest[4];
    WP = W;
#ifdef UNRAVEL
    FA(1); FB(1); FC(1); FD(1); FE(1); FT(1); FA(1); FB(1); FC(1); FD(1);
    FE(1); FT(1); FA(1); FB(1); FC(1); FD(1); FE(1); FT(1); FA(1); FB(1);
    FC(2); FD(2); FE(2); FT(2); FA(2); FB(2); FC(2); FD(2); FE(2); FT(2);
    FA(2); FB(2); FC(2); FD(2); FE(2); FT(2); FA(2); FB(2); FC(2); FD(2);
    FE(3); FT(3); FA(3); FB(3); FC(3); FD(3); FE(3); FT(3); FA(3); FB(3);
    FC(3); FD(3); FE(3); FT(3); FA(3); FB(3); FC(3); FD(3); FE(3); FT(3);
    FA(4); FB(4); FC(4); FD(4); FE(4); FT(4); FA(4); FB(4); FC(4); FD(4);
    FE(4); FT(4); FA(4); FB(4); FC(4); FD(4); FE(4); FT(4); FA(4); FB(4);
    m->digest[0] += E;
    m->digest[1] += T;
    m->digest[2] += A;
    m->digest[3] += B;
    m->digest[4] += C;
#else /* !UNRAVEL */
#ifdef UNROLL_LOOPS
    FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1);
    FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1); FG(1);
    FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2);
    FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2); FG(2);
    FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3);
    FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3); FG(3);
    FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4);
    FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4); FG(4);
#else /* !UNROLL_LOOPS */
    for (i =  0; i < 20; ++i) { FG(1); }
    for (i = 20; i < 40; ++i) { FG(2); }
    for (i = 40; i < 60; ++i) { FG(3); }
    for (i = 60; i < 80; ++i) { FG(4); }
#endif /* !UNROLL_LOOPS */
    m->digest[0] += A;
    m->digest[1] += B;
    m->digest[2] += C;
    m->digest[3] += D;
    m->digest[4] += E;
#endif /* !UNRAVEL */
	}

/* change endianness of data */

static void
maybeByteReverse( Juint *buffer, int count ) {
    int i;
    Juint in;

	if( comLittleEndian() ) {
		count /= sizeof(Juint);
		for (i = 0; i < count; ++i) {
			in = *buffer;
			*buffer++ = ((in << 24) & 0xff000000) | ((in <<  8) & 0x00ff0000) |
				((in >>  8) & 0x0000ff00) | ((in >> 24) & 0x000000ff);
			}
		}
	}

/* initialize the SHA digest */

static COMDECL
shaUpdate( IMsgDigest mi, Jubyte *buffer, Jint count ) {
	Sha		m = (Sha)mi;
    int 		i;

    if( (m->countLo + ((Juint) count << 3)) < m->countLo ) {
        ++m->countHi;
		}
    m->countLo += (Juint) count << 3;
    m->countHi += (Juint) count >> 29;
    if (m->local) {
        i = SHA_BLOCKSIZE - m->local;
        if (i > count) {
            i = count;
			}	
        memcpy(((Jubyte *) m->data) + m->local, buffer, i);
        count -= i;
        buffer += i;
        m->local += i;
        if (m->local == SHA_BLOCKSIZE) {
            maybeByteReverse(m->data, SHA_BLOCKSIZE);
            shaTransform(m);
			}
		else {
            return 0;
			}
		}
    while( count >= SHA_BLOCKSIZE ) {
        memcpy( m->data, buffer, SHA_BLOCKSIZE );
        buffer += SHA_BLOCKSIZE;
        count -= SHA_BLOCKSIZE;
        maybeByteReverse( m->data, SHA_BLOCKSIZE );
        shaTransform( m );
		}
    memcpy( m->data, buffer, count );
    m->local = count;
	return 0;
	}

static COMDECL
shaBegin( IMsgDigest mi ) {
	Sha		m = (Sha)mi;
	
    m->digest[0] = 0x67452301L;
    m->digest[1] = 0xefcdab89L;
    m->digest[2] = 0x98badcfeL;
    m->digest[3] = 0x10325476L;
    m->digest[4] = 0xc3d2e1f0L;
    m->countLo = 0;
    m->countHi = 0;
    m->local = 0;
	return 0;
	}

static COMDECL
shaEnd( IMsgDigest mi, MDID id ) {
	Sha	m = (Sha)mi;
    int 	count;
    Juint 	lo_bit_count, hi_bit_count;

    lo_bit_count = m->countLo;
    hi_bit_count = m->countHi;
    count = (int) ((lo_bit_count >> 3) & 0x3f);
    ((Jubyte *) m->data)[count++] = 0x80;
    if (count > SHA_BLOCKSIZE - 8) {
        memset(((Jubyte *) m->data) + count, 0, SHA_BLOCKSIZE - count);
        maybeByteReverse( m->data, SHA_BLOCKSIZE );
        shaTransform( m );
        memset( (Jubyte *) m->data, 0, SHA_BLOCKSIZE - 8 );
		}
	else {
        memset( ((Jubyte *) m->data) + count, 0, SHA_BLOCKSIZE - 8 - count );
		}
    maybeByteReverse( m->data, SHA_BLOCKSIZE );
    m->data[14] = hi_bit_count;
    m->data[15] = lo_bit_count;
    shaTransform( m );
	memcpy( (char *)id, (char *)m->digest, sizeof(*id) );
	return 0;
	}

static COMDECL
shaConstruct( IMsgDigest mi ) {
	Sha		sha = (Sha)mi;
	
	shaBegin( mi );
	sha->intf.vtbl = &dsha;
	return 0;
	}

static COMDECL
shaDestruct( IMsgDigest mi ) {
	return 0;
	}

static COMDECLS
shaDump( IMsgDigest mi, FILE* fp, char* bp ) {
	return bp;
	}

struct _DMsgDigest dsha = {
	query,
	addref,
	release,
	shaDump,
	shaDestruct,
	shaBegin,
	shaEnd,
	shaUpdate,
    };

extern IMsgDigest
comNewShaDigest( IMem mem ) {
	Sha		m;

	if( mem == NULL )
		mem = comHeap();
	m = comAlloc( mem, sizeof(*m) );
	if( m != NULL ) {
		shaConstruct( &m->intf );
		return &m->intf;
		}
	return NULL;
	}

