/********************************************************************
 *		File			:		ss.c
 *		Description		:		simple spreadsheet
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <assert.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "metautil.h"

/*
 *	This piece of code collect runtime and is modified
 *	from Berkeley Code Base util_cpu_time
 */
#ifdef IBM_WATC		/* IBM Waterloo-C compiler (same as bsd 4.2) */
#define void int
#define BSD
#endif

#if defined(BSD) || defined(SYSV)
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif

#if defined(UNIX60) || defined(UNIX100) || defined(__CYGWIN32__)
#include <sys/types.h>
#include <sys/times.h>
#endif

#ifdef vms		/* VAX/C compiler -- times() with 100 HZ clock */
#include <types.h>
#include <time.h>
#endif


static long 
util_cpu_time( void ) {
    long t = 0;

#if defined(BSD) || defined(SYSV)
    struct rusage rusage;
    (void) getrusage(RUSAGE_SELF, &rusage);
    t = (long) rusage.ru_utime.tv_sec*1000 + rusage.ru_utime.tv_usec/1000;
#endif

#ifdef IBMPC
    long ltime;
    (void) time(&ltime);
    t = ltime * 1000;
#endif

#ifdef UNIX60			/* times() with 60 Hz resolution */
    struct tms buffer;
    times(&buffer);
    t = buffer.tms_utime * 16.6667;
#endif

#ifdef UNIX100
    struct tms buffer;		/* times() with 100 Hz resolution */
    times(&buffer);
    t = buffer.tms_utime * 10;
#endif

#ifdef __CYGWIN32__
    /* Works under Windows NT but not Windows 95. */
    struct tms buffer;		/* times() with 1000 Hz resolution */
    times(&buffer);
    t = buffer.tms_utime;
#endif

#ifdef vms
    tbuffer_t buffer;	/* times() with 100 Hz resolution */
    times(&buffer);
    t = buffer.proc_user_time * 10;
#endif

    return t;
	}

extern Jerror
ssConstruct(
	Spreadsheet h, IMem mem, Jint nrows, char** descriptor
	) {
	Jint		ncols, i;
	char*		token;
	char		buffer[80];
	
	for( ncols = 0; descriptor[ncols] != NULL; ncols ++ );
	h->headers = comAlloc( mem, ncols*sizeof(char*) );
	h->types = comAlloc( mem, ncols*sizeof(char) );
	h->formats = comAlloc( mem, ncols*sizeof(char*) );
	
	for( i = 0; i < ncols; i ++ ) {
		strcpy( buffer, descriptor[i] );
		token = strtok( buffer, "%" );
		h->types[i] = token[0];

		token = strtok( NULL, "%" );
		h->headers[i] = comAlloc( mem, strlen(token)+1 );
		strcpy( h->headers[i], token );

		token = strtok( NULL, "%" );
		h->formats[i] = comAlloc( mem, strlen(token)+2 );
		strcpy( &h->formats[i][1], token );
		h->formats[i][0] = '%';
		}
	h->mem = mem;
	h->nrows = nrows;
	h->ncols = ncols;
	h->data = comAlloc( mem, sizeof(*h->data)*nrows*ncols );
	memset( h->data, '\0', sizeof(*h->data)*nrows*ncols );
	if( h->data == NULL )
		return COM_E_OUTOFMEMORY;
	h->maxrow = 0;
	h->timeTop = 0;
	return 0;
	}

extern Jerror
ssDestruct( Spreadsheet h ) {
	comFree( h->mem, h->data );
	comFree( h->mem, h->types );
	comFree( h->mem, h->headers );
	comFree( h->mem, h->formats );
	return 0;
	}

extern Jerror
ssSetColByIndex( Spreadsheet h, Jint col, Jint base, Jint step ) {
	Jint		i;
	Jint*		pcol;

	pcol = &h->data[col*h->nrows].i;
	for( i = 0; i < h->nrows; i ++ ) {
		*pcol ++= base;
		base += step;
		}
	return 0;
	}

extern Jerror
ssSetColByIndexLog( Spreadsheet h, Jint col, Jint base, Jint step ) {
	Jint		i;
	Jint*		pcol;

	pcol = &h->data[col*h->nrows].i;
	for( i = 0; i < h->nrows; i ++ ) {
		*pcol ++= 1 << (base);
		base += step;
		}
	return 0;
	}


extern Jerror
ssDumpCSV( Spreadsheet h, char* file, Jboolean append ) {
	FILE*		fp;
	Jint		row, i;

	if( append )
		fp = fopen( file, "a+" );
	else
		fp = fopen( file, "w" );
	if( fp == NULL )
		return -1;
	for( i = 0; i < h->ncols; i ++ ) {
		if( i > 0 )
			fprintf( fp, "," );
		fprintf( fp, "%s", h->headers[i] );
		}
	fprintf( fp, "\n" );
	for( row = 0; row <= h->maxrow; row ++ ) {
		for( i = 0; i < h->ncols; i ++ ) {
			if( i > 0 )
				fprintf( fp, "," );
			switch( h->types[i] ) {
			case 's' :
				fprintf( fp, h->formats[i], h->data[i*h->nrows+row].s );
				break;
			case 'i' :
				fprintf( fp, h->formats[i], h->data[i*h->nrows+row].i );
				break;
			case 'f' :
				fprintf( fp, h->formats[i], h->data[i*h->nrows+row].f );
				break;
			case 't' :	/* for time */
				fprintf( fp, "%d.%02d", h->data[i*h->nrows+row].i/1000,
						 (h->data[i*h->nrows+row].i%1000)/10
						 );
				break;
				}
			}
		fprintf( fp, "\n" );
		}
	fclose( fp );
	return 0;
	}

extern Jerror
ssDumpGrace( Spreadsheet h, char* file, Jboolean append ) {
	return ssDumpCSV( h, file, append );
	}

extern Jerror
ssDumpLatex( Spreadsheet h, char* file, Jboolean append ) {
	FILE*		fp;
	Jint		row, i;
	
	if( append )
		fp = fopen( file, "a+" );
	else
		fp = fopen( file, "w" );
	if( fp == NULL )
		return -1;
	for( row = 0; row <= h->maxrow; row ++ ) {
		for( i = 0; i < h->ncols; i ++ ) {
			if( i > 0 )
				fprintf( fp, "&" );
			switch( h->types[i] ) {
			case 's' :
				fprintf( fp, h->formats[i], h->data[i*h->nrows+row].s );
				break;
			case 'i' :
				fprintf( fp, h->formats[i], h->data[i*h->nrows+row].i );
				break;
			case 'f' :
				fprintf( fp, h->formats[i], h->data[i*h->nrows+row].f );
				break;
			case 't' :
				fprintf( fp, "%d.%02d", h->data[i*h->nrows+row].i/1000,
						 (h->data[i*h->nrows+row].i%1000)/10
						 );
				}
			}
		fprintf( fp, " \\hline\n" );
		}
	fclose( fp );
	return 0;
	}

extern Jerror
ssSum( Spreadsheet h, Jint row, Jint col ) {
	Jint	i;
	
	switch( h->types[col] ) {
	case 'i' :
	case 't' : {
		Jint*	sum;
		Jint*	pcol;
		
		pcol = &h->data[col*h->nrows].i;
		sum = pcol + row;
		for( i = 0, *sum = 0; i < row; i ++ )
			*sum += *pcol ++;
		break;
		}
	case 'f' : {
		Jfloat*	sum;
		Jfloat*	pcol;
		
		pcol = &h->data[col*h->nrows].f;
		sum = pcol + row;
		for( i = 0, *sum = 0; i < row; i ++ )
			*sum += *pcol ++;
		break;
		}
	default	:
		assert( 0 );
		}
	ssEnsure( h, row );
	return 0;
	}

extern Jerror
ssAverage( Spreadsheet h, Jint row, Jint col ) {
	ssSum( h, row, col );
	switch( h->types[col] ) {
	case 'i' :
	case 't' : 
		ssSetI( h, row, col, ssGetI( h, row, col ) / row );
		break;
	case 'f' :
		ssSetF( h, row, col, ssGetF( h, row, col ) / row );
		break;
		}
	return 0;
	}

extern Jerror
ssBeginTime( Spreadsheet h ) {
	assert( h->timeTop < SS_MAX_COUNTER );
	h->time[h->timeTop ++] = util_cpu_time();
	return 0;
	}

extern Jerror
ssEndTime( Spreadsheet h, Jint row, Jint col ) {
	assert( h->timeTop > 0 );
	ssSetI( h, row, col, (util_cpu_time() - h->time[--h->timeTop]) );
	return 0;
	}





