/********************************************************************
 *		File			:		vector.c
 *		Description		:		vector
 *		Author			:		Jianwen Zhu
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "metautil.h"

extern Jerror
vecConstruct( Vector m, IMem mem, Jint maxCount ) {
	Jint	i;

	m->data = comAlloc( mem, maxCount*sizeof(IGeneric) );
	if( m->data == NULL )
		return COM_E_OUTOFMEMORY;
	for( i = 0; i < maxCount; i ++ )
		m->data[i] = NULL;
	m->count = 0;
	m->mem = comRef( mem );
	m->maxCount = maxCount;
	return 0;
	}

extern Jerror
vecDestruct( Vector m ) {
	IMem	mem = m->mem;
	
	comFree( mem, m->data );
	m->count = 0;
	m->mem = NULL;
	m->maxCount = 0;
	return 0;
	}

#define max(a, b)	((a) > (b) ? (a) : (b))
					
extern Jerror
vecExpand( Vector m, Jint count ) {
	if( count > m->maxCount ) {
		IGeneric*		data;
		Jint			maxCount;
		IGeneric*		p1 = m->data, *end = p1 + m->count, *p2;
		IMem			mem = m->mem;
		
		maxCount = max( (count >> 2), 16 ) + count;
		data = comAlloc( mem, maxCount*sizeof(IGeneric) );
		if( data == NULL )
			return COM_E_OUTOFMEMORY;
		for( p2 = data; p1 < end; p1 ++, p2 ++ )
			*p2 = *p1;
		end = data + maxCount;
		while( p2 < end )
			*p2 ++ = NULL;
		comFree( mem, m->data );
		m->data = data;
		m->maxCount = maxCount;
		}
	if( m->count < count )
		m->count = count;
	return 0;
	}

extern Jerror
vecSetElement( Vector v, Jint i, IGeneric d ) {
	vecExpand( v, i + 1 );
	v->data[i] = d;
	return 0;
	}

extern Jerror
vecAddElement( Vector v, IGeneric d ) {
	vecExpand( v, v->count + 1 );
	v->data[v->count-1] = d;
	return 0;
	}

extern Jerror
vecRemoveElement( Vector v, Jint i ) {
	v->count --;
	for( ; i < v->count; i ++ )
		v->data[i] = v->data[i+1];
	return 0;
	}

extern Jerror
vecPush( Vector v, IGeneric d ) {
	Jint	count = v->count;
	
	vecExpand( v, count + 1 );
	v->data[count] = d;
	return 0;
	}

extern IGeneric
vecPop( Vector v ) {
	return v->data[--v->count];
	}


typedef struct _VecIter {
	struct _ISequence	intf;
	Vector				vec;
	Jint				cur;
	}* VecIter;

static COMDECL
_vecIterQuery( ISequence mi, const ComIID iid, IGeneric *i ) {
	if( COM_SAME_IIDNAME( iid, UnKnown )
		|| COM_SAME_IIDNAME( iid, Iterator )
		|| COM_SAME_IIDNAME( iid, Sequence ) ) {
		*i = mi;
		return 0;
		}
	*i = 0;
	return COM_E_NOINTERFACE;
	}

static COMDECLU
_vecIterAddRef( ISequence ci ) {
	return 1;
	}

static COMDECLU
_vecIterRelease( ISequence ci ) {
	return 1;
	}

static COMDECL
_vecIterReset( ISequence iti ) {
	VecIter	it = (VecIter)iti;

	it->cur = 0;
	return 0;
	}

static Jboolean	COMCALL
_vecIterHasMoreElements( ISequence iti ) {
	VecIter	it = (VecIter)iti;

	return it->cur < vecGetSize( it->vec );
	}

static IGeneric COMCALL
_vecIterNextElement( ISequence iti ) {
	VecIter		it = (VecIter)iti;

	return vecGetElement( it->vec, it->cur++ );
	}

static COMDECL
_vecIterPushElement( ISequence iti, IMem mem, IGeneric data ) {
	VecIter	it = (VecIter)iti;

	vecPush( it->vec, data );
	
	return 0;
	}

static IGeneric COMCALL
_vecIterPopElement( ISequence iti ) {
	VecIter	it = (VecIter)iti;

	return vecPop( it->vec );
	}

static COMDECL
_vecIterAppendElement( ISequence iti, IMem mem, IGeneric data ) {
	VecIter	it = (VecIter)iti;
	
	vecAddElement( it->vec, data );

	return 0;
	}

static COMDECL
_vecIterRemoveElement( ISequence iti, IGeneric data ) {
	VecIter	it = (VecIter)iti;
	Jint	i;

	for( i = 0; i < vecGetSize(it->vec); i ++ )
		if( vecGetElement(it->vec,i) == data )
			vecRemoveElement( it->vec, i );
	return 0;
	}

static COMDECL
_vecIterRemoveAllElements( ISequence iti ) {
	VecIter	it = (VecIter)iti;
	
	it->vec->count = 0;
	return 0;
	}

static struct _DSequence	iterVtbl = {
	_vecIterQuery, _vecIterAddRef, _vecIterRelease,
	_vecIterReset, _vecIterHasMoreElements, _vecIterNextElement,
	_vecIterPushElement, _vecIterPopElement,
	_vecIterAppendElement, _vecIterRemoveElement, _vecIterRemoveAllElements
    };

extern Jerror
vecConstructSequence( Vector l, ISequence mem ) {
	VecIter	it = (VecIter)mem;

	assert( sizeof(*it) <= ACCESSOR_SIZE );
	it->vec = l;
	it->cur = 0;
	it->intf.vtbl = &iterVtbl;
	
	return 0;
	}


