/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

// BDD Class Heap, ordered by numVars of class.  BDD with 
// larger numVars poped first.

#include <assert.h>

#include "fbddutil.h"
#include "metautil.h"

COM_DECLARE_CLASS( BchManager );
struct _BchManager
{
  Model *model;
  struct _Heap heap;
  struct _HashTable ht;
  char active;
};

#define bddcDecomposable(model, bddc) ( \
  bddc->type!=PRIMARY_INPUT && bddc->type!=PRIMARY_OUTPUT \
  && bddc->numVars>2 \
  && !isBasicGateBdd(model, bddc->bdd) )

BchManager _bchm;

BchManager 
bchNewManager(Model *model)
{
  BchManager bchm =
    (BchManager)comAlloc(_fbddHeap, sizeof(struct _BchManager));
  
  _bchm = bchm;

  bchm->model = model;
  heapConstruct( &bchm->heap, _fbddHeap, 100 );
  hashDefaultConstruct( &bchm->ht, _fbddHeap);
  bchm->active = TRUE;

  return bchm;
}

void
bchDelManager(BchManager bchm)
{
  assert(heapIsEmpty(&bchm->heap));

  while(!heapIsEmpty(&bchm->heap)){
    HeapCell hc = heapDelMin(&bchm->heap);
    comFree(heapMem(&bchm->heap), hc);
  }
  heapDelAll(&bchm->heap);
  hashDestruct(&bchm->ht);

  bchm->active = FALSE;

  comFree(_fbddHeap, bchm);
}

char 
bchIsActive(BchManager bchm)
{
  if (bchm == NULL)
    return FALSE;
  return bchm->active;
}

void
bchPush(BchManager bchm, BddClass *bddc)
{
  HeapCell cell;
  ListCell cBchBddc;
  List lBchs;

  if (!bddcDecomposable(bchm->model, bddc))
    return;

  assert(bddc->cBchBddc == NULL);

  // get lBchs
  lBchs = (List)hashGet(&bchm->ht, (void*)bddc->numVars);
  if (lBchs == NULL){
    HashCell hashCell;
    // construct lBchs
    lBchs = (List)comAlloc(_fbddHeap, sizeof(struct _List));
    listConstruct0(lBchs, _fbddHeap);

    // insert into hash table
    hashNewCell(&bchm->ht, hashCell, (void*)bddc->numVars, lBchs);
    hashPutCell(&bchm->ht, hashCell);

    // last cell is pointer to hashCell
    listAppend(lBchs, bddc->numVars);

    // insert list into heap
    cell = heapNewCell( &bchm->heap, 0 - bddc->numVars, lBchs );
    heapInsertElement( &bchm->heap, cell );
  }

  // insert bddc to lBchs
  listNewCell(_fbddHeap, cBchBddc, bddc); 
  listAddElement(lBchs, cBchBddc);
  bddc->cBchBddc = cBchBddc;

}

BddClass *
bchPop(BchManager bchm)
{
  BddClass *bddcResult = NULL;

  HeapCell heapCell = NULL;
  List lBchs;
  ListCell cBddc;
  
  // find non empty list
  
  while(!heapIsEmpty(&bchm->heap)){
    heapCell = heapFindMin( &bchm->heap );
    lBchs = heapData(List, heapCell);
    assert(listGetSize(lBchs) != 0);
    if (listGetSize(lBchs)==1){
      HashCell hc;
      int numVars;
      
      // get hash cell
      cBddc = listPop(lBchs);
      numVars = listData(int, cBddc);
      listDelCell(_fbddHeap, cBddc);
      // destroy list
      comFree(_fbddHeap, lBchs);
      
      // remove from hash table
      hc = hashRemove(&bchm->ht, (void*)numVars);
      comFree(_fbddHeap, hc);
      
      // remove from heap
      heapCell = heapDelMin( &bchm->heap );
      comFree(_fbddHeap, heapCell);
      
      heapCell = NULL;
    }
    else
      break;
  }
  
  if (heapCell != NULL){
    // remove bddc from lBchs
    cBddc = listPop(lBchs);
    bddcResult = listData(BddClass*, cBddc);
    bddcResult->cBchBddc = NULL;
    listDelCell(_fbddHeap, cBddc);
  }

  return bddcResult;
}

void
bchRemove(BchManager bchm, BddClass *bddc)
{
  ListCell cBchBddc = bddc->cBchBddc;

  if (cBchBddc != NULL){
    // remove from list.  but keep empty lists in heap.
    List lBchs;
    assert(bchm->active);

    lBchs = (List)hashGet(&bchm->ht, (void*)bddc->numVars);
    assert(lBchs != NULL); 
    assert(listGetSize(lBchs)>1);
    listRemoveElement(lBchs, bddc->cBchBddc);
    comFree(_fbddHeap, bddc->cBchBddc);
    bddc->cBchBddc = NULL;
  }
}

int
bchEmpty(BchManager bchm)
{
  return heapIsEmpty(&bchm->heap);
}
