/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
  This file contains functions for copying BDDs, folding BDDs 
  (variable reordering), and obtaining properties of the BDD.
 */
#include "stdlib.h"
#include "assert.h"
#include "time.h"

#include "metautil.h"

#include "cudd.h"
#include "fbddutil.h"


/*****************************
           Copy BDD
 *****************************/
typedef struct _CopyBddData
{
  DdManager *srcManager, *dstManager;
  DdNode *nReplace, *nReplaceWith;
  int *oldIndexToNew;
  int copyType;
  Nml nml;
  int replaceRootNewIndex;

  // Used in getChildWithCut()
  char cutType;
  DdNode *danglingEdgesReplaceWith;
  int cutLevel;
  
  // Used in markUsedLevels()
  int topLevel;
  int *levelToIndex;
  int maxIndex, varCount;

  DdNode *nReplaceRoot, *nReplaceThen, *nReplaceElse;
  char replaceComplement;

} CopyBddData;

// Requires: Relative variable order consistency.
// If variable u is lower in the variable order than v 
// in the source bdd, then u must be lower than v in the destination bdd.
static DdNode *
copyBddHelper(CopyBddData *cbd, DdNode *srcNode)
{
  DdNode *result;

  if (Cudd_Regular(srcNode) == Cudd_Regular(cbd->nReplace)){
    if (srcNode == cbd->nReplace)
      srcNode = cbd->nReplaceWith;
    else{
      // srcNode is complement of nReplaceWith
      // replace srcNode with complement of nReplaceWith
      if (Cudd_IsComplement(cbd->nReplaceWith))
	srcNode = Cudd_Regular(cbd->nReplaceWith);
      else
	srcNode = Cudd_Complement(cbd->nReplaceWith);
    }
  }

  if (cbd->copyType == CT_NESTED_BDD){
    DdNode *oldSrcNode = srcNode;
    char foundMatch = FALSE;
    if (Cudd_Regular(srcNode) == Cudd_Regular(cbd->nReplaceThen) 
	&& Cudd_Regular(cbd->nReplaceElse) == Cudd_Regular(cbd->nReplaceThen)){
      // special case
      // then and else nodes are complements of each other
      if (Cudd_IsComplement(cbd->nReplaceThen)){
	assert(Cudd_IsComplement(cbd->nReplaceElse)==FALSE);
	srcNode = Cudd_ReadLogicZero(cbd->srcManager);	
      }
      else{
	assert(Cudd_IsComplement(cbd->nReplaceElse));
	srcNode = Cudd_ReadOne(cbd->srcManager);	
      }
      foundMatch = TRUE;
    }
    else if (Cudd_Regular(srcNode) == Cudd_Regular(cbd->nReplaceThen)){
      if (Cudd_IsComplement(cbd->nReplaceThen))
	srcNode = Cudd_ReadLogicZero(cbd->srcManager);
      else
	srcNode = Cudd_ReadOne(cbd->srcManager);
      foundMatch = TRUE;
    }
    else if (Cudd_Regular(srcNode) == Cudd_Regular(cbd->nReplaceElse)){
      if (Cudd_IsComplement(cbd->nReplaceElse))
	srcNode = Cudd_ReadOne(cbd->srcManager);
      else
	srcNode = Cudd_ReadLogicZero(cbd->srcManager);
      foundMatch = TRUE;
    }

    if (foundMatch && Cudd_IsComplement(oldSrcNode)){
      // reverse 
      if (srcNode == Cudd_ReadOne(cbd->srcManager))
	srcNode = Cudd_ReadLogicZero(cbd->srcManager);
      else{
	assert(srcNode == Cudd_ReadLogicZero(cbd->srcManager));
	srcNode = Cudd_ReadOne(cbd->srcManager);
      }
    }
  }

  if (Cudd_IsConstant(srcNode)){
    if (Cudd_IsComplement(srcNode)){
      // srcNode is constant 0
      result = Cudd_ReadLogicZero(cbd->dstManager);
      Cudd_Ref(result);
    }
    else{
      // srcNode is constant 1
      result = Cudd_ReadOne(cbd->dstManager);
      Cudd_Ref(result);
    }
  }
  else{
    DdNode *tNode, *eNode;
    DdNode *curNode;
    DdNode *matchNode;
    int newIndex;

    if (cbd->copyType == CT_NESTED_BDD_REMAINDER 
	&& Cudd_Regular(srcNode) == Cudd_Regular(cbd->nReplaceRoot))
      newIndex = cbd->replaceRootNewIndex;
    else if (cbd->oldIndexToNew == NULL)
      newIndex = Cudd_NodeReadIndex(srcNode);
    else
      newIndex = cbd->oldIndexToNew[Cudd_NodeReadIndex(srcNode)];

    if ((matchNode = nmlGetNode(&cbd->nml, srcNode)) == NULL){
      // node not encountered before
      assert(newIndex != -1);
      
      if (cbd->copyType == CT_NESTED_BDD_REMAINDER 
	  && Cudd_Regular(srcNode) == Cudd_Regular(cbd->nReplaceRoot)){
	tNode = copyBddHelper( cbd, cbd->nReplaceThen );
	eNode = copyBddHelper( cbd, cbd->nReplaceElse );
      }
      else{
	tNode = copyBddHelper( cbd, Cudd_T(srcNode) );
	eNode = copyBddHelper( cbd, Cudd_E(srcNode) );
      }

      curNode = Cudd_bddIthVar(cbd->dstManager, newIndex);
      Cudd_Ref(curNode);

      assert(tNode != NULL);
      assert(eNode != NULL);
      assert(curNode != NULL);
      result = Cudd_bddIte(cbd->dstManager, curNode, tNode, eNode);
      assert(result);
      Cudd_Ref(result);
      Cudd_RecursiveDeref(cbd->dstManager, tNode);
      Cudd_RecursiveDeref(cbd->dstManager, eNode);
      Cudd_RecursiveDeref(cbd->dstManager, curNode);
      
      if (Cudd_IsComplement(srcNode)){
	DdNode *nTmp = Cudd_Not(result);
	Cudd_Ref(nTmp);
	Cudd_RecursiveDeref(cbd->dstManager, result);
	result = nTmp;
      }

      // must keep result (bdd) alive if it is to be 
      // used in matchings later
      Cudd_Ref(result); 
      nmlPutNode(&cbd->nml, srcNode, result);
    }
    else{
      // node already encountered
      result = matchNode;
      Cudd_Ref(result);
    }
  }
  return result;
}

// Performs two functions simultaneously. Replaces dangling edges (edges 
// on the cut that point to a node) with _cutEdgesReplaceWith.  And returns
// a copy of this Bdd, possibly in another DdManager.
// CT_REPLACE_ONE_EDGES: one edges originating above the cut line
// are redirected to sibling nodes.
// CT_REPLACE_ZERO_EDGES: zero edges originating above the cut line
// are redirected to sibling nodes.
// CT_REPLCE_DANGLING_EDGES: dangling edges are redireced to 
// danglingEdgesReplaceWith.
// valueWRTRoot: indicates whether this value is taken as it's regular or 
// complement with respect to the root node (not the node that points to it).


static DdNode *
getChildWithCut(CopyBddData *cbd, DdNode *nParent, char valueWRTRoot, 
		char dir, char *childValueWRTRoot)
{
  DdNode *nChild, *nResult;

  if (dir == THEN_EDGE)
    nChild = Cudd_T(nParent);
  else if (dir == ELSE_EDGE)
    nChild = Cudd_E(nParent);

  if (Cudd_ReadPerm(cbd->srcManager, Cudd_NodeReadIndex(nParent))
      <=cbd->cutLevel){
    if (cbd->cutType==CT_REPLACE_ONE_EDGES && Cudd_IsConstant(nChild) 
	&& ((Cudd_IsComplement(nChild)==FALSE && valueWRTRoot==VAL_REGULAR) 
	    || 
	    (Cudd_IsComplement(nChild)==TRUE 
	     && valueWRTRoot==VAL_COMPLEMENT ))){
      // redirect edge to sibling.
      if (dir == THEN_EDGE)
	nResult = Cudd_E(nParent);
      else if (dir == ELSE_EDGE)
	nResult = Cudd_T(nParent);
    }
    else if (cbd->cutType==CT_REPLACE_ZERO_EDGES && Cudd_IsConstant(nChild) && 
	     ((Cudd_IsComplement(nChild)==FALSE 
	       && valueWRTRoot==VAL_COMPLEMENT) 
	      || (Cudd_IsComplement(nChild)==TRUE 
	      && valueWRTRoot==VAL_REGULAR) )){
      // redirect edge to sibling
      if (dir == THEN_EDGE)
	nResult = Cudd_E(nParent);
      else if (dir == ELSE_EDGE)
	nResult = Cudd_T(nParent);
    }
    else if(cbd->cutType==CT_REPLACE_DANGLING_EDGES && 
	    Cudd_IsConstant(nChild)==FALSE &&
	    Cudd_ReadPerm(cbd->srcManager, Cudd_NodeReadIndex(nChild))
	    >cbd->cutLevel){
      // redirect dangling edges to danglingEdgesReplaceWith
      nResult = cbd->danglingEdgesReplaceWith;
      if (Cudd_IsComplement(cbd->danglingEdgesReplaceWith))
	*childValueWRTRoot = VAL_COMPLEMENT;
      else
	*childValueWRTRoot = VAL_REGULAR;
      return Cudd_Regular(nResult);
    }
    else{
      // return usual
      nResult = nChild;
    }
  }
  else
    nResult = nChild;

  if (Cudd_IsComplement(nResult)){
    *childValueWRTRoot = 
      (valueWRTRoot==VAL_REGULAR?VAL_COMPLEMENT:VAL_REGULAR);
  }
  else{
    *childValueWRTRoot = valueWRTRoot;
  }

  return Cudd_Regular(nResult);
}

DdNode *
copyBddHelperWithCut(CopyBddData *cbd, DdNode *srcNode, char valueWRTRoot)
{
  DdNode *result;

  assert(srcNode != NULL);
  assert(srcNode == Cudd_Regular(srcNode));

  if (Cudd_IsConstant(srcNode)){
    if (valueWRTRoot == VAL_REGULAR){
      // srcNode is constant 0
      result = Cudd_ReadOne(cbd->dstManager);
      Cudd_Ref(result);
    }
    else{
      // srcNode is constant 1
      result = Cudd_ReadLogicZero(cbd->dstManager);
      Cudd_Ref(result);
    }
  }
  else{
    DdNode *tNode, *eNode;
    DdNode *curNode;
    DdNode *matchNode;
    int newIndex;
    DdNode *nNmlKey;
    char childValueWRTRoot;
    
    if (cbd->oldIndexToNew == NULL)
      newIndex = Cudd_NodeReadIndex(srcNode);
    else
      newIndex = cbd->oldIndexToNew[Cudd_NodeReadIndex(srcNode)];

    // hack: srcNode always comes in regular form.  In node lookup,
    // need to distinguish between VAL_COMPLEMENT and VAL_REGULAR.  Use
    // the last bit of srcNode for this purpose.
    nNmlKey = srcNode;
    if (valueWRTRoot == VAL_COMPLEMENT){
      unsigned int key;
      key = (unsigned int)nNmlKey;
      key |= 1;
      nNmlKey = (DdNode*)key;
    }
      
    if ((matchNode = nmlGetNode(&cbd->nml, nNmlKey)) == NULL){
      // node not encountered before
      DdNode *nChild;

      assert(newIndex != -1);
      nChild = getChildWithCut(cbd, srcNode, valueWRTRoot, THEN_EDGE, 
			       &childValueWRTRoot);
      tNode = copyBddHelperWithCut( cbd, nChild, childValueWRTRoot );  

      nChild = getChildWithCut(cbd, srcNode, valueWRTRoot, ELSE_EDGE, 
			       &childValueWRTRoot);
      eNode = copyBddHelperWithCut(cbd,  nChild, childValueWRTRoot );  

      curNode = Cudd_bddIthVar(cbd->dstManager, newIndex);
      Cudd_Ref(curNode);

      result = Cudd_bddIte(cbd->dstManager, curNode, tNode, eNode);
      assert(result);
      Cudd_Ref(result);
      Cudd_RecursiveDeref(cbd->dstManager, tNode);
      Cudd_RecursiveDeref(cbd->dstManager, eNode);
      Cudd_RecursiveDeref(cbd->dstManager, curNode);
      
      Cudd_Ref(result);
      nmlPutNode(&cbd->nml, nNmlKey, result);
    }
    else{
      // node already encountered
      result = matchNode;
      Cudd_Ref(result);
    }
  }
  return result;
}

static void
clearNmlBddReferences(Nml *nml, DdManager *man)
{
  int i, j;
  HashCell cell;
  hashForeach( nml->hashTable, i, j, cell ){
    DdNode *n = hashData(DdNode*, cell);
    Cudd_RecursiveDeref(man, n);
  }
}

DdNode*
rmGetReplacement(DdNode *nReplace, int pos)
{
  if (pos == THEN_EDGE)
    return Cudd_T(nReplace);
  else if (pos == ELSE_EDGE)
    return Cudd_E(nReplace);
  else
    assert(FALSE);
}

static void
markUsedLevels(CopyBddData *cbd, DdNode *node)
{

  if (Cudd_Regular(node) == Cudd_Regular(cbd->nReplace)){
    if (node == cbd->nReplace)
      node = cbd->nReplaceWith;
    else{
      // srcNode is complement of _nReplaceWith
      // replace srcNode with complement of _nReplaceWith
      if (Cudd_IsComplement(cbd->nReplaceWith))
	node = Cudd_Regular(cbd->nReplaceWith);
      else
	node = Cudd_Complement(cbd->nReplaceWith);
    }
  }
  
  if (cbd->copyType == CT_NESTED_BDD){
    if (Cudd_Regular(node) == Cudd_Regular(cbd->nReplaceThen)){
      if (Cudd_IsComplement(node))
	node = Cudd_ReadLogicZero(cbd->srcManager);
      else
	node = Cudd_ReadOne(cbd->srcManager);
    }
    else if (Cudd_Regular(node) == Cudd_Regular(cbd->nReplaceElse)){
      if (Cudd_IsComplement(node))
	node = Cudd_ReadOne(cbd->srcManager);
      else
	node = Cudd_ReadLogicZero(cbd->srcManager);
    }
  }
  
  if (cbd->copyType == CT_NESTED_BDD_REMAINDER 
      && Cudd_Regular(node) == Cudd_Regular(cbd->nReplaceRoot)){
     if (!nmlGetMark(&cbd->nml, Cudd_Regular(node))){
       nmlPutMark(&cbd->nml, Cudd_Regular(node), 1);
       cbd->varCount++;
       markUsedLevels(cbd, cbd->nReplaceThen);
       markUsedLevels(cbd, cbd->nReplaceElse);
     }
  }
  else if (Cudd_IsConstant(node) == FALSE){
    if (!nmlGetMark(&cbd->nml, Cudd_Regular(node))){
      int level = Cudd_ReadPerm(cbd->srcManager, Cudd_NodeReadIndex(node));
      // node has not been visited before
      nmlPutMark(&cbd->nml, Cudd_Regular(node), 1);
      
      // levels grow down
      if (cbd->levelToIndex[level - cbd->topLevel] == -1){
	cbd->levelToIndex[level - cbd->topLevel] = 
	  Cudd_NodeReadIndex(node); // mark as used level
	cbd->varCount++;

	if (Cudd_NodeReadIndex(node) > cbd->maxIndex)
	  cbd->maxIndex = Cudd_NodeReadIndex(node);
      }

      markUsedLevels(cbd, rmGetReplacement(node, THEN_EDGE));
      markUsedLevels(cbd, rmGetReplacement(node, ELSE_EDGE));
    }
  }
}

static DdNode *
_bddCopyAndReIndex(CopyBddData *cbd, DdManager *srcManager, DdNode *srcNode, 
		   DdManager *dstManager, 
		   int *numVars, int **oldIndexToNew, int *arraySize)
{
  int i,j;
  DdNode *result;
  int levelToIndexSize;
  clock_t t;
  int *levelToIndex2 = NULL;

  t = util_cpu_time();

  cbd->srcManager = srcManager;
  cbd->dstManager = dstManager;
  cbd->oldIndexToNew = *oldIndexToNew;

  // Use hash table for determining uniqueness of indexes.
  // Don't want to put entries in heap more than once.

  // get _maxIndex, and _varCount and fill the heap with variable 
  // indexes, ordered with lower level indexes first.
  cbd->maxIndex=0;
  cbd->varCount =0;
  cbd->topLevel = Cudd_ReadPerm(cbd->srcManager, Cudd_NodeReadIndex(srcNode));

  if (cbd->topLevel != CUDD_MAXINDEX){
    levelToIndexSize = Cudd_ReadSize(cbd->srcManager) - cbd->topLevel;
    cbd->levelToIndex = (int*)comAlloc(_fbddHeap,sizeof(int)*levelToIndexSize);
    for (i=0; i<levelToIndexSize; i++)
      cbd->levelToIndex[i] = -1;
  }
  else{
    levelToIndexSize = 0;
  }

  nmlConstruct(&cbd->nml);
  markUsedLevels(cbd, srcNode); // updates _oldIndexToNew
  nmlDestruct(&cbd->nml);

  *arraySize = cbd->maxIndex+1;
  *numVars = cbd->varCount;
  *oldIndexToNew = (int*)comAlloc(_fbddHeap,sizeof(int)*(*arraySize));
  cbd->oldIndexToNew = *oldIndexToNew;
  for (i=0; i<*arraySize; i++)
    cbd->oldIndexToNew[i] = -1;

  // populate levelToIndex2
  {
    int numVars2 = Cudd_ReadSize(dstManager);
    levelToIndex2 = (int*)comAlloc(_fbddHeap,sizeof(int)*(*numVars));

#ifdef DEBUG_COMPILE
    for (j=0; j<*numVars; j++)
      levelToIndex2[j] = -1;
#endif

    for(j=0; j<numVars2; j++){
      int level = numVars2 - Cudd_ReadPerm(dstManager, j) - 1;
      if ( level < (*numVars) )
	levelToIndex2[level] = j;
    }

#ifdef DEBUG_COMPILE
    // check no levelToIndex2 values repeat
    for (j=0; j<*numVars; j++){
      int k;
      assert (levelToIndex2[j] != -1);
      for (k=j+1; k<*numVars; k++){
	assert(levelToIndex2[j] != levelToIndex2[k]);
      }
    }
#endif

  }
  
  j=0;
  if (cbd->copyType == CT_NESTED_BDD_REMAINDER){
    int replaceRootLevel = 
      Cudd_ReadPerm(cbd->srcManager, Cudd_NodeReadIndex(cbd->nReplaceRoot)) 
      - cbd->topLevel;
    cbd->replaceRootNewIndex = -1;

    assert (replaceRootLevel < levelToIndexSize);

    for (i=levelToIndexSize-1; i>=0; i--){
      if (i == replaceRootLevel){
	// level replaceRootLevel is for the variable generated by the
	// extracted nested BDD
	assert(levelToIndex2[j] != -1);
	cbd->replaceRootNewIndex = levelToIndex2[j];
	j++;
      }
      if (cbd->levelToIndex[i] != -1){

	assert (j<*numVars);
	cbd->oldIndexToNew[cbd->levelToIndex[i]] = levelToIndex2[j];
	j++;
      }
    }
    assert(cbd->replaceRootNewIndex != -1);
  }
  else{
    for (i=levelToIndexSize-1; i>=0; i--){
      if (cbd->levelToIndex[i] != -1){
	assert (j<*numVars);
	cbd->oldIndexToNew[cbd->levelToIndex[i]] = levelToIndex2[j];
	j++;
      }
    }
  }
  comFree(_fbddHeap,levelToIndex2);

  //  if (cbd->topLevel != CUDD_CONST_INDEX){
  if (cbd->topLevel != CUDD_MAXINDEX){
    comFree(_fbddHeap,cbd->levelToIndex);

    // want index 0 starting at bottom, set this up if not already done so
    if (Cudd_ReadVars(dstManager,0) == NULL){
      assert(FALSE);
      for (i=0; i<*numVars; i++)
	Cudd_bddNewVarAtLevel(dstManager,0);
    }
  }
  nmlConstruct(&cbd->nml);

  if (cbd->cutType == CT_NONE){
    result = copyBddHelper(cbd, srcNode);
    clearNmlBddReferences(&cbd->nml, dstManager);
  }
  else{
    result = copyBddHelperWithCut(cbd, Cudd_Regular(srcNode), 
	  (Cudd_IsComplement(srcNode))?VAL_COMPLEMENT:VAL_REGULAR);
    clearNmlBddReferences(&cbd->nml, dstManager);
  }
  nmlDestruct(&cbd->nml);  

  tBddCopy += util_cpu_time()-t;

  return result;
}

DdNode *
bddCopyAndReIndex(DdManager *srcManager, DdNode *srcNode, 
		  DdManager *dstManager, 
		  int *numVars, int **oldIndexToNew, int *arraySize,
		  DdNode *nReplace, DdNode *nReplaceWith)
{
  CopyBddData cbd;

  cbd.copyType = CT_REGULAR;
  cbd.nReplace = nReplace;
  cbd.nReplaceWith = nReplaceWith;
  cbd.replaceComplement = FALSE;

  cbd.cutType=CT_NONE;
  cbd.danglingEdgesReplaceWith=NULL;
  cbd.cutLevel=10000;

  return _bddCopyAndReIndex(&cbd, srcManager, srcNode,
			    dstManager,
			    numVars, oldIndexToNew, arraySize);
}


DdNode *
bddCopyAndReIndexNestedBdd(DdManager *srcManager, DdNode *srcNode, 
			   DdManager *dstManager, 
			   int *numVars, int **oldIndexToNew, int *arraySize,
			   int copyType, 
			   DdNode *nReplaceRoot, 
			   DdNode *nReplaceThen, DdNode *nReplaceElse,
			   int *nestedBddIndex)
{
  DdNode *result;
  CopyBddData cbd;

  cbd.copyType = copyType;
  cbd.nReplaceRoot = nReplaceRoot;
  cbd.nReplaceThen = nReplaceThen;
  cbd.nReplaceElse = nReplaceElse;
  cbd.nReplace = NULL;
  cbd.nReplaceWith = NULL;

  cbd.cutType=CT_NONE;
  cbd.danglingEdgesReplaceWith=NULL;
  cbd.cutLevel=10000;

  result =  _bddCopyAndReIndex(&cbd, srcManager, srcNode,
			       dstManager,
			       numVars, oldIndexToNew, arraySize);
  if (cbd.copyType == CT_NESTED_BDD_REMAINDER)
    *nestedBddIndex = cbd.replaceRootNewIndex;

  cbd.nReplaceThen = NULL;
  cbd.nReplaceElse = NULL;

  return result;
}

DdNode *
bddCopyAndReIndexWithCut(DdManager *srcManager, DdNode *srcNode, 
		  DdManager *dstManager, 
		  int *numVars, int **oldIndexToNew, int *arraySize,
		  char cutType, DdNode *replaceWith, int cutLevel)
{
  DdNode *nResult;
  CopyBddData cbd;

  cbd.cutType = cutType;
  cbd.danglingEdgesReplaceWith = replaceWith;
  cbd.cutLevel = cutLevel;

  cbd.copyType = CT_REGULAR;
  cbd.nReplace = NULL;
  cbd.nReplaceWith = NULL;
  cbd.replaceComplement = FALSE;

  nResult = _bddCopyAndReIndex(&cbd, srcManager, srcNode, dstManager, 
			       numVars, oldIndexToNew, arraySize);

  cbd.cutType = CT_NONE;

  return nResult;
}

DdNode *
bddCopy(DdManager* srcManager, DdNode *srcNode, DdManager *dstManager, 
	int *oldIndexToNew, DdNode *nReplace, DdNode *nReplaceWith)

{
  DdNode *result;
  clock_t t;
  CopyBddData cbd;


  t = util_cpu_time();

  cbd.nReplace = nReplace;
  cbd.nReplaceWith = nReplaceWith;
  cbd.srcManager = srcManager;
  cbd.dstManager = dstManager;
  cbd.oldIndexToNew = oldIndexToNew;
  cbd.copyType = CT_NONE;

  cbd.cutType = CT_NONE;
  cbd.danglingEdgesReplaceWith = NULL;
  cbd.cutLevel = 10000;

  nmlConstruct(&cbd.nml);
  result = copyBddHelper(&cbd, srcNode);
  clearNmlBddReferences(&cbd.nml, dstManager);
  nmlDestruct(&cbd.nml);

  tBddCopy += util_cpu_time()-t;

  return(result);
}

/************************************
          Bdd Folding
 ************************************/
static int 
bddGetSupportSize(DdManager *manager, DdNode *node)
{
  int i=0;
  DdNode *nTop;
  nTop = Cudd_Support(manager, node);
  Cudd_Ref(nTop);

  node = nTop;
  if (node == NULL)
    return 0;
  else{
    while (Cudd_IsConstant(node)==FALSE){
      i++;
      node = Cudd_T(node);
    }
  }

  Cudd_RecursiveDeref(manager, nTop);

  return i;
}

/*
  Function: foldBddClass()
  Desc: 
  Does not Dereference topNode.
*/
// extern ManArray *_manArrayFold;
DdNode *
foldBddClass(Model *model, DdNode *topNode,
	     int *NumVars, int **oldIndexToNew, int *arraySize,
	     DdNode *nReplace, DdNode *nReplaceWith)
{
  DdManager *manager = model->ddManager;
  ManArray *manArrayFold = &model->manArrayFold;
  DdManager *tempManager = NULL;
  DdNode *node;
  int *oldIndexToNew1, *oldIndexToNew2;
  int numVars1, numVars2;
  int arraySize1, arraySize2;
  int i;
  int *permute = NULL;
  clock_t t1, t2;
  DdNode *nTemp;
  int maxNumVars;
  int permuteMaxIndex;
  int bddSize;

  t1 = util_cpu_time();

  maxNumVars = bddGetSupportSize(manager, topNode);
  tempManager = getDdManager(manArrayFold, maxNumVars);

  for (i=0; i<maxNumVars; i++){
    Cudd_bddIthVar(tempManager, i);
  }

  // copy bdd out of folded bdd, so that we can reduce it, without 
  // affecting other variables.
  node = bddCopyAndReIndex(manager, topNode, 
			   tempManager, 
			   &numVars1, &oldIndexToNew1, &arraySize1,
			   nReplace, nReplaceWith);
  
  //  Cudd_AutodynEnable(tempManager, CUDD_REORDER_TYPE);
  permuteMaxIndex = -1;
  for (i=0; i<arraySize1; i++)
    if (oldIndexToNew1[i]>permuteMaxIndex)
      permuteMaxIndex = oldIndexToNew1[i];

  permute = 
    (int*)comAlloc(_fbddHeap,sizeof(int) * (Cudd_ReadSize(tempManager)));
  for (i=0; i<Cudd_ReadSize(tempManager); i++)
    permute[i] = i;

  for (i=0; i<=permuteMaxIndex; i++)
    permute[i] = i;

  // reduce heap
  bddSize = Cudd_DagSize(node);
  t2 = util_cpu_time();
  
  if (bddSize-1 > numVars1)
    Cudd_ReduceHeap(tempManager, CUDD_REORDER_TYPE, numVars1+1);

  tReduceHeap += util_cpu_time() - t2;

  // copy reduced bdd into folded bdd.
  nTemp = bddCopyAndReIndex(tempManager, node, 
			    manager, 
			    &numVars2, &oldIndexToNew2, &arraySize2,
			    nReplace, nReplaceWith);
  Cudd_RecursiveDeref(tempManager, node);
  node = nTemp;

  releaseDdManager(manArrayFold);

  *oldIndexToNew = (int*)comAlloc(_fbddHeap,sizeof(int)*arraySize1);

  for (i=0; i<arraySize1; i++){
    if (oldIndexToNew1[i] == -1)
      (*oldIndexToNew)[i] = -1;
    else{
      assert(oldIndexToNew1[i] <= permuteMaxIndex);
      (*oldIndexToNew)[i] = oldIndexToNew2[permute[oldIndexToNew1[i]]];
    }
  }
  
  comFree(_fbddHeap,permute);
  comFree(_fbddHeap,oldIndexToNew1);
  comFree(_fbddHeap,oldIndexToNew2);
  
  *NumVars = numVars1;
  assert(maxNumVars == numVars1);
  *arraySize = arraySize1;

  assert(*arraySize >= 0);

  tFoldBddClass += util_cpu_time() - t1;

  return node;
}

// Does not deref bdd.
DdNode*
foldBddUpdateOldIndexToNew(model, bdd, oldIndexToNew, arraySize)
     Model *model;
     DdNode *bdd;
     int **oldIndexToNew;
     int arraySize;
{
  DdNode *nTmp;
  int *oldIndexToNewTmp1;
  int *oldIndexToNewTmp2;
  int numVarsTmp, arraySizeTmp;
  int i;

  nTmp = foldBddClass(model, bdd,
		      &numVarsTmp, &oldIndexToNewTmp1, &arraySizeTmp,
		      NULL, NULL);
  bdd = nTmp;

  oldIndexToNewTmp2 = (int*)comAlloc(_fbddHeap,sizeof(int)*arraySize);
  for (i=0; i<arraySize; i++){
    if ((*oldIndexToNew)[i]==-1)
      oldIndexToNewTmp2[i] = -1;
    else
      oldIndexToNewTmp2[i] = oldIndexToNewTmp1[(*oldIndexToNew)[i]];
  }
  comFree(_fbddHeap,*oldIndexToNew);
  *oldIndexToNew = oldIndexToNewTmp2;
  comFree(_fbddHeap,oldIndexToNewTmp1);

  return bdd;
}

/************************************
          Bdd Properties
 ************************************/
typedef struct _BddPropertiesData
{
  char nodeIndexMark[MAX_BDD_VARS];
  Nml nmlNodeMark;
  int nodeCount_withComplement;
  int nodeCount_noComplement;
  int numVars;
} BddPropertiesData;

static void
bddGetPropertiesHelper(BddPropertiesData *bpd, DdNode *node)
{
  int mark;
  int putMark;

  if (Cudd_IsConstant(node))
    return;

  if (Cudd_IsComplement(node))
    putMark=2;
  else
    putMark=1;

  if ((mark=nmlGetMark(&bpd->nmlNodeMark, Cudd_Regular(node)))==0){
    nmlPutMark(&bpd->nmlNodeMark, Cudd_Regular(node), putMark);
    bpd->nodeCount_withComplement++;
    bpd->nodeCount_noComplement++;

    // update _numVars
    if (bpd->nodeIndexMark[Cudd_NodeReadIndex(node)]==0){
      bpd->nodeIndexMark[Cudd_NodeReadIndex(node)]=1;
      bpd->numVars++;
    }

    bddGetPropertiesHelper(bpd, getComplementAdjustedNode(Cudd_T(node), node));
    bddGetPropertiesHelper(bpd, getComplementAdjustedNode(Cudd_E(node), node));
  }
  else if(mark==1){
    // previously encountered regular version of node
    if (putMark==2){
      putMark = 3;
      nmlPutMark(&bpd->nmlNodeMark, Cudd_Regular(node), putMark);
      bpd->nodeCount_noComplement++;

      bddGetPropertiesHelper(bpd, 
			     getComplementAdjustedNode(Cudd_T(node), node));
      bddGetPropertiesHelper(bpd, 
			     getComplementAdjustedNode(Cudd_E(node), node));
    }
  }
  else if(mark==2){
    // previously encountered complement version of node
    if (putMark==1){
      putMark = 3;
      nmlPutMark(&bpd->nmlNodeMark, Cudd_Regular(node), putMark);
      bpd->nodeCount_noComplement++;
      bddGetPropertiesHelper(bpd, 
			     getComplementAdjustedNode(Cudd_T(node), node));
      bddGetPropertiesHelper(bpd, 
			     getComplementAdjustedNode(Cudd_E(node), node));
    }
  }
}

void
bddGetProperties(DdNode *node, BddProperties *bp)
{
  int i;
  BddPropertiesData bpd;

  for (i=0; i<MAX_BDD_VARS; i++)
    bpd.nodeIndexMark[i]=0;

  bpd.nodeCount_withComplement = 0;
  bpd.nodeCount_noComplement = 0;
  bpd.numVars = 0;

  nmlConstruct(&bpd.nmlNodeMark);
  bddGetPropertiesHelper(&bpd, node);
  nmlDestruct(&bpd.nmlNodeMark);

  bp->nodeCount_withComplement = bpd.nodeCount_withComplement;
  bp->nodeCount_noComplement = bpd.nodeCount_noComplement;
  bp->numVars = bpd.numVars;
}

// returns -1, decomposition 1 is better
// returns 1, decomposition 2 is better
int
bddCmpCost(bpOrig, bpA1, bpB1, bpC1, bpA2, bpB2, bpC2)
     BddProperties *bpOrig;
     BddProperties *bpA1, *bpB1, *bpC1;
     BddProperties *bpA2, *bpB2, *bpC2;
{
  BddProperties bp1, bp2;

  bp1.numVars = 0;
  bp1.nodeCount_withComplement = 0;
  bp1.nodeCount_noComplement = 0;

  bp2.numVars = 0;
  bp2.nodeCount_withComplement = 0;
  bp2.nodeCount_noComplement = 0;

  // cost based on node count
  bp1.nodeCount_withComplement = bpA1->nodeCount_withComplement 
    + bpB1->nodeCount_withComplement;
  bp1.nodeCount_noComplement = bpA1->nodeCount_noComplement 
    + bpB1->nodeCount_noComplement;

  bp2.nodeCount_withComplement = bpA2->nodeCount_withComplement 
    + bpB2->nodeCount_withComplement;
  bp2.nodeCount_noComplement = bpA2->nodeCount_noComplement 
    + bpB2->nodeCount_noComplement;

  if (bpC1 != NULL){
    bp1.nodeCount_withComplement += bpC1->nodeCount_withComplement;
    bp1.nodeCount_noComplement += bpC1->nodeCount_noComplement;
  }
  
  if (bpC2 != NULL){
    bp2.nodeCount_withComplement += bpC2->nodeCount_withComplement;
    bp2.nodeCount_noComplement += bpC2->nodeCount_noComplement;
  }


  return (bp1.nodeCount_withComplement + bp1.nodeCount_noComplement)
    - (bp2.nodeCount_withComplement + bp2.nodeCount_noComplement);
}


/************************
       Other
 ************************/
// variables used in bddiEval
void
printBdd(DdNode *node)
{
  printf("%d-%d ", Cudd_NodeReadIndex(node), Cudd_IsComplement(node));
  if (Cudd_IsConstant(node) == FALSE){
    printf("then:%d else:%d\n", 
	   Cudd_NodeReadIndex(Cudd_T(node)), Cudd_NodeReadIndex(Cudd_E(node)));
    printBdd(Cudd_T(node));
    printBdd(Cudd_E(node));
  }
  else{
    if (Cudd_IsComplement(node))
      printf("constant: %g \n", Cudd_V(node)-1);
    else
      printf("constant: %g \n", Cudd_V(node));
  }
}

DdNode*
bddComputeOperation(DdManager *ddManager, extractType et, 
		    DdNode *n1, DdNode *n2)
{
  DdNode *bddResult;

  switch(et){
  case ET_AND:
    bddResult = Cudd_bddAnd(ddManager, n1, n2);
    break;
  case ET_OR:
    bddResult = Cudd_bddOr(ddManager, n1, n2);
    break;
  case ET_AND01:
    bddResult = Cudd_bddAnd(ddManager, Cudd_Not(n1), n2);
    break;
  case ET_AND10:
    bddResult = Cudd_bddAnd(ddManager, n1, Cudd_Not(n2));
    break;
  case ET_XOR:
    bddResult = Cudd_bddXor(ddManager, n1, n2);
    break;
  default:
    assert(FALSE);
  }
  return bddResult;
}
