/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
  File: decomp.c
  Desc: Functions for performing decomposition using various BDD-based 
  techniques.  If sharing extraction is activated, it is interleaved 
  with decomposition.
*/

#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include "metautil.h"

#include "fbdd.h"
#include "fbddutil.h"
#include "fbddopt.h"
#include "decomp.h"

#define NUM_DECOMP_TYPES 15

typedef enum{DT_NONE, DT_ALG, DT_BOOL, DT_BDD_DOM, DT_NESTED_BDD, 
	       DT_MUX, DT_NOT, DT_BXNOR, DT_AND, DT_OR, DT_XNOR, 
	       DT_B2VD, DT_ERROR} decompType;

clock_t _tDecomp=0;
clock_t _tDecompAlg=0;
clock_t _tDecompBool=0;
clock_t _tDecompMux=0;
clock_t _tDecompBdd=0;
clock_t _tDecompNest=0;

static int _numDecomps[NUM_DECOMP_TYPES];
static int _numFoldedDecomps[NUM_DECOMP_TYPES];
static int _totalDecomps = 0;
static int _totalFoldedDecomps = 0;

/**** Function prototypes of decomposition algorithms ****/
static DecompTransform* decompNot(DecompManager dm, BddClass *bddc);
static DecompTransform* decompMux(DecompManager dm, BddClass *bddc);
static DecompTransform* decompAlgebraic(DecompManager dm, BddClass *bddc);
static DecompTransform* mux2VarDecompose(DecompManager dm, BddClass *bddc);
DecompTransform*        decompBoolean(DecompManager dm, BddClass *bddc);
// DecompTransform*        decompBoolXnor(DecompManager dm, BddClass *bddc);
DecompTransform*        decompNestedBdd(DecompManager dm, BddClass *bddc);
DecompTransform*        decompBddDom(DecompManager dm, BddClass *bddc);
DecompTransform*        decompB2vd(DecompManager dm, BddClass *bddc);
void                    totalMuxDecompose(DecompManager dm, BddClass *bddc);

void boolDecompPrint();

void
decompCBAssign(IDecompCallBack cb,
	       void *objCallee,
	       void (*beforeRemoveBddi)( void *obj, BddInstance *bddi ),
	       void (*afterAddBddi)( void *obj, BddInstance *bddi ),
	       void (*afterDecomposeBddc)( void *obj, void *data ))
{
  cb->objCallee = objCallee;
  cb->beforeRemoveBddi = beforeRemoveBddi;
  cb->afterAddBddi = afterAddBddi;  
  cb->afterDecomposeBddc = afterDecomposeBddc;
}

void
decompAddBddc(DecompManager dm, BddClass *bddc)
{
  bchPush(dm->bchm, bddc);
}

void
decompRemBddc(DecompManager dm, BddClass *bddc)
{
  if (bddc->dt != NULL)
    dtDestruct(bddc->dt);
  bchRemove(dm->bchm, bddc);
}

/* 
   func: decomposeBddc()
   description: Applies a set of decomposition algorithms on bddc 
   and returns the decomposition that results in the best area.  
   bddc->dt is assigned this return value.
*/
static DecompTransform *
decomposeBddc(DecompManager dm, BddClass *bddc)
{
  ListCell cDecompAlgorithm;
  decompType dType = DT_NONE;
  DecompTransform *dtBest;
  BddProperties bpOrig;
  DecompTransform* (*decompAlgBest)(DecompManager dm, BddClass *bddc);
  
  bddGetProperties(bddc->bdd, &bpOrig);

  if (bddc->dt != NULL)
    return bddc->dt; // already computed

  dtBest = NULL;
  listForeach(&dm->lDecompAlgs, cDecompAlgorithm){
    DecompTransform *dt;

    DecompTransform* (*decompAlgorithm)(DecompManager dm, BddClass *bddc) 
      = listData(void*, cDecompAlgorithm);
    
    if ((decompAlgorithm==decompMux || decompAlgorithm==decompNot)
	&& dtBest != NULL)
      // use mux decomposition as a last resort.
      dt = NULL;
    else
      dt = (*decompAlgorithm)(dm, bddc);

    if (dt!=NULL 
	&& (bddc->numVars<=3 
	    || decompAlgorithm == decompB2vd
	    || dtCostValid(&bpOrig, dt)
	    || (dtBest==NULL && (decompAlgorithm==decompNot 
				 || decompAlgorithm==decompMux)))){
      // dt != NULL
      if (dtBest==NULL){
	dtBest = dt;
	decompAlgBest = decompAlgorithm;
	if (decompAlgBest == decompAlgebraic)
	  break;
      }
      else if(dtBest!=NULL && (dtCostCmp(&bpOrig, dt, dtBest) > 0 )){
	dtDestruct(dtBest);
	dtBest = dt;
	decompAlgBest = decompAlgorithm;
	if (decompAlgBest == decompAlgebraic)
	  break;
      }
      else{
	if (dt != NULL)
	  dtDestruct(dt);
      }
    }
    else{
      if (dt != NULL)
	dtDestruct(dt);
    }
  }

  if (decompAlgBest == decompAlgebraic){
    dType = DT_ALG;
    strcpy(dtBest->dType, "alg");
  }
  else if (decompAlgBest == decompBoolean){
    dType = DT_BOOL;
    strcpy(dtBest->dType, "bool");
  }
  else if (decompAlgBest == decompNestedBdd){
    dType = DT_NESTED_BDD;
    strcpy(dtBest->dType, "nest");
  }
  else if (decompAlgBest == decompBddDom){
    dType = DT_BDD_DOM;
    strcpy(dtBest->dType, "bdd");
  }
  else if (decompAlgBest == decompMux){
    dType = DT_MUX;
    strcpy(dtBest->dType, "mux");
  }
  else if (decompAlgBest == decompNot){
    dType = DT_NOT;    
    strcpy(dtBest->dType, "not");
  }
  /*
  else if (decompAlgBest == decompBoolXnor){
    dType = DT_BXNOR;    
    strcpy(dtBest->dType, "bxnor");
  }
  */
  else if (decompAlgBest == decompB2vd){
    dType = DT_B2VD;    
    strcpy(dtBest->dType, "b2vd");
  }
  else{
    assert(FALSE);
  }

  _numFoldedDecomps[dType]++;
  _numDecomps[dType] += bddcNumInsts(bddc);

  bddc->dt = dtBest;
  return dtBest;
}

/*
  FUNC: decompMux()
  DESC: Performs variable mux decomposition.
*/
//char 
static DecompTransform *
decompMux(DecompManager dm, BddClass *bddc)
{
  // bddcMux: the mux bdd
  // bddcA: input 1 of the mux
  // bddcB: input 0 of the mux

  Model *model = dm->fm->model;
  clock_t t;

  DdNode *nAnd;
  DdNode *bddGate;
  BddClass *bddcGate;
  DdManager *ddManager = model->ddManager;

  DecompTransform *dt;
  int *newIndexToOld;
  int numGate;

  if (Cudd_IsComplement(bddc->bdd))
    return NULL;

  t = util_cpu_time();

  // if one cofactor contains the other, then the mux gate can be simplified.
  nAnd = Cudd_bddAnd(ddManager, Cudd_T(bddc->bdd), Cudd_E(bddc->bdd));
  Cudd_Ref(nAnd);
  if ( nAnd == Cudd_T(bddc->bdd)){
    // Cudd_E contains Cudd_T.
    // Cudd_T => Cudd_E 
    // replace mux with (select)'(else) + (then).
    DdNode *f, *a, *b, *s;
    
    // index 2: select signal
    // index 1: signal 'a'
    // index 0: signal 'b'
    // f = sa + s'b
    
    b = Cudd_bddIthVar(ddManager, 0);
    Cudd_Ref(b);
    
    a = Cudd_bddIthVar(ddManager, 1);
    Cudd_Ref(a);
    
    s = Cudd_bddIthVar(ddManager, 2);
    Cudd_Ref(s);
    s = Cudd_Complement(s);
    
    f = Cudd_bddAnd(ddManager, s, b);
    Cudd_Ref(f);
    Cudd_RecursiveDeref(ddManager, b);
    Cudd_RecursiveDeref(ddManager, s);
    
    s = Cudd_bddOr(ddManager, f, a);
    Cudd_Ref(s);
    Cudd_RecursiveDeref(ddManager, f);
    Cudd_RecursiveDeref(ddManager, a);
    
    bddGate = s;
  }
  else if ( nAnd == Cudd_E(bddc->bdd))
  {
    // Cudd_T contains Cudd_E.
    // Cudd_E => Cudd_T 
    // replace mux with (else) + (select)(then).
    
    DdNode *f, *a, *b, *s;
    
    // index 2: select signal
    // index 1: signal 'a'
    // index 0: signal 'b'
    // f = sa + s'b
    
    b = Cudd_bddIthVar(ddManager, 0);
    Cudd_Ref(b);
    
    a = Cudd_bddIthVar(ddManager, 1);
    Cudd_Ref(a);
    
    s = Cudd_bddIthVar(ddManager, 2);
    Cudd_Ref(s);
    
    f = Cudd_bddAnd(ddManager, s, a);
    Cudd_Ref(f);
    Cudd_RecursiveDeref(ddManager, a);
    Cudd_RecursiveDeref(ddManager, s);
    
    s = Cudd_bddOr(ddManager, f, b);
    Cudd_Ref(s);
    Cudd_RecursiveDeref(ddManager, f);
    Cudd_RecursiveDeref(ddManager, b);
    
    bddGate = s;
  }
  else{
    // For variables in nAnd not shared by both Then and Else child,
    // set the variable to '0'.

    bddGate = model->bddcMux->bdd;
    Cudd_Ref(bddGate);
    bddcGate = model->bddcMux;
  }
  Cudd_RecursiveDeref(ddManager, nAnd);

  dt = dtConstruct(ddManager);

  // add mux gate to decomposition
  numGate = -1;
  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*3);
  // b signal
  if (Cudd_E(bddc->bdd) != model->nodeSF){
    newIndexToOld[0] = numGate; 
    numGate--;
  }
  else{
    newIndexToOld[0] = Cudd_NodeReadIndex(Cudd_E(bddc->bdd));
  }
  // a signal
  if (Cudd_T(bddc->bdd) != model->nodeSF){
    newIndexToOld[1] = numGate; 
    numGate--;
  }
  else{
    newIndexToOld[1] = Cudd_NodeReadIndex(Cudd_T(bddc->bdd));
  }
  newIndexToOld[2]=Cudd_NodeReadIndex(bddc->bdd); // select signal
  dtAddGate(dt, newIndexToOld, bddGate, 3, 3, FALSE);

  // should not assert false below, other wise there existed a 1 or 0
  // dominator and an algebraic decomposition should have been performed.

  // Negative Cofactor
  if ( Cudd_E(bddc->bdd) != model->nodeSF){
    Cudd_Ref(Cudd_E(bddc->bdd));
    dtAddMuxInput(model, dt, Cudd_E(bddc->bdd));
  }

  // Positive Cofactor
  if ( Cudd_T(bddc->bdd) != model->nodeSF)
    Cudd_Ref(Cudd_T(bddc->bdd));
    dtAddMuxInput(model, dt, Cudd_T(bddc->bdd));
  
    _tDecompMux += util_cpu_time()-t;

  return dt;
}

/*
  FUNC: mux2VarDecompose()
  DESC: Decompose 2 input mux gates into an interconnection of AND and OR
  gates.  MUX = sa + s'b.
 */
static DecompTransform *
mux2VarDecompose(DecompManager dm, BddClass *bddc)
{
  clock_t t;
  Model *model = dm->fm->model;
  DdManager *ddManager = model->ddManager;

  DdNode *bddOr;

  DecompTransform *dt;
  int *newIndexToOld;

  assert(bddc == model->bddcMux || bddc == model->bddcMux2);

  t = util_cpu_time();
  
  dt = dtConstruct(ddManager);

  bddOr = model->bddcOr->bdd;
  Cudd_Ref(bddOr);

  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*2);
  newIndexToOld[0] = -1;
  newIndexToOld[1] = -2;
  dtAddGate(dt, newIndexToOld, bddOr, 2, 2, FALSE);

  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*2);
  if (bddc==model->bddcMux)
    newIndexToOld[0] = 1;
  else
    newIndexToOld[0] = 0;
  newIndexToOld[1] = 2;
  Cudd_Ref(model->bddcAnd->bdd);
  dtAddGate(dt, newIndexToOld, model->bddcAnd->bdd, 2, 2, FALSE);
 
  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*2);
  if (bddc==model->bddcMux)
    newIndexToOld[0] = 0;
  else
    newIndexToOld[0] = 1;
  newIndexToOld[1] = 2;
  Cudd_Ref(model->nodeAND10);
  dtAddGate(dt, newIndexToOld, model->nodeAND10, 2, 2, FALSE);
  
  _tDecompMux += util_cpu_time()-t;

  bddc->dt = dt;
  return dt;
}

/*
  FUNC: decompNot()
  DESC: Extract a NOT gate out the top of bddc.
 */
static DecompTransform *
decompNot(DecompManager dm, BddClass *bddc)
{
  Model *model = dm->fm->model;
  DdNode *nodeA;
  int i;
  DdManager *ddManager = model->ddManager;
  DecompTransform *dt;
  int *newIndexToOld;

  if (!Cudd_IsComplement(bddc->bdd))
    return FALSE;

  nodeA = Cudd_Regular(bddc->bdd);
  Cudd_Ref(nodeA);

  dt = dtConstruct(ddManager);
  
  // construct dtg for NOT gate
  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*1);
  newIndexToOld[0] = -1;
  Cudd_Ref(model->nodeNOT);
  dtAddGate(dt, newIndexToOld, model->nodeNOT, 1, 1, FALSE);

  // construct dtg for NOT input
  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*bddc->numVars);
  for (i=0; i<bddc->numVars; i++)
    newIndexToOld[i] = i;
  dtAddGate(dt, newIndexToOld, nodeA, bddc->numVars, bddc->numNodes, FALSE);

  return dt;
}

/*
  FUNC: decompAlg()
  DESC: Performs disjunctive bi-decomposition where the operand is either
  an AND, OR or XNOR gate.
*/
static DecompTransform*
decompAlgebraic(DecompManager dm, BddClass *bddc)
{
  // Search for variable (level) where all paths must go through.
  // Check if this level corresponds to an algebraic AND, OR or XNOR dominator
  Model *model = dm->fm->model;
  DdNode *dominator = NULL; // the dominator closest to the centre of the bdd
  int domLevel;
  decompType domDecompType;

  int optLevel; // optimal level to perform decomp
  int rootLevel;
  int searchDepth; // the level of the lowest variable on the search wavefront.

  int numNodesFoundAtSearchDepth;
  DdNode *searchDepthFirstNode = NULL;

  char extPathOne = FALSE; // TRUE if an external path to one exists
  char extPathZero = FALSE; // TRUE if an external path to zero exists

  //  char andOrDecompNotPossible;
  //  char foundComplementEdge; /* found compliment edge pointing to node 
  //			       at search Depth */
  DdManager *ddManager = model->ddManager;

  DecompTransform *dt;
  int nodeCount;

  Nml nml; // node mark lookup
  DdNode *nChild;
  int i;

  static struct _Heap heapNodeLevel;  // nodes sorted by level.
  HeapCell heapCell;

  clock_t t;

  if ( bddc->bdd == NULL || Cudd_IsConstant(bddc->bdd) )
    return NULL;

  t = util_cpu_time();

  // search depth starts at one level below from the root
  rootLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddc->bdd));
  searchDepth = rootLevel + 1;
  optLevel = searchDepth-1 + bddc->numVars/2;

  numNodesFoundAtSearchDepth = 0;

  heapConstruct(&heapNodeLevel, _fbddHeap, 100);
  heapCell = 
    heapNewCell(&heapNodeLevel, 
		Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddc->bdd)), 
		Cudd_Regular(bddc->bdd));
  heapInsertElement(&heapNodeLevel, heapCell);

  nmlConstruct(&nml);
  while(// still more levels left
	searchDepth<=rootLevel + bddc->numVars-1 
	
	// not cannot find a dominator closer to the optimum level
	&& !(dominator!=NULL && searchDepth>optLevel 
	     && (abs(domLevel - optLevel) < (searchDepth-optLevel)))
	
	// no path to zero and path to one
	&& !(extPathZero && extPathOne)
	
	// frontier not empty
	&& !heapIsEmpty(&heapNodeLevel)
	){
    DdNode *node;    
    char negated;
    int nodeLevel;
    
    // pop node closest to the root in the bdd.
    heapCell = heapDelMin(&heapNodeLevel);
    node = heapData(DdNode*, heapCell);
    comFree(heapMem(&heapNodeLevel), heapCell);

    if (node == NULL)
      break;

    if (Cudd_IsComplement(node))
      negated = TRUE;
    else
      negated = FALSE;
    nodeLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(node));

    if (nodeLevel == searchDepth){
      decompType tempDecompType = DT_NONE;
      // The current node being popped is at searchDepth.  This implies
      // that all other nodes in the heap are also at searchDepth.  Otherwise,
      // if a node in the heap is lower than searchDepth, then searchDepth 
      // would have been set to the level of that node.  If a node in the heap
      // is higher, then it would have been popped instead of node.

      if ( numNodesFoundAtSearchDepth == 1){
	
	if (extPathOne == FALSE && extPathZero == TRUE)
	  tempDecompType = DT_AND;
	else if (extPathOne == TRUE && extPathZero == FALSE)
	  tempDecompType = DT_OR;
	else
	  assert(FALSE);
	
      }
      else if(numNodesFoundAtSearchDepth == 2
	      && (extPathOne == FALSE && extPathZero == FALSE)
	      && searchDepthFirstNode != NULL){
	// found an XOR gate
	assert(searchDepthFirstNode == Cudd_Regular(node));
	tempDecompType = DT_XNOR;
      }

      if (tempDecompType != DT_NONE){
	// found a dominator
	dominator = node;
	domDecompType = tempDecompType;
	domLevel = searchDepth;
	
	if (domDecompType==DT_XNOR && Cudd_IsComplement(dominator))
	  dominator = Cudd_Regular(dominator);
      }
    }

    // Process Children
    nChild = Cudd_T(node);
    for (i=0; i<2; i++){
      int childNegated;

      if ((Cudd_IsComplement(nChild) && negated)
	  || (!Cudd_IsComplement(nChild) && !negated))
	childNegated = FALSE;
      else
	childNegated = TRUE;
      
      if (childNegated)
	nChild = Cudd_Complement(nChild);
      else
	nChild = Cudd_Regular(nChild);
      
      if (Cudd_IsConstant(nChild)){
	if (Cudd_IsComplement(nChild))
	  extPathZero = TRUE;
	else
	  extPathOne = TRUE;
      }
      else{
	int nChildLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(nChild));

	if (nmlGetMark(&nml, nChild) == 0){
	  // First time visited node.

	  if (nChildLevel == searchDepth){
	    numNodesFoundAtSearchDepth++;
	    if (Cudd_Regular(nChild) != searchDepthFirstNode)
	      searchDepthFirstNode = NULL;
	  }
	  else if (nChildLevel > searchDepth){
	    searchDepth = nChildLevel;
	    
	    // reset variables for new searchDepth
	    numNodesFoundAtSearchDepth = 1;
	    searchDepthFirstNode = Cudd_Regular(nChild);
	  }
	  else{
	    // nChildLevel above searchDevelop
	  }
	  
	  // set new mark
	  nmlPutMark(&nml, nChild, 1);
	  heapCell = 
	    heapNewCell(&heapNodeLevel, 
			Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(nChild)), 
			nChild);
	  heapInsertElement(&heapNodeLevel, heapCell);
	}
      }
      nChild = Cudd_E(node);
    }
    nmlRemove(&nml, Cudd_Regular(node));
  }
  
  // Destroy Heap
  while (!heapIsEmpty(&heapNodeLevel)){
    HeapCell heapCell = heapDelMin(&heapNodeLevel);
    comFree(heapMem(&heapNodeLevel), heapCell);
  }
  heapDelAll(&heapNodeLevel);

  nmlDestruct(&nml);

  if (dominator == NULL)
    return NULL;

  dt = dtConstruct(ddManager);

  { 
    BddClass *bddcGate;
    DdNode *nodeA, *nodeB, *nodeTmp;
    int numVarsA, numVarsB;
    int *oldIndexToNewA, *oldIndexToNewB;
    int arraySizeA, arraySizeB; // the size of oldIndexToNewA 
    int j;
    int *newIndexToOld;

    if (domDecompType == DT_XNOR)
      nodeA = Cudd_Regular(dominator);
    else
      nodeA = dominator;
    Cudd_Ref(nodeA);

    numVarsA = Cudd_NodeReadIndex(nodeA)+1;
    oldIndexToNewA = (int*)comAlloc(_fbddHeap,sizeof(int)*numVarsA);
    for (j=0; j<numVarsA; j++)
      oldIndexToNewA[j]=j;
    arraySizeA = numVarsA;

    numVarsB = bddc->numVars - numVarsA;
    oldIndexToNewB = (int*)comAlloc(_fbddHeap,sizeof(int)*bddc->numVars);
    for (j=0; j<numVarsA; j++)
      oldIndexToNewB[j] = -1;
    for (j=numVarsA; j<bddc->numVars; j++)
      oldIndexToNewB[j] = j-numVarsA;
    arraySizeB = bddc->numVars;

    if (!Cudd_IsComplement(bddc->bdd)){
      switch( domDecompType ){
      case DT_AND: 
	bddcGate = model->bddcAnd;

	comFree(_fbddHeap,oldIndexToNewB);
	nodeTmp = bddCopyAndReIndexWithCut(
                    ddManager, bddc->bdd, ddManager, 
		    &numVarsB, &oldIndexToNewB, &arraySizeB,
		    CT_REPLACE_DANGLING_EDGES, Cudd_ReadOne(ddManager),
		    domLevel-1);
	comFree(_fbddHeap,oldIndexToNewB);
	nodeB = bddCopyAndReIndex(ddManager, nodeTmp, ddManager, 
				 &numVarsB, &oldIndexToNewB, &arraySizeB,
				 NULL, NULL);
	Cudd_RecursiveDeref(ddManager, nodeTmp);

	break;
      case DT_OR:
	bddcGate = model->bddcOr;

	comFree(_fbddHeap,oldIndexToNewB);
	nodeTmp = bddCopyAndReIndexWithCut(
                    ddManager, bddc->bdd, ddManager, 
		    &numVarsB, &oldIndexToNewB, &arraySizeB,
		    CT_REPLACE_DANGLING_EDGES, Cudd_ReadLogicZero(ddManager),
		    domLevel-1);
	comFree(_fbddHeap,oldIndexToNewB);
	nodeB = bddCopyAndReIndex(ddManager, nodeTmp, ddManager, 
				 &numVarsB, &oldIndexToNewB, &arraySizeB,
				 NULL, NULL);
	Cudd_RecursiveDeref(ddManager, nodeTmp);

	break;
      case DT_XNOR:
	nodeB = bddCopy(ddManager, bddc->bdd, ddManager, oldIndexToNewB,
			dominator, Cudd_ReadOne(ddManager));
	bddcGate = model->bddcXnor;
	break;
      default:
	assert(FALSE);
      }
    }
    else{
      switch( domDecompType ){
      case DT_AND: 
	bddcGate = model->bddcNand;
	comFree(_fbddHeap,oldIndexToNewB);
	nodeTmp = bddCopyAndReIndexWithCut(
                    ddManager, Cudd_Regular(bddc->bdd), ddManager, 
		    &numVarsB, &oldIndexToNewB, &arraySizeB,
		    CT_REPLACE_DANGLING_EDGES, Cudd_ReadOne(ddManager),
		    domLevel-1);
	comFree(_fbddHeap,oldIndexToNewB);
	nodeB = bddCopyAndReIndex(ddManager, nodeTmp, ddManager, 
				 &numVarsB, &oldIndexToNewB, &arraySizeB,
				 NULL, NULL);
	Cudd_RecursiveDeref(ddManager, nodeTmp);


	break;
      case DT_OR:
	bddcGate = model->bddcNor;
	comFree(_fbddHeap,oldIndexToNewB);
	nodeTmp = bddCopyAndReIndexWithCut(
                    ddManager, Cudd_Regular(bddc->bdd), ddManager, 
		    &numVarsB, &oldIndexToNewB, &arraySizeB,
		    CT_REPLACE_DANGLING_EDGES, Cudd_ReadLogicZero(ddManager),
		    domLevel-1);
	comFree(_fbddHeap,oldIndexToNewB);
	nodeB = bddCopyAndReIndex(ddManager, nodeTmp, ddManager, 
				 &numVarsB, &oldIndexToNewB, &arraySizeB,
				 NULL, NULL);
	Cudd_RecursiveDeref(ddManager, nodeTmp);

	break;
      case DT_XNOR:
	nodeB = bddCopy(ddManager, Cudd_Regular(bddc->bdd), ddManager, 
			oldIndexToNewB,
			dominator, Cudd_ReadOne(ddManager));
	bddcGate = model->bddcXor;
	break;
      default:
	assert(FALSE);
      }
    }

    assert (numVarsB == bddc->numVars - numVarsA);
    assert (arraySizeB == bddc->numVars);
  
    // construct dtg for root gate.
    {
      int gateNum = -1;
      Cudd_Ref(bddcGate->bdd);
      newIndexToOld = 
	(int*)comAlloc(_fbddHeap, sizeof(int)*2);
      if (nodeA != model->nodeSF){
	newIndexToOld[0] = gateNum; // input A
	gateNum--;
      }
      else{
	Cudd_RecursiveDeref(ddManager, nodeA);
	newIndexToOld[0] = getOldIndexFromNew(oldIndexToNewA, arraySizeA, 0);
      }
      if (nodeB != model->nodeSF){
	newIndexToOld[1] = gateNum; // input B
	gateNum--;
      }
      else{
	Cudd_RecursiveDeref(ddManager, nodeB);
	newIndexToOld[1] = getOldIndexFromNew(oldIndexToNewB, arraySizeB, 0);
      }
      dtAddGate(dt, newIndexToOld, bddcGate->bdd, 2, 2, FALSE);
    }

    assert(Cudd_IsConstant(nodeA) == FALSE);
    assert(Cudd_IsConstant(nodeB) == FALSE);
    

    if (nodeA != model->nodeSF){
      // construct dtg for nodeA
      newIndexToOld = 
	(int*)comAlloc(_fbddHeap, sizeof(int)*numVarsA);
      //	dtCreateNewIndexToOld(numVarsA);

#ifdef DEBUG_COMPILE
      for (j=0; j<numVarsA; j++)
	newIndexToOld[j] = -1;
#endif
      for (j=0; j<arraySizeA; j++){
	assert(oldIndexToNewA[j] != -1);
	newIndexToOld[oldIndexToNewA[j]]=j;
      }
#ifdef DEBUG_COMPILE
      for (j=0; j<numVarsA; j++)
	assert(newIndexToOld[j] != -1);
#endif
      nodeCount = Cudd_DagSize(nodeA)-1;
      dtAddGate(dt, newIndexToOld, nodeA, numVarsA, nodeCount, TRUE);
    }

    // construct dtg for nodeB
    if (nodeB != model->nodeSF){
      newIndexToOld = 
	(int*)comAlloc(_fbddHeap, sizeof(int)*numVarsB);
	// dtCreateNewIndexToOld(numVarsB);
#ifdef DEBUG_COMPILE
      for (j=0; j<numVarsB; j++)
	newIndexToOld[j] = -1;
#endif
      for (j=0; j<arraySizeB; j++){
	if (oldIndexToNewB[j] != -1)
	  newIndexToOld[oldIndexToNewB[j]]=j;
      }
#ifdef DEBUG_COMPILE
      for (j=0; j<numVarsB; j++)
	assert(newIndexToOld[j] != -1);
#endif
      nodeCount = Cudd_DagSize(nodeB)-1;
      dtAddGate(dt, newIndexToOld, nodeB, numVarsB, nodeCount, TRUE);
    }

    comFree(_fbddHeap,oldIndexToNewA);
    comFree(_fbddHeap,oldIndexToNewB);
  }  
  
  _tDecompAlg += util_cpu_time()-t;

  return dt;
}

/*
  FUNC: decompNewManager()
  DESC: Initialize heap (bchHeap) with BddClasses that need to be 
  decomposed.  Must be called before decomposeModel() and decomposeEnd().
 */
DecompManager 
decompNewManager(FbddManager fm)
{
  DecompManager dm = 
    (DecompManager)comAlloc(_fbddHeap, sizeof(struct _DecompManager));
  Model *model = fm->model;
  IMem mem;
  struct _List lBddcs;
  ListCell cBddc;
  
  int i;
  
  dm->fm = fm;
  assert(fm->dm == NULL);
  fm->dm = dm;

  listConstruct0(&dm->lCBs, _fbddHeap);
  fbddCBAssign(&dm->cbFbdd, dm,
	       (void (*)( void *obj, BddClass *bddc ))decompAddBddc,
	       (void (*)( void *obj, BddClass *bddc ))decompRemBddc);
  cbAdd(dm->fm, dm->cFbddCallBack, &dm->cbFbdd);
  
  for (i=0; i<NUM_DECOMP_TYPES; i++){
    _numDecomps[i] = 0;
    _numFoldedDecomps[i] = 0;
  }
  
  dm->bchm = bchNewManager(model);

  // put starting bddcs in heap
  mem = _fbddHeap;
  listConstruct0(&lBddcs, mem);
  modelBddcList(mem, model, &lBddcs);
  
  listForeach(&lBddcs, cBddc){
    BddClass *bddc = listData(BddClass*, cBddc);
    
    if (bddc->type == INTERNAL_VARIABLE){
      if (bclGet(model->bclm, bddc->bdd) == NULL)
	bclPut(model->bclm, bddc->bdd, bddc);
      
      bchPush(dm->bchm, bddc);
    }
  }
  listDelAll(&lBddcs);
  
  // Initialize List of Decomposition Algorithms
  listConstruct0(&dm->lDecompAlgs, _fbddHeap);
  listAppend(&dm->lDecompAlgs, decompAlgebraic);
  listAppend(&dm->lDecompAlgs, decompBoolean);
  listAppend(&dm->lDecompAlgs, decompBddDom);
  listAppend(&dm->lDecompAlgs, decompNestedBdd);
  listAppend(&dm->lDecompAlgs, decompB2vd);
  listAppend(&dm->lDecompAlgs, decompMux);
  listAppend(&dm->lDecompAlgs, decompNot);
  //  listAppend(&dm->lDecompAlgs, decompBoolXnor);

  return dm;
}

/*
  FUNC: decompDelManager()
  DESC: 
 */
void
decompDelManager(DecompManager dm)
{
  ListCell cell;
  Model *model = dm->fm->model;

  bchDelManager(dm->bchm);
  cbRemove(dm->fm, dm->cFbddCallBack);

  listDelAll(&dm->lDecompAlgs);

  listForeach(&(model->lOutputs), cell)
    bddiCollapseGates(_fbddHeap, 
		      listData(BddInstance*,cell), model->bddcSF);
  
  listDelAll(&dm->lDecompAlgs);

  dm->fm->dm = NULL;
  comFree(_fbddHeap, dm);
}

/* FUNC: decompRun()
   DESC: Break circuit down into basic gates.
*/
void
decompRun(DecompManager dm)
{
  Model *model = dm->fm->model;
  long t;
  
  t = util_cpu_time();

  while( bchEmpty(dm->bchm)==FALSE ){
    DecompTransform *dt;
    BddClass *bddc;
    BddProperties bpOrig;
    
    int numInsts;
    
    // get heap with largest number of variables
    bddc = bchPop(dm->bchm);
    if (bddc == NULL)
      break;
    assert(Cudd_Regular(bddc->bdd)->ref > 0);
    
    if (bddcNumInsts(bddc)==0){
      assert(FALSE);
    }
    
    bddGetProperties(bddc->bdd, &bpOrig);
    
    numInsts = bddcNumInsts(bddc);
    
    _totalDecomps += bddcNumInsts(bddc);
    _totalFoldedDecomps++;
    
    dt = decomposeBddc(dm, bddc);
    
    assert(dt != NULL);
    
    // If decomposition algorithms in lDecompAlgs perform poorly,
    // then decompose the bddc into mux gates using the BDD structure.
    if (dt==NULL || !dtCostValid(&bpOrig, dt)){
      totalMuxDecompose(dm, bddc);
    }
    else{
      dtCommit(dm, bddc);
    }
    
    // decompose mux gates
    if (bddcNumInsts(model->bddcMux)>0){
      mux2VarDecompose(dm, model->bddcMux);
      dtCommit(dm, model->bddcMux);
    }
    if (bddcNumInsts(model->bddcMux2)>0){
      mux2VarDecompose(dm, model->bddcMux2);
      dtCommit(dm, model->bddcMux2);
    }

    _tDecomp += util_cpu_time()-t;
    
    // If activated, perform sharing extraction.
    cbRun(IDecompCallBack, dm, afterDecomposeBddc, 0);
    
    t = util_cpu_time();
  }
  _tDecomp += util_cpu_time()-t;
}

void
decompDump()
{
  printf("Decomposition\n");
  printf("number of folded decompositions:   %5d\n", _totalFoldedDecomps);
  printf("number of regular decompositions:  %5d\n", _totalDecomps);
  printf("\n");
}

clock_t
decompGetTime()
{
  return _tDecomp;
}
