/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
  FILE: decompBool.c
  DESC: 
*/

#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"
#include "decomp.h"

typedef enum{DT_AND, DT_OR} decompType;

typedef struct _BddEdge
{
  DdNode *node;
  char pos; // 1 for if, 2 for else
} BddEdge; 

extern clock_t _tDecompBool;
static int _numBoolDisjunct=0;

static BddEdge *
beCreate(DdNode *node, char pos)
{
  BddEdge *be;
  be = (BddEdge*)comAlloc(_fbddHeap, sizeof(BddEdge));
  be->node = node;
  be->pos = pos;

  return be;
}

static void
getHorzCutAfterLevel(DdManager *ddManager, DdNode *bdd, int level, 
	      List lNodeEdges, List lOneEdges, List lZeroEdges)
{
  DdNode *nCurr;
  int topLevelOffset;
  struct _LevelFirstSearch lfs;

  topLevelOffset = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bdd));
  lfsConstruct(&lfs, ddManager, bdd);

  while ((nCurr=lfsNextNode(&lfs)) != NULL){
    int i;
    char pos;
    int nodeLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(nCurr));

    if (Cudd_IsConstant(nCurr) || nodeLevel-topLevelOffset > level)
      break;

    assert(nCurr == Cudd_Regular(nCurr));
    
    for (i=0; i<2; i++){
      DdNode *nChild;
      if (i==0){
	nChild = Cudd_T(nCurr);
	pos = THEN_EDGE;
      }
      else{
	nChild = Cudd_E(nCurr);
	pos = ELSE_EDGE;
      }

      if (Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(nChild))-topLevelOffset
	  > level || Cudd_IsConstant(nChild)){
	// edge points to node below level
	BddEdge *be = beCreate(nCurr, pos);
	if (Cudd_IsConstant(nChild)){
	  if (Cudd_IsComplement(nChild)){
	    listAppend(lZeroEdges, be);
	  }
	  else{
	    listAppend(lOneEdges, be);
	  }
	}
	else{
	  // edge points to a node
	  listAppend(lNodeEdges, be);
	}
      }
    }
  }
  lfsDestruct(&lfs);
}

DecompTransform *
decompBoolean(DecompManager dm, BddClass *bddc)
{
  Model *model = dm->fm->model;
  const int MAX_VARS = 99999;
  int numVars = bddc->numVars;
  int level;

  char canDecompose = FALSE;

  int bestLevel;
  char bestDecompType;

  DdNode *bestBddA=NULL;
  int *bestOldIndexToNewA=NULL;
  int bestArraySizeA;
  int bestNumVarsA;

  int topLevel;
  DdNode *bestBddB=NULL;
  int *bestOldIndexToNewB=NULL;
  int bestArraySizeB;
  int bestNumVarsB;

  int i;

  DdManager *tmpManager;
  DdManager *gblManager = model->ddManager;

  clock_t t;

  BddProperties bpBestA, bpBestB, bpBestC;
  //  BddProperties bpMuxA, bpMuxB, bpMux, 
  BddProperties bpA, bpB, bpAndOr;
  BddProperties bpOrigBdd;

  DecompTransform *dt;

  t = util_cpu_time();

  bpAndOr.nodeCount_withComplement = 2;
  bpAndOr.nodeCount_noComplement = 2;
  bpAndOr.numVars = 2;

  // get bdd properties of original bdd
  bddGetProperties(bddc->bdd, &bpOrigBdd);

  tmpManager = getDdManager(&model->manArrayDecompose, bddc->numVars);

  for (i=0; i<bddc->numVars; i++)
    Cudd_bddIthVar(tmpManager, i);
  
  topLevel = Cudd_ReadPerm(gblManager, Cudd_NodeReadIndex(bddc->bdd));
  
  // initialize best to mux
  bpBestA.nodeCount_withComplement = MAX_VARS;
  bpBestA.nodeCount_noComplement = MAX_VARS;
  bpBestA.numVars = MAX_VARS;
  bpBestB.nodeCount_withComplement = MAX_VARS;
  bpBestB.nodeCount_noComplement = MAX_VARS;
  bpBestB.numVars = MAX_VARS;
  bpBestC.nodeCount_withComplement = MAX_VARS;
  bpBestC.nodeCount_noComplement = MAX_VARS;
  bpBestC.numVars = MAX_VARS;

  for (level=0; level<numVars-1; level++){
    DdNode *bddA, *bddB;
    int *oldIndexToNewA, *oldIndexToNewB;
    int numVarsA, numVarsB;
    int arraySizeA, arraySizeB;
    int j;

    for (j=0; j<2; j++){ // try AND and OR decomposition
      char cutType;

      DdNode *nBddAReplaceWith;
      if (j==0){
	// AND decomp
	nBddAReplaceWith = model->bddcLogicOne->bdd;
	cutType = CT_REPLACE_ZERO_EDGES;
      }
      else{
	// OR decomp
	nBddAReplaceWith = model->bddcLogicZero->bdd;
	cutType = CT_REPLACE_ONE_EDGES;
      }

      // Computing cost of bddA
      bddA = bddCopyAndReIndexWithCut(gblManager, bddc->bdd, tmpManager, 
				   &numVarsA, &oldIndexToNewA, &arraySizeA,
				   CT_REPLACE_DANGLING_EDGES, nBddAReplaceWith,
				   level+topLevel);
      bddGetProperties(bddA, &bpA);

      // Computing cost of bddB
      bddB = bddCopyAndReIndexWithCut(gblManager, bddc->bdd, tmpManager, 
				      &numVarsB, &oldIndexToNewB, &arraySizeB,
				      cutType, NULL, 
				      level+topLevel);
      bddGetProperties(bddB, &bpB);

      if (Cudd_IsConstant(bddA)==FALSE && Cudd_IsConstant(bddB)==FALSE 
	  && bddCmpCost(&bpOrigBdd, &bpA, &bpB, 
			&bpAndOr, &bpBestA, &bpBestB, &bpBestC) < 0
	  ){

	if (bestBddA != NULL){
	  // free old resources
	  comFree(_fbddHeap,bestOldIndexToNewA);
	  comFree(_fbddHeap,bestOldIndexToNewB);
	  Cudd_RecursiveDeref(tmpManager, bestBddA);
	  Cudd_RecursiveDeref(tmpManager, bestBddB);
	}
	canDecompose = TRUE;
	bestLevel = level;
	bestDecompType = (j==0?DT_AND:DT_OR);
	bestOldIndexToNewA = oldIndexToNewA;
	bestOldIndexToNewB = oldIndexToNewB;
	bestBddA = bddA;
	bestBddB = bddB;
	bestArraySizeA = arraySizeA;
	bestArraySizeB = arraySizeB;      
	bestNumVarsA = numVarsA;
	bestNumVarsB = numVarsB;
      }
      else{
	comFree(_fbddHeap,oldIndexToNewA);
	comFree(_fbddHeap,oldIndexToNewB);
	Cudd_RecursiveDeref(tmpManager, bddA);
	Cudd_RecursiveDeref(tmpManager, bddB);
      }
    }
  }
  
  if (canDecompose == FALSE){
    if (canDecompose == TRUE){
      comFree(_fbddHeap,bestOldIndexToNewA);
      comFree(_fbddHeap,bestOldIndexToNewB);
      Cudd_RecursiveDeref(tmpManager, bestBddA);
      Cudd_RecursiveDeref(tmpManager, bestBddB);
    }
    dt = NULL;
  }
  else{
    int *oldIndexToNewA2, *oldIndexToNewB2;
    int numVarsA2, numVarsB2;
    int arraySizeA2, arraySizeB2;
    int numVarsTmp, arraySizeTmp, *oldIndexToNewTmp;

    DdNode *bddTmp;
    int i;
    
    int *newIndexToOld;

    dt = dtConstruct(model->ddManager);

    // Construct bddcA
    bddTmp = bddCopyAndReIndex(tmpManager, bestBddA, gblManager, 
			       &numVarsA2, &oldIndexToNewA2, &arraySizeA2,
			       NULL, NULL);
    Cudd_RecursiveDeref(tmpManager, bestBddA);
    bestBddA = foldBddClass(model, bddTmp,
			    &numVarsTmp, &oldIndexToNewTmp, &arraySizeTmp,
			    NULL, NULL);
    Cudd_RecursiveDeref(gblManager, bddTmp);
    mergeOldIndexToNew(oldIndexToNewA2, arraySizeA2, 
		       oldIndexToNewTmp, arraySizeTmp);

    // Construct bddcB
    bddTmp = bddCopyAndReIndex(tmpManager, bestBddB, gblManager, 
			       &numVarsB2, &oldIndexToNewB2, &arraySizeB2,
			       NULL, NULL);
    Cudd_RecursiveDeref(tmpManager, bestBddB);
    bestBddB = foldBddClass(model, bddTmp,
			    &numVarsTmp, &oldIndexToNewTmp, &arraySizeTmp,
			    NULL, NULL);
    Cudd_RecursiveDeref(gblManager, bddTmp);
    mergeOldIndexToNew(oldIndexToNewB2, arraySizeB2, 
		       oldIndexToNewTmp, arraySizeTmp);
    
    // construct bestOldIndexToNewA
    for(i=0; i<bestArraySizeA; i++){
      if (bestOldIndexToNewA[i] != -1 && bestOldIndexToNewA[i]<arraySizeA2 )
	bestOldIndexToNewA[i] = oldIndexToNewA2[bestOldIndexToNewA[i]];
      else
	bestOldIndexToNewA[i] = -1;
      assert(bestOldIndexToNewA[i] < 10000);
    }
    comFree(_fbddHeap,oldIndexToNewA2);

    // construct bestOldIndexToNewB
    for(i=0; i<bestArraySizeB; i++){
      if (bestOldIndexToNewB[i] != -1 && bestOldIndexToNewB[i] < arraySizeB2)
	bestOldIndexToNewB[i] = oldIndexToNewB2[bestOldIndexToNewB[i]];
      else
	bestOldIndexToNewB[i] = -1;
      assert(bestOldIndexToNewB[i] < 10000);
    }
    comFree(_fbddHeap,oldIndexToNewB2);

    // construct dtg for and/or gate
    {
      int numGate = -1;
      newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*2);
      // dtCreateNewIndexToOld(2);
      if (bestBddA != model->bddcSF->bdd){
	newIndexToOld[0] = numGate;
	numGate--;
      }
      else{
	newIndexToOld[0] = 
	  getOldIndexFromNew(bestOldIndexToNewA, bestArraySizeA, 0);
	assert(newIndexToOld[0] != -1);
	Cudd_RecursiveDeref(gblManager, bestBddA);
      }
      if (bestBddB != model->bddcSF->bdd){
	newIndexToOld[1] = numGate;
	numGate--;
      }
      else{
	newIndexToOld[1] = 
	  getOldIndexFromNew(bestOldIndexToNewB, bestArraySizeB, 0);
	assert(newIndexToOld[1] != -1);
	Cudd_RecursiveDeref(gblManager, bestBddB);
      }

      if (bestDecompType == DT_AND){
	Cudd_Ref(model->bddcAnd->bdd);
	dtAddGate(dt, newIndexToOld, model->bddcAnd->bdd, 2, 2, FALSE);
      }
      else{
	Cudd_Ref(model->bddcOr->bdd);
	dtAddGate(dt, newIndexToOld, model->bddcOr->bdd, 2, 2, FALSE);
      }
    }

    // make dtg A
    if (bestBddA != model->bddcSF->bdd){
      newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVarsA2);
#ifdef DEBUG_COMPILE
      for (i=0; i<numVarsA2; i++)
	newIndexToOld[i] = -1;
#endif
      for (i=0; i<bestArraySizeA; i++)
	if (bestOldIndexToNewA[i] != -1) 
	  newIndexToOld[bestOldIndexToNewA[i]] = i;
      dtAddGate(dt, newIndexToOld, bestBddA, numVarsA2, 
		Cudd_DagSize(bestBddA)-1, FALSE);
    }

    if (bestBddB != model->bddcSF->bdd){
      newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVarsB2);
#ifdef DEBUG_COMPILE
      for (i=0; i<numVarsB2; i++)
	newIndexToOld[i] = -1;
#endif
      for (i=0; i<bestArraySizeB; i++)
	if (bestOldIndexToNewB[i] != -1) 
	  newIndexToOld[bestOldIndexToNewB[i]] = i;
      dtAddGate(dt, newIndexToOld, bestBddB, numVarsB2, 
		Cudd_DagSize(bestBddB)-1, FALSE);
    }

    comFree(_fbddHeap,bestOldIndexToNewA);
    comFree(_fbddHeap,bestOldIndexToNewB);
  }

  releaseDdManager(&model->manArrayDecompose);

  _tDecompBool += util_cpu_time() - t;

  return dt;
}

void
boolDecompPrint()
{
  printf("boolNumDisjunct:%d\n", _numBoolDisjunct);
}
