/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"
#include "decomp.h"

#define VAR_PAIR_WINDOW 2

extern clock_t tDecompBool;

DecompTransform *
decompB2vd(DecompManager dm, BddClass *bddc)
{
  Model *model = dm->fm->model;
  DdManager *gblManager = model->ddManager;
  DecompTransform *dt=NULL;
  int i, j;
  int w = VAR_PAIR_WINDOW;

  DdNode *bddRemainderBest = NULL;
  int iBest, jBest;
  extractType etBest;
  int costBest;
  int cost;

  DdNode *bddRemainder;
  int dagSize;
  int dagSizeBest;

  int extractorIndex = Cudd_NodeReadIndex(bddc->bdd)+1;
  int origDagSize = Cudd_DagSize(bddc->bdd);
  
  DdNode *bddExtractor;
  DdNode *nExtractorIndex = Cudd_bddIthVar(gblManager, extractorIndex);

  Cudd_Ref(nExtractorIndex);

  // find best extractor
  for (i=0; i<bddc->numVars; i++){
    int min = comMin(i+w, bddc->numVars-1);
    for (j=i+1; j<min; j++){
      int k;
      DdNode *nI, *nJ;
      nI = Cudd_bddIthVar(gblManager, i);
      Cudd_Ref(nI);
      nJ = Cudd_bddIthVar(gblManager, j);
      Cudd_Ref(nJ);
      for (k=0; k<5; k++){
	DdNode *bddR1, *bddR2;

	// compute extraction
	switch(k){
	case ET_OR: //a+b
	  bddExtractor = Cudd_bddOr(gblManager, nI, nJ);
	  break;
	case ET_AND10: //ab'
	  bddExtractor = Cudd_bddAnd(gblManager, nI, Cudd_Not(nJ));
	  break;
	case ET_AND01: //a'b
	  bddExtractor = Cudd_bddAnd(gblManager, Cudd_Not(nI), nJ);
	  break;
	case ET_AND: //ab
	  bddExtractor = Cudd_bddAnd(gblManager, nI, nJ);      
	  break;
	case ET_XOR:
	  bddExtractor = Cudd_bddXor(gblManager, nI, nJ);
	  break;
	default:
	  assert(FALSE);
	}
	Cudd_Ref(bddExtractor);
	
	bddR1 = Cudd_bddRestrict(gblManager, bddc->bdd, bddExtractor);
	Cudd_Ref(bddR1);
	bddR2 = 
	  Cudd_bddRestrict(gblManager, bddc->bdd, Cudd_Not(bddExtractor));
	Cudd_Ref(bddR2);
	
	bddRemainder = Cudd_bddIte(gblManager, nExtractorIndex, bddR1, bddR2);
	Cudd_Ref(bddRemainder);
	Cudd_RecursiveDeref(gblManager, bddR1);
	Cudd_RecursiveDeref(gblManager, bddR2);
	Cudd_RecursiveDeref(gblManager, bddExtractor);
	  
	dagSize = Cudd_DagSize(bddRemainder);
	cost = dagSize;
	if (k == ET_XOR) // adjust for xor
	  cost++;

	// compare with best
	if (dagSize < origDagSize){
	  if (bddRemainderBest == NULL ||
	      cost < costBest){
	    if (bddRemainderBest != NULL)
	      Cudd_RecursiveDeref(gblManager, bddRemainderBest);
	    bddRemainderBest = bddRemainder;
	    iBest = i;
	    jBest = j;
	    etBest = k;
	    costBest = cost;
	    dagSizeBest = dagSize;
	  }
	  else{
	    Cudd_RecursiveDeref(gblManager, bddRemainder);
	  }
	}
	else{
	  Cudd_RecursiveDeref(gblManager, bddRemainder);
	}
      }
      Cudd_RecursiveDeref(gblManager, nI);
      Cudd_RecursiveDeref(gblManager, nJ);
    }
  }

  Cudd_RecursiveDeref(gblManager, nExtractorIndex);

  if (bddRemainderBest != NULL){
    int *oldIndexToNew;
    int numVars;
    int arraySize;
    int *newIndexToOld;
    DdNode *bddTmp;
    DdNode *nI, *nJ;

    dt = dtConstruct(model->ddManager);

    bddTmp = bddRemainderBest;
    bddRemainderBest = foldBddClass(model, bddTmp,
			    &numVars, &oldIndexToNew, &arraySize,
			    NULL, NULL);
    Cudd_RecursiveDeref(gblManager, bddTmp);
    dagSize = Cudd_DagSize(bddRemainderBest);

    newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVars);
    for (i=0; i<arraySize; i++){
      if (oldIndexToNew[i] != -1){
	//	printf("i:%d oldIndexToNew[i]:%d\n", i, oldIndexToNew[i]);
	newIndexToOld[oldIndexToNew[i]] = i;
      }
    }
    assert(extractorIndex);
    newIndexToOld[oldIndexToNew[extractorIndex]] = -1;
    dtAddGate(dt, newIndexToOld, bddRemainderBest, numVars, 
	      Cudd_DagSize(bddRemainderBest)-1, FALSE);
    
    // construct dtg for extractor
    nI = Cudd_bddIthVar(gblManager, 0);
    Cudd_Ref(nI);
    nJ = Cudd_bddIthVar(gblManager, 1);
    Cudd_Ref(nJ);
    bddExtractor = bddComputeOperation(gblManager, etBest, nI, nJ);
    Cudd_Ref(bddExtractor);
    Cudd_RecursiveDeref(gblManager, nI);
    Cudd_RecursiveDeref(gblManager, nJ);
    
    newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*2);
    newIndexToOld[0] = iBest;
    newIndexToOld[1] = jBest;

    if (etBest == ET_XOR)
      dtAddGate(dt, newIndexToOld, bddExtractor, 
		2, 3, FALSE);
    else
      dtAddGate(dt, newIndexToOld, bddExtractor, 
		2, 2, FALSE);

    comFree(_fbddHeap,oldIndexToNew);

    if (dm->fm->options[opt_verb])
      printf("b2vd success:%d numVars:%d size:%d et:%d\n", 
	     (dt!=NULL), bddc->numVars, 
	     Cudd_DagSize(bddRemainderBest)-1, etBest);
  }
  else{
    if (dm->fm->options[opt_verb])
      printf("b2vd success:%d numVars:%d\n", (dt!=NULL), bddc->numVars);
  }
  
  return dt;
}
