/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"
#include "decomp.h"

extern clock_t _tDecompBdd;

typedef struct _BddDomDecomp
{
  DdNode *node;
  int varCount;
  int nodeCount;
  float nodeRepRatio;
  float varRepRatio;
} BddDomDecomp;

static void
incrNumEdges(Nml *nml, DdNode *node)
{
  if (!Cudd_IsConstant(node)){
    int numEdges;
    numEdges = nmlGetMark(nml, Cudd_Regular(node));
    if (numEdges<0)
      numEdges--;      
    else
      numEdges++;
    nmlPutMark(nml, Cudd_Regular(node), numEdges);
  }
}

static void
decrNumEdges(Nml *nml, DdNode *node)
{
  if (!Cudd_IsConstant(node)){
    int numEdges;
    numEdges = nmlGetMark(nml, Cudd_Regular(node));
    if (numEdges > 0){
      numEdges = 0-numEdges;
    }
    numEdges--;
    nmlPutMark(nml, Cudd_Regular(node), numEdges);
  }
}

// returns a value of -1 if node is not a bdd dominator
// If a node is not dominated by node, then its numEdges entry is negative
static BddDomDecomp
getBddDomDecompValue(Nml *nmlNumInEdges, DdManager *ddManager, DdNode *node, 
		     int numVars, int topLevel,
		     int *numNodesWithIndexGbl, int *numNodesWithIndexLcl)
{
  BddDomDecomp result;
  struct _Nml nmlNumEdges; // num edges found starting from node
  LevelFirstSearch lfs;
  DdNode *nCurr;
  int oldLevel=-1;
  int curLevel;
  int repNodeCount;
  char isDom;
  int i;

  nmlConstruct(&nmlNumEdges);

  result.node = node;
  result.nodeCount = 0;
  repNodeCount =0;
  result.varCount =0;

  lfsConstruct(&lfs, ddManager, node);
  while((nCurr=lfsNextNode(&lfs)) != NULL){
    if (Cudd_IsConstant(nCurr))
      continue;

    result.nodeCount++;
    curLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(nCurr));
    numNodesWithIndexLcl[curLevel-topLevel]++;
    if (oldLevel != curLevel){ // advanced to another variable
      result.varCount++;
      oldLevel = curLevel;
    }

    if (nCurr != node 
	&& nmlGetMark(&nmlNumEdges, Cudd_Regular(nCurr)) 
	!= nmlGetMark(nmlNumInEdges, Cudd_Regular(nCurr))){

      // number of edges to nCurr from node does not match
      // number of edges to nCurr from root.  node is not a bdd dominator.
      repNodeCount++;

      decrNumEdges(&nmlNumEdges, Cudd_Regular(Cudd_T(nCurr)));
      decrNumEdges(&nmlNumEdges, Cudd_Regular(Cudd_E(nCurr)));
      
      result.node = NULL;
      result.nodeCount = -1;
      result.varCount = -1;
      break;
    }else{
      // increment the number of edges to Cudd_T and Cudd_E
      incrNumEdges(&nmlNumEdges, Cudd_Regular(Cudd_T(nCurr)));
      incrNumEdges(&nmlNumEdges, Cudd_Regular(Cudd_E(nCurr)));
    }
  }
  lfsDestruct(&lfs);
  nmlDestruct(&nmlNumEdges);  

  // if decomp is not value, return .node=NULL
  // check that 'node' dominators the nodes below it.
  isDom = TRUE;
  for (i=0; i<numVars; i++)
    if (numNodesWithIndexLcl[i]!=0 &&
	numNodesWithIndexLcl[i]!=numNodesWithIndexGbl[i]){
      isDom = FALSE;
      break;
    }

  if ( isDom==FALSE || result.varCount <= 1)
    result.node = NULL;
  else{
    result.nodeRepRatio = (float)repNodeCount/(float)result.nodeCount;
    if (result.nodeRepRatio > 0.03)
      result.node = NULL;
  }
  return result;
}

static DecompTransform *
bddDomCommit(Model *model, DdManager *gblManager, DdManager *tmpManager, 
	     BddClass *bddcOld, 
	     DdNode *nRoot, DdNode *nDom,
	     int *oldIndexToNew, int arraySize)
{
  // A stands for extracted bdd
  int numVarsA, arraySizeA;
  int *oldIndexToNewA;

  // B stands for root bdd
  int numVarsB, arraySizeB;
  int *oldIndexToNewB;

  DdNode *bddExt, *bddRoot, *bddTmp;
  DdNode *newVar;

  DecompTransform *dt;
  int *newIndexToOld;
  int j;

  int newVarIndex;

  bddExt = bddCopyAndReIndex(tmpManager, nDom, gblManager, 
			     &numVarsA, &oldIndexToNewA, &arraySizeA,
			     NULL, NULL);
  bddTmp = foldBddUpdateOldIndexToNew(model, bddExt, 
				      &oldIndexToNewA, arraySizeA);
  Cudd_RecursiveDeref(gblManager, bddExt);
  bddExt = bddTmp;

  // ERROR Increasing numberof vars in manager.
  newVar = Cudd_bddNewVarAtLevel(tmpManager, 
				 Cudd_ReadPerm(tmpManager, 
					       Cudd_NodeReadIndex(nDom))+1);
  newVarIndex = Cudd_NodeReadIndex(newVar);

  Cudd_Ref(newVar);
  bddRoot = bddCopyAndReIndex(tmpManager, nRoot, gblManager, 
			     &numVarsB, &oldIndexToNewB, &arraySizeB,
			     nDom, newVar);
  Cudd_RecursiveDeref(tmpManager, newVar);
  bddTmp = foldBddUpdateOldIndexToNew(model, bddRoot, 
				      &oldIndexToNewB, arraySizeB);
  Cudd_RecursiveDeref(gblManager, bddRoot);
  bddRoot = bddTmp;

  dt = dtConstruct(model->ddManager);
  // create dtg for root
  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVarsB);
  // dtCreateNewIndexToOld(numVarsB);
#ifdef DEBUG_COMPILE
  for (j=0; j<numVarsB; j++)
    newIndexToOld[j] = -1;
#endif
  for (j=0; j<arraySize; j++){
    if (oldIndexToNew[j] != -1 &&
	oldIndexToNew[j] < arraySizeB && 
	oldIndexToNewB[oldIndexToNew[j]] != -1)
      newIndexToOld[oldIndexToNewB[oldIndexToNew[j]]] = j;
  }
  assert(oldIndexToNewB[newVarIndex] != -1);
  newIndexToOld[oldIndexToNewB[newVarIndex]] = -1; // assign input for ext.
  dtAddGate(dt, newIndexToOld, bddRoot, numVarsB, 
	    Cudd_DagSize(bddRoot)-1, TRUE);

  // create dtg for extracted bdd
  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVarsA);
#ifdef DEBUG_COMPILE
  for (j=0; j<numVarsA; j++)
    newIndexToOld[j] = -1;
#endif
  for (j=0; j<arraySize; j++){
    if (oldIndexToNew[j] != -1 &&
	oldIndexToNew[j] < arraySizeA && 
	oldIndexToNewA[oldIndexToNew[j]] != -1)
      newIndexToOld[oldIndexToNewA[oldIndexToNew[j]]] = j;
  }
  dtAddGate(dt, newIndexToOld, bddExt, numVarsA, 
	    Cudd_DagSize(bddExt)-1, TRUE);

  // Create bddClasses bddcA and bddcB
  comFree(_fbddHeap,oldIndexToNewA);
  comFree(_fbddHeap,oldIndexToNewB);

  return dt;
}

DecompTransform *
decompBddDom(DecompManager dm, BddClass *bddc)
{
  Model *model = dm->fm->model;
  Nml nmlNumInEdges;
  DdGen *gen;
  DdNode *nCurr;
  BddDomDecomp currDecomp;
  BddDomDecomp bestDecomp;
  DdManager *tmpManager;
  int numVars, arraySize;
  int *oldIndexToNew;
  DdNode *bddTmp;
  int i;
  char result;
  int *numNodesWithIndexGbl;
  int *numNodesWithIndexLcl;
  int topLevel;
  DdManager *ddManager = model->ddManager;
  DecompTransform *dt;

  clock_t t = util_cpu_time();

  tmpManager = getDdManager(&model->manArrayDecompose, bddc->numVars);

  for (i=0; i<bddc->numVars; i++)
    Cudd_bddIthVar(tmpManager, i);

  bddTmp = bddCopyAndReIndex(ddManager, bddc->bdd, tmpManager, 
			     &numVars, &oldIndexToNew, &arraySize,
			     NULL, NULL);

  topLevel = Cudd_ReadPerm(tmpManager, Cudd_NodeReadIndex(bddTmp));
  numNodesWithIndexGbl = comAlloc(_fbddHeap, 
				  sizeof(int)*bddc->numVars);
  numNodesWithIndexLcl = comAlloc(_fbddHeap, 
				  sizeof(int)*bddc->numVars);
  for (i=0; i<bddc->numVars; i++)
    numNodesWithIndexGbl[i]=0;

  nmlConstruct(&nmlNumInEdges);
  
  // populate nmlNumEdges
  Cudd_ForeachNode(tmpManager, bddTmp, gen, nCurr){
    assert(Cudd_IsComplement(nCurr) == FALSE);
    if (Cudd_IsConstant(nCurr)==FALSE){
      numNodesWithIndexGbl[Cudd_ReadPerm(tmpManager, 
			 Cudd_NodeReadIndex(nCurr))-topLevel]++;
      incrNumEdges(&nmlNumInEdges, Cudd_Regular(Cudd_T(nCurr)));
      incrNumEdges(&nmlNumInEdges, Cudd_Regular(Cudd_E(nCurr)));
    }
  }

  // initialize bestDecomp
  bestDecomp.varCount = -1;
  bestDecomp.nodeCount = -1;
  bestDecomp.node = NULL;

  // Find best bddDom decomposition in bddTmp
  //  _nmlTotal = &nmlNumInEdges;

  Cudd_ForeachNode(tmpManager, bddTmp, gen, nCurr){
    if (Cudd_Regular(nCurr) != Cudd_Regular(bddTmp)){
      int j;

      for (j=0; j<bddc->numVars; j++)
	numNodesWithIndexLcl[j]=0;

      currDecomp = 
	getBddDomDecompValue(&nmlNumInEdges, 
			     tmpManager, nCurr, bddc->numVars, topLevel, 
			     numNodesWithIndexGbl, numNodesWithIndexLcl);

      // Priority, 1) node repeat ratio is lower.
      // 2) varCount is higher
      // 3) nodeCount is higher
      if (bestDecomp.node == NULL && currDecomp.node != NULL)
	bestDecomp = currDecomp;
      else if (currDecomp.node != NULL && 
	  (currDecomp.nodeRepRatio < bestDecomp.nodeRepRatio || 
	  (currDecomp.nodeRepRatio == bestDecomp.nodeRepRatio &&
	   (currDecomp.nodeCount > bestDecomp.nodeCount ||
	    (currDecomp.varCount > bestDecomp.varCount 
	     && currDecomp.nodeCount == bestDecomp.nodeCount))
	   )))
	bestDecomp = currDecomp;
    }
  }

  nmlDestruct(&nmlNumInEdges);
  comFree(_fbddHeap, numNodesWithIndexGbl);
  comFree(_fbddHeap, numNodesWithIndexLcl);

  if (bestDecomp.node == NULL || bestDecomp.varCount <= 1){
    result = FALSE;
    dt = NULL;
  }
  else{
    // Commit decomposition

    dt = 
      bddDomCommit(model, ddManager, tmpManager, bddc, 
		   bddTmp, bestDecomp.node, 
		   oldIndexToNew, arraySize);
    result = TRUE;
  }
  comFree(_fbddHeap,oldIndexToNew);

  Cudd_RecursiveDeref(tmpManager, bddTmp);

  releaseDdManager(&model->manArrayDecompose);

  _tDecompBdd += util_cpu_time() - t;

  return dt;
}
