/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <assert.h>
#include <time.h>
#include "fbddutil.h"
#include "fbddopt.h"
#include "decomp.h"

/************************************
     SubBddSearchFrontier Object
*************************************/
typedef struct _SubBddSearchFrontier
{
  IMem arena;

  // sorts nodes on the frontier (with all inedges accounted for)
  // in nondecreasing order of level.
  struct _Heap heap; 
  
  Nml nmlFInEdgeTotal;
  Nml nmlFInEdgeCount;

  Nml nmlInEdgeTotal; // total number of in edges for each node.
  Nml nmlInEdgeCount; // number of in edges encountered so far for each node.

  int *varNodeCountTotal;
  int *varNodeCountCount;

  DdManager *ddManager;
  BddClass *bddcOld;
  int frontierCount;

  DdNode *nBestRoot, *nBestThen, *nBestElse;
  int bestNodeCount;
  int bestOverlap;
  
  Nml nmlNodeMark; // used by calcVarNodeCountTotal

} SubBddSearchFrontier;

extern clock_t _tDecompNest;

static DdNode*
sbsfHeapPop(SubBddSearchFrontier *sbsf)
{
  HeapCell hc;
  DdNode *node;

  if (heapIsEmpty(&sbsf->heap))
    return NULL;

  hc = heapDelMin(&sbsf->heap);
  node = heapData(DdNode*, hc);
  comFree(sbsf->arena, hc);

  return node;
}

static void
sbsfCalcInEdgeTotal(SubBddSearchFrontier *sbsf, DdNode *node)
{
  DdNode *nTmp;
  int count;

  count = nmlGetMark(&sbsf->nmlInEdgeTotal, node);
  count++;
  nmlPutMark(&sbsf->nmlInEdgeTotal, node, count);

  if (Cudd_IsConstant(node))
    return;

  if (Cudd_IsConstant(node)==FALSE && count==1){
    nTmp = getComplementAdjustedNode(Cudd_T(node), node);
    sbsfCalcInEdgeTotal(sbsf, nTmp);
    
    nTmp = getComplementAdjustedNode(Cudd_E(node), node);
    sbsfCalcInEdgeTotal(sbsf, nTmp);
  }
}

static void
sbsfCalcFInEdgeTotal(SubBddSearchFrontier *sbsf, DdNode *node)
{
  DdNode *nTmp;
  int count;

  count = nmlGetMark(&sbsf->nmlFInEdgeTotal, Cudd_Regular(node));
  count++;
  nmlPutMark(&sbsf->nmlFInEdgeTotal, Cudd_Regular(node), count);

  if (Cudd_IsConstant(node))
    return;

  if (Cudd_IsConstant(node)==FALSE && count==1){
    nTmp = getComplementAdjustedNode(Cudd_T(node), node);
    sbsfCalcFInEdgeTotal(sbsf, nTmp);
    
    nTmp = getComplementAdjustedNode(Cudd_E(node), node);
    sbsfCalcFInEdgeTotal(sbsf, nTmp);
  }
}

static void
calcVarNodeCountTotalHelper(SubBddSearchFrontier *sbsf, DdNode *node)
{
  if (nmlGetMark(&sbsf->nmlNodeMark, Cudd_Regular(node))==0){
    nmlPutMark(&sbsf->nmlNodeMark, Cudd_Regular(node), 1);

    // first time encountering this node
    if (Cudd_IsConstant(node)==FALSE){
      sbsf->varNodeCountTotal[Cudd_NodeReadIndex(node)]++;

      calcVarNodeCountTotalHelper(sbsf, Cudd_T(node));
      calcVarNodeCountTotalHelper(sbsf, Cudd_E(node));
    }
  }
}

static void
calcVarNodeCountTotal(SubBddSearchFrontier *sbsf, DdNode *node)
{
  nmlConstruct(&sbsf->nmlNodeMark);
  calcVarNodeCountTotalHelper(sbsf, node);
  nmlDestruct(&sbsf->nmlNodeMark);
}

static void
sbsfInit(SubBddSearchFrontier *sbsf, DdManager *manager, BddClass *bddc)
{
  int i;

  sbsf->ddManager = manager;
  sbsf->arena = _fbddHeap;
  sbsf->bddcOld = bddc;

  nmlConstruct(&sbsf->nmlFInEdgeTotal);
  sbsfCalcFInEdgeTotal(sbsf, bddc->bdd);

  sbsf->varNodeCountTotal = 
    (int*)comAlloc(sbsf->arena, sizeof(int)*sbsf->bddcOld->numVars);
  for (i=0; i<sbsf->bddcOld->numVars; i++){
    sbsf->varNodeCountTotal[i] = 0;
  }
  calcVarNodeCountTotal(sbsf, bddc->bdd);

  sbsf->varNodeCountCount = 
    (int*)comAlloc(sbsf->arena, sizeof(int)*sbsf->bddcOld->numVars);
}

static void
sbsfPrepare(SubBddSearchFrontier *sbsf, DdNode *node)
{
  int i;

  heapConstruct( &sbsf->heap, sbsf->arena, 64);

  nmlConstruct(&sbsf->nmlInEdgeTotal);
  sbsfCalcInEdgeTotal(sbsf, Cudd_Regular(node));

  nmlConstruct(&sbsf->nmlInEdgeCount);
  nmlConstruct(&sbsf->nmlFInEdgeCount);

  sbsf->frontierCount = 0;
  sbsf->nBestRoot = NULL;
  sbsf->nBestThen = NULL;
  sbsf->nBestElse = NULL;
  sbsf->bestNodeCount = -1;

  for (i=0; i<sbsf->bddcOld->numVars; i++){
    sbsf->varNodeCountCount[i] = 0;
  }
}

static void
sbsfReset(SubBddSearchFrontier *sbsf)
{
  heapDelAll(&sbsf->heap);
  nmlDestruct(&sbsf->nmlInEdgeTotal);
  nmlDestruct(&sbsf->nmlInEdgeCount);
  nmlDestruct(&sbsf->nmlFInEdgeCount);
}

static void
sbsfDestruct(SubBddSearchFrontier *sbsf)
{
  nmlDestruct(&sbsf->nmlFInEdgeTotal);
  comFree(sbsf->arena, sbsf->varNodeCountCount);
  comFree(sbsf->arena, sbsf->varNodeCountTotal);
}

static void
sbsfAddNode(SubBddSearchFrontier *sbsf, DdNode *node)
{
  int nInEdgeCount, nInEdgeTotal;
  
  nInEdgeTotal = nmlGetMark(&sbsf->nmlInEdgeTotal, node);
  assert(nInEdgeTotal > 0);
  nInEdgeCount = nmlGetMark(&sbsf->nmlInEdgeCount, node);

  if (nInEdgeCount == 0)
    sbsf->frontierCount++;

  nInEdgeCount++;
  assert(nInEdgeCount <= nInEdgeTotal);
  nmlPutMark(&sbsf->nmlInEdgeCount, node, nInEdgeCount);

  if (nInEdgeCount == nInEdgeTotal && !Cudd_IsConstant(node)){
    int fInEdgeTotal = nmlGetMark(&sbsf->nmlFInEdgeTotal, Cudd_Regular(node));
    int fInEdgeCount = nmlGetMark(&sbsf->nmlFInEdgeCount, Cudd_Regular(node));

    assert(fInEdgeCount <= fInEdgeTotal);
    if (fInEdgeTotal == fInEdgeCount){
      HeapCell hc;

      //  All in edges accounted for.  Place node in heap.
      hc = 
	heapNewCell( &sbsf->heap, 
		     Cudd_ReadPerm(sbsf->ddManager,Cudd_NodeReadIndex(node)), 
		     node );
      heapInsertElement( &sbsf->heap, hc );
    }
  }
}

/*****************************************
        SubBddDecompose Functions
 *****************************************/
static void
sbdSearchFromNode(SubBddSearchFrontier *sbsf, DdNode *node)
{
  DdNode *nTmp;
  DdNode *nCurr;
  int numNodes;
  Nml nmlNodeMark;
  
  nmlConstruct(&nmlNodeMark);

  // initialize the frontier with the THEN and ELSE edge of node.
  sbsf->varNodeCountCount[Cudd_NodeReadIndex(node)]++;

  nTmp = getComplementAdjustedNode(Cudd_T(node), node);
  nmlIncrMark(&sbsf->nmlFInEdgeCount, Cudd_Regular(nTmp));
  sbsfAddNode(sbsf, nTmp);

  nTmp = getComplementAdjustedNode(Cudd_E(node), node);
  nmlIncrMark(&sbsf->nmlFInEdgeCount, Cudd_Regular(nTmp));
  sbsfAddNode(sbsf, nTmp);  

  numNodes = 1;

  while ((nCurr=sbsfHeapPop(sbsf)) != NULL){
    int mark;
    sbsf->frontierCount--; // remove nCurr from frontier count.
    numNodes++;

    mark = nmlGetMark(&nmlNodeMark, Cudd_Regular(nCurr));
    if (mark==0){
      nmlPutMark(&nmlNodeMark, Cudd_Regular(nCurr), 1);
      sbsf->varNodeCountCount[Cudd_NodeReadIndex(nCurr)]++;
      
      nmlIncrMark(&sbsf->nmlFInEdgeCount, Cudd_Regular(Cudd_T(nCurr)));
      nmlIncrMark(&sbsf->nmlFInEdgeCount, Cudd_Regular(Cudd_E(nCurr)));
    }

    // Add the then and else edge of nCurr to sbsf.
    nTmp = getComplementAdjustedNode(Cudd_T(nCurr), nCurr);
    sbsfAddNode(sbsf, nTmp);

    nTmp = getComplementAdjustedNode(Cudd_E(nCurr), nCurr);
    sbsfAddNode(sbsf, nTmp);

    assert(sbsf->frontierCount >= 2);
    if (sbsf->frontierCount == 2 
	&& !(Cudd_Regular(node)==Cudd_Regular(sbsf->bddcOld->bdd) 
	     && Cudd_IsConstant(Cudd_E(nCurr)) 
	     && Cudd_IsConstant(Cudd_T(nCurr)))
	){ 
      int overlap=0;
      int numVarsExt=0;
      int i;

      // compute overlap of support between the extracted and remainder bdds.
      for (i=0; i<sbsf->bddcOld->numVars; i++){
	if (sbsf->varNodeCountCount[i] != 0){
	  if (sbsf->varNodeCountCount[i] != sbsf->varNodeCountTotal[i]){
	    overlap++;
	  }
	  numVarsExt++;
	}
      }

      if ( !(overlap>0 && numNodes<3))
      {
	// found a sub Bdd.
	sbsf->nBestRoot = Cudd_Regular(node);
	sbsf->nBestThen =  getComplementAdjustedNode(Cudd_T(nCurr), nCurr);
	sbsf->nBestElse = getComplementAdjustedNode(Cudd_E(nCurr), nCurr);
	sbsf->bestOverlap = overlap;
	sbsf->bestNodeCount = numNodes;
	assert(sbsf->nBestThen != sbsf->nBestElse);
      }
    }
  }
  nmlDestruct(&nmlNodeMark);
}

DecompTransform *
decompNestedBdd(DecompManager dm, BddClass *bddc)
{
  Model *model = dm->fm->model;
  LevelFirstSearch lfs;
  DdNode *nCurr;
  char result;

  DdNode *nBestRoot = NULL;
  DdNode *nBestThen = NULL;
  DdNode *nBestElse = NULL;
  int bestNodeCount = -1;
  int i;

  DecompTransform *dt;
  int *newIndexToOld;

  SubBddSearchFrontier sbsf;
  
  clock_t t = util_cpu_time();

  lfsConstruct(&lfs, model->ddManager, bddc->bdd);
  sbsfInit(&sbsf, model->ddManager, bddc);
  while((nCurr=lfsNextNode(&lfs)) != NULL){
    if(Cudd_IsConstant(nCurr))
      break;
    sbsfPrepare(&sbsf, Cudd_Regular(nCurr));
    sbdSearchFromNode(&sbsf, Cudd_Regular(nCurr));

    if (sbsf.nBestRoot != NULL
	&& sbsf.bestNodeCount > bestNodeCount){
      nBestRoot = sbsf.nBestRoot;
      nBestThen = sbsf.nBestThen;
      nBestElse = sbsf.nBestElse;
      bestNodeCount = sbsf.bestNodeCount;
    }
    sbsfReset(&sbsf);
  }
  lfsDestruct(&lfs);
  sbsfDestruct(&sbsf);

  if (nBestRoot != NULL){
    DdNode *nTmp, *nNestedBdd, *nRemainder;
    DdManager *tmpManager;

    // A : nested Bdd
    // B : remainder
    int numVarsA1, numVarsA2;
    int arraySizeA1, arraySizeA2;
    int *oldIndexToNewA1, *oldIndexToNewA2;

    int numVarsB1, numVarsB2;
    int arraySizeB1, arraySizeB2;
    int *oldIndexToNewB1, *oldIndexToNewB2;

    int nestedBddIndex;

    // found sub BDD, extract it.
    tmpManager = getDdManager(&model->manArrayDecompose, bddc->numVars+1);
  
    nTmp =
      bddCopyAndReIndexNestedBdd(model->ddManager, 
				 Cudd_Regular(nBestRoot), 
				 tmpManager, 
				 &numVarsA1, 
				 &oldIndexToNewA1, &arraySizeA1,
				 CT_NESTED_BDD, 
				 NULL, 
				 nBestThen, nBestElse,
				 NULL);
    Cudd_ReduceHeap(tmpManager, CUDD_REORDER_TYPE, 0);
    nNestedBdd = bddCopyAndReIndex(tmpManager, nTmp,
				   model->ddManager,
				   &numVarsA2, 
				   &oldIndexToNewA2, &arraySizeA2,
				   NULL, NULL);
    Cudd_RecursiveDeref(tmpManager, nTmp);
    assert(numVarsA1 == numVarsA2);

    nTmp =
      bddCopyAndReIndexNestedBdd(model->ddManager, 
				 bddc->bdd, 
				 tmpManager, 
				 &numVarsB1, 
				 &oldIndexToNewB1, &arraySizeB1,
				 CT_NESTED_BDD_REMAINDER, 
				 nBestRoot, 
				 nBestThen, nBestElse,
				 &nestedBddIndex);
    Cudd_ReduceHeap(tmpManager, CUDD_REORDER_TYPE, 0);
    nRemainder = bddCopyAndReIndex(tmpManager, nTmp,
				   model->ddManager,
				   &numVarsB2, &oldIndexToNewB2, &arraySizeB2,
				   NULL, NULL);
    Cudd_RecursiveDeref(tmpManager, nTmp);
    assert(numVarsB1 == numVarsB2);

    releaseDdManager(&model->manArrayDecompose);

    dt = dtConstruct(model->ddManager);

    // construct dtg for remainder
    newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVarsB1);
    for (i=0; i<numVarsB2; i++)
      newIndexToOld[i] = -1;
    for (i=0; i<arraySizeB1; i++){
      if (oldIndexToNewB1[i] != -1 
	  && oldIndexToNewB1[i] < arraySizeB2
	  && oldIndexToNewB2[oldIndexToNewB1[i]] != -1)
	newIndexToOld[oldIndexToNewB2[oldIndexToNewB1[i]]] = i;
    }
    newIndexToOld[oldIndexToNewB2[nestedBddIndex]] = -1;
    dtAddGate(dt, newIndexToOld, nRemainder, numVarsB2, 
	      Cudd_DagSize(nRemainder), TRUE);

    // construct dtg for extracted bdd
    newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVarsA2);
    for (i=0; i<numVarsA2; i++)
      newIndexToOld[i] = -1;
    for (i=0; i<arraySizeA1; i++){
      if (oldIndexToNewA1[i] != -1 
	  && oldIndexToNewA1[i] < arraySizeA2
	  && oldIndexToNewA2[oldIndexToNewA1[i]] != -1)
	newIndexToOld[oldIndexToNewA2[oldIndexToNewA1[i]]] = i;
    }
    dtAddGate(dt, newIndexToOld, nNestedBdd, numVarsA2, 
	      Cudd_DagSize(nNestedBdd), TRUE);

    comFree(_fbddHeap,oldIndexToNewA1);
    comFree(_fbddHeap,oldIndexToNewA2);
    comFree(_fbddHeap,oldIndexToNewB1);
    comFree(_fbddHeap,oldIndexToNewB2);

    result = TRUE;
  }
  else{
    result = FALSE;
    dt = NULL;
  }

  _tDecompNest += util_cpu_time() - t;

  return dt;
}
