/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"
#include "decomp.h"


static BddInstance *
tmdBddi(DecompManager dm, DdNode *bdd)
{
  Model *model = dm->fm->model;
  BddInstance *bddi; 
  BddInstance *bddiS; // select signal bddi
  int i;
  char isTopNode;

  if(bdd == dm->bddTopNode)
    isTopNode = TRUE;
  else
    isTopNode = FALSE;

  if (Cudd_IsConstant(bdd)){
    if (Cudd_IsComplement(bdd)){
      return model->bddiLogicZero;
    }
    else{
      return model->bddiLogicOne;
    }
  }

  // check for existing bddi replacement
  if ((bddi = hashGet(&dm->htBddToBddi, bdd)) != NULL){
    return bddi;
  }
  else if (Cudd_IsComplement(bdd) && 
	   ((bddi=hashGet(&dm->htBddToBddi, Cudd_Regular(bdd))) != NULL)){
    // use shared bddi, and add NOT gate on top (done in code later).
  }
  else{
    bddiS = dm->oldInputVars[Cudd_NodeReadIndex(bdd)];
    
    if (Cudd_T(bdd) == model->bddcLogicOne->bdd
	&& Cudd_E(bdd) == model->bddcLogicZero->bdd){
      assert(!isTopNode);
      bddi = bddiS;
    }
    else{
      if(Cudd_T(bdd) == model->bddcLogicOne->bdd){
	if (isTopNode && !Cudd_IsComplement(bdd))
	  bddcAddInstance(model->bddcOr, bddi=dm->bddiOrig);
	else
	  bddi = bddcNewInstance(model->bddcOr);
	bddi->inputVars[0] = bddiS;
	bddi->inputVars[1] = tmdBddi(dm, Cudd_E(bdd));
      }
      else if(Cudd_T(bdd) == model->bddcLogicZero->bdd){
	if (isTopNode && !Cudd_IsComplement(bdd))
	  bddcAddInstance(model->bddcNor, bddi=dm->bddiOrig);
	else
	  bddi = bddcNewInstance(model->bddcNor);
	bddi->inputVars[0] = bddiS;
	bddi->inputVars[1] = tmdBddi(dm, Cudd_Not(Cudd_E(bdd)));
      }
      else if(Cudd_E(bdd) == model->bddcLogicOne->bdd){
	if (isTopNode && !Cudd_IsComplement(bdd))
	  bddcAddInstance(model->bddcNand, bddi=dm->bddiOrig);
	else
	  bddi = bddcNewInstance(model->bddcNand);
	bddi->inputVars[0] = bddiS;
	bddi->inputVars[1] = tmdBddi(dm, Cudd_Not(Cudd_T(bdd)));
      }
      else if(Cudd_E(bdd) == model->bddcLogicZero->bdd){
	if (isTopNode && !Cudd_IsComplement(bdd))
	  bddcAddInstance(model->bddcAnd, bddi=dm->bddiOrig);
	else
	  bddi = bddcNewInstance(model->bddcAnd);
	bddi->inputVars[0] = bddiS;
	bddi->inputVars[1] = tmdBddi(dm, Cudd_T(bdd));
      }
      else if(Cudd_Regular(Cudd_T(bdd)) == Cudd_Regular(Cudd_E(bdd))){
	if (isTopNode && !Cudd_IsComplement(bdd))
	  bddcAddInstance(model->bddcXnor, bddi=dm->bddiOrig);
	else
	  bddi = bddcNewInstance(model->bddcXnor);    
	bddi->inputVars[0] = bddiS;
	bddi->inputVars[1] = tmdBddi(dm, Cudd_T(bdd));
      }
      else{
	if (isTopNode && !Cudd_IsComplement(bdd))
	  bddcAddInstance(model->bddcMux, bddi=dm->bddiOrig);
	else
	  bddi = bddcNewInstance(model->bddcMux);    
	bddi->inputVars[0] = tmdBddi(dm, Cudd_E(bdd));
	bddi->inputVars[1] = tmdBddi(dm, Cudd_T(bdd));
	bddi->inputVars[2] = bddiS;
      }
      
      for (i=0; i<bddi->bddClass->numVars; i++){
	listAppend(&bddi->inputVars[i]->lFanout, bddi);
      }
      assert(isBasicGate(model, bddi->bddClass));
      hashPut(&dm->htBddToBddi, Cudd_Regular(bdd), bddi);
    }
  }

  if (Cudd_IsComplement(bdd)){
    BddInstance *bddiNew;
    if (isTopNode){
      bddcAddInstance(model->bddcNot, dm->bddiOrig);
      bddiNew = dm->bddiOrig;

      bddiNew->inputVars[0] = bddi;
      bddi = bddiNew;
      listAppend(&bddi->inputVars[0]->lFanout, bddi);
    }
    else {
      bddiNew = bddcNewInstance(model->bddcNot);

      bddiNew->inputVars[0] = bddi;
      bddi = bddiNew;
      listAppend(&bddi->inputVars[0]->lFanout, bddi);
    }
    hashPut(&dm->htBddToBddi, bdd, bddi);
    listAppend(&dm->lSweepBddis, bddi);
  }

  if (isTopNode)
    assert(bddi == dm->bddiOrig);

  return bddi;
}

void
totalMuxDecompose(DecompManager dm, BddClass *bddc)
{
  Model *model = dm->fm->model;

  listConstruct0(&dm->lSweepBddis, _fbddHeap);

  while(listGetSize(&bddc->lInsts)>0){
    ListCell cBddi = listGetHead(&bddc->lInsts);
    BddInstance *bddi = listData(BddInstance*, cBddi);
    int i;

    // run call backs before destroying bddi.
    cbRun(IDecompCallBack, dm, beforeRemoveBddi, bddi);

    hashDefaultConstruct(&dm->htBddToBddi, _fbddHeap);

    for (i=0; i<bddi->bddClass->numVars; i++){
      int result;
      result = listDelData(&bddi->inputVars[i]->lFanout, 
			   bddi);
      assert(result);
    }

    dm->bddiOrig = bddi;
    dm->oldInputVars = bddi->inputVars;
    bddi->inputVars = NULL;
    dm->bddTopNode = dm->bddiOrig->bddClass->bdd;


    Cudd_Ref(dm->bddTopNode);
    bclRemoveBddi(model->bclm, model, dm->bddiOrig);
    
    tmdBddi(dm, dm->bddTopNode);
    
    assert(isBasicGate(model, dm->bddiOrig->bddClass));

    Cudd_RecursiveDeref(model->ddManager, dm->bddTopNode);
    comFree(_fbddHeap, dm->oldInputVars);

    hashDestruct(&dm->htBddToBddi);
  }

  bclRemove(model->bclm, bddc->bdd);
  Cudd_RecursiveDeref(model->ddManager, bddc->bdd);
  bddcDestroyKeepInstances(bddc);

  sweepIncr(dm->fm->sm, _fbddHeap, &dm->lSweepBddis);
}
