/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <assert.h>

#include "fbddutil.h"
#include "fbddopt.h"
#include "decomp.h"

typedef struct _DecompTransformGate
{
  // newIndexToOld[i]=j: if a positive value, then variable i
  // of the new gates should be assigned with variable j in the old gate.
  // if a negative value, then variable i should point to the bddi
  // of the (magnitude) jth child DecomposeTransformGate.
  int *newIndexToOld;
  DdNode *bdd;
  BddProperties bp;
  int nodeCount;
  int numVars;

  // TRUE if Cudd_Reduce() is required on the bdd when committing.
  char reduceRequired;
} DecompTransformGate;


void    dtSetHead(DecompTransform *dt, int *newIndexToOld, 
		  DdNode *bdd, char reduceRequired);
int     dtGetCost(DecompTransform *dt);
void    dtgReduce(Model *model, DecompTransform *dt, 
		  DecompTransformGate *dtg);
#define dtGetNumDTGs(dt) (listGetSize(&dt->lDTGs))
#define max(x, y) (x>y?x:y)

static DecompTransformGate*
dtgCreate(int *newIndexToOld, DdNode *bdd, int numVars,
	  int nodeCount, char reduceRequired)
{
  DecompTransformGate *dtgNew;
  IMem mem = _fbddHeap;

  dtgNew = (DecompTransformGate*)comAlloc(mem, sizeof(DecompTransformGate));
  dtgNew->newIndexToOld = newIndexToOld;
  dtgNew->bdd = bdd;
  dtgNew->numVars = numVars;
  dtgNew->nodeCount = nodeCount;
  dtgNew->reduceRequired = reduceRequired;
  
  return dtgNew;
}

// Does not Cudd_RecursiveDeref the BDDs
static void
dtgDestruct(DecompTransformGate *dtg, DdManager *ddManager)
{
  IMem mem = _fbddHeap;

  Cudd_RecursiveDeref(ddManager, dtg->bdd);
  comFree(mem, dtg->newIndexToOld);
  comFree(mem, dtg);
}

DecompTransform *
dtConstruct(DdManager *ddManager)
{
  IMem mem = _fbddHeap;
  DecompTransform *dtNew;
  dtNew = (DecompTransform*)comAlloc(mem, sizeof(DecompTransform));
  dtNew->dType = (char*)comAlloc(mem, sizeof(char)*5);
  dtNew->ddManager = ddManager;
  listConstruct0(&dtNew->lDTGs, _fbddHeap);

  return dtNew;
}

void 
dtDestruct(DecompTransform *dt)
{
  while(listGetSize(&dt->lDTGs)>0){
    ListCell cell = listPop(&dt->lDTGs);
    DecompTransformGate *dtg 
      = listData(DecompTransformGate*, cell);
    dtgDestruct(dtg, dt->ddManager);
    listDelCell(_fbddHeap, cell);
  }
  comFree(_fbddHeap, dt->dType);
  comFree(_fbddHeap, dt);
}

void 
dtAddGate(DecompTransform *dt, 
	  int *newIndexToOld, DdNode *bdd, int numVars,
	  int nodeCount, char reduceRequired)
{
  int i;
  DecompTransformGate *dtg;

  assert(reduceRequired == TRUE || reduceRequired == FALSE);

#ifdef DEBUG_COMPILE
  // non head dtg cannot have newIndexToOld entry of -1.
  if (listGetSize(&dt->lDTGs) != 0)
    for (i=0; i<numVars; i++)
      assert(newIndexToOld[i] != -1);
#endif

  dtg = dtgCreate(newIndexToOld, bdd, numVars, 
		  nodeCount, reduceRequired);

  assert(dtg->reduceRequired == TRUE || dtg->reduceRequired == FALSE);

  bddGetProperties(bdd, &dtg->bp);

  listAppend(&dt->lDTGs, dtg);
}

void
dtAddMuxInput(Model *model, DecompTransform *dt, DdNode *bdd)
{
  DdManager *ddManager = model->ddManager;
  int nodeCount;
  int i;
  int *newIndexToOld;
  int *oldIndexToNew;
  int numVars;
  int arraySize;
  DdNode *bddFolded;

  Cudd_Ref(bdd);
  bddFolded = foldBddClass(model, bdd,
			   &numVars, &oldIndexToNew, &arraySize,
			   NULL, NULL);
  Cudd_RecursiveDeref(ddManager, bdd);

  nodeCount = Cudd_DagSize(bddFolded)-1;

  newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*numVars);
#ifdef DEBUG_COMPILE
  for (i=0; i<numVars; i++)
    newIndexToOld[i] = -1;
#endif

  for (i=0; i<arraySize; i++){
    if (oldIndexToNew[i] != -1)
      newIndexToOld[oldIndexToNew[i]] = i;
  }

#ifdef DEBUG_COMPILE
  for (i=0; i<numVars; i++)
    assert(newIndexToOld[i] != -1);
#endif

  comFree(_fbddHeap,oldIndexToNew);

  dtAddGate(dt, newIndexToOld, bddFolded, numVars, nodeCount, FALSE);
}

// returns +1 if dt1 is better
// returns -1 if dt2 is better
int 
dtCostCmp(BddProperties *bpOrig, DecompTransform *dt1, DecompTransform *dt2)
{
  if (dt1 == NULL)
    return -1;
  else if(dt2 == NULL)
    return 1;
  else{
    ShareCost sc1, sc2;
    ListCell cDtg;

    char ncChangeSmall1 = FALSE;
    char ncChangeSmall2 = FALSE;
    int cost1, cost2;

    cost1=0; cost2=0;

    sc1.shareBdd = (void*)0x4; // dummy
    sc1.nNodeChange = 0;
    sc1.nVarChange = 0;

    sc2.shareBdd = (void*)0x5; // dummy
    sc2.nNodeChange = 0;
    sc2.nVarChange = 0;

    listForeach(&dt1->lDTGs, cDtg){
      DecompTransformGate *dtg = listData(DecompTransformGate *, cDtg);
      sc1.nNodeChange += dtg->nodeCount;
      sc1.nVarChange += dtg->numVars;

      if ((bpOrig->nodeCount_withComplement
	   -bpOrig->nodeCount_withComplement*1/10
	   > dtg->bp.nodeCount_withComplement))
	ncChangeSmall1 = TRUE;
      //      maxVar1 = max(maxVar1, dtg->numVars);
    }

    listForeach(&dt2->lDTGs, cDtg){
      DecompTransformGate *dtg = listData(DecompTransformGate *, cDtg);
      sc2.nNodeChange += dtg->nodeCount;
      sc2.nVarChange += dtg->numVars;

      if ((bpOrig->nodeCount_withComplement
	   -bpOrig->nodeCount_withComplement*1/10
	   > dtg->bp.nodeCount_withComplement))
        ncChangeSmall2 = TRUE;
    }

    if (ncChangeSmall1 && ncChangeSmall2){
      // both small improvement.
      return scCmp(&sc1, &sc2);      
    }
    else if (ncChangeSmall1){
      // 1 small improvement, 2 is better
      return -1;
    }
    else if (ncChangeSmall2){
      // 2 small improvement, 1 is better
      return 1;
    }
    else
      return scCmp(&sc1, &sc2);
  }
}

int 
dtGetCost(DecompTransform *dt)
{
  int cost = 0;
  ListCell cDtg;

  listForeach(&dt->lDTGs, cDtg){
    DecompTransformGate *dtg = listData(DecompTransformGate *, cDtg);
    cost += dtg->nodeCount + dtg->numVars;
  }

  return cost;
}

// ensure the derivatives of the decomposition each have less nodes
// and less or equal number of variables.
char
dtCostValid(BddProperties *bpOrig, DecompTransform *dt)
{
  ListCell cDtg;
  int numBigDTGs=0;
  assert(dt != NULL);

  listForeach(&dt->lDTGs, cDtg){
    DecompTransformGate *dtg = listData(DecompTransformGate *, cDtg);

    if (!(dtg->bp.nodeCount_withComplement < bpOrig->nodeCount_withComplement
	  || (dtg->bp.nodeCount_withComplement == 
	      bpOrig->nodeCount_withComplement
	      && (dtg->bp.nodeCount_noComplement 
		  < bpOrig->nodeCount_noComplement)))
	|| dtg->bp.numVars > bpOrig->numVars)
      return FALSE;

    // Compare with totalMuxDecompose().
    // invalid if more than two decompTransformGates are 
    // bigger than 90% of the original gate
    // if cost invalid, will default to totalMuxDecompose
    if (100*dtg->bp.nodeCount_withComplement 
	> 90*bpOrig->nodeCount_withComplement
	&& 100*dtg->bp.nodeCount_noComplement 
	> 90*bpOrig->nodeCount_noComplement)
      numBigDTGs++;
    if (numBigDTGs>=2){
      return FALSE;
    }
  }
  return TRUE;
}

void 
dtgReduce(Model *model, DecompTransform *dt, DecompTransformGate *dtg)
{
  DdManager *ddManager = model->ddManager;
  int *oldIndexToNew;
  int *newIndexToOld;
  int numVars, arraySize;
  DdNode *bddFolded;
  int i;

  bddFolded = foldBddClass(model, dtg->bdd,
			   &numVars, &oldIndexToNew, &arraySize,
			   NULL, NULL);
  Cudd_RecursiveDeref(ddManager, dtg->bdd);
  dtg->bdd = bddFolded;

  assert(numVars == dtg->numVars);

  newIndexToOld = 
    (int*)comAlloc(_fbddHeap, sizeof(int)*dtg->numVars);
  for (i=0; i<numVars; i++)
    newIndexToOld[i] = -1;

  for (i=0; i<arraySize; i++){
    if (oldIndexToNew[i] != -1)
      newIndexToOld[oldIndexToNew[i]] = dtg->newIndexToOld[i];
  }

  comFree(_fbddHeap, dtg->newIndexToOld);
  dtg->newIndexToOld = newIndexToOld;
  dtg->reduceRequired = FALSE;

  comFree(_fbddHeap,oldIndexToNew);
}

// Commits a decomp transform
void 
dtCommit(DecompManager dm, BddClass *bddc)
{
  Model *model = dm->fm->model;
  DecompTransform *dt = bddc->dt;
  DecompTransformGate *dtg;
  BddClass **newBddcs;
  BddInstance **newBddis; // index matches with index of newBddcs
  IMem arena = _fbddHeap;
  ListCell cDTG;
  int nDtg;

  int numDTGs = dtGetNumDTGs(dt);

  assert(dt != NULL); 

  if (dm->fm->options[opt_verb])
    printf("decomp type:%4s numVars:%3d numNodes:%3d\n", bddc->dt->dType, 
	   bddc->numVars, bddc->numNodes);
  
  newBddcs = 
    (BddClass**)comAlloc(arena, sizeof(BddClass*)*dtGetNumDTGs(dt));
  newBddis = 
    (BddInstance**)comAlloc(arena, sizeof(BddInstance*)*dtGetNumDTGs(dt));
  for (nDtg=0; nDtg<dtGetNumDTGs(dt); nDtg++){
    newBddcs[nDtg]=NULL;
    newBddis[nDtg]=NULL;
  }

  nDtg=0;
  listForeach(&dt->lDTGs, cDTG){
    dtg = listData(DecompTransformGate*, cDTG);
    if (dtg->reduceRequired == TRUE)
      dtgReduce(model, dt, dtg);
    Cudd_Ref(dtg->bdd); // bddc is accepted, add ref so that 
                        // bdd is preserved when dtg is destroyed.

    newBddcs[nDtg] = bddcCreateForBdd(model, 
				      dtg->bdd, dtg->numVars, 
				      &model->nextBddcId);

    assert(newBddcs[nDtg] != bddc);
    nDtg++;
  }

  if (dm->fm->options[opt_verb]){
    int i;
    for (i=0; i<dtGetNumDTGs(dt); i++)
      printf("          gate:%d numVars:%3d numNodes:%3d\n", i, 
	     newBddcs[i]->numVars, newBddcs[i]->numNodes);
  }
  
  // For each instance of bddc, create the child bddis.
  // Process children first.
  if (dm->fm->options[opt_nf]){
    assert(bddcNumInsts(bddc) == 1);
  }
  while(listGetSize(&bddc->lInsts)>0){
    ListCell cBddi = listPop(&bddc->lInsts);
    BddInstance *bddi = listData(BddInstance*, cBddi);
    BddInstance **oldInputVars = bddi->inputVars;
    int i;

    comFree(_fbddHeap, cBddi);
    bddi->inputVars = NULL;

    cbRun(IDecompCallBack, dm, beforeRemoveBddi, bddi);
    
    // children no longer fannout to bddi
    assert(bddi->bddClass->type == INTERNAL_VARIABLE);
    for (i=0; i<bddi->bddClass->numVars; i++){
      int result;
      result = listDelData(&oldInputVars[i]->lFanout, 
				    bddi);
      assert(result);
    }
    
    nDtg = listGetSize(&dt->lDTGs);
    listForeachBackward(&dt->lDTGs, cDTG){
      int j;
      BddClass *bddcDTG;
      BddInstance *bddiDTG;
      char isHead;
      BddInstance **inputVars;
      
      nDtg--;
      dtg = listData(DecompTransformGate*, cDTG);
      bddcDTG = newBddcs[nDtg];

      if (cDTG == listGetHead(&dt->lDTGs))
	isHead = TRUE;
      else
	isHead = FALSE;

      // create inputVars for new bddi
      inputVars = 
	(BddInstance**)comAlloc(_fbddHeap, 
				sizeof(BddInstance*)*dtg->numVars);
      for (j=0; j<dtg->numVars; j++){
	if (dtg->newIndexToOld[j] >= 0){
	  int newIndex = j;
	  int oldIndex = dtg->newIndexToOld[j];
	  inputVars[newIndex] = oldInputVars[oldIndex];
	}
	else{
	  inputVars[j] = newBddis[0-dtg->newIndexToOld[j]];
	  assert(inputVars[j] != NULL);
	}
      }
      
      // check if instance with same support already exists
      bddiDTG = NULL;

      // create new instance
      if (!isHead)
	bddiDTG = bddcNewInstance(bddcDTG);
      else{
	// is head
	bddcAddInstance(bddcDTG, bddi);
	bddiDTG = bddi;
      }
      comFree(_fbddHeap, bddiDTG->inputVars);
      bddiDTG->inputVars = inputVars;

      for (j=0; j<bddiDTG->bddClass->numVars; j++)
	listAppend(&inputVars[j]->lFanout, bddiDTG);

      newBddis[nDtg] = bddiDTG;

      cbRun(IDecompCallBack, dm, afterAddBddi, bddiDTG);
    }
    comFree(_fbddHeap, oldInputVars);
  }
 
  assert(bddcNumInsts(bddc)==0);

  if (!isBasicGate(model, bddc)){
    bclRemove(model->bclm, bddc->bdd);
    Cudd_RecursiveDeref(model->ddManager, bddc->bdd);

    bddcDestroyKeepInstances(bddc);
  }
  else{
    if (bddc->dt != NULL){
      dtDestruct(bddc->dt);
      bddc->dt = NULL;
    }
  }

  // Sweep SF and Not gates.
  if (dm->fm->options[FBDD_OPT_SE_DISABLE]==FALSE){
    struct _List lSweepBddis;
    int j;

    listConstruct0(&lSweepBddis, _fbddHeap);
    for (j=numDTGs-1; j>=0; j--){
      if (newBddis[j]->bddClass->numVars == 1)
	listAppend(&lSweepBddis, newBddis[j]);
    }
    sweepIncr(dm->fm->sm, arena, &lSweepBddis);
  }
  comFree(arena, newBddcs);
  comFree(arena, newBddis);
}
