/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "cuddInt.h"
#include "cudd.h"

#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"

struct _ElimManager
{
  FbddManager fm;

  /* Call Back Instances */
  struct _ISweepCallBack cbSweep ;
  ListCell cSweepCallBack;

  // MergePairClass to List of instances
  struct _HashTable htMPCs; 

  // MergePairInstance to ListCell in MPC lis
  struct _HashTable htMPIs; 

  // MergePairClasses ordered by size
  struct _Heap heapMPCs;         

  //  static struct _HashTable htRemDepMPIs;
  //  struct _List lRemDepMPIs;
  List mpiRemoveKeepList;
  
  DdManager *tmpManager;
  ElimMode elimMode;
  clock_t tLimit;
  int maxBddNodeChange;
  char changed;
};

typedef struct _MergePairClass MergePairClass;
struct _MergePairClass
{
  BddClass* bddcParent;
  BddClass* bddcChild;
  DdNode*   bddSupportConfig;
  int       pos;
  int       shareSize;
};

typedef struct _MergePairInstance MergePairInstance;
struct _MergePairInstance
{
  BddInstance* bddiParent;
  BddInstance* bddiChild;
  int          pos;
};

typedef struct _InputConfigCollapseInfo InputConfigCollapseInfo;
struct _InputConfigCollapseInfo
{
  BddClass *bddcCollapsed;
  int *oldIndexToNew;
  char canSimplify;
};

IHashScheme hashMPCScheme;
IHashScheme hashMPIScheme;

/* Function Prototypes */
static MergePairInstance*  mpiCreate(BddInstance *bddiParent, 
				     BddInstance *bddiChild, int pos);
static void                mpiDestruct(MergePairInstance *mpi);
static int                 mpiNumSS(BddInstance *bddiParent, 
				    BddInstance *bddiChild, int pos);

static void elimAddBddiMPs(ElimManager em, BddInstance *bddi);
static void elimRemMPsRootedAtBddi(ElimManager em, BddInstance *bddiParent);
static void elimRemMPsEndedAtBddi(ElimManager em, BddInstance *bddiParent);
static BddClass *elimCollapseMPC(ElimManager em, MergePairClass *mpc);

static InputConfigCollapseInfo* 
icciCollapse(ElimManager em, MergePairInstance *mpi);

static char checkMPC(ElimManager em, List lMPIs);
static int  checkMPCsMPIs(ElimManager em, Model *model);

static char  isValidMP(BddInstance *bddiParent, BddInstance *bddiChild);
static void
getCollapsedInputVars(MergePairInstance *mpi, BddInstance **inputVars);

/************************
     Global Variables
 ************************/
static int _nre = 0; // number of regular eliminations
static int _nfe = 0; // number of folded eliminations
static int _numNewMPIAborted=0;

/**********************************
      MergePairClass
***********************************/
static MergePairClass *
mpcCreate(ElimManager em, 
	  BddInstance *bddiParent, BddInstance *bddiChild, int pos)
{
  MergePairClass *mpcResult 
    = (MergePairClass*)comAlloc(_fbddHeap, sizeof(MergePairClass));

  mpcResult->bddSupportConfig
    = bddiGetSupportConfig(em->fm->model->ddManager, bddiParent, bddiChild, 
		       &(mpcResult->shareSize));

  if (em->fm->options[opt_nf])
    mpcResult->shareSize = mpiNumSS(bddiParent, bddiChild, pos);

  mpcResult->bddcParent = bddiParent->bddClass;
  mpcResult->bddcChild = bddiChild->bddClass;
  mpcResult->pos = pos;

  return mpcResult;
}

static void
mpcDestruct(ElimManager em, MergePairClass *mpc)
{
  if (mpc->bddSupportConfig != NULL)
    Cudd_RecursiveDeref(em->fm->model->ddManager, mpc->bddSupportConfig);

  comFree(_fbddHeap, mpc);
}

/**********************************
      MergePairInstance
***********************************/
static MergePairInstance *
mpiCreate(BddInstance *bddiParent, BddInstance *bddiChild, int pos)
{
  MergePairInstance *mpiResult = 
    (MergePairInstance*)comAlloc(_fbddHeap, sizeof(MergePairInstance));

  mpiResult->bddiParent = bddiParent;
  mpiResult->bddiChild = bddiChild;
  mpiResult->pos = pos;
  
  return mpiResult;
}

static void
mpiDestruct(MergePairInstance *mpi)
{
  comFree(_fbddHeap, mpi);
}

static void
mpiAdd(ElimManager em, BddInstance *bddiParent, BddInstance *bddiChild)
{
  List lMPIs;
  ListCell cNew;
  HeapCell hcMPIs;
  int pos = bddiGetPosChildInParent(bddiParent, bddiChild);
  MergePairInstance *mpi; 
  MergePairClass *mpc;

  if (isValidMP(bddiParent, bddiChild)==FALSE ||
      (em->elimMode==FBDD_ELIMMODE_CHILD_SINGLE_SUPPORT 
       && bddiChild->bddClass->numVars>1))
    return;

  mpi = mpiCreate(bddiParent, bddiChild, pos);
  mpc = mpcCreate(em, bddiParent, bddiChild, pos);
  
  if ((hashGet(&em->htMPIs, mpi) != NULL)
      || (em->elimMode==FBDD_ELIMMODE_WITH_SS 
	  && mpc->shareSize==0
      && !(bddiChild->bddClass->numVars==1 && bddiNumFanout(bddiChild)==1))){
    mpcDestruct(em, mpc);
    mpiDestruct(mpi);
    return;
  }

  if (em->fm->options[opt_nf])
    hcMPIs=hashGet(&em->htMPCs, mpi);
  else
    hcMPIs=hashGet(&em->htMPCs, mpc);

  if (hcMPIs == NULL){
    int minVarSize;

    // first time encountering this MPC, create list
    lMPIs = (List)comAlloc(_fbddHeap, sizeof(struct _List));
    listConstruct0(lMPIs, _fbddHeap);

    if (mpc->bddcChild->numVars == 1)
      minVarSize = 1;
    else{
      minVarSize = mpc->bddcParent->numVars + mpc->bddcChild->numVars;
    }

    hcMPIs = heapNewCell( &em->heapMPCs, (int)mpc, lMPIs );
    heapInsertElement( &em->heapMPCs, hcMPIs );

    if (em->fm->options[opt_nf])
      hashPut(&em->htMPCs, mpi, hcMPIs);
    else{
      assert(hcMPIs != NULL);
      hashPut(&em->htMPCs, mpc, hcMPIs);
    }
  }
  else{
    mpcDestruct(em, mpc);
  }

  lMPIs = heapData(List, hcMPIs);
  
  // add this instance to the MPC
  listNewCell(_fbddHeap, cNew, mpi);
  listAppendElement(lMPIs, cNew);
  
  // record mapping between MPI and it's ListCell in the MPC
  assert(cNew != NULL);
  hashPut(&em->htMPIs, mpi, cNew);
}

// Remove a merge pair instance from htMPIs and htMPC.
static void
mpiRemove(ElimManager em, MergePairInstance *mpiDel)
{
  MergePairClass mpcDel;
  MergePairClass *mpcTemp;
  ListCell cDelInst;
  List lMPIs;
  HashCell hcMPI;
  HeapCell heapCell;
  
  assert(isValidMP(mpiDel->bddiParent, mpiDel->bddiChild));

  if ((hcMPI=hashRemove(&em->htMPIs, mpiDel)) == NULL)
    return;

  cDelInst = hashData(ListCell, hcMPI);

  mpcTemp = mpcCreate(em, mpiDel->bddiParent, mpiDel->bddiChild, mpiDel->pos);
  mpcDel = *mpcTemp;
  mpcDestruct(em, mpcTemp);

  if(em->fm->options[opt_nf])
    // mpc is based on mpiDel
    heapCell = (HeapCell)hashGet(&em->htMPCs, mpiDel); 
  else
    heapCell = (HeapCell)hashGet(&em->htMPCs, &mpcDel);

  assert(heapCell != NULL);
  lMPIs = heapData(List, heapCell);
  assert(lMPIs != NULL);

  assert(listGetSize(lMPIs) > 0);
  listDelElement(lMPIs, cDelInst);

  // if MPC is empty, remove it.
  if(listGetSize(lMPIs)==0){
    HeapCell hcMPIs;
    HashCell hc;

    // remove entry from heapMPCs
    if (em->fm->options[opt_nf])
      hcMPIs = (HeapCell)hashGet(&em->htMPCs, mpiDel);
    else
      hcMPIs = (HeapCell)hashGet(&em->htMPCs, &mpcDel);
    assert(heapData(List, hcMPIs)==lMPIs);

    assert(heapIsEmpty(&em->heapMPCs)==FALSE);
    heapRemoveElement(&em->heapMPCs, hcMPIs);
    comFree(heapMem(&em->heapMPCs), hcMPIs);

    // remove from hash table
    if (em->fm->options[opt_nf])
      hc = hashRemove(&em->htMPCs, mpiDel);
    else{
      MergePairClass *mpc;
      hc = hashRemove(&em->htMPCs, &mpcDel);
      mpc = hashKey(MergePairClass*, hc);
      mpcDestruct(em, mpc);
    }
    comFree(em->htMPCs.mem, hc);

    // free list
    if (lMPIs != em->mpiRemoveKeepList){
      listDelAll(lMPIs);
      comFree(_fbddHeap, lMPIs);
    }
  }
  
  // remove MPI
  comFree(_fbddHeap, (MergePairInstance*)heapKey(hcMPI));
  comFree(em->htMPIs.mem, hcMPI);
}

int
mpiNumSS(BddInstance *bddiParent, BddInstance *bddiChild, int pos)
{
  int sNumVars, lNumVars, i;
  BddInstance **sInputVars, **lInputVars;
  BddInstance *sBddi, *lBddi;
  int result=0;

  if (bddiParent->bddClass->numVars < bddiChild->bddClass->numVars){
    sBddi = bddiParent;
    lBddi = bddiChild;
  }
  else{
    sBddi = bddiChild;
    lBddi = bddiParent;
  }

  sNumVars = sBddi->bddClass->numVars;
  sInputVars = sBddi->inputVars;
  lNumVars = lBddi->bddClass->numVars;
  lInputVars = lBddi->inputVars;

  // mark input vars
  for (i=0; i<sNumVars; i++)
    sInputVars[i]->mark |= 0x1;

  // count number of SS
  for (i=0; i<lNumVars; i++)
    if ((lInputVars[i]->mark&0x1) != 0)
      result++;

  // clear marks
  for (i=0; i<sNumVars; i++)
    sInputVars[i]->mark &= 0xfffffffe;

  return result;
}

#define mpiHasSS(bddiParent, bddiChild, pos) \
          (mpiNumSS(bddiParent, bddiChild, pos)>0)


/**********************************
    Input Config Collapse Info
***********************************/
static InputConfigCollapseInfo*
icciCreate(IMem mem, BddClass *bddcCollapsed, int *oldIndexToNew, 
	   char canSimplify)
{
  InputConfigCollapseInfo *icci;

  icci = 
    (InputConfigCollapseInfo*)comAlloc(mem, sizeof(InputConfigCollapseInfo));
  icci->oldIndexToNew = oldIndexToNew;
  icci->bddcCollapsed = bddcCollapsed;

  return icci;
}

static void
icciFree(IMem mem, InputConfigCollapseInfo *icci)
{
  comFree(mem, icci->oldIndexToNew);
  comFree(mem, icci);
}

static InputConfigCollapseInfo*
icciCollapse(ElimManager em, MergePairInstance *mpi)
{
  BddClass *bddcSimplified, *bddcCollapsed;
  InputConfigCollapseInfo *icci;
  MergePairClass mpc;
  BddInstance **inputVars;
  int *inputConfig;
  int *oldIndexToNew;
  int numVars;
  int i;

  numVars = mpi->bddiParent->bddClass->numVars
    + mpi->bddiChild->bddClass->numVars - 1;

  inputVars = (BddInstance**)comAlloc(_fbddHeap, sizeof(BddInstance*)*numVars);
  getCollapsedInputVars(mpi, inputVars);
  inputConfig = icCompute(_fbddHeap, numVars, FALSE, inputVars);

  mpc.bddcParent = mpi->bddiParent->bddClass;
  mpc.bddcChild = mpi->bddiChild->bddClass;
  mpc.pos = mpi->pos;

  oldIndexToNew = (int*)comAlloc(_fbddHeap, sizeof(int)*numVars);
  for (i=0; i<numVars; i++)
    oldIndexToNew[i] = -1;
  bddcCollapsed = elimCollapseMPC(em, &mpc);
  bddcSimplified = sweepBddc(em->fm->sm, bddcCollapsed, 
			     inputConfig, TRUE, oldIndexToNew);
  //  assert(bddcNumInsts(bddcCollapsed)==0);
  if (bddcNumInsts(bddcCollapsed)==0 
      && bddcSimplified != bddcCollapsed
      && isBasicGate(em->fm->model, bddcCollapsed)==FALSE){
    bclRemove(em->fm->model->bclm, bddcCollapsed->bdd);
    bddcDestruct(bddcCollapsed);
  }

  icci = icciCreate(_fbddHeap, bddcSimplified, oldIndexToNew, 
		    icCanSimplify(inputConfig));

  comFree(_fbddHeap, inputVars);
  comFree(_fbddHeap, inputConfig);

  return icci;
}

/***********************************
       Pair Merging Functions
************************************/
// A valid merge pair is one where the bddiChild has no more than 1 fanout
// and the resulting bddi after merging has no more than MAX_BDD_VARS inputs.
static char
isValidMP(BddInstance *bddiParent, BddInstance *bddiChild)
{
  if (bddiParent->bddClass->type != INTERNAL_VARIABLE 
      || bddiChild->bddClass->type != INTERNAL_VARIABLE
      || bddiChild->bddClass->numVars == 0 // this is handled by const prop.
      || bddiParent->bddClass->numVars + bddiChild->bddClass->numVars - 1 
      > IDEAL_MAX_BDD_VARS) 
    return FALSE;
  else
    return TRUE;
}

static char
varChangeAcceptable(int nvOldParent, int nvOldChild, 
		    int nvCollapsed, char dupChildLogic)
{
  if (dupChildLogic){
    return (nvCollapsed<nvOldParent);
  }
  else{
    return (nvCollapsed<nvOldParent+nvOldChild-1);
  }
}

static char
nodeChangeAcceptable(ElimManager em, 
		     int ncParent, int ncChild, 
		     int ncCollapsed, char duplicateChild, 
		     char relax)
{
  int maxDeltaBddNodes = em->maxBddNodeChange;

  if (ncCollapsed > MAX_BDD_SIZE)
    // only accept if it's smaller or equal to original parent or child.
    if (ncCollapsed > ncParent || ncCollapsed > ncChild)
      return FALSE;

  if (duplicateChild){
    if (ncCollapsed - ncParent <= maxDeltaBddNodes)
      return TRUE;
    else
      return FALSE;
  }
  else{
    if (ncCollapsed - (ncParent+ncChild) <= maxDeltaBddNodes)
      return TRUE;
    else 
      return FALSE;
  }
}

// Remove all MergePairs that have bddiChild as the child component of the
// merge pair.
void
elimRemMPsEndedAtBddi(ElimManager em, BddInstance *bddiChild)
{
  ListCell cFanout;
  MergePairInstance mpi;

  if (bddiChild->bddClass->type == PRIMARY_OUTPUT)
    return;

  mpi.bddiChild = bddiChild;

  listForeach(&bddiChild->lFanout, cFanout){
    BddInstance *bddiParent = listData(BddInstance*, cFanout);
    int pos = -1;
    int i;
    int numVars = bddiParent->bddClass->numVars;

    // find pos where bddiChild connects to bddiParent.
    for (i=0; i<numVars; i++){
      if (bddiParent->inputVars[i] == bddiChild){
	pos = i;
	break;
      }
    }
    assert(pos != -1);

    mpi.bddiParent = bddiParent;
    mpi.pos = pos;

    if (isValidMP(mpi.bddiParent, mpi.bddiChild))
      mpiRemove(em, &mpi);
  }
}

// remove merge pairs that have bddiParent as their parent component of 
// the merge pair.
void
elimRemMPsRootedAtBddi(ElimManager em, BddInstance *bddiParent)
{
  int numVars;
  MergePairInstance mpiDel;
  int i;

  if (bddiParent->bddClass->type == PRIMARY_INPUT)
    return;

  numVars = bddiParent->bddClass->numVars;
  mpiDel.bddiParent = bddiParent;
  for (i=0; i<numVars; i++){
    mpiDel.bddiChild = bddiParent->inputVars[i];
    mpiDel.pos = i;

    if (isValidMP(mpiDel.bddiParent, mpiDel.bddiChild))
      mpiRemove(em, &mpiDel);
  }
}

// remove all merge pairs that are dependent on bddi's in mpi.
static void
elimRemDepMPIs(ElimManager em, MergePairInstance *mpi)
{
  // Remove merge pairs where bddiParent is the child.
  elimRemMPsEndedAtBddi(em, mpi->bddiParent);

  // Remove merge pairs where bddiParent is the parent.
  elimRemMPsRootedAtBddi(em, mpi->bddiParent);

  // Remove merge pairs where bddiChild is the parent.
  elimRemMPsRootedAtBddi(em, mpi->bddiChild);
}

static BddClass *
elimCollapseMPC(ElimManager em, MergePairClass *mpc)
{
  Model *mdl = em->fm->model;
  int numParentVars;
  int numChildVars;
  BddClass *bddcResult;
  DdNode *nResult;
  DdNode *nTmpParent;
  DdNode *nTmpChild;
  DdNode *nTmpCollapse;
  int *oldIndexToNew;
  int i;

  numParentVars = mpc->bddcParent->numVars;
  numChildVars = mpc->bddcChild->numVars;

  // copy parent BDD to em->tmpManager
  oldIndexToNew = (int*)comAlloc(_fbddHeap, sizeof(int)*numParentVars);
  for (i=0; i<=mpc->pos; i++)
    oldIndexToNew[i] = i;
  for (i=mpc->pos+1; i<numParentVars; i++)
    oldIndexToNew[i] = i + numChildVars;

  nTmpParent = bddCopy(mdl->ddManager, mpc->bddcParent->bdd, em->tmpManager,
                       oldIndexToNew, NULL, NULL);

  comFree(_fbddHeap, oldIndexToNew);

  // copy child BDD to _tmpManager
  oldIndexToNew = (int*)comAlloc(_fbddHeap, sizeof(int)*numChildVars);
  for (i=0; i<numChildVars; i++)
    oldIndexToNew[i] = i+mpc->pos+1;

  nTmpChild = bddCopy(mdl->ddManager, mpc->bddcChild->bdd, em->tmpManager,
                      oldIndexToNew, NULL, NULL);

  comFree(_fbddHeap, oldIndexToNew);

  // collapse
  nTmpCollapse = 
    Cudd_bddCompose(em->tmpManager, nTmpParent, nTmpChild, mpc->pos);
  Cudd_Ref(nTmpCollapse);
  Cudd_RecursiveDeref(em->tmpManager, nTmpParent);
  Cudd_RecursiveDeref(em->tmpManager, nTmpChild);

  assert(nTmpCollapse != NULL);

  // _tmpManager now has numParentVars + numChildVars - 1 variable in use.
  // The variable missing now is mpc->pos.
  // copy back to mdl->ddManager.
  oldIndexToNew =
    (int*)comAlloc(_fbddHeap, sizeof(int)*(numChildVars+numParentVars));
  for (i=0; i<mpc->pos; i++)
    oldIndexToNew[i] = i;
  oldIndexToNew[mpc->pos] = 0;
  for (i=mpc->pos+1; i<numChildVars+numParentVars; i++)
    oldIndexToNew[i] = i-1;
  nResult = bddCopy(em->tmpManager, nTmpCollapse, mdl->ddManager,
                    oldIndexToNew, NULL, NULL);
  comFree(_fbddHeap, oldIndexToNew);

  Cudd_RecursiveDeref(em->tmpManager, nTmpCollapse);

  // Does this bdd already exist?
  bddcResult = 
    bddcCreateForBdd(mdl, nResult, numChildVars+numParentVars-1,
		     &mdl->nextBddcId);

  return bddcResult;
}

Jint
mpcHeapCmp(IOrderScheme si, Jint keysize, void* weights,
	   void* v1, void* v2)
{
  int cost1, cost2;
  MergePairClass *mpc1 = ((MergePairClass **)v1)[0];
  MergePairClass *mpc2 = ((MergePairClass **)v2)[0];

  if (mpc1 == (void*)0x80000000)
    return FALSE;
  else if (mpc2 == (void*)0x80000000)
    return TRUE;

  assert(mpc2 != (void*)0x80000000);

  if (mpc1->bddcChild->numVars == 1)
    cost1 = 1;
  else
    cost1 = mpc1->bddcParent->numVars + mpc1->bddcChild->numVars
      - mpc1->shareSize;

  if (mpc2->bddcChild->numVars == 1)
    cost2 = 1;
  else
    cost2 = mpc2->bddcParent->numVars + mpc2->bddcChild->numVars
      - mpc2->shareSize;

  if (cost1 > cost2)
    return TRUE;
  else if (cost1 < cost2)
    return FALSE;
  else{
    if (mpc1->shareSize > mpc2->shareSize)
      return FALSE;
    else if (mpc1->shareSize < mpc2->shareSize)
      return TRUE;
    else{
      if (mpc1->bddcChild->numVars > mpc2->bddcChild->numVars)
	return TRUE;
      else if (mpc1->bddcChild->numVars < mpc2->bddcChild->numVars)
	return FALSE;
      else{
	if (mpc1->bddcParent->numVars > mpc2->bddcParent->numVars)
	  return TRUE;
	else if (mpc1->bddcParent->numVars < mpc2->bddcParent->numVars)
	  return FALSE;
	else{
	  if (mpc1->pos > mpc2->pos)
	    return TRUE;
	  else if (mpc1->pos < mpc2->pos)
	    return FALSE;
	  else
	    return FALSE;
	}
      }
    }
  }
}

static struct  _DOrderScheme    _mpcHeapSchemeVtbl = {
        NULL, NULL, NULL, mpcHeapCmp};
static struct _IOrderScheme     _mpcHeapScheme = { &_mpcHeapSchemeVtbl };
IOrderScheme                    mpcHeapScheme = &_mpcHeapScheme;

static
void
elimMPCreateEntries(ElimManager em, 
		   BddInstance *bddiParent, BddInstance *bddiChild, int pos)
{
  MergePairClass *mpc;
  MergePairInstance *mpi;
  char addNewMP=FALSE;
  
  if ((em->elimMode==FBDD_ELIMMODE_CHILD_SINGLE_SUPPORT 
       && bddiChild->bddClass->numVars>1))
    return;

  mpc = mpcCreate(em, bddiParent, bddiChild, pos);
  mpi = mpiCreate(bddiParent, bddiChild, pos);

  if (hashGet(&em->htMPIs, mpi) != NULL
      || (em->elimMode==FBDD_ELIMMODE_WITH_SS && mpc->shareSize==0
      && !(bddiChild->bddClass->numVars==1 && bddiNumFanout(bddiChild)==1))){
    mpcDestruct(em, mpc);
    mpiDestruct(mpi);
    return;
  }

  mpcDestruct(em, mpc);
  mpiDestruct(mpi);
  mpc = NULL;
  mpi = NULL;

  addNewMP = TRUE;

  if (addNewMP == FALSE){
    _numNewMPIAborted++;
  }
  else{
    mpiAdd(em, bddiParent, bddiChild);
  }
}

// Requires bddi does not fanout to the same function more than once.
void
elimAddBddiMPs(ElimManager em, BddInstance *bddi)
{
  int pos;
  int i;
  int numVars;
  ListCell cFanout;
  
  // Make MPIs with fanout
  listForeach(&bddi->lFanout, cFanout){
    BddInstance *bddiFanout = listData(BddInstance*, cFanout);
    char foundChild = FALSE;
    
    numVars = bddiFanout->bddClass->numVars;
    
    if (isValidMP(bddiFanout, bddi)){
      // find position
      for (i=0; i<numVars; i++){
	if (bddiFanout->inputVars[i] == bddi){
	  assert(foundChild == FALSE);
	  foundChild = TRUE;
	  pos = i;
	}
      }
      assert(foundChild);
      elimMPCreateEntries(em, bddiFanout, bddi, pos);
    }
  }
  
  // Make MPIs with input.
  numVars = bddi->bddClass->numVars;
  for(i=0; i<numVars; i++){
    BddInstance *bddiChild = bddi->inputVars[i];
    
    if (isValidMP(bddi, bddiChild))
      elimMPCreateEntries(em, bddi, bddiChild, i);
  }
}

static int
bddiAllFanoutMPValid(BddInstance *bddi)
{
  ListCell cFanout;
  
  // cannot collapse bddi into a prim output.
  listForeach(&bddi->lFanout, cFanout){
    BddInstance *bddiFanout = listData(BddInstance*, cFanout);
    
    if (!isValidMP(bddiFanout, bddi))
      return FALSE;
  }
  return TRUE;
}

// Does not update MPIs except through destroyNetwork.
static void
removeFanout(ElimManager em, IMem mem, 
	     BddInstance *bddiChild, BddInstance *bddiParent)
{
  int nDel;

  nDel = listDelDataAll(&bddiChild->lFanout, bddiParent);
  assert(nDel==1);

  if (bddiNumFanout(bddiChild)==0){
    sweepDanglingNetwork(em->fm->sm, bddiChild);
  }
}

// don't add mp attached to the following nodes.
// dt stands for don't touch.
BddInstance *_dtParent, *_dtChild;
static void
removeFanout_UpdateSFMPs(ElimManager em, IMem mem, 
			 BddInstance *bddiChild, 
			 BddInstance *bddiParent)
{
  removeFanout(em, mem, bddiChild, bddiParent);

  /*
  if (bddiNumFanout(bddiChild)==1){
    BddInstance *bddiFanout = listData(BddInstance*,bddiChild->lFanout.head);
    if (bddiFanout != _dtChild && bddiFanout != _dtParent)
      mpiAdd(_em, bddiFanout, bddiChild);
    //      createMPsEndedAtBddi(bddiChild);
  }
  */
}

// Requires model to have been sweeped.
// collapse __mpi into a single gate of type bddcCollapsed
static void
elimCommitMPI(ElimManager em, MergePairInstance *__mpi, 
	      InputConfigCollapseInfo *icci)
{
  Model *model = em->fm->model;
  BddInstance **parentInputVars;
  int parentNumVars = __mpi->bddiParent->bddClass->numVars;
  int childNumVars = __mpi->bddiChild->bddClass->numVars;
  int i,j;
  BddInstance *bddiParent = __mpi->bddiParent;
  char dupChildLogic;
  int *oldIndexToNew = icci->oldIndexToNew;
  BddInstance **newInputVars;

  // copy __mpi to mpi, because mpi will be deallocated
  MergePairInstance _mpi;
  MergePairInstance *mpi = &_mpi;
  *mpi = *__mpi;

  assert(mpi->bddiParent->bddClass->type == INTERNAL_VARIABLE);
  assert(mpi->bddiChild->bddClass->type == INTERNAL_VARIABLE);
  assert(mpi->bddiParent->bddClass->numVars > 0);
  assert(mpi->bddiChild->bddClass->numVars > 0);

  assert(mpi->bddiParent->inputVars[mpi->pos] == mpi->bddiChild);

  dupChildLogic = (bddiNumFanout(mpi->bddiChild) > 1);

  if (em->fm->options[opt_nf]){
    if (!isBasicGateBdd(model, icci->bddcCollapsed->bdd))
      assert(bddcNumInsts(icci->bddcCollapsed)==0);
  }

  // remove dependant merge pairs
  elimRemMPsEndedAtBddi(em, mpi->bddiParent);
  elimRemMPsRootedAtBddi(em, mpi->bddiParent);
  if (!dupChildLogic)
    elimRemMPsRootedAtBddi(em, mpi->bddiChild);

  listDelElement(&mpi->bddiParent->bddClass->lInsts, 
		     mpi->bddiParent->cBddcInst);
  
  // Do not add MPIs for these nodes when calling 
  // removeFanout_UpdateSFMPs.
  _dtParent = mpi->bddiParent;
  _dtChild = mpi->bddiChild;

  parentInputVars = bddiParent->inputVars;
  bddiParent->inputVars = NULL;

  // Assign new input variables of bddiParent.
  if (icci->bddcCollapsed->numVars > 0)
    newInputVars = 
      (BddInstance**)comAlloc(_fbddHeap, sizeof(BddInstance*)*
			      icci->bddcCollapsed->numVars);
  else
    newInputVars = NULL;

#ifdef DEBUG_COMPILE
  for (i=0; i<icci->bddcCollapsed->numVars; i++)
    newInputVars[i] = NULL;
#endif

  for (i=0; i<mpi->pos; i++){
    if (oldIndexToNew[i] != -1){
      newInputVars[icci->oldIndexToNew[i]] 
	= parentInputVars[i];
    }
    else{
      removeFanout_UpdateSFMPs(em, _fbddHeap, 
			       parentInputVars[i], mpi->bddiParent);
    }
  }

  for (i=mpi->pos+1; i<parentNumVars; i++){
    if (oldIndexToNew[i+childNumVars-1] != -1){
      newInputVars[icci->oldIndexToNew[i+childNumVars-1]] 
	= parentInputVars[i];
    }
    else{
      removeFanout_UpdateSFMPs(em, _fbddHeap, 
			       parentInputVars[i], mpi->bddiParent);
    }
  }

  for (j=0; j<childNumVars; j++){
    BddInstance *bddiLeaf;
    ListCell cNew;

    if (oldIndexToNew[mpi->pos+j] != -1){
      newInputVars[icci->oldIndexToNew[mpi->pos+j]] 
	= mpi->bddiChild->inputVars[j]; 
    }

    bddiLeaf = mpi->bddiChild->inputVars[j];
    // Update the fanout of the leaves of bddiChild to point to bddiParent
    // instead of bddiChild.

    // fanout from child fannin to bddiCollapsed
    if (oldIndexToNew[mpi->pos+j] != -1){
      listNewCell(_fbddHeap, cNew, mpi->bddiParent);
      listAppendElement(&bddiLeaf->lFanout, cNew);
    }
  }

  // check 
#ifdef DEBUG_COMPILE
  for (j=0; j<icci->bddcCollapsed->numVars; j++){
    assert(newInputVars[j] != NULL);
  }
#endif

  if (mpi->bddiParent->bddClass != icci->bddcCollapsed
      && bddcNumInsts(mpi->bddiParent->bddClass)==0 
      && isBasicGate(model, mpi->bddiParent->bddClass)==FALSE){
    bclRemove(model->bclm, mpi->bddiParent->bddClass->bdd);
    bddcDestruct(mpi->bddiParent->bddClass);
  }

  //  removeFanout(em, _fbddHeap, mpi->bddiChild, mpi->bddiParent);

  // Disconnect child from parent
  listDelDataAll(&mpi->bddiChild->lFanout, mpi->bddiParent);

  // Create Bddi for collapsed function
  bddcAddInstance(icci->bddcCollapsed, mpi->bddiParent);
  if (mpi->bddiParent->inputVars != NULL)
    comFree(_fbddHeap, mpi->bddiParent->inputVars);
  mpi->bddiParent->inputVars = newInputVars;

  // Destroy child if dangling
  if (bddiNumFanout(mpi->bddiChild)==0){
    sweepDanglingNetwork(em->fm->sm, mpi->bddiChild);
  }

  // Add the new merge pairs.
  elimAddBddiMPs(em, mpi->bddiParent);

  comFree(_fbddHeap, parentInputVars);

  em->changed=TRUE;
}

static void
getCollapsedInputVars(MergePairInstance *mpi, BddInstance **inputVars)
{
  int parentNumVars = mpi->bddiParent->bddClass->numVars;
  int childNumVars = mpi->bddiChild->bddClass->numVars;
  int i;

  for (i=0; i<mpi->pos; i++)
    inputVars[i] = mpi->bddiParent->inputVars[i];

  for (i=mpi->pos+1; i<parentNumVars; i++)
    inputVars[i+childNumVars-1] = mpi->bddiParent->inputVars[i];

  for (i=0; i<childNumVars; i++)
    inputVars[mpi->pos+i] = mpi->bddiChild->inputVars[i];
}

static void
constPropHelper(ElimManager em, BddInstance *bddi)
{
  ListCell cFanout;

  while(TRUE){
    BddInstance *bddiFanout;
    struct _List lDanglingOutputs;
    ListCell cOutput;
    char done=FALSE;

    // get fanout to internal variable
    listForeach(&bddi->lFanout, cFanout){
      bddiFanout = listData(BddInstance*, cFanout);
      if (bddiFanout->bddClass->type == INTERNAL_VARIABLE)
	break;
    }
    if (bddiFanout->bddClass->type != INTERNAL_VARIABLE)
      // no more fanouts to internal variables
      break;

    elimRemMPsRootedAtBddi(em, bddiFanout);
    elimRemMPsEndedAtBddi(em, bddiFanout);

    listConstruct0(&lDanglingOutputs, _fbddHeap);
    sweepBddi(em->fm->sm, bddiFanout, &lDanglingOutputs);

    if (bddiNumFanout(bddi)==0)
      done=TRUE;
    
    listForeach(&lDanglingOutputs, cOutput){
      sweepDanglingNetwork(em->fm->sm, listData(BddInstance*, cOutput));
    }
    listDelAll(&lDanglingOutputs);

    elimAddBddiMPs(em, bddiFanout);
    
   if (bddiFanout->bddClass->numVars == 0)
      constPropHelper(em, bddiFanout);

    if (done)
      break;
  }
}

static void
constProp(ElimManager em, BddInstance *bddi)
{
  constPropHelper(em, bddi);
}

ElimManager
elimNewManager(FbddManager fm, ElimMode elimMode)
{
  ElimManager em = comAlloc(_fbddHeap, sizeof(struct _ElimManager));
  Model *model = fm->model;

  int i;

  IMem mem;
  struct _List lInsts;
  ListCell cInst;

  fm->em = em;
  em->fm = fm;

  if (em->fm->sm != NULL){
    sweepCBAssign(&em->cbSweep, em,
       (void (*)( void *obj, BddInstance *bddi ))elimRemMPsRootedAtBddi,
		  NULL);
    cbAdd(em->fm->sm, em->cSweepCallBack, &em->cbSweep);
  }

  em->elimMode = elimMode;
  em->mpiRemoveKeepList = NULL;
  em->maxBddNodeChange = -1;
  em->changed = FALSE;

  // Initialize htMPCs and htMPIs
  hashDefaultConstruct(&em->htMPCs, _fbddHeap);
  if (em->fm->options[opt_nf])
    em->htMPCs.scheme = hashMPIScheme; // every MPI is a unique MPC
  else
    em->htMPCs.scheme = hashMPCScheme;

  hashDefaultConstruct(&em->htMPIs, _fbddHeap);
  em->htMPIs.scheme = hashMPIScheme;

  //  listConstruct0(&em->lRemDepMPIs, _fbddHeap);

  heapConstruct0( &em->heapMPCs, _fbddHeap, 2000, 1, NULL, mpcHeapScheme);

  // Initalize DdManager
  em->tmpManager = newDdManager();
  for (i=0; i<MAX_BDD_VARS+1; i++)
    Cudd_bddNewVarAtLevel(em->tmpManager, 0);

  // Initialize htMPC and htMPI
  mem = _fbddHeap;
  listConstruct0(&lInsts, mem);
  modelBddiList(mem, model, &lInsts);

  listForeach(&lInsts, cInst){
    BddInstance *bddi = listData(BddInstance*, cInst);
    int numVars;
    int nChild;
    BddInstance *bddiChild;

    if (bddi->bddClass->type == PRIMARY_INPUT
	|| bddi->bddClass->type == LATCH)
      continue;

    numVars = bddi->bddClass->numVars;
    for(nChild=0; nChild<numVars; nChild++){
      bddiChild = bddi->inputVars[nChild];
      
      if ( isValidMP(bddi, bddiChild) == FALSE )
	continue;

      elimMPCreateEntries(em, bddi, bddiChild, nChild);
    }
  }
  listDelAll(&lInsts);

  return em;
}

void
elimDelManager(ElimManager em)
{
  int i, j;
  HashCell cell;

  if (em->fm->sm != NULL)
    cbRemove(em->fm->sm, em->cSweepCallBack);

  hashForeach( &em->htMPCs, i, j, cell ){
    List lMPIs = heapData(List, hashData(HeapCell, cell));
    MergePairClass *mpc = hashKey(MergePairClass *, cell);
    mpcDestruct(em, mpc);
    while(listGetSize(lMPIs)>0){
      ListCell cMPI = listPop(lMPIs);
      MergePairInstance *mpi = listData(MergePairInstance*, cMPI);
      comFree(_fbddHeap, cMPI);
      mpiDestruct(mpi);
    }
    comFree(_fbddHeap, lMPIs);
  }
  hashDestruct(&em->htMPCs);
  hashDestruct(&em->htMPIs);

  heapDelAll( &em->heapMPCs);

  Cudd_Quit(em->tmpManager);

  em->fm->em = NULL;
  comFree(_fbddHeap, em);
}

char
elimRun(ElimManager em, int maxBddNodeChange, clock_t tLimit)
{
  Model *model = em->fm->model;
  clock_t t;

  em->maxBddNodeChange = maxBddNodeChange;
  if (tLimit <= 0)
    em->tLimit = 100000 + util_cpu_time();    
  else
    em->tLimit = tLimit + util_cpu_time();

  em->changed = FALSE;

  while(heapIsEmpty(&em->heapMPCs)==FALSE){
    List lMPIs = heapData(List, heapFindMin(&em->heapMPCs));
    ListCell cMPI;
    MergePairClass mpc;
    MergePairInstance *mpi;
    BddClass *bddcOldParent, *bddcOldChild;
    int canSimplify;
    int numVars;
    BddClass *bddcSimplified;
    InputConfigCollapseInfo *icci;
    int shareSize;
    DdNode *bddSupportConfig;
    int oldNre;
    char bddcSimplifiedTooLarge;

    assert(listGetSize(lMPIs)>0);

    mpi = listData(MergePairInstance*, listGetHead(lMPIs));

    mpc.bddcParent = mpi->bddiParent->bddClass;
    mpc.bddcChild = mpi->bddiChild->bddClass;
    mpc.pos = mpi->pos;

    assert(mpc.bddcParent->numVars != 0);
    assert(bddcNumInsts(mpc.bddcParent) != 0);
    assert(bddcNumInsts(mpc.bddcChild) != 0);

    bddcOldParent = mpc.bddcParent;
    bddcOldChild = mpc.bddcChild;

    numVars = mpi->bddiParent->bddClass->numVars
      + mpi->bddiChild->bddClass->numVars - 1;

    t = util_cpu_time();

    bddSupportConfig = 
      bddiGetSupportConfig(model->ddManager, 
		       mpi->bddiParent, mpi->bddiChild,
		       &shareSize);

     // compute inputConfig
    icci = icciCollapse(em, mpi);
    bddcSimplified = icci->bddcCollapsed;
    canSimplify = icci->canSimplify;

    if (bddcSimplified->numNodes > MAX_BDD_SIZE)
      bddcSimplifiedTooLarge = TRUE;
    else
      bddcSimplifiedTooLarge = FALSE;

    tCollapse += util_cpu_time() - t;

    if (em->fm->options[opt_verb]){
      printf("pm nv1:%d nv2:%d nvA:%d nn1:%d nn2:%d nnA:%d\n", 
	     mpc.bddcParent->numVars, mpc.bddcChild->numVars, 
	     bddcSimplified->numVars,
	     mpc.bddcParent->numNodes, mpc.bddcChild->numNodes,
	   bddcSimplified->numNodes);
    }
    _nfe++;

    oldNre = _nre;

    if (em->fm->options[opt_nf])
      assert(listGetSize(lMPIs)==1);
    
    em->mpiRemoveKeepList = lMPIs;
    while((cMPI=listGetHead(lMPIs)) != NULL){
      MergePairInstance *mpi = listData(MergePairInstance*, cMPI);

      int ncOldParent = mpi->bddiParent->bddClass->numNodes;
      int ncOldChild = mpi->bddiChild->bddClass->numNodes;
      int dupChildLogic = (bddiNumFanout(mpi->bddiChild) > 1);
      int nvOldParent = mpi->bddiParent->bddClass->numVars;
      int nvOldChild = mpi->bddiChild->bddClass->numVars;

      if (em->fm->options[opt_nf])
	assert(_nre == oldNre);
      _nre ++;

      assert(mpi->bddiParent->bddClass == mpc.bddcParent
	     && mpi->bddiChild->bddClass == mpc.bddcChild
	     && mpi->pos == mpc.pos);

      assert(!(mpc.bddcChild->numVars==1 && bddcSimplified==NULL));
      
      if (bddcSimplified!=NULL 
	  && ((em->elimMode == FBDD_ELIMMODE_TOTAL && canSimplify==FALSE 
	       && bddcSimplified->numNodes<MAX_BDD_SIZE) /* disjunctive*/
	      ||
	      (nodeChangeAcceptable(em, ncOldParent, ncOldChild, 
				    bddcSimplified->numNodes, dupChildLogic, 
				    varChangeAcceptable(nvOldParent, 
							nvOldChild, 
							bddcSimplified->numVars, 
							dupChildLogic))
	       || mpc.bddcChild->numVars==1
	       || (dupChildLogic==FALSE && mpc.bddcParent->numVars==1))) 
	  ){
	BddInstance *bddiParent;
	// Lots of these are left trailing.
	
	bddiParent = mpi->bddiParent;
	elimCommitMPI(em, mpi, icci); // mpi will be freed in pmExec.
	// bddiParent->bddClass will change.
	
	if (bddiParent->bddClass->numVars == 0){
	  // the resulting bddi is a constant.
	  elimRemMPsEndedAtBddi(em, bddiParent);
	  constProp(em, bddiParent);
	}

	if (em->fm->options[opt_verb])
	  printf("success: %d\n", dupChildLogic );
      }
      else{
	// collapsing will cause the node count to increase more than
	// allowed. abort collapsing this mpi.

	if (em->fm->options[opt_verb])
	  printf("failed: %d\n", dupChildLogic );
	
	mpiRemove(em, mpi);
	
	// check if full fannout elmination will help.
	/*
	  if (bddcSimplified!=NULL && bddiNumFanout(mpi->bddiChild)>1){
	  elimBddi(model, mpi->bddiChild);	    
	  }
	*/
      }
    }
    icciFree(_fbddHeap, icci);
    em->mpiRemoveKeepList = NULL;
    if (listGetSize(lMPIs)==0)
      comFree(_fbddHeap, lMPIs);

    if (bddcNumInsts(bddcSimplified) == 0 
	&& isBasicGate(model, bddcSimplified)==FALSE){
      bclRemove(model->bclm, bddcSimplified->bdd);
      bddcDestruct(bddcSimplified);
    }

    if (em->tLimit < util_cpu_time()
	|| (bddcSimplifiedTooLarge && em->elimMode == FBDD_ELIMMODE_TOTAL)){
      // Reached time limit or 
      // BDDs getting too large
      break;
    }
  }
  return em->changed;
}

void
elimDump()
{
  printf("Elimination\n");
  printf("number folded eliminations:  %5d\n", _nfe);
  printf("number regular eliminations: %5d\n", _nre);
  printf("\n");
}

/*****************************
       Debug Functions
 *****************************/
// returns 1 if ok
// returns 0 if inconsistant
int
checkMPCsMPIs(ElimManager em, Model *model)
{
  IMem arena;
  struct _List lInsts;
  ListCell cInst;
  
  arena = _fbddHeap;
  listConstruct0(&lInsts, arena);
  modelBddiList(arena, model, &lInsts);
  listForeach(&lInsts, cInst){
    BddInstance *bddi = listData(BddInstance*, cInst);
    int numVars;
    int nChild;
    BddInstance *bddiChild;

    if (bddi->bddClass->type == PRIMARY_INPUT
	|| bddi->bddClass->type == LATCH)
      continue;

    numVars = bddi->bddClass->numVars;
    for(nChild=0; nChild<numVars; nChild++){
      MergePairClass *mpc;
      HeapCell heapCell;
      MergePairInstance mpi;

      bddiChild = bddi->inputVars[nChild];

      // Do not create merge pair if the child has more than one fanout
      // or the collapsed support set has more than MAX_BDD_VARS.
      if ( isValidMP(bddi, bddiChild) == FALSE )
	continue;

      mpi.bddiParent = bddi;
      mpi.bddiChild = bddiChild;
      mpi.pos = nChild;
      if (hashGet(&em->htMPIs, &mpi) == NULL)
	continue;

      mpc = mpcCreate(em, bddi, bddiChild, nChild);
      if (em->fm->options[opt_nf]){
	assert(FALSE);
	heapCell = (HeapCell)hashGet(&em->htMPCs, mpc);
      }
      else
	heapCell = (HeapCell)hashGet(&em->htMPCs, mpc);

      assert(heapCell != NULL);

      mpcDestruct(em, mpc);
    }
  }

  return TRUE;
}

static char
checkMPC(ElimManager em, List lMPIs)
{
  char result = TRUE;
  IMem mem = _fbddHeap;
  ListCell cMPI;
  BddInstance ** inputVars;
  InputConfig inputConfig;
  MergePairInstance *mpi;
  int numVars;
  DdNode *bddSupportConfig;
  DdNode *bddSupportConfigTest;
  int shareSize;

  mpi = listData(MergePairInstance*, listGetHead(lMPIs));

  bddSupportConfig = 
    bddiGetSupportConfig(em->fm->model->ddManager, 
		     mpi->bddiParent, mpi->bddiChild,
		     &shareSize);

  numVars = mpi->bddiParent->bddClass->numVars
    + mpi->bddiChild->bddClass->numVars - 1;

  // compute inputConfig
  inputVars = (BddInstance**)comAlloc(mem, sizeof(BddInstance*)*numVars);
  getCollapsedInputVars(mpi, inputVars);
  inputConfig = icCompute(mem, numVars, FALSE, inputVars);
  printf("check mpc\n");
  printArray(inputConfig, numVars+1);

  listForeach(lMPIs, cMPI){
    InputConfig inputConfigTest;
    BddInstance **inputVarsTest;
    int supportConfigMatched;
    int i;
    
    mpi = listData(MergePairInstance*, cMPI);
    
    bddSupportConfigTest = 
      bddiGetSupportConfig(em->fm->model->ddManager, 
		       mpi->bddiParent, mpi->bddiChild,
		       &shareSize);
    supportConfigMatched = (bddSupportConfig == bddSupportConfigTest);
    assert(supportConfigMatched);
    Cudd_RecursiveDeref(em->fm->model->ddManager, bddSupportConfigTest);

    inputVarsTest 
      = (BddInstance**)comAlloc(mem, sizeof(BddInstance*)*numVars);
    getCollapsedInputVars(mpi, inputVarsTest);
    inputConfigTest = icCompute(mem, numVars, FALSE, inputVarsTest);
    
    printArray(inputConfigTest, numVars+1);
    
    for(i=0; i<numVars+1; i++){
      if (inputConfigTest[i] != inputConfig[i]){
	printf("supportConfigMatched:%d\n", supportConfigMatched);
	assert(FALSE);
	result = FALSE;
	break;
      }
    }

    comFree(mem, inputConfigTest);
    comFree(mem, inputVarsTest);
  }
  Cudd_RecursiveDeref(em->fm->model->ddManager, bddSupportConfig);

  printf("check mpc done!\n");

  return result;
}

/*******************************
     MPI AND MPC Hash Schemes
********************************/
static Juint COMCALL 
_mpcHashCode( IHashScheme si, HashTable h, void* key)
{
  MergePairClass *mpc = key;
  
  return (((unsigned int)(mpc->bddcParent)>>2)*PRIME1
	  + ((unsigned int)(mpc->bddcChild)>>2)*PRIME2
	  + mpc->pos*PRIME3
	  + ((unsigned int)(mpc->bddSupportConfig))*PRIME4
	  );
}

static Jboolean COMCALL 
_mpcEquals(IHashScheme si, HashTable h, void* key, void *another)
{
  MergePairClass *mpc1 = (MergePairClass*)key;
  MergePairClass *mpc2 = (MergePairClass*)another;

  if (mpc1->bddcParent == mpc2->bddcParent
      && mpc1->bddcChild == mpc2->bddcChild
      && mpc1->pos == mpc2->pos
      && mpc1->bddSupportConfig == mpc2->bddSupportConfig)
    return TRUE;
  else
    return FALSE;
}

static Juint COMCALL 
_mpiHashCode( IHashScheme si, HashTable h, void* key)
{
  MergePairInstance *mpi = key;

  return (((unsigned int)(mpi->bddiParent)>>2)*PRIME1 
	  + ((unsigned int)(mpi->bddiChild)>>2)*PRIME2 );
}

static Jboolean COMCALL 
_mpiEquals(IHashScheme si, HashTable h, void* key, void *another)
{
  MergePairInstance *mpi1 = (MergePairInstance*)key;
  MergePairInstance *mpi2 = (MergePairInstance*)another;

  if (mpi1->bddiParent == mpi2->bddiParent
      && mpi1->bddiChild == mpi2->bddiChild)
    return TRUE;
  else
    return FALSE;
}

static struct  _DHashScheme	_hashMPCSchemeVtbl = {
  _hashQuery, _hashAddref, _hashRelease,
  _mpcHashCode, _mpcEquals
};

static struct  _DHashScheme	_hashMPISchemeVtbl = {
  _hashQuery, _hashAddref, _hashRelease,
  _mpiHashCode, _mpiEquals
};

static struct _IHashScheme	_hashMPCScheme = { &_hashMPCSchemeVtbl };
IHashScheme			hashMPCScheme = &_hashMPCScheme;

static struct _IHashScheme	_hashMPIScheme = { &_hashMPISchemeVtbl };
IHashScheme			hashMPIScheme = &_hashMPIScheme;

