/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include "fbddutil.h"
#include "fbddopt.h"

// Timers
clock_t tTotal=0;
clock_t tSimplify=0;
clock_t tSweep=0;
clock_t tElim=0;

clock_t tSE=0;

clock_t tBddCopy=0;
clock_t tReduceHeap=0;
clock_t tFoldBddClass=0;
clock_t tCollapse=0;

FbddManager _fm;

void
fbddOptimize(FbddManager fm)
{ 
  Model *model = fm->model;
  int modelCostBefore, modelCostAfter;
  clock_t t;
  int ncCost, bestCost, worstCost;
  SeManager sem;

  /***********************************
   *       Simplify Bddcs            *
   ***********************************/
  printf("Simplify...\n");
  fbddSimplify(fm);

  /*******************
   *       Sweep     *
   *******************/
  sweepRun(fm->sm);
  
  /* Share Extract (find disjunctive extractors before they are collapsed 
     away. */
  if (!fm->options[FBDD_OPT_SE_DISABLE]){
    SeManager sem = seNewManager(fm, fm->options[FBDD_OPT_SE_MODE]);

    printf("Sharing extraction...\n");
    seRun(sem, 2);
    seDelManager(sem);
  }
  
  /********************
   *    Eliminate     *
   ********************/
  // Attempt total (aggressive) elimination, if the result is worse,
  // restore original and perform sharing extraction and 
  // 'normal' elimination.
  if (fm->options[FBDD_OPT_ELIM_DISABLE] == FALSE){
    clock_t tFirstElim = util_cpu_time();
    ElimManager em;

    printf("Eliminate...\n");
    t = util_cpu_time();
    em = elimNewManager(fm, FBDD_ELIMMODE_AGGRESSIVE);
    elimRun(em, -1, 0);    
    elimDelManager(em);
    tElim += util_cpu_time() - t;
    tFirstElim = util_cpu_time() - tFirstElim;

    if (!fm->options[opt_fast]){
      int ncBefore, bestBefore, worstBefore;
      int ncAfter, bestAfter, worstAfter;
      
      modelBackup(model);

      modelGetCost(model, &ncBefore, &bestBefore, &worstBefore);
      modelCostBefore = bestBefore;
      
      t = util_cpu_time();
      printf("Aggressive eliminate...");
      fflush(stdout);
      em = elimNewManager(fm, FBDD_ELIMMODE_AGGRESSIVE);
      elimRun(em, 100, (3*tFirstElim)/2);    
      elimDelManager(em);
      tElim += util_cpu_time()-t;

      modelGetCost(model, &ncAfter, &bestAfter, &worstAfter);
      modelCostAfter = worstAfter;
      
      if (fm->options[opt_verb]){
	printf("before nc:%6d best:%6d worst:%6d\n",
	       ncBefore, bestBefore, worstBefore);
	printf(" after nc:%6d best:%6d worst:%6d\n",
	       ncAfter, bestAfter, worstAfter);
      }

      if (modelCostAfter >= modelCostBefore && !fm->options[opt_gblelim]){
	// Abort global elimination
	printf(" aborted.\n");
	modelSwapBackup(model);
      }
      else
	printf(" successful.\n");
      modelClearBackup(model);
    }
    sweepDanglingLatchNetworks(fm->sm);
  }
  
  modelGetCost(model, &ncCost, &bestCost, &worstCost);
  if (fm->options[opt_verb]){
    printf("after elim nc:%6d best:%6d worst:%6d\n",
	   ncCost, bestCost, worstCost);
  }

  // Sweep remaining buffers or constants.
  if (bddcNumInsts(model->bddcSF)>0 || bddcNumInsts(model->bddcNot)>0
      || bddcNumInsts(model->bddcLogicOne) 
      || bddcNumInsts(model->bddcLogicZero)){
    sweepRun(fm->sm);
  }

  /******************************************
   *   Sharing Extraction / Decomposition   *
   ******************************************/
  if (!fm->options[FBDD_OPT_SE_DISABLE] 
      && !fm->options[FBDD_OPT_DECOMP_DISABLE])
    printf("Sharing Extraction / Decomposition...\n");
  else if(!fm->options[FBDD_OPT_SE_DISABLE])
    printf("Sharing Extraction...\n");
  else if (!fm->options[FBDD_OPT_DECOMP_DISABLE])
    printf("Decomposition...\n");
  
  if (!fm->options[FBDD_OPT_SE_DISABLE]){
    sem = seNewManager(fm, fm->options[FBDD_OPT_SE_MODE]);
    seRun(sem, 0);
  }
  
  /********************
   *  Decomposition   *
   ********************/
  if (fm->options[FBDD_OPT_DECOMP_DISABLE] == FALSE){
    DecompManager dm;
    dm = decompNewManager(fm);
    decompRun(dm);
    decompDelManager(dm);
  }

  if (!fm->options[FBDD_OPT_SE_DISABLE])
    seDelManager(sem);

  // Perform Sweep
  sweepRun(fm->sm);
  
  if (fm->options[opt_verb]){
    int nc, best, worst;
    modelGetCost(model, &nc, &best, &worst);
    printf("model cost before nes: nc:%d best:%d worst:%d\n",
	   nc, best, worst);
    printf("model lit count:%d\n", modelLitCount(model));
  }

  /******************************
   * Equivalent Node Extraction *
   ******************************/
  if (!fm->options[FBDD_OPT_SE_DISABLE]){
    sem = seNewManager(fm, fm->options[FBDD_OPT_SE_MODE]);
    seRun(sem, 0);
    seDelManager(sem);
  }

  if (fm->options[opt_verb]){
    int nc, best, worst;
    modelGetCost(model, &nc, &best, &worst);
    printf("model cost after nes: nc:%d best:%d worst:%d\n",
	   nc, best, worst);
    printf("model lit count:%d\n", modelLitCount(model));
  }
}

/**************************
         FbddManager
***************************/
FbddManager 
fbddNewManager()
{
  FbddManager fm;
  int i;

  //  _fbddHeap = newArena(comHeap());
  _fbddHeap = comHeap();

  fm = (FbddManager)comAlloc(_fbddHeap, 
			     sizeof(struct _FbddManager));
  fm->model = NULL;
  fm->sm = sweepNewManager(fm);
  fm->dm = NULL;
  fm->sem = NULL;
  fm->em = NULL;
  listConstruct0( &fm->lCBs, _fbddHeap);

  // Set default options
  for (i=0; i<NUM_OPTIONS; i++)
    fm->options[i] = FALSE;
  
  fm->options[FBDD_OPT_MAXEXTSIZE] = 2;
  fm->options[FBDD_OPT_SE_MODE] = FBDD_SE_MODE_FAST;

  _fm = fm;

  return fm;
}

void        
fbddDelManager(FbddManager fm)
{
  sweepDelManager(fm->sm);

  bclDelManager(fm->model->bclm);
  fm->model->bclm = NULL;

  if (fm->model != NULL)
    modelDestroy(fm->model);
  comFree(_fbddHeap, fm);

  /*
  comDestruct(_fbddHeap);
  comFree(comHeap(), _fbddHeap);
  */
}


void
fbddSetElimDisable(FbddManager fm, char val)
{
  fm->options[FBDD_OPT_ELIM_DISABLE] = val;
}

void
fbddSetDecompDisable(FbddManager fm, char val)
{
  fm->options[FBDD_OPT_DECOMP_DISABLE] = val;
}

void
fbddSetSeDisable(FbddManager fm, char val)
{
  fm->options[FBDD_OPT_SE_DISABLE] = val;
}

void fbddSetMaxExtSize(FbddManager fm, int val)
{
  fm->options[FBDD_OPT_MAXEXTSIZE] = val;
}

void
fbddSetSeMode(FbddManager fm, SeMode mode)
{
  fm->options[FBDD_OPT_SE_MODE]= mode;
}

void 
fbddFinalizeOptions(FbddManager fm)
{
  fm->model = modelCreate("");

  // Initialize bcl (BddClassLookup) data structure. 
  // Bcl maps a bdd with its BddClass.
  fm->model->bclm = bclNewManager();
  populateBclWithBasicGates(fm->model->bclm, fm->model);
} 

void
fbddCBAssign(IFbddCallBack cb, void *objCallee,
	     void (*onNewBddc)( void *obj, BddClass *bddc ),
	     void (*onDelBddc)( void *obj, BddClass *bddc ))
{
  cb->objCallee = objCallee;
  cb->onNewBddc = onNewBddc;
  cb->onDelBddc = onDelBddc;
}
