/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <assert.h>

#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"

/*********************************
          Global Variables
 *********************************/
int maxExtSize;
IMem _fbddHeap;
int verbose = 0;

void
printArray(int *array, int numElements)
{
  int i;
  for (i=0; i<numElements; i++)
    printf("%d,", array[i]);
  printf("\n");
}

BddClass *
bddcCreate(char *name, DdNode *bdd, bddcType type, int numVars)
{
  BddClass *newBddc;
  int i;

  newBddc = comAlloc(_fbddHeap, sizeof(BddClass));

  newBddc->mark = 0;

  if (name != NULL){
    newBddc->name = comAlloc(_fbddHeap, sizeof(char)*(strlen(name)+1));
    strcpy(newBddc->name, name);
  }
  else
    newBddc->name = NULL;

  newBddc->bdd = bdd;

#ifdef DEBUG_COMPILE
  if (numVars>1 && bdd != NULL){
    assert(Cudd_NodeReadIndex(bdd)==numVars-1);
  }
#endif

  if (bdd != NULL)
    newBddc->numNodes = Cudd_DagSize(bdd) - 1;
  else
    newBddc->numNodes = -1;

  newBddc->type = type;
  assert(numVars < MAX_BDD_VARS);
  newBddc->numVars = numVars;
  listConstruct0( &newBddc->lInsts, _fbddHeap );
  
  newBddc->lSeels = 
    (List*)comAlloc(_fbddHeap, sizeof(List)*(maxExtSize-1));
  for (i=0; i<=maxExtSize-2; i++){
    newBddc->lSeels[i] = NULL;
  }
  newBddc->dt = NULL;
  newBddc->cBchBddc = NULL;

  return newBddc;
}

BddInstance *
bddcNewInstance(BddClass *bddc)
{
  BddInstance *newBddi;
  ListCell cell;
  int i;

  newBddi = comAlloc(_fbddHeap, sizeof(BddInstance));
  assert(newBddi != NULL);
  newBddi->lSees = 
    (struct _List *)comAlloc(_fbddHeap, sizeof(struct _List)*(maxExtSize-1));

  newBddi->bddClass = bddc;
  newBddi->inputVars = 
    comAlloc(_fbddHeap, sizeof(BddInstance*)*(bddc->numVars));
  for (i=0; i<bddc->numVars; i++)
    newBddi->inputVars[i] = NULL;
  newBddi->id = 0;

  listConstruct0(&newBddi->lFanout, _fbddHeap);
  for (i=0; i<=maxExtSize-2; i++)
    listConstruct0(&newBddi->lSees[i], _fbddHeap);

  listNewCell(_fbddHeap, cell, newBddi);
  listAppendElement(&bddc->lInsts, cell);
  newBddi->cBddcInst = cell;
  newBddi->mark = 0;

  return newBddi;
}

// reusing memory for one instance for another.
// used to maintain the output nodes when decomposing instances.
// don't free the inputVars since they may be used.
void
bddcAddInstance(BddClass *bddc, BddInstance* bddi)
{
  ListCell cell;
  int i;

  bddi->bddClass = bddc;
  assert(bddi->inputVars == NULL);
  if (bddc->numVars > 0)
    bddi->inputVars = 
      comAlloc(_fbddHeap, sizeof(BddInstance*)*(bddc->numVars));
  else
    bddi->inputVars = NULL;

  for (i=0; i<bddc->numVars; i++)
    bddi->inputVars[i] = NULL;
  bddi->mark = 0x0;

  // add to bddc
  listNewCell(_fbddHeap, cell, bddi);
  listAppendElement(&bddc->lInsts, cell);
  bddi->cBddcInst = cell;
}

BddInstance *
bddiCopyConstruct(BddInstance *bddi)
{
  BddInstance *bddiNew;
  int i;

  bddiNew = bddcNewInstance(bddi->bddClass);
  for (i=0; i<bddi->bddClass->numVars; i++)
    bddiNew->inputVars[i] = bddi->inputVars[i];

  return bddiNew;
}

// does not deref bdd from DdManager
void 
bddcDestroyKeepInstances(BddClass* bddc)
{
  // don't destroy all instances of this class
  // these instances may have become instances of other 
  // classes.
  cbRun(IFbddCallBack, _fm, onDelBddc, bddc);

  if (bddc->name != NULL)
    comFree(_fbddHeap, bddc->name);

  listDelAll(&bddc->lInsts);

  comFree(_fbddHeap, bddc->lSeels);
  comFree(_fbddHeap, bddc);
}

// does not deref bdd from DdManager
void
bddcDestruct(BddClass* bddc)
{
  BddInstance *bddi;
  ListCell cell;

  while( listGetSize(&bddc->lInsts)>0){
    cell = listPop(&bddc->lInsts);
    bddi = (BddInstance*)(cell->data);
    comFree(_fbddHeap, cell);
    bddiDestruct(bddi);
  }

  bddcDestroyKeepInstances(bddc);
}

typedef struct _CollapseGatesData
{
  Nml nml;
  BddClass *bddcCollapse;
  IMem mem;
} CollapseGatesData;

// removes all instances of bddcCollapse
static void
bddiCollapseGatesHelper(CollapseGatesData *cgd, BddInstance *bddi)
{
  // does not free memory for fwding gates
  int i;
  Nml *nml = &cgd->nml;
  BddClass *bddcCollapse = cgd->bddcCollapse;

  if (bddi == NULL 
      || bddi->bddClass->type != PRIMARY_INPUT)
    return;

  if (nmlGetBddiMark(nml, bddi) == 1)
    return;
  
  nmlPutBddiMark(nml, bddi, 1);
  
  if (bddi->bddClass->type == PRIMARY_OUTPUT)
    bddiCollapseGatesHelper(cgd, bddi->inputVars[0]);
  else{
    for (i=0; i<bddi->bddClass->numVars; i++){
      while (bddi->inputVars[i]->bddClass == bddcCollapse){
	BddInstance *bddiCollapse = bddi->inputVars[i];
	BddInstance *bddiNonForwarding;

	bddiNonForwarding = bddi->inputVars[i]->inputVars[0];
	while(bddiNonForwarding->bddClass == bddcCollapse)
	  bddiNonForwarding = bddiNonForwarding->inputVars[0];

	bddi->inputVars[i] = bddiNonForwarding;
	listAppend(&bddiNonForwarding->lFanout, bddi);
	listDelData(&bddiCollapse->lFanout, bddi);

	// destruct unreferenced nodes from network
	while ( bddiNumFanout(bddiCollapse)==0 ) {
	  BddInstance *bddiInput = bddiCollapse->inputVars[0];

	  assert(bddiCollapse->bddClass == bddcCollapse);
	  listDelData(&bddiInput->lFanout, bddiCollapse);

	  listDelElement(&bddiCollapse->bddClass->lInsts, 
			     bddiCollapse->cBddcInst);
	  bddiDestruct(bddiCollapse);

	  bddiCollapse = bddiInput;
	}
	
      }
      bddiCollapseGatesHelper(cgd, bddi->inputVars[i]);
    }
  }
}

void
bddiCollapseGates(IMem mem, BddInstance *bddi, BddClass *bddcCollapse)
{
  CollapseGatesData cgd;
  cgd.bddcCollapse = bddcCollapse;
  cgd.mem = mem;
  nmlConstruct(&cgd.nml);
  bddiCollapseGatesHelper(&cgd, bddi);
  nmlDestruct(&cgd.nml);
}

static int
bddiNumGatesHelper(Nml *nmlNumGates, BddInstance *bddi)
{
  int i;
  int accum=0;

  assert(bddi);

  if (nmlGetBddiMark(nmlNumGates, bddi) 
      || bddi->bddClass->type == PRIMARY_INPUT)
    return 0;
  
  nmlPutBddiMark(nmlNumGates, bddi, 1);
  
  if (bddi->bddClass->type == PRIMARY_OUTPUT)
    return bddiNumGatesHelper(nmlNumGates, bddi->inputVars[0]);
  else{
    accum=1;
    for (i=0; i<bddi->bddClass->numVars; i++)
      accum += bddiNumGatesHelper(nmlNumGates, bddi->inputVars[i]);
  }
  return accum;
}

int
modelNumGates(Model *model)
{
  static Nml nmlNumGates;
  int result = 0;
  ListCell cell;

  nmlConstruct(&nmlNumGates);
  listForeach(&(model->lOutputs), cell){
    result += bddiNumGatesHelper(&nmlNumGates, listData(BddInstance*,cell));
  }
  nmlDestruct(&nmlNumGates);

  return result;
}

// Does not initialize lBddcs
// List does not include primary inputs and outputs
void
modelBddcList(IMem mem, Model *model, List lBddcs)
{
  ListCell cell;
  struct _List lBddis;

  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);

  listForeach(&(lBddis), cell){
    BddInstance *bddi;
    bddi = listData(BddInstance*,cell);
    
    if (bddi->bddClass->type != PRIMARY_INPUT
	&& bddi->bddClass->type != PRIMARY_OUTPUT
	&& bddcGetMark(bddi->bddClass, 0) == FALSE
	){
      ListCell cNew;
      BddClass *bddc = bddi->bddClass;

      // BddClass not encountered before.  Mark BddClass.
      bddcSetMark(bddi->bddClass, 0, TRUE);

      // Add BddClass to list
      listNewCell(mem, cNew, bddc);
      listAppendElement(lBddcs, cNew);
    }
  }
  listDelAll(&lBddis);

  listForeach(lBddcs, cell){
    bddcSetMark(listData(BddClass*, cell), 0, FALSE);
  }
}

void
modelGetCost(Model *model, int *nc, int *bestCase, int *worstCase)
{
  IMem arena;
  ListCell cell;
  struct _List lBddis;

  *nc = 0;
  *worstCase = 0;
  *bestCase = 0;

  arena = _fbddHeap;

  listConstruct0(&lBddis, arena);
  modelBddiList(arena, model, &lBddis);

  listForeach(&lBddis, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    if (bddi->bddClass->type == INTERNAL_VARIABLE){
      BddClass *bddc = bddi->bddClass;
      *nc += bddc->numNodes;
      *bestCase += (bddc->numVars-1)*2;
      *worstCase += bddc->numNodes*3;
    }
  }
  listDelAll(&lBddis);
}


int
modelCountRepeatingNOTs(Model *model)
{
  ListCell cBddi;
  int count = 0;

  listForeach(&model->bddcNot->lInsts, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    if (bddi->inputVars[0]->bddClass == model->bddcNot)
      count++;
  }
  
  return count;
}

/*
  function: modelGetNodeCount()
  desc: Returns the number of BDD nodes used to represent the circuit.
*/
int
modelGetNodeCount(Model *model)
{
  int nc, best, worst;

  modelGetCost(model, &nc, &best, &worst);

  return nc;
}

Model *
modelCreate(char *name)
{
  Model *model;
  DdManager *ddManager;
  int i;
  model = comAlloc(_fbddHeap, sizeof(Model));

  strcpy(model->name, name);

  listConstruct0(&model->lOutputs, _fbddHeap);
  listConstruct0(&model->lInputs, _fbddHeap);
  listConstruct0(&model->lBOutputs, _fbddHeap);

  //  model->numInputVars=0;
  model->bclm = NULL;
  model->ddManager = newDdManager();
  ddManager = model->ddManager;

  // Create variables so that index 0 is at the bottom. 
  // (by default, it's at the top)
  // DANGER: maximum number of bdd variables may become greater than 
  // size of lInputs.
  for (i=0; i<MAX_BDD_VARS; i++)
    Cudd_bddNewVarAtLevel(model->ddManager, 0);

  model->nextBddcId = 0;

  manArrayConstruct(&model->manArrayFold);
  manArrayConstruct(&model->manArrayDecompose);

  model->oldIndexToNewSF = (int*)comAlloc(_fbddHeap, sizeof(int)*MAX_BDD_VARS);
  for (i=0; i<MAX_BDD_VARS; i++)
    model->oldIndexToNewSF[i] = i;

  // build Logic1 and Logic0 classes
  model->nodeLogicZero = buildLogicZero(ddManager);
  model->nodeLogicOne = buildLogicOne(ddManager);
  model->nodeSF = buildSF(ddManager);
  model->nodeMUX = buildMUX(ddManager);
  model->nodeMUX2 = buildMUX2(ddManager);
  model->nodeAND = buildAND(ddManager);
  model->nodeOR = buildOR(ddManager);
  model->nodeXNOR = buildXNOR(ddManager);
  model->nodeNOT = buildNOT(ddManager);  
  model->nodeNAND = Cudd_Complement(model->nodeAND);
  Cudd_Ref(model->nodeNAND);
  model->nodeNOR = Cudd_Complement(model->nodeOR);
  Cudd_Ref(model->nodeNOR);
  model->nodeXOR = Cudd_Complement(model->nodeXNOR);
  Cudd_Ref(model->nodeXOR);
  model->nodeAND01 = buildAND01(ddManager);
  model->nodeAND10 = buildAND10(ddManager);
  
  // build mux bddc
  model->bddcLogicZero = 
    bddcCreate("zero", model->nodeLogicZero, INTERNAL_VARIABLE, 0);
  model->bddcLogicOne = 
    bddcCreate("one", model->nodeLogicOne, INTERNAL_VARIABLE, 0);
  model->bddcSF = bddcCreate("sf", model->nodeSF, INTERNAL_VARIABLE, 1);
  model->bddcMux = bddcCreate("mux", model->nodeMUX, INTERNAL_VARIABLE, 3);
  model->bddcMux2 = bddcCreate("mux2", model->nodeMUX2, INTERNAL_VARIABLE, 3);
  model->bddcAnd = bddcCreate("and", model->nodeAND, INTERNAL_VARIABLE, 2);
  model->bddcOr = bddcCreate("or", model->nodeOR, INTERNAL_VARIABLE, 2);
  model->bddcXnor = bddcCreate("xnor", model->nodeXNOR, INTERNAL_VARIABLE, 2);
  model->bddcNot = bddcCreate("not", model->nodeNOT, INTERNAL_VARIABLE, 1);
  model->bddcNand = bddcCreate("nand", model->nodeNAND, INTERNAL_VARIABLE, 2);
  model->bddcNor = bddcCreate("nor", model->nodeNOR, INTERNAL_VARIABLE, 2);
  model->bddcXor = bddcCreate("xor", model->nodeXOR, INTERNAL_VARIABLE, 2);

  model->bddcLatchFe = bddcCreate("LATCH_FE", NULL, LATCH, 2);
  model->bddcLatchRe = bddcCreate("LATCH_RE", NULL, LATCH, 2);
  model->bddcLatchAh = bddcCreate("LATCH_AH", NULL, LATCH, 2);
  model->bddcLatchAl = bddcCreate("LATCH_AL", NULL, LATCH, 2);
  model->bddcLatchAs = bddcCreate("LATCH_AS", NULL, LATCH, 2);
  model->bddcPrimaryOut = 
    bddcCreate("PRIMARY_OUTPUT", NULL, PRIMARY_OUTPUT, 1);
  model->bddcPrimaryIn = bddcCreate("PRIMARY_INPUT", NULL, PRIMARY_INPUT, 0);
  model->bddcClock = bddcCreate("CLOCK", NULL, PRIMARY_INPUT, 0);

  model->bddiLogicOne = NULL;
  model->bddiLogicZero = NULL;

  model->bddiDummy = bddcNewInstance(model->bddcPrimaryOut);

  // keep alive
  listAppend(&model->bddiDummy->lFanout, model->bddiDummy); 
  model->bddiGenericClk = NULL;

  return model;
}

void
modelDestroy(Model *model)
{
  struct _List lBddcs;
  ListCell cBddc;

  // destroy bddis and bddcs
  listConstruct0(&lBddcs, _fbddHeap);
  modelBddcList(_fbddHeap, model, &lBddcs);
  listForeach(&lBddcs, cBddc){
    BddClass *bddc = listData(BddClass*, cBddc);
    if (!isBasicGate(model, bddc))
      bddcDestruct(bddc);
  }
  listDelAll(&lBddcs);

  bddcDestruct(model->bddcPrimaryOut);
  bddcDestruct(model->bddcPrimaryIn);
  bddcDestruct(model->bddcLogicZero);
  bddcDestruct(model->bddcLogicOne);
  bddcDestruct(model->bddcSF);
  bddcDestruct(model->bddcMux);
  bddcDestruct(model->bddcMux2);
  bddcDestruct(model->bddcAnd);
  bddcDestruct(model->bddcOr);
  bddcDestruct(model->bddcXnor);
  bddcDestruct(model->bddcNot);
  bddcDestruct(model->bddcNand);
  bddcDestruct(model->bddcNor);
  bddcDestruct(model->bddcXor);
  bddcDestruct(model->bddcLatchFe);
  bddcDestruct(model->bddcLatchRe);
  bddcDestruct(model->bddcLatchAh);
  bddcDestruct(model->bddcLatchAl);
  bddcDestruct(model->bddcLatchAs);
  bddcDestruct(model->bddcClock);

  listDelAll(&model->lOutputs);
  listDelAll(&model->lInputs);
  assert(listGetSize(&model->lBOutputs) == 0);

  Cudd_Quit(model->ddManager);

  manArrayDestruct(&model->manArrayFold);
  manArrayDestruct(&model->manArrayDecompose);

  comFree(_fbddHeap, model->oldIndexToNewSF);

  comFree(_fbddHeap, model);
}

/*
  FUNC: modelComputeFanout()
  DESC:Computes list of pointers between a gate and its fanout.
  Populates BddInstance->lFanout with pointers to fanout gates.
 */
void
modelComputeFanout(Model *model)
{
  struct _List lBddis;
  ListCell cBddi;
  IMem mem = _fbddHeap;

  listConstruct0(&lBddis, mem);
  modelAllBddiList(mem, model, &lBddis);
  
  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);

    if (bddi->bddClass->type != PRIMARY_OUTPUT)
      listDelAll(&bddi->lFanout);
    listConstruct0(&bddi->lFanout, _fbddHeap);
  }  

  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    int childPos;
    int numVars = bddi->bddClass->numVars;

    if (bddi->bddClass->type == PRIMARY_INPUT)
      continue;

    for(childPos=0; childPos<numVars; childPos++){
      ListCell cNew;
      
      BddInstance *bddiChild = bddi->inputVars[childPos];
      listNewCell(_fbddHeap, cNew, bddi);
      listAppendElement(&bddiChild->lFanout, cNew);
    }
  }
  listDelAll(&lBddis);
}

typedef struct _MblData
{
  List list;
  IMem arena;
} MblData;
/*
  FUNC: modelBddiListHelper()
  DESC: Returns list of BddInstance's found in circuit.
 */
static void
modelBddiListHelper(MblData *mblData, BddInstance *bddi)
{
  ListCell newCell;
  int i;

  List list = mblData->list;
  IMem mblArena = mblData->arena;

  if (bddiGetMark(bddi, 12))
    return;
  bddiSetMark(bddi, 12, TRUE);

  listNewCell(mblArena, newCell, bddi);
  listAppendElement(list, newCell);

  // process children
  if (bddi->bddClass->type != PRIMARY_INPUT){
    for (i=0; i<bddi->bddClass->numVars; i++)
      modelBddiListHelper(mblData, bddi->inputVars[i]);
  }
}

void
modelBddiList(IMem mem, Model *model, List list)
{
  ListCell cell;
  MblData mblData;

  mblData.list = list;
  mblData.arena = mem;
  listForeach( &(model->lOutputs), cell ){
    modelBddiListHelper(&mblData, listData(BddInstance*, cell));
  }
  
  // remove bddi marks
  listForeach( list, cell ){
    BddInstance *bddi = listData(BddInstance*, cell);
    bddiSetMark(bddi,12,FALSE);
  }
}

// like modelBddiList() but also includes bddis in dangling networks.
void
modelAllBddiList(IMem mem, Model *model, List list)
{
  ListCell cell;
  MblData mblData;
  int i;

  mblData.list = list;
  mblData.arena = mem;
  listForeach( &(model->lOutputs), cell ){
    modelBddiListHelper(&mblData, listData(BddInstance*, cell));
  }
  for (i=0; i<5; i++){
    BddClass *bddcLatch;
    ListCell cBddi;

    switch(i){
    case 0: bddcLatch = model->bddcLatchFe; break;
    case 1: bddcLatch = model->bddcLatchRe; break;
    case 2: bddcLatch = model->bddcLatchAh; break;
    case 3: bddcLatch = model->bddcLatchAl; break;
    case 4: bddcLatch = model->bddcLatchAs; break;
    }
    listForeach(&bddcLatch->lInsts, cBddi){
      BddInstance *bddi = listData(BddInstance*, cBddi);
      modelBddiListHelper(&mblData, bddi);
    }
  }

  // remove bddi marks
  listForeach( list, cell ){
    BddInstance *bddi = listData(BddInstance*, cell);
    bddiSetMark(bddi,12,FALSE);
  }
}


void
bddiDestruct(BddInstance *bddi)
{
  comFree(_fbddHeap, bddi->inputVars);
  listDelAll(&bddi->lFanout);
  if (bddi->id != NULL)
    comFree(_fbddHeap, bddi->id);

  bddiFreeSeeList(bddi);

  comFree(_fbddHeap, bddi->lSees);
  comFree( _fbddHeap, bddi );
}

// Returns the position of bddiChild in bddiParent.
// Returns -1 if failed.
int bddiGetPosChildInParent(BddInstance *bddiParent, BddInstance *bddiChild)
{
  int i;
  int numVars = bddiParent->bddClass->numVars;

  for (i=0; i<numVars; i++){
    if (bddiParent->inputVars[i] == bddiChild){
      return i;
    }
  }
  return -1;
}

// Literal Count.
// Requires the model be decomposed
int
modelLitCount(Model *model)
{
  IMem mem;
  struct _List lBddis;
  ListCell cBddi;
  int litCount = 0;

  mem = _fbddHeap;

  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);
  
  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    BddClass *bddc = bddi->bddClass;

    if (bddc == model->bddcMux || bddc == model->bddcXnor 
	|| bddc == model->bddcXor){
      litCount += 4;
    }
    else if (bddc == model->bddcSF || bddc == model->bddcNot){
      ; // add nothing
    }
    else if (bddc == model->bddcLogicOne || bddc == model->bddcLogicZero){
      ; // add nothing
    }
    else{
      // all other gates on a decomposed network has a literal count of 2
      litCount += 2;
    }
  }

  listDelAll(&lBddis);

  return litCount;
}

// Returns the number of bddi's with multi fanout
int
modelCountMFBddis(Model *model)
{
  int numMFBddis = 0;
  IMem arena = _fbddHeap;
  struct _List lBddis;
  ListCell cBddi;

  listConstruct0(&lBddis, _fbddHeap);
  modelBddiList(arena, model, &lBddis);
  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    if (bddi->bddClass->type != PRIMARY_INPUT
	&& bddi->bddClass->type != PRIMARY_OUTPUT
	&& bddiNumFanout(bddi) > 1)
      numMFBddis++;
  }
  listDelAll(&lBddis);

  return numMFBddis;
}

void
modelPrintGates(Model *model)
{
  int numLatches = 0;
  int numPriInputs = 0;
  int numPriOutputs = 0;
  int numGates[MAX_BDD_VARS];
  int i;
  int area=0;

  IMem arena = _fbddHeap;
  struct _List lBddcs;
  ListCell cBddc;

  for (i=0; i<MAX_BDD_VARS; i++)
    numGates[i] = 0;

  listConstruct0(&lBddcs, arena);
  modelBddcList(arena, model, &lBddcs);
  listForeach(&lBddcs, cBddc){
    BddClass *bddc = listData(BddClass*, cBddc);
    if (bddc->type == PRIMARY_INPUT)
      numPriInputs += bddcNumInsts(bddc);
    else if (bddc->type == PRIMARY_OUTPUT)
      numPriOutputs += bddcNumInsts(bddc);
    else if (bddc->type == LATCH)
      numLatches += bddcNumInsts(bddc);
    else{
      numGates[bddc->numVars] += bddcNumInsts(bddc);
      if (bddc == model->bddcXor || bddc == model->bddcXnor)
	area += 7*bddcNumInsts(bddc);
      else 
	area += 4*bddcNumInsts(bddc);
    }
  }
  listDelAll(&lBddcs);

  printf("Gate Composition\n");
  printf("Primary Inputs:   %6d\n", listGetSize(&model->lInputs));
  printf("Primary Outputs:  %6d\n", listGetSize(&model->lOutputs));
  printf("Latches:          %6d\n", numLatches);
  for (i=0; i<MAX_BDD_VARS; i++)
    if (numGates[i]>0)
      printf("%3d Input:        %6d\n", i, numGates[i]);
  printf("AREA: %d\n", area);
}

// check no repeating inputs
// and no cycles
int
checkBddi(BddInstance *bddi)
{
  int numVars = bddi->bddClass->numVars;
  int i;
  int result = TRUE;

  for (i=0; i<numVars; i++){
    assert(bddi->inputVars[i] != NULL);
    if ((bddi->inputVars[i]->mark&0x8) != 0) 
      assert(FALSE);
    if (bddi->inputVars[i] == NULL){
      result = FALSE;
      break;
    }
    bddi->inputVars[i]->mark |= 0x8;
  }

  for (i=0; i<numVars; i++){
    bddi->inputVars[i]->mark -= 0x8;
  }

  return result;
}

/*
  FUNC: checkBddcsBddis()
  DESC: Check consistancy between BddClasses and BddInstances.  Check all
  BddInstances found in BddClasses are reachable.  Check no repeating inputs.
  
 */
int
checkBddcsBddis(Model *model)
{
  struct _List lBddcs;
  struct _List lBddis;
  ListCell cBddi;
  ListCell cBddc;

  int result = TRUE;

  IMem mem = _fbddHeap;

  listConstruct0(&lBddcs, mem);
  listConstruct0(&lBddis, mem);

  modelBddcList(mem, model, &lBddcs);
  modelBddiList(mem, model, &lBddis);

  // check there are no bddi's in a bddc that don't have a node 
  // in the dag.

  // delete all Bddis
  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    List list = &bddi->bddClass->lInsts;
    ListCell cell;
    int found = FALSE;

    assert(checkBddi(bddi));
    
    // check that a bddi can be found in it's bddc.
    listForeach(list, cell){
      BddInstance *b = listData(BddInstance*, cell);

      if (b == bddi){
	assert(cell == bddi->cBddcInst);
	assert(found == FALSE);
	found = TRUE;
      }
    }
    if (found == FALSE){
      printf("bddi:%x not found in bddc:%x\n", 
	     (unsigned int)bddi, 
	     (unsigned int)bddi->bddClass);
    }
    assert(found);
  }

  listForeach(&lBddcs, cBddc){
    BddClass *bddc = listData(BddClass*, cBddc);
    listForeach(&bddc->lInsts, cBddi){
      BddInstance *bddi = listData(BddInstance*, cBddi);
      if (checkBddiReachable(model, bddi)==FALSE){
	result = FALSE;
	assert(FALSE);
	break;
      }
    }
  }

  listDelAll(&lBddcs);
  listDelAll(&lBddis);

  return result;
}

int
checkFaninFanout(Model *model)
{
  struct _List lBddis;
  ListCell cBddi;
  IMem mem = _fbddHeap;

  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);  

  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    ListCell cBddiParent;
    int i;

    // check bddiParents point to bddi
    if (bddi->bddClass->type != PRIMARY_OUTPUT){
      listForeach(&bddi->lFanout, cBddiParent){
	BddInstance *bddiParent = listData(BddInstance*, cBddiParent);
	int found = FALSE;
	
	for (i=0; i<bddiParent->bddClass->numVars; i++){
	  if (bddiParent->inputVars[i] == bddi){
	    found = TRUE;
	    break;
	  }
	}
	if (found == FALSE){
	  assert(found);
	  return FALSE;
	}
      }
    }

    //check bddiChild points to bddi
    if (bddi->bddClass->type != PRIMARY_INPUT){
      for (i=0; i<bddi->bddClass->numVars; i++){
	BddInstance *bddiChild = bddi->inputVars[i];
	ListCell cFanout;
	int found = FALSE;
	
	listForeach(&bddiChild->lFanout, cFanout){
	  if (listData(BddInstance*, cFanout)==bddi){
	    found = TRUE;
	    break;
	  }
	}
	if (found == FALSE){
	  assert(FALSE);
	  return FALSE;
	}
      }
    }
  }
  listDelAll(&lBddis);  

  return TRUE;
}

int
checkBddiReachable(Model *model, BddInstance *bddi)
{
  struct _List lBddis;
  ListCell cBddi;
  int result = FALSE;

  IMem mem = _fbddHeap;

  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);  

  listForeach(&lBddis, cBddi){
    BddInstance *bddi2 = listData(BddInstance*, cBddi);
    if (bddi2 == bddi){
      result = TRUE;
      break;
    }
  }
  listDelAll(&lBddis);

  return result;
}

int 
bddiReachable(Model *model, BddInstance *bddi)
{
  return checkBddiReachable(model, bddi);
}

int
checkAllBddiReachable(Model *model)
{
  struct _List lBddis;
  ListCell cBddi;
  int result = FALSE;

  IMem mem = _fbddHeap;

  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);  

  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    bddiReachable(model, bddi);
  }
  listDelAll(&lBddis);

  return result;
}

int
checkNoConsts(Model *model)
{
  IMem mem = _fbddHeap;
  struct _List lBddcs;
  BddClass *bddc;
  ListCell cBddc;

  listConstruct0(&lBddcs, mem);
  modelBddcList(mem, model, &lBddcs);
  listForeach(&lBddcs, cBddc){
    bddc = listData(BddClass*, cBddc);
    if (bddc->numVars == 0){
      ListCell cBddi;
      // check that all fanouts are primary outputs
      listForeach(&bddc->lInsts, cBddi){
	ListCell cBddiFanout;
	BddInstance *bddi = listData(BddInstance*, cBddi);
	listForeach(&bddi->lFanout, cBddiFanout){
	  BddInstance *bddiFanout = listData(BddInstance*, cBddiFanout);
	  if (bddiFanout->bddClass->type != PRIMARY_OUTPUT)
	    return FALSE;
	}
      }
    }
  }

  return TRUE;
}

int
checkSimplified(Model *model)
{
  int result;
  IMem mem = _fbddHeap;
  struct _List lBddis;
  BddInstance *bddi;
  ListCell cBddi;

  result = TRUE;

  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);
  listForeach(&lBddis, cBddi){
    InputConfig ic;
    bddi = listData(BddInstance*, cBddi);
    
    if (bddi->bddClass->type != INTERNAL_VARIABLE)
      continue;

    ic = icCompute(mem, bddi->bddClass->numVars, TRUE, bddi->inputVars);
    if (icNumSS(ic)>0){
      result = FALSE;
      comFree(mem, ic);
      break;
    }
    comFree(mem, ic);
  }
  listDelAll(&lBddis);

  return result;
}


DdManager *
newDdManager()
{
  DdManager *ddManager = Cudd_Init(0, 0, CUDD_UNIQUE_SLOTS, 
				      BDS_BDD_STARTUP_CACHE,
				      BDS_BDD_STARTUP_MEMORY);

  Cudd_SetMaxGrowth(ddManager, 1.20);
  ddManager->maxGrowth = 1.20;
  ddManager->siftMaxVar = 1000000;
  ddManager->siftMaxSwap = 1000000000;
  return ddManager;
}

#define MAN_ARRAY_SIZE (20)
#define MAN_VAR_SIZE (10)
void
manArrayConstruct(ManArray *manArray)
{
  int i;

  manArray->ddManagers
    = (DdManager**)comAlloc(_fbddHeap, sizeof(DdManager*)*MAN_ARRAY_SIZE);
  for (i=0; i<MAN_ARRAY_SIZE; i++)
    manArray->ddManagers[i] = NULL;
  manArray->activeManager = NULL;
}

void
manArrayDestruct(ManArray *manArray)
{
  int i;
  DdManager **ddManagers = manArray->ddManagers;
  for (i=0; i<MAN_ARRAY_SIZE; i++)
    if (ddManagers[i] != NULL)
      Cudd_Quit(ddManagers[i]);
  comFree(_fbddHeap, manArray->ddManagers);
}

DdManager *
getDdManager(ManArray *manArray, int numVars)
{
  DdManager **ddManagers = manArray->ddManagers;
  int index = numVars / MAN_VAR_SIZE;

  assert(manArray->activeManager == NULL);

  if (index > MAN_ARRAY_SIZE-1)
    index = MAN_ARRAY_SIZE-1;
  if (ddManagers[index] == NULL){
    ddManagers[index] = newDdManager();
  }

  {
    int l;
    l = Cudd_ReadNodeCount(ddManagers[index]);
    assert(l==1);
    l = Cudd_ReadSize(ddManagers[index]);
    if(l < numVars){
      int i;
      for (i=0; i<numVars; i++){
	Cudd_bddIthVar(ddManagers[index], i);
      }
    }
  }

  // check ddManager does not have more variables than it should have
  if (index > MAN_ARRAY_SIZE-1)
    ;
  else{
    int s = (index+1) * MAN_VAR_SIZE;
    assert(s > 0);
  }

  manArray->activeManager = ddManagers[index];
  return ddManagers[index];
}

void
releaseDdManager(ManArray *manArray)
{
  assert(manArray->activeManager != NULL);

  manArray->activeManager = NULL;
}

// updateBddisForRoot
void
updateBddisForRoot(Model *model, BddInstance *bddi, BddClass *bddc,
		   
		   int *oldIndexToNewA1, int arraySizeA1,
		   BddInstance **oldInputVars,

		   int *oldIndexToNewA2, int arraySizeA2,
		   BddInstance **oldInputVars2,

		   int *oldIndexToNewB1, int arraySizeB1,
		   int additionalPos, BddInstance *additionalBddi)
{
  int oldNumVars;
  int j;

  oldNumVars = bddi->bddClass->numVars;
  bclRemoveBddi(model->bclm, model, bddi);

  bddcAddInstance(bddc, bddi);
  assert(oldNumVars>=arraySizeA1);
  for (j=0; j<oldNumVars; j++)
    if (j<arraySizeA1
	&& oldIndexToNewA1[j] != -1
	&& oldIndexToNewA1[j]<arraySizeB1
	&& oldIndexToNewB1[oldIndexToNewA1[j]] != -1){
      assert(oldIndexToNewB1[oldIndexToNewA1[j]]>=0);
      assert(oldIndexToNewB1[oldIndexToNewA1[j]]
	     < bddi->bddClass->numVars);
      assert(bddi->inputVars[oldIndexToNewB1[oldIndexToNewA1[j]]] == NULL);
      bddi->inputVars[oldIndexToNewB1[oldIndexToNewA1[j]]] = oldInputVars[j];
    }
    else{
      // variable j no longer is in the support of bddi2
      listDelData(&oldInputVars[j]->lFanout, bddi);
    }

  if (oldIndexToNewA2 != NULL)
    // support can come from a source different from the original function 
    // if created under sharing extraction.
    for (j=0; j<arraySizeA2; j++)
      if (j<arraySizeA2
	  && oldIndexToNewA2[j] != -1
	  && oldIndexToNewA2[j]<arraySizeB1
	  && oldIndexToNewB1[oldIndexToNewA2[j]] != -1){
	assert(oldIndexToNewB1[oldIndexToNewA2[j]]>=0);
	
	if (bddi->inputVars[oldIndexToNewB1[oldIndexToNewA2[j]]] != NULL)
	  assert(bddi->inputVars[oldIndexToNewB1[oldIndexToNewA2[j]]]
		 == oldInputVars2[j]);
	else{
	  bddi->inputVars[oldIndexToNewB1[oldIndexToNewA2[j]]] 
	    = oldInputVars2[j];
	  
	  listAppend(&oldInputVars2[j]->lFanout, bddi);
	}
      }

  assert(oldIndexToNewB1[additionalPos] >= 0 && 
	 oldIndexToNewB1[additionalPos]<bddi->bddClass->numVars);
  if (oldIndexToNewB1[additionalPos] != -1){
    assert(bddi->inputVars[oldIndexToNewB1[additionalPos]] == NULL);
    
    bddi->inputVars[oldIndexToNewB1[additionalPos]] = additionalBddi;
  }

#ifdef DEBUG_COMPILE
  for (j=0; j<bddi->bddClass->numVars; j++)
    assert(bddi->inputVars[j] != NULL);
#endif
}

BddInstance *
createBddiForExt(Model *model, BddClass *bddc, 
		 BddInstance **oldInputVars, int oldNumVars,
		 int *oldIndexToNewA, int arraySizeA, 

		 BddInstance **oldInputVars2, int oldNumVars2,
		 int *oldIndexToNewA2, int arraySizeA2, 

		 int *oldIndexToNewB, int arraySizeB,
		 BddInstance *bddiFanout1, BddInstance *bddiFanout2)
{
  BddInstance *bddiExt;
  int j;

  bddiExt = bddcNewInstance(bddc);
  for (j=0; j<arraySizeA; j++)
    if ( oldIndexToNewA[j]!=-1 
	 && oldIndexToNewA[j]<arraySizeB
	 && oldIndexToNewB[oldIndexToNewA[j]] != -1){
      assert(oldIndexToNewB[oldIndexToNewA[j]]>=0);
      assert(oldIndexToNewB[oldIndexToNewA[j]]<bddc->numVars);
      assert(bddiExt->inputVars[oldIndexToNewB[oldIndexToNewA[j]]]
	     == NULL);
      bddiExt->inputVars[oldIndexToNewB[oldIndexToNewA[j]]]
	= oldInputVars[j];
    }

  if (oldIndexToNewA2 != NULL)
    for (j=0; j<arraySizeA2; j++)
      if ( oldIndexToNewA2[j]!=-1 
	   && oldIndexToNewA2[j]<arraySizeB
	   && oldIndexToNewB[oldIndexToNewA2[j]] != -1){
	assert(oldIndexToNewB[oldIndexToNewA2[j]]>=0);
	assert(oldIndexToNewB[oldIndexToNewA2[j]]<bddc->numVars);

	if (bddiExt->inputVars[oldIndexToNewB[oldIndexToNewA2[j]]] != NULL)
	  assert(bddiExt->inputVars[oldIndexToNewB[oldIndexToNewA2[j]]]
		 == oldInputVars2[j]);
	else
	  bddiExt->inputVars[oldIndexToNewB[oldIndexToNewA2[j]]]
	    = oldInputVars2[j];
      }

  if (bddiFanout1 != NULL)
    listAppend(&bddiExt->lFanout, bddiFanout1);
  if (bddiFanout2 != NULL)
    listAppend(&bddiExt->lFanout, bddiFanout2);
  
#ifdef DEBUG_COMPILE
  for (j=0; j<bddc->numVars; j++)
    assert(bddiExt->inputVars[j] != NULL);
#endif

  // update fanout of support
  for (j=0; j<bddc->numVars; j++)
    listAppend(&bddiExt->inputVars[j]->lFanout,
		   bddiExt);

  return bddiExt;
}

int 
getOldIndexFromNew(int *oldIndexToNew, int arraySize, int newIndex)
{
  int i;
  for (i=0; i<arraySize; i++)
    if (oldIndexToNew[i] == newIndex)
      return i;
  return -1;
}

// creates a BddClass for the Bdd if a BddClass does
// not already exist for it.
BddClass*
bddcCreateForBdd(Model *model, DdNode *node, 
		 int numVars, int *nextId)
{
  DdManager *ddManager = model->ddManager;
  BclManager bclm = model->bclm;
  BddClass *bddc;
  char name[30];
  
  // create BddClass for node if it does not already exist
  bddc = bclGet(bclm, node);
  if (bddc == NULL){
    // BddClass did not exist before
    sprintf(name, "cell_%d", (*nextId)++);
    bddc = bddcCreate(name, node, INTERNAL_VARIABLE, numVars);
    bclPut(bclm, node, bddc); 
    
    cbRun(IFbddCallBack, _fm, onNewBddc, bddc);
  }
  else{
    // bdd already exists, don't need 2 references to it.
    Cudd_RecursiveDeref(ddManager, node);
  }

  return bddc;
}

static int 
_log2(int n)
{
  if (n<1){
    assert(FALSE);
    return -1;
  }
  else if (n==1)
    return 0;
  else if (n==2)
    return 1;
  else if (n>2 && n<=4)
    return 2;
  else if (n>4 && n<=8)
    return 3;
  else if (n>8 && n<=16)
    return 4;
  else if (n>16 && n<=32)
    return 5;
  else if (n>32 && n<=64)
    return 6;
  else if (n>64 && n<=128)
    return 7;
  else if (n>128 && n<=256)
    return 8;
  else if (n>256 && n<=512)
    return 9;
  else{
    assert(FALSE);
    return -1;
  }
}

// computes a BDD representation of a number using
// variables between lowerBound and upperBound
static DdNode*
computeBddNumberFromRange(DdManager *ddManager, 
			  int num, int lowerBound, int upperBound)
{
  DdNode *bddResult;
  int numVars = upperBound-lowerBound+1;
  int i;
  
  bddResult = Cudd_ReadOne(ddManager);
  Cudd_Ref(bddResult);

  for (i=0; i<numVars; i++){
    DdNode *nVar, *bddTemp;
    nVar = Cudd_bddIthVar(ddManager, lowerBound+i);
    Cudd_Ref(nVar);

    if ((num & 0x1) == 1){
      // this bit is true
    }
    else{
      // bit is false
      nVar = Cudd_Complement(nVar);
    }
    bddTemp = Cudd_bddAnd(ddManager, nVar, bddResult);
    Cudd_Ref(bddTemp);
    Cudd_RecursiveDeref(ddManager, nVar);
    Cudd_RecursiveDeref(ddManager, bddResult);
    bddResult = bddTemp;
			    
    num = num >> 1;
  }

  return bddResult;
}

DdNode*
bddiGetSupportConfig(DdManager *ddManager, 
		     BddInstance *bddiParent, BddInstance *bddiChild,
		     int *shareSize)
{
  IMem mem = _fbddHeap;
  int nvParent = bddiParent->bddClass->numVars;
  int nvChild = bddiChild->bddClass->numVars;
  int logNvParent = _log2(nvParent);
  int logNvChild = _log2(nvChild);
  int i,j;
  DdNode *bddResult, *bddSS;
  DdNode *bddTemp;
  DdNode *bddParentVar, *bddChildVar;

  BddInstance **inputVars;

  int *inputConfig;
  
  inputVars
    = (BddInstance**)comAlloc(mem, sizeof(BddInstance*)*(nvParent+nvChild));

  *shareSize=0;
  
  for (i=0; i<nvParent; i++)
    inputVars[i] = bddiParent->inputVars[i];
  for (i=0; i<nvChild; i++)
    inputVars[nvParent+i] = bddiChild->inputVars[i];

  inputConfig = icCompute(mem, nvParent+nvChild, FALSE, inputVars);

  bddResult = Cudd_ReadLogicZero(ddManager);
  Cudd_Ref(bddResult);

  for(i=1+nvParent; i<1+nvParent+nvChild; i++)
    if ((j=inputConfig[i]) != i-1){

      (*shareSize)++;
      // variable 
      switch (j){
      case CONST_ONE:
	break;
      case CONST_ZERO:
	break;
      default:
	// input i-1-nvParent of child matches with input j of parent
	bddParentVar = computeBddNumberFromRange(ddManager,
						 j, 0, logNvParent-1);
	bddChildVar = computeBddNumberFromRange(ddManager,
						i-1-nvParent, 
						logNvParent,
						logNvChild+logNvParent-1);
	bddSS = Cudd_bddAnd(ddManager, bddParentVar, bddChildVar);
	Cudd_Ref(bddSS);
	Cudd_RecursiveDeref(ddManager, bddParentVar);
	Cudd_RecursiveDeref(ddManager, bddChildVar);

	// add cube to bddResult
	bddTemp = Cudd_bddOr(ddManager, bddSS, bddResult);
	Cudd_Ref(bddTemp);
	Cudd_RecursiveDeref(ddManager, bddSS);
	Cudd_RecursiveDeref(ddManager, bddResult);
	bddResult = bddTemp;

     	break;
      }
    }

  comFree(mem, inputConfig);
  comFree(mem, inputVars);
  
  return bddResult;
}

BddInstance *
copyNetwork(HashTable htBddiToCopy, BddInstance *bddi)
{
  BddInstance *bddiNew;
  int i;

  if (bddi->bddClass->type == PRIMARY_INPUT)
    return bddi;
  
  if ((bddiNew=hashGet(htBddiToCopy, bddi)) == NULL){
    bddiNew = bddcNewInstance(bddi->bddClass);
    hashPut(htBddiToCopy, bddi, bddiNew);
    for (i=0; i<bddi->bddClass->numVars; i++){
      bddiNew->inputVars[i] = copyNetwork(htBddiToCopy, bddi->inputVars[i]);
      listAppend(&bddiNew->inputVars[i]->lFanout, bddiNew);
    }
    if (bddi->bddClass->type == PRIMARY_OUTPUT){
      bddiNew->id =
	(char*)comAlloc(_fbddHeap, sizeof(char)*(strlen(bddi->id)+1));
      strcpy(bddiNew->id, bddi->id);
    }
    bddiNew->mark = bddi->mark;
  }
  return bddiNew;
}

void
modelBackup(Model *model)
{
  ListCell cOutput;
  struct _HashTable htBddiToCopy;

  hashDefaultConstruct(&htBddiToCopy, _fbddHeap);

  listConstruct0(&model->lBOutputs, _fbddHeap);

  listForeach(&model->lOutputs, cOutput){
    BddInstance *bddiCopy;
    BddInstance *bddiOutput = listData(BddInstance*,cOutput);

    bddiCopy = copyNetwork(&htBddiToCopy, bddiOutput);
    listAppend(&model->lBOutputs, bddiCopy);
  }
  hashDestruct(&htBddiToCopy);
}

void
modelClearBackup(Model *model)
{
  ListCell cOutput;
  
  // destroy existing model
  listForeach(&model->lBOutputs, cOutput){
    BddInstance *bddiOutput = listData(BddInstance*,cOutput);
    sweepDanglingNetwork(_fm->sm, bddiOutput);
  }
  listDelAll(&model->lBOutputs);
}

void
modelSwapBackup(Model *model)
{
  struct _List lOld;
  lOld = model->lOutputs;

  // swap active and backup outputs
  model->lOutputs = model->lBOutputs;
  model->lBOutputs = lOld;
}

void
modelRestore(Model *model)
{
  modelSwapBackup(model);
  modelClearBackup(model);
}

int
modelGetArea(Model *model)
{
  int smallGates = bddcNumInsts(model->bddcAnd)+bddcNumInsts(model->bddcNand)
    +bddcNumInsts(model->bddcOr)+bddcNumInsts(model->bddcNor);
  
  int largeGates = bddcNumInsts(model->bddcXor)
    +bddcNumInsts(model->bddcXnor)
    +bddcNumInsts(model->bddcMux)
    +bddcNumInsts(model->bddcMux2);
  
  return (4*smallGates + 7*largeGates);
}

// returns 0 if equal
// +1 if sc1 is better.
// -1 if sc2 is better.
int
scCmp(ShareCost *sc1, ShareCost *sc2)
{
  if (sc1->shareBdd==NULL)
  //  if (sc1->shareBdd==NULL || sc1->nNodeChange>0)
  //  if (sc1->shareBdd==NULL || sc1->nVarChange>0 || sc1->nNodeChange>0)
    return -1;

  if (sc2->shareBdd==NULL)
    return 1;

  if (sc1->nNodeChange + sc1->nVarChange 
      < sc2->nNodeChange + sc2->nVarChange){
    return 1;
  }
  else
    return -1;
}


void
mergeOldIndexToNew(int *oldIndexToNew1, int arraySize1, 
		   int *oldIndexToNew2, int arraySize2)
{
  int i;
  for (i=0; i<arraySize1; i++){
    if (oldIndexToNew1[i]!=-1 && oldIndexToNew1[i]<arraySize2)
      oldIndexToNew1[i] = oldIndexToNew2[oldIndexToNew1[i]];
  }
  comFree(_fbddHeap,oldIndexToNew2);
}
