/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <assert.h>

#include "cuddInt.h"
#include "cudd.h"
#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"
#include "fbdd.h"

void yyparse(void);

static void
_fbddPrintStats(FbddManager fm)
{
  Model *model = fm->model;

  printf("SUMMARY\n");
  printf("=======\n");

  printf("Runtime\n");
  printf("time sweep:         %10d\n", (int)tSweep);
  printf("time eliminate:     %10d\n", (int)tElim);
  printf("time share extract: %10d\n", (int)seGetTime());
  printf("time decompose:     %10d\n", (int)decompGetTime());
  printf("total time:         %10d\n", (int)tTotal);
  printf("\n");

  elimDump();
  seDump(NULL);
  decompDump();
  
  modelPrintGates(model);
  printf("\n");
}

int
fbddReadParameters(FbddManager fm, int argc, char* argv[])
{
  int i;

  if (argc > 1){
    printf("Options\n");
    printf("=======\n");
    if (argc == 2)
      // default options
      printf("    defaults options\n");

    for (i=1; i<argc-1; i++){
      printf("    %s\n", argv[i]);
      if (strcmp(argv[i], "-maxextsize")==0){
	i++;
	fm->options[FBDD_OPT_MAXEXTSIZE] = atoi(argv[i]);
	maxExtSize = atoi(argv[i]);
	if (maxExtSize > 5 || maxExtSize <2){
	  printf("Error: maxExtSize [%d] invalid!", maxExtSize);
	  exit(1);
	}
      }
      else if (strcmp(argv[i], "-ne")==0) // no eliminiation
	fm->options[FBDD_OPT_ELIM_DISABLE] = TRUE;
      else if (strcmp(argv[i], "-nd")==0) // no decomposition
	fm->options[FBDD_OPT_DECOMP_DISABLE] = TRUE;
      else if (strcmp(argv[i], "-nse")==0)
	fm->options[FBDD_OPT_SE_DISABLE] = TRUE; 
      else if (strcmp(argv[i], "-seexact")==0)
	fm->options[FBDD_OPT_SE_MODE] = FBDD_SE_MODE_EXACT;
      else if (strcmp(argv[i], "-f")==0) // no gbl elim (fast)
	fm->options[opt_fast] = TRUE;
      else if (strcmp(argv[i], "-verb")==0)
	fm->options[opt_verb] = TRUE; 
      else if (strcmp(argv[i], "-gblelim")==0)
	fm->options[opt_gblelim] = TRUE; 

      else if (strcmp(argv[i], "-nf")==0) // no folding
	fm->options[opt_nf] = TRUE;
      else if (strcmp(argv[i], "-ns")==0) // no sweep
	fm->options[opt_nsweep] = TRUE; 
      else if (strcmp(argv[i], "-rd")==0) // read decomposed
	fm->options[opt_rd] = TRUE;
      else if (strcmp(argv[i], "-printdata")==0) // run time fast
	fm->options[opt_printdata] = TRUE; 

      else if (strcmp(argv[i], "-tweconjext")==0){
	fm->options[opt_conj_ext] = TRUE;
      }
      else if (strcmp(argv[i], "")==0)
	; // ignore
      else{
	printf("error: unknown option %s!\n", argv[i]);
	exit(1);
      }
    }
  }
  else{
    return 0;
  }

  #ifndef FBDD_RETIRE
  maxExtSize = fm->options[FBDD_OPT_MAXEXTSIZE];
  #endif

  return TRUE;
}

static void
strippath(char *str)
{
  char *lastSlash=str; 
  char *curr = str;
  while(*curr != 0) {
    if (*curr == '/'){
      curr++;
      lastSlash = curr;
    }
    else
      curr++;
  }
  
  // remove path
  curr = str;
  while(TRUE){
    *curr = *lastSlash;
    if (*lastSlash == 0)
      break;
    curr++;
    lastSlash++;
  }

  return lastSlash;
}

int
main( int argc, char* argv[] )
{
  FbddManager fm;
  char *strFileBlifIn;
  char strFileBlifOut[50];
  char strFileVerilogOut[50];
  int i;

  const char strUsage[500] = "Usage: fbdd [-options] ckt.blif\n\
    -maxextsize size     find extractors up to size 'size' (default 2)\n\
    -seexact             perform exact sharing extraction \n\
    -sefast              perform fast sharing extraction (default) \n\
    -ne                  skip elimination \n\
    -nd                  skip decomposition \n\
    -nse                 skip sharing extraction \n";
  
  tTotal = util_cpu_time();

  printf("University of Toronto, FBDD 1.0\n");

  comBegin();

  fm = fbddNewManager();

  /*********************
   *  Read Parameters  *
   *********************/
  if (fbddReadParameters(fm, argc, argv)){
    time_t today;

    fbddFinalizeOptions(fm);

    printf("Synthesis report for %s\n", argv[argc-1]);
    time(&today);
    printf("%s", ctime(&today));
    printf("\n");
  }
  else{
    printf(strUsage);
    return 0;
  }
  /*  Generate Input / Ouput File Names  */
  strFileBlifIn = argv[argc-1];
  for (i=strlen(strFileBlifIn)-1; i>=0; i--){
    if (strFileBlifIn[i] == '.'){
      strFileBlifIn[i]=0;
      break;
    }
  }
  strcpy(strFileBlifOut, strFileBlifIn);
  strippath(strFileBlifOut); 
  strcat(strFileBlifOut, "_opt.blif");
  strcpy(strFileVerilogOut, strFileBlifIn);
  strippath(strFileVerilogOut); 
  strcat(strFileVerilogOut, "_opt.v");
  strFileBlifIn[strlen(strFileBlifIn)] = '.'; // put '.blif' back

 /********************************
   *          Read Blif           *
   ********************************/
  printf("Reading Circuit...\n");
  if (!fbddReadBlif(fm, strFileBlifIn)){
    printf("Error: File not found [%s]!\n", strFileBlifIn);
    return 0;
  }

  /********************************
   *         Optimize             *
   ********************************/
  fbddOptimize(fm);
  tTotal = util_cpu_time() - tTotal;

  /********************************
   *         Write Output         *
   ********************************/
  printf("Write circuit...\n\n");
  fbddWriteBlif(fm, strFileBlifOut);
  fbddWriteVerilog(fm, strFileVerilogOut);
  _fbddPrintStats(fm);

  /* Free memory */
  fbddDelManager(fm);
  comEnd();

  return 0;
}

