/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
  The purpose of this file is to provide fast 
  lookup of node markings.  Ideally, node markings should
  be a part of the DdNode structure but to use CUDD as is,
  I am using this lookup table instead.
  nml = node markings lookup.

  This used in two places
  1) When finding Algebraic dominators
  2) When copying bdds
*/

#include <stdlib.h>

#include "metautil.h"
#include "fbddutil.h"

// extern IHashScheme hashSPtrScheme;

void
nmlConstruct(Nml *nml)
{
  nml->arena = _fbddHeap;
  nml->hashTable = comAlloc(nml->arena, sizeof (struct _HashTable));
  hashDefaultConstruct(nml->hashTable, nml->arena);
  //  nml->hashTable->scheme = hashSPtrScheme;
}

void
nmlDestruct(Nml *nml)
{
  hashDestruct( nml->hashTable );
  comFree(nml->arena, nml->hashTable);
}

DdNode*
nmlGetNode(Nml *nml, DdNode *node)
{
  return (DdNode*)(hashGet(nml->hashTable, node));
}

void
nmlPutNode(Nml *nml, DdNode *node, DdNode *nodeB)
{
  HashCell hc;

  hc = hashRemove(nml->hashTable, (void*)node);
  if (hc != NULL)
    comFree(nml->hashTable->mem, hc);

  hashPut(nml->hashTable, (void*)node, (void*)nodeB);
}

int
nmlGetMark(Nml *nml, DdNode *node)
{
  return (int)(hashGet(nml->hashTable, node));
}

void
nmlPutMark(Nml *nml, DdNode *node, int marking)
{
  HashCell hc;

  hc = hashRemove(nml->hashTable, (void*)node);
  if (hc != NULL)
    comFree(nml->hashTable->mem, hc);
  hashPut(nml->hashTable, (void*)node, (void*)marking);
}

void
nmlIncrMark(Nml *nml, DdNode *node)
{
  //  int i = (int)hashGet(nml->hashTable, node);
  int i=0;
  HashCell hc = hashRemove(nml->hashTable, (void*)node);
  if (hc != NULL){
    i = hashData(int, hc);
    comFree(nml->hashTable->mem, hc);
  }

  hashPut(nml->hashTable, node, (void*)(i+1));
}

long
nmlGetLMark(Nml *nml, DdNode *node)
{
  //  printf("nmlGetMark\n");
  long *l;
  l = hashGet(nml->hashTable, node);

  if (l == NULL)
    return -1;
  else
    return *l;
}

void
nmlPutLMark(Nml *nml, DdNode *node, long marking)
{
  //  printf("nmlPutMark\n");
  long *l;
  HashCell hc;

  hc = hashRemove(nml->hashTable, (void*)node);
  if (hc != NULL){
    l = hashData(long*, hc);
    comFree(nml->hashTable->mem, hc);
  }
  else{
    l = (long*)comAlloc( nml->arena, sizeof(long) );
  }

  *l = marking;
			  
  hashPut(nml->hashTable, (void*)node, (void*)l);
}

int
nmlGetBddiMark(Nml *nml, BddInstance *bddi)
{
  return (int)(hashGet(nml->hashTable, bddi));
}

void
nmlPutBddiMark(Nml *nml, BddInstance *bddi, int marking)
{
  HashCell hc;
  if ((hc=hashRemove(nml->hashTable, bddi)) != NULL){
    comFree(nml->hashTable->mem, hc);
  }

  hashPut(nml->hashTable, (void*)bddi, (void*)marking);
}

void
nmlRemove(Nml *nml, DdNode *node)
{
  HashCell hc;
  hc = hashRemove(nml->hashTable, (void*)node);
  if (hc != NULL)
    comFree(nml->hashTable->mem, hc);
}
