/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "cuddInt.h"
#include "cudd.h"

#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"
#include "reader.h"

#define BOOL char
#define TRUE 1
#define FALSE 0

typedef struct _Consumer
{
  BddInstance *bddi;
  int pos; // the index into bddi->inputVars[] that consumes this signal.
} Consumer;

struct _Signal
{
  char *name;

  BddInstance *bddiProducer;
  struct _List lConsumers;
};

static DdManager *_manager;
static Model *_model;
static struct _HashTable __htSignals;
static HashTable _htSignals = &__htSignals;
static IMem _arena;
static DdNode *_nodeMUX;

int yyparse();
extern FILE *yyin; // the blifFile stream, defaults to stdin.

// Called at the end of the 'logic_gate' rule.
// Create bddi (and bddc if it doesn't already exist) for this logic gate.
// And update the signals in _htSignals affected by this logic gate.

struct _List lCubes;
static struct _List lSSCubeSets; // List of cubesets with sharedSupport

typedef struct _SSCubes{
  struct _BitSet bs;
  struct _List lCubes;
  ListCell listCell; // cell in lSSCubeSets
  char genSingleGate;
} SSCubes;

Signal *
sigCreate(IMem imem, char *name)
{
  Signal *signal = (Signal*)comAlloc( imem, sizeof(Signal) );
  signal->bddiProducer = NULL;
  listConstruct0(&(signal->lConsumers), _arena);
  signal->name = (char*)comAlloc(imem, sizeof(char)*(strlen(name)+1));
  strcpy(signal->name, name);

  return signal;
}

void
sigDestruct(IMem mem, Signal *signal)
{
  comFree(mem, signal->name);
  listDelAll(&signal->lConsumers);
  comFree(mem, signal);
}

// Parses and constructs the Model for strBlifFile.
// Returns TRUE if successful, FALSE otherwise.
char
fbddReadBlif(FbddManager fm, char* strBlifFile)
{
  FILE *blifFile;
  Model *model = fm->model;

  if ((blifFile = fopen(strBlifFile, "r")) == NULL){
    return 0;
  }

  _model = model;
  yyin = blifFile; // directs bison to parse the file pointed to by blifFile.
  yyparse();

  fclose(blifFile);

  modelComputeFanout(model);

  // remove forwarding gates
  /*
  {
    ListCell cell;
    listForeach(&(model->lOutputs), cell)
      bddiCollapseGates(_fbddHeap, 
			listData(BddInstance*,cell), model->bddcSF);
  }

  modelComputeFanout(model);
  */

  sweepDanglingLatchNetworks(fm->sm);

  return TRUE;
}

void
blifAddPrimaryOutput(char *name)
{
  Signal *signal;
  Consumer *consumer;
  BddInstance *bddi;
  
  bddi = bddcNewInstance(_model->bddcPrimaryOut);
  listAppend(&_model->lOutputs, bddi);
  
  bddi->id = (char*)comAlloc(_fbddHeap, sizeof(char)*(strlen(name)+1));
  strcpy(bddi->id, name);
  
  // make entry in _htSignals
  signal = sigCreate(_arena, bddi->id);
  signal->bddiProducer = NULL;
  consumer = comAlloc(_arena, sizeof(Consumer));
  consumer->bddi = bddi;
  consumer->pos = 0;
  
  listAppend(&signal->lConsumers, consumer);
  hashPut(_htSignals, bddi->id, signal);
}

void
blifAddPrimaryInput(char *name)
{
  Signal *signal;
  BddInstance *bddi;
  
  bddi = bddcNewInstance(_model->bddcPrimaryIn);
  listAppend(&_model->lInputs, bddi);
  
  bddi->id = (char*)comAlloc(_fbddHeap, sizeof(char)*(strlen(name)+1));
  strcpy(bddi->id, name);
  
  // make entry in _htSignals
  signal = sigCreate(_arena, bddi->id);
  signal->bddiProducer = bddi; 
  hashPut(_htSignals, bddi->id, signal);
}

void
blifAddClock(char *name)
{
  Signal *signal;
  BddInstance *bddi;

  bddi = bddcNewInstance(_model->bddcClock);
  listAppend(&_model->lInputs, bddi);
  
  bddi->id = (char*)comAlloc(_fbddHeap, sizeof(char)*(strlen(name)+1));
  strcpy(bddi->id, name);
  
  // make entry in _htSignals
  signal = sigCreate(_arena, bddi->id);
  signal->bddiProducer = bddi; 
  hashPut(_htSignals, bddi->id, signal);
}

void
blifSetModel(Model *model)
{
  _model = model;
}

void
blifBeginModelInit(char *modelName)
{
  _arena = _fbddHeap;

  // _htSignals provides fast lookup from signal name to signal producer 
  // and consumers.  This information is used in the end to properly 
  // connect the bddi's.
  hashDefaultConstruct(_htSignals, _arena);
  _htSignals->scheme = hashStringScheme;

  strcpy(_model->name, modelName);
  _manager = _model->ddManager;

  listConstruct0(&(_model->lOutputs), _arena);
  listConstruct0(&(_model->lInputs), _arena);
}

// Called immediately after the ".model" command is scanned.
// Performs Initialization and creates bddi and _htSignal entries for 
// each variable in lInputs and lOutputs.
// modelName: the model's name.
// lInputs, lOutputs: a linked list of strings.  These strings 
// will be destroyed so a copy of the string must be made if the 
// strings are needed later.
void
blifModelBegin(char* modelName, List lInputs, List lOutputs, List lClocks)
{
  ListCell cell;
  
  blifBeginModelInit(modelName);

  listForeach(lOutputs, cell){
    char* name = listData(char*, cell);
    blifAddPrimaryOutput(name);
  }

  // Create Instances and _htSignals entries for primary inputs
  //  printf("Constructing Primary IN Bddis\n");
  listForeach(lInputs, cell){
    char* name = listData(char*, cell);
    blifAddPrimaryInput(name);
  }
  //  printf("Done: Constructing Primary IN Bddis\n");

  if (lClocks != NULL){
    listForeach(lClocks, cell){
      char* name = listData(char*, cell);
      blifAddClock(name);
    }
  }

  _nodeMUX = buildMUX(_manager);
}

// called at the end of the 'model' rule
void
blifModelEnd()
{
  int i, j;
  HashCell hcell;
  ListCell lcell;
  Consumer *consumer;

  // Connect the instances
  hashForeach(_htSignals, i, j, hcell){
    // connect consumers to producers
    Signal *signal = hashData(Signal*, hcell);

    if (signal->bddiProducer == NULL){
      printf("Error: producer not assiged for signal %s!\n", signal->name);
      exit(1);
    }
    assert(listGetSize(&signal->lConsumers)>0);
    listForeach(&(signal->lConsumers), lcell){
      consumer = listData(Consumer*, lcell);
      consumer->bddi->inputVars[consumer->pos] = signal->bddiProducer;
      comFree(_arena, consumer);
    }
    sigDestruct(_arena, signal);
  }

  hashDestruct(_htSignals);

  // check that all bddi's have their inputVars assigned
#ifdef DEBUG_COMPILE
  listForeach(&(_model->lOutputs), lcell){
    BddInstance *bddi = listData(BddInstance*, lcell);
    int i;

    for (i=0; i<bddi->bddClass->numVars; i++)
      assert(bddi->inputVars[i] != NULL);
  }
#endif
}

// Called immediately after a '.names' command is scanned.
void
blifLogicGateBegin()
{
  listConstruct0(&lCubes, _arena);
  listConstruct0(&lSSCubeSets, _arena);
}

static void
createConsumersForInputVars(BddInstance *bddi, 
			    char *cube, BitSet bs, List lSignals)
{
  int nInputVar=0; // index into inputVars
  int i=0;
  ListCell cId;
  char *id;

  listForeach(lSignals, cId){
    char isMember;
    id = listData(char*, cId);

    if (i == listGetSize(lSignals)-1)
      // last signal is the output signal and is not processed here
      break;

    if (cube != NULL){
      if (cube[i] == '1' || cube[i] == '0')
	isMember = TRUE;
      else
	isMember = FALSE;
    }
    else {
      assert (bs != NULL);

      if (bsetIsMember(bs, i))
	isMember = TRUE;
      else
	isMember = FALSE;
    }

    if (isMember){
      Signal *signal;
      ListCell newCell;
      Consumer *consumer = comAlloc(_arena, sizeof(Consumer));

      if ((signal=hashGet(_htSignals, id)) == NULL){
	// new internal signal.  (non primary input/output).
	signal = sigCreate(_arena, id);
	hashPut(_htSignals, signal->name, signal);
      }
      
      // 'signal' is consumed by bddi at bddi->numVars[nInputVar].
      consumer->bddi = bddi;
      consumer->pos = nInputVar;
      listNewCell(_arena, newCell, consumer);
      listAppendElement(&signal->lConsumers, newCell);

      nInputVar++;
    }
    i++;
  }
}

DdNode *
bddCreateFromCube(char *cube, BitSet bsUsedInputVars, int *numVars)
{
  DdNode *nAnd, *nVar, *nTmp;
  int i=0;
  char c;

  *numVars = 0;
  // create AND function for each product term
  nAnd = Cudd_ReadOne(_manager);
  Cudd_Ref(nAnd);
  while((c=cube[i]) != ' '
	&& c != '\\'){
    if (c=='1'){
      nVar = Cudd_bddIthVar(_manager,*numVars);
      Cudd_Ref(nVar);
      nTmp = Cudd_bddAnd(_manager, nVar, nAnd);
      Cudd_Ref(nTmp);
      Cudd_RecursiveDeref(_manager,nAnd);
      Cudd_RecursiveDeref(_manager,nVar);
      (*numVars)++;
    }
    else if(c=='0'){
      nVar = Cudd_bddIthVar(_manager,*numVars);
      Cudd_Ref(nVar);
      nTmp = Cudd_bddAnd(_manager,Cudd_Not(nVar),nAnd);
      Cudd_Ref(nTmp);
      Cudd_RecursiveDeref(_manager,nAnd);
      Cudd_RecursiveDeref(_manager,nVar);
      (*numVars)++;
    }
    else if(c=='-'){
      // do nothing, function does not depend on this variable
      nTmp = nAnd;
      if (bsUsedInputVars!=NULL && bsetIsMember(bsUsedInputVars, i))
	(*numVars)++;
      }
    else{
      // error
      perror("Error reading input file!");
      exit(1);
    }
    nAnd = nTmp;
    i++;
  }
  if (c == '\\'){
    i=i+2;
    c=cube[i];
  }
  return nAnd;
}

DdNode *
bddCreateFromCubeSet(List lCubes, BitSet bsUsedInputVars, int *numVars)
{
  ListCell cCube;
  DdNode *nOr;
  DdNode *nAnd, *nTmp;

  *numVars = bsetSize(bsUsedInputVars);
  nOr = Cudd_ReadLogicZero(_manager);
  Cudd_Ref(nOr);
  listForeach(lCubes, cCube){
    char *cube = listData(char*, cCube);
    int dummy;
    nAnd = bddCreateFromCube(cube, bsUsedInputVars, &dummy);
    nTmp = Cudd_bddOr(_manager, nAnd, nOr);
    Cudd_Ref(nTmp);
    Cudd_RecursiveDeref(_manager, nAnd);    
    Cudd_RecursiveDeref(_manager, nOr);
    nOr = nTmp;
  }

  return nOr;
}

void
updateSignalProducer(char *id, BddInstance *bddi)
{
  Signal *signal;

  if ((signal=hashGet(_htSignals, id)) == NULL){
    // new internal signal.  (non primary input/output).
    signal = sigCreate(_arena, id);
    assert(strcmp(id, signal->name) == 0);
    hashPut(_htSignals, signal->name, signal);
  }
  
  // 'signal' is produced by 'bddi'
  if (signal->bddiProducer != NULL){
    printf("Error: more than one producer of signal %s\n", id);
    exit(1);
  }
  
  signal->bddiProducer = bddi;
}

#define LOGIC_ZERO 1
#define LOGIC_ONE 2
char _zeroInputGate;

/*
  Requires the following to be set up:
  lSignals : (single) output signal is last in list.
  lCubes
 */
static void
blifPartitionLogicGateEnd(List lSignals)
{
  int numSignals;
  int i;
  SSCubes *ssc;
  struct _BitSet bsResult;
  ListCell cCube;
  ListCell cSSCubes;
  SSCubes *sscSmall; // all small cube sets are put in here.
  char c;

  numSignals = listGetSize(lSignals)-1; // num input signals

  listForeach(&lCubes, cCube){
    char *cube = listData(char*, cCube);
    
    ssc = comAlloc(_arena, sizeof(SSCubes));
    ssc->genSingleGate = FALSE;
    listConstruct0(&ssc->lCubes, _arena);
    listAppend(&ssc->lCubes, cube);

    // compute ssc->bs
    bsetConstruct(&ssc->bs, _arena, numSignals);
    i=0;
    while((c=cube[i]) != ' '
	  && c != '\\'){
      if (c=='1' || c=='0'){
	bsetInclude(&ssc->bs, i);
      }
      else if(c=='-'){
	bsetExclude(&ssc->bs, i);
      }
      i++;
    }
    assert( i==numSignals );
    
    if (c == '\\'){
      i=i+2;
      c=cube[i];
    }
    i++;
    if ((c=cube[i]) != '1'){
      // cube 
      printf("cube output column not 1!\n");
      exit(1);
    }
    
    // if this ssc->bs cube shares support with other cubesets,
    // then it becomes a part of that cube set.
    bsetConstruct(&bsResult, _arena, numSignals);
    
    // Remove all SSCubes from lSSCubeSets that have support overlapping 

    // with ssc and merge them with ssc.
    cSSCubes = lSSCubeSets.head;
    while(cSSCubes != NULL){
      ListCell cNext = cSSCubes->next;
      SSCubes *sscCurr = listData(SSCubes*, cSSCubes);
      bsetIntersect(&ssc->bs, &sscCurr->bs, &bsResult);
      if (bsetIsEmpty(&bsResult) == FALSE){
	// cubes in ssc overlap with cubes in sscCurr.
	// merge the two.
	bsetUnite(&ssc->bs, &sscCurr->bs, &bsResult);      
	bsetCopy(&bsResult, &ssc->bs);
	
	// move cubes from sscCurr to ssc
	listConcat(&ssc->lCubes, &sscCurr->lCubes);
	
	// remove sscCurr from list
	listRemoveElement(&lSSCubeSets, sscCurr->listCell);

	// free sscCurr
	bsetDestruct(&sscCurr->bs);
	comFree(_arena, sscCurr->listCell);
	comFree(_arena, sscCurr);
      }
      cSSCubes = cNext;
    }
    // add ssc to lSSCubeSets;
    {
      ListCell cNew;
      listNewCell(_arena, cNew, ssc);
      ssc->listCell = cNew;
      listAppendElement(&lSSCubeSets, cNew);
    }
    bsetDestruct(&bsResult);
  }
  listDelAll(&lCubes);

  // find small cube sets in lSSCubeSets and merge them into a single cube set.
  sscSmall = comAlloc(_arena, sizeof(SSCubes));
  bsetConstruct(&sscSmall->bs, _arena, numSignals);
  bsetMakeEmpty(&sscSmall->bs);
  listConstruct0(&sscSmall->lCubes, _arena);
  bsetConstruct(&bsResult, _arena, numSignals);
  cSSCubes = lSSCubeSets.head;
  while(cSSCubes != NULL){
    ListCell cNext = cSSCubes->next;
    SSCubes *sscCurr = listData(SSCubes*, cSSCubes);
    if ( bsetSize(&sscCurr->bs) < IDEAL_MAX_BDD_VARS){
      bsetUnite(&sscSmall->bs, &sscCurr->bs, &bsResult);      
      bsetCopy(&bsResult, &sscSmall->bs);

      // add to sscSmall
      listConcat(&sscSmall->lCubes, &sscCurr->lCubes);
      
      // destroy sscCurr
      listRemoveElement(&lSSCubeSets, sscCurr->listCell);
      bsetDestruct(&sscCurr->bs);
      comFree(_arena, sscCurr->listCell);
      comFree(_arena, sscCurr);
    }
    cSSCubes = cNext;
  }
  bsetDestruct(&bsResult);
  if (listGetSize(&sscSmall->lCubes) != 0){
    ListCell cNew;
    listNewCell(_arena, cNew, sscSmall);
    sscSmall->listCell = cNew;
    sscSmall->genSingleGate = TRUE;
    listAppendElement(&lSSCubeSets, cNew);
  }
  else{
    bsetDestruct(&sscSmall->bs);
    comFree(_arena, sscSmall);
  }

  // form gates for each cube set.
  // OR the cube sets together.
  {
    struct _List lOrInputBddis;
    DdNode *bddOr;
    BddClass *bddc;
    ListCell cell;
    BddInstance *bddi;

    listConstruct0(&lOrInputBddis, _arena);

    listForeach(&lSSCubeSets, cSSCubes){
      SSCubes *ssc = listData(SSCubes*, cSSCubes);

      // if num cubes is less than numvars, then read cubes decomposed.
      if ( ssc->genSingleGate==FALSE 
	  && listGetSize(&ssc->lCubes) < bsetSize(&ssc->bs)
	  && listGetSize(&ssc->lCubes)>1){

	// this is a large gate (because genSingleGate is FALSE)
	// make each cube into a single gate
	ListCell cCube;
	listForeach(&ssc->lCubes, cCube){
	  char *cube = listData(char*, cCube);
	  int numVars;
	  DdNode *bdd;
	  bdd = bddCreateFromCube(cube, NULL, &numVars);
	  bddc = 
	    bddcCreateForBdd(_model, bdd, numVars, 
			     &_model->nextBddcId);
	  bddi = bddcNewInstance(bddc);
	  listAppend(&lOrInputBddis, bddi);
	  createConsumersForInputVars(bddi, cube, NULL, lSignals);
	}
      }
      else{
	// make cubeset into a single gate
	DdNode *bdd;
	int numVars;
	bdd = bddCreateFromCubeSet(&ssc->lCubes, &ssc->bs, &numVars);
	bddc = bddcCreateForBdd(_model, bdd, numVars, 
				&_model->nextBddcId);
	bddi = bddcNewInstance(bddc);
	listAppend(&lOrInputBddis, bddi);
	createConsumersForInputVars(bddi, NULL, &ssc->bs, lSignals);
      }
      // free ssc
      bsetDestruct(&ssc->bs);
      listForeach(&ssc->lCubes, cCube){
	char *cube = listData(char*, cCube);
	comFree(_arena, cube);
      }
      listDelAll(&ssc->lCubes);
      comFree(_arena, ssc);
    }
    listDelAll(&lSSCubeSets);

    if (listGetSize(&lOrInputBddis)>1){
      bddOr = buildOROfSize(_manager, listGetSize(&lOrInputBddis));
      
      // the final bdd is pointed to by bddOr
      bddc = bddcCreateForBdd(_model, bddOr, listGetSize(&lOrInputBddis), 
			       &_model->nextBddcId);
      
      bddi = bddcNewInstance(bddc);
      
      // set inputs of the OR gate
      i=0;
      listForeach(&lOrInputBddis, cell){
	bddi->inputVars[i] = listData(BddInstance*, cell);
	i++;
      }
      assert(i==bddc->numVars);
    }
    else if (listGetSize(&lOrInputBddis)==1){
      bddi = listData(BddInstance*, lOrInputBddis.head);
    }
    else{
      DdNode *bdd;

      assert(listGetSize(&lCubes)==0);

      if (_zeroInputGate == LOGIC_ONE)
	bdd = Cudd_ReadOne(_manager);
      else if (_zeroInputGate == LOGIC_ZERO)
	bdd = Cudd_ReadLogicZero(_manager);

      bddc = bddcCreateForBdd(_model, bdd, 0,
			      &_model->nextBddcId);
      bddi = bddcNewInstance(bddc);
    }
    
    // update producer of signal.
    {
      char *id;
      id = listData(char*,lSignals->tail);
      
      if (listGetSize(&lOrInputBddis)==1)
	updateSignalProducer(id, listData(BddInstance*, lOrInputBddis.head));
      else
	updateSignalProducer(id, bddi);	
    }
    listDelAll(&lOrInputBddis);
  }
}

void
blifLogicGateEnd(List lSignals)
{
  blifPartitionLogicGateEnd(lSignals);
}

static void
blifPartitionAddCube(char *cube)
{
  char *new;
  new = comAlloc(_arena, sizeof(char)*(strlen(cube)+1));
  strcpy(new, cube);
  listAppend(&lCubes, new);
}

void
blifAddCube(char *cube)
{
  if (strcmp(cube,"1")==0)
    blifConstOne();
  else
    blifPartitionAddCube(cube);
}

void
blifConstOne()
{
  _zeroInputGate = LOGIC_ONE;
}

void
blifConstZero()
{
  _zeroInputGate = LOGIC_ZERO;
}

void
blifLatch(char *input, char *output, char *type, char *control, int val)
{
  BddInstance *bddi;
  struct _List lSignals;
  struct _BitSet bs;
  int valMark;
  int i;

  if (type==NULL || type[0]==0 || strcmp(type, "fe")==0)
    bddi = bddcNewInstance(_model->bddcLatchFe);
  else if (strcmp(type, "re")==0)  
    bddi = bddcNewInstance(_model->bddcLatchRe);
  else if (strcmp(type, "ah")==0)  
    bddi = bddcNewInstance(_model->bddcLatchAh);
  else if (strcmp(type, "al")==0)  
    bddi = bddcNewInstance(_model->bddcLatchAl);
  else if (strcmp(type, "as")==0)  
    bddi = bddcNewInstance(_model->bddcLatchAs);
  else{
    printf("Unknown latch type: %s!\n", type);
    exit(1);
  }
  
  valMark = val;

  valMark = valMark << 8;
  bddi->mark = bddi->mark | valMark;

  listConstruct0(&lSignals, _arena);
  listAppend(&lSignals, input);
  if (control != NULL && control[0] != 0){
    listAppend(&lSignals, control);
  }
  else{
    if (_model->bddiGenericClk == NULL){
      // construct generic global clock
      _model->bddiGenericClk = bddcNewInstance(_model->bddcClock);
      _model->bddiGenericClk->id = (char*)comAlloc(_fbddHeap, 
				 sizeof(char)*(strlen("__generic_clock__")+1));
      strcpy(_model->bddiGenericClk->id, "__generic_clock__");
    }
    bddi->inputVars[1]  = _model->bddiGenericClk;
  }
  listAppend(&lSignals, output);

  updateSignalProducer(output, bddi);  

  bsetConstruct(&bs, _arena, listGetSize(&lSignals)-1);
  for(i=0; i<listGetSize(&lSignals)-1; i++)
    bsetInclude(&bs, i);
  createConsumersForInputVars(bddi, NULL, &bs, &lSignals);
  bsetDestruct(&bs);

  listDelAll(&lSignals);
}




