/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
  File: twoLiteralExtract.c
  Desc: Provides functions for performing sharing extraction.
  The functions here will maintain the integrity between bcl, bddcs and bddis.
 */

#include <assert.h>
#include <time.h>

#include "fbddutil.h"
#include "fbddopt.h"
#include "fbdd.h"

#include "se.h"

typedef struct _SEExtractorLocal
{
  int *vars;
  extractType et;
} SEExtractorLocal;

// Two variable extractor list element.
// An element of BddInstance->lSees[].
typedef struct _SEExtractorListElement
{
  SEExtractor *see;
  ListCell cell;
} SEExtractorListElement;

/**************************************
       Function Prototypes
 *************************************/
static void checkSeeles(Model *model);
#define getSeeCost(see,size) (size-(size-1)*listGetSize(&see->lBddis))
#define findSeeList(sem, see, size) (hashGet(&sem->htSEELists[size-2],  \
					  (void*)(2-getSeeCost(see, size))))
#define bittest(i, j) ((i&(1<<j))!=0)
#define bitset(i, pos) (i=(i|(1<<pos)))

/*** Sharing Extraction Remainder ***/
SERemainder*              serCreate(SeManager sem, DdNode *bdd, int numVars, 
				    int *oldIndexToNew, int arraySize);
void                      serDestruct(SeManager sem, SERemainder *remainder);

/*** Sharing Extraction Extractor Local ***/
static SEExtractorListElement* 
                          seeleConstruct(SEExtractor *see, ListCell cell);
static void               seeleDestruct(SEExtractorListElement *seele);
static SEExtractorLocal*  seelConstruct(int var1, int var2, extractType et);
static SEExtractorLocal*  seelConstructOfSize(int *vars, extractType et);
static void               seelDestruct(SEExtractorLocal *seel);
static SEExtractorLocal*  seelComputeForVarPair(DdManager *ddManager, 
						BddClass *bddc, 
						DdNode *n1, DdNode *n2);
static SEExtractorLocal*  seelComputeForVarPairConj(DdManager *ddManager, 
						    BddClass *bddc, 
						    DdNode *n1, DdNode *n2);
static SEExtractorLocal*  seelComputeForVarComb(DdManager *ddManager, 
						 BddClass *bddc, 
						 int *varCombination,
						 char size);
static void               seelAddAllCombsInList(List lSeels, 
						List lExpandedSeels);
static void               seelEnumForRemainder(SeManager sem, 
					       BddInstance *bddiR, 
					       BddInstance *bddiOrig,
					       BddInstance *bddiShare);
static void               seelEnumForBddcOfSizeDisj(SeManager sem,
						    BddClass *bddc,
						    int extSize);
static void               seelEnumForBddcDisj(SeManager sem, BddClass *bddc);
static void               seelEnumForBddcConj(SeManager sem, BddClass *bddc);
static void               seelEnumForBddc(SeManager sem, BddClass *bddc);

/*** Sharing Extraction Extractor ***/
static SEExtractor*  seeCreate(BddInstance *bddi1, BddInstance *bddi2, 
			       extractType et);
static SEExtractor*  seeCreateOfSize(BddInstance **bddis, extractType et);
static SEExtractor*  seeCopyConstruct(SEExtractor *see, int size);
static void          seeDestruct(SEExtractor* see);
static List          seeGetList(SeManager sem, SEExtractor *see, int size);
static void          seeAddBddi(SeManager sem, int size, SEExtractor *see, 
				BddInstance *bddi);
static BddInstance*  seeCreateExtractorBddi(SeManager sem, SEExtractor *see, 
					    int extSize);
static void          seeCreateRemainderBddi(SeManager sem, SEExtractor *see, 
					    int extSize, BddInstance *bddi, 
					    BddInstance *bddiShared);
static void          seeExecuteExtraction(SeManager sem, SEExtractor *see, 
					  int extSize);
static DdNode*       seeComputeRemainderBddDisj(DdManager *ddManager, 
						BddInstance *bddi,
						SEExtractor *see, 
						int extractIndex, int extSize);
static DdNode*       seeComputeRemainderBddConj(DdManager *ddManager, 
						BddInstance *bddi,
						SEExtractor *see, 
						int extractIndex, int extSize);
static DdNode*       seeComputeRemainderBdd(SeManager sem, 
					    BddInstance *bddi,
					    SEExtractor *see, 
					    int extractIndex, 
					    int extSize);

/*** Sharing Extraction ***/
static void    seAddExtractor(SeManager sem, BddInstance *bddi1, 
			      BddInstance *bddi2, 
			      BddInstance *bddi, extractType et);
static void    seAddExtractorOfSize(SeManager sem, int size, 
				    BddInstance **bddis, 
				    BddInstance *bddi, extractType et);
void           seAddBddiExtractors(SeManager sem, BddInstance *bddi);
void           seRemBddiExtractors(SeManager sem, BddInstance *bddi);
static void    seRemBddiExtractorsKeepSees(SeManager sem,
					   BddInstance *bddi, List lDeadSees);
static void    seRemBddiExtractorsUsingBddiShare(SeManager sem,
						 BddInstance *bddi, 
						 List lDeadSees, 
						 BddInstance *bddiShare);

/******** BDD Operations and Conversion **********/
static DdNode*  bddIntToCube(DdManager* ddManager, int cube, 
			     int numVars, int *varMap);
static DdNode*  bddIntToBdd(DdManager* ddManager, unsigned int func, 
			    int numVars);
static DdNode*  getLocalNodeFromBddi(DdManager *ddManager,
				     BddInstance *bddiParent, 
				     BddInstance *bddiChild);

/***** Aligning and Normalizing Extractors for Matching *******/
static void     seeAlign(BddInstance **bddi1, BddInstance **bddi2, 
			 extractType *et);
static void     seeAlignOfSize(int size, BddInstance **bddis, extractType *et);
static void     normalizeFunc(unsigned int *val, int numVars);

/****** Enumerating Variable Combinations *********/
static int      vcombInit(int *comb, int combSize, 
			  int *activeVars, int activeVarsSize);
static int      vcombIncr(int *comb, int combSize,
			  int *activeVars, int activeVarsSize);
static void     vcombComputeActiveVars(BddClass *bddc, int extSize, 
				       int **activeVars, int *numActiveVars);

/*********************
   Global Variables 
 ********************/
static const IHashScheme hashSeeScheme[]; 

static int _numExtractionsOfSize[4] = {0};
static int _numSharingsOfSize[4] = {0};

// counters for number of regular sharing extraction computations.
static int _nfse1=0; // computing seels
static int _nrse1=0;
static int _nfse2=0; // computing remainders
static int _nrse2=0;

// Runtime variables.
static clock_t _tSe=0;
static clock_t _tCofactors=0;
static clock_t _tHeap=0;

/****************************************
      Sharing Extraction Remainder
*****************************************/
SERemainder*
serCreate(SeManager sem, DdNode *bdd, int numVars, 
	  int *oldIndexToNew, int arraySize)
{
  SERemainder* remainder = 
    (SERemainder*)comAlloc(_fbddHeap, sizeof(SERemainder));

  remainder->bdd = bdd;
  Cudd_Ref(bdd);
  remainder->oldIndexToNew = oldIndexToNew;
  remainder->arraySize = arraySize;
  remainder->numVars = numVars;

  return remainder;
}

void
serDestruct(SeManager sem, SERemainder *remainder)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  if (remainder != NULL){
    Cudd_RecursiveDeref(ddManager, remainder->bdd);
    comFree(_fbddHeap,remainder->oldIndexToNew);
    comFree(_fbddHeap, remainder);
  }
}

/***********************************************
   Sharing Extraction Extractor List Element
***********************************************/
static SEExtractorListElement*
seeleConstruct(SEExtractor *see, ListCell cell)
{
  SEExtractorListElement *seele 
    = (SEExtractorListElement*)comAlloc(_fbddHeap, sizeof(SEExtractor));

  seele->see = see;
  seele->cell = cell;

  return seele;
}

static void
seeleDestruct(SEExtractorListElement *seele)
{
  comFree(_fbddHeap, seele);
}

/***************************************
   Sharing Extraction Extractor Local
****************************************/
static SEExtractorLocal *
seelConstruct(int var1, int var2, extractType et)
{
  SEExtractorLocal *seel = 
    (SEExtractorLocal*)comAlloc(_fbddHeap, sizeof(SEExtractorLocal));

  assert(var1 < IDEAL_MAX_BDD_VARS);
  assert(var2 < IDEAL_MAX_BDD_VARS);

  seel->vars = (int*)comAlloc(_fbddHeap, sizeof(int)*2);
  seel->vars[0] = var1;
  seel->vars[1] = var2;
  seel->et = et;

  return seel;
}

static SEExtractorLocal *
seelConstructOfSize(int *vars, extractType et)
{
  SEExtractorLocal *seel =
    (SEExtractorLocal*)comAlloc(_fbddHeap, sizeof(SEExtractorLocal));

  seel->vars = vars;
  seel->et = et;

  return seel;
}

static void
seelDestruct(SEExtractorLocal *seel)
{
  if (seel != NULL){
    comFree(_fbddHeap, seel->vars);
    comFree(_fbddHeap, seel);
  }
}

static SEExtractorLocal *
seelComputeForVarPair(DdManager *ddManager, BddClass *bddc, 
		       DdNode *n1, DdNode *n2)
{
  int k;
  DdNode *bddA, *bddB, *bddC, *bddD;
  DdNode *bddCFA, *bddCFB, *bddCFC, *bddCFD;

  int i = Cudd_NodeReadIndex(n1);
  int j = Cudd_NodeReadIndex(n2);

  DdNode *nI = n1;
  DdNode *nJ = n2;

  SEExtractorLocal *seel = NULL;
  
  bddA = Cudd_bddAnd(ddManager, Cudd_Not(nI), Cudd_Not(nJ));
  Cudd_Ref(bddA);
  bddB = Cudd_bddAnd(ddManager, nI, Cudd_Not(nJ));
  Cudd_Ref(bddB);
  bddC = Cudd_bddAnd(ddManager, Cudd_Not(nI), nJ);
  Cudd_Ref(bddC);
  bddD = Cudd_bddAnd(ddManager, nI, nJ);
  Cudd_Ref(bddD);
  
  bddCFA = Cudd_Cofactor(ddManager, bddc->bdd, bddA);
  Cudd_Ref(bddCFA);
  bddCFB = Cudd_Cofactor(ddManager, bddc->bdd, bddB);
  Cudd_Ref(bddCFB);
  bddCFC = Cudd_Cofactor(ddManager, bddc->bdd, bddC);
  Cudd_Ref(bddCFC);
  bddCFD = Cudd_Cofactor(ddManager, bddc->bdd, bddD);
  Cudd_Ref(bddCFD);
  
  if (!(bddCFA==bddCFB && bddCFB==bddCFC && bddCFC==bddCFD)){

    for (k=0; k<5; k++){
      char prediction=FALSE;
      extractType et;
      
      switch(k){
      case 0:
	et = ET_AND;
	prediction = (bddCFA == bddCFC && bddCFA == bddCFB);
	break;
      case 1:
	et = ET_OR;
	prediction = (bddCFD == bddCFC && bddCFD == bddCFB);
	break;
      case 2:
	et = ET_AND01;
	prediction = (bddCFA == bddCFB && bddCFD == bddCFB);
	break;
      case 3:
	et = ET_AND10;
	prediction = (bddCFA == bddCFC && bddCFD == bddCFC);
	break;
      case 4:
	et = ET_XOR;
	prediction = (bddCFB == bddCFC && bddCFA == bddCFD);
	break;
      }
      
      if (prediction){
	seel = seelConstruct(i, j, et);
      }
    }
  }
  Cudd_RecursiveDeref(ddManager, bddA);
  Cudd_RecursiveDeref(ddManager, bddB);
  Cudd_RecursiveDeref(ddManager, bddC);
  Cudd_RecursiveDeref(ddManager, bddD);
  
  Cudd_RecursiveDeref(ddManager, bddCFA);
  Cudd_RecursiveDeref(ddManager, bddCFB);
  Cudd_RecursiveDeref(ddManager, bddCFC);
  Cudd_RecursiveDeref(ddManager, bddCFD);

  return seel;
}
 
static SEExtractorLocal *
seelComputeForVarPairConj(DdManager *ddManager, BddClass *bddc, 
		       DdNode *n1, DdNode *n2)
{
  DdNode *bddTmp;
  int k;

  DdNode *bddExtractor;
  DdNode *bddR1, *bddR2, *bddR;
  int sizeBefore, sizeAfter;
  int et;

  int i = Cudd_NodeReadIndex(n1);
  int j = Cudd_NodeReadIndex(n2);

  DdNode *nI = n1;
  DdNode *nJ = n2;

  SEExtractorLocal *seel = NULL;
  
  sizeBefore = Cudd_DagSize(bddc->bdd);

  for (k=0; k<5; k++){
    et = k;
    switch(k){
    case ET_OR: //a+b
      bddExtractor = Cudd_bddOr(ddManager, nI, nJ);
      break;
    case ET_AND10: //ab'
      bddExtractor = Cudd_bddAnd(ddManager, nI, Cudd_Not(nJ));
      break;
    case ET_AND01: //a'b
      bddExtractor = Cudd_bddAnd(ddManager, Cudd_Not(nI), nJ);
      break;
    case ET_AND: //ab
      bddExtractor = Cudd_bddAnd(ddManager, nI, nJ);      
      break;
    case ET_XOR:
      bddExtractor = Cudd_bddXor(ddManager, nI, nJ);
      break;
    default:
      assert(FALSE);
    }
    Cudd_Ref(bddExtractor);

    bddR1 = Cudd_bddRestrict(ddManager, bddc->bdd, bddExtractor);
    Cudd_Ref(bddR1);
    bddR2 = Cudd_bddRestrict(ddManager, bddc->bdd, Cudd_Not(bddExtractor));
    Cudd_Ref(bddR2);

    if (bddc->bdd != bddR1 && bddc->bdd != bddR2){
      bddTmp = Cudd_bddIthVar(ddManager, Cudd_NodeReadIndex(bddc->bdd)+1);
      Cudd_Ref(bddTmp);

      bddR = Cudd_bddIte(ddManager, bddTmp, bddR1, bddR2);
      Cudd_Ref(bddR);
      Cudd_RecursiveDeref(ddManager, bddTmp);

      assert(bddR != bddc->bdd);

      sizeAfter = Cudd_DagSize(bddR);
      if (sizeAfter < sizeBefore){
	seel = seelConstruct(i, j, et);
      }
      Cudd_RecursiveDeref(ddManager, bddR);
    }
    Cudd_RecursiveDeref(ddManager, bddR1);
    Cudd_RecursiveDeref(ddManager, bddR2);
    Cudd_RecursiveDeref(ddManager, bddExtractor);
  }
  return seel;
}

/***********************************************
       Sharing Extraction Extractor
***********************************************/
static SEExtractor *
seeCreate(BddInstance *bddi1, BddInstance *bddi2, extractType et)
{
  SEExtractor* result;
  result = (SEExtractor*)comAlloc(_fbddHeap, sizeof(SEExtractor));
  listConstruct0(&result->lBddis, _fbddHeap);

  result->bddis = (BddInstance**)comAlloc(_fbddHeap, sizeof(BddInstance*)*2);
  result->bddis[0] = bddi1;
  result->bddis[1] = bddi2;
  result->et = et;
  result->lc = NULL;

  return result;
}

static SEExtractor *
seeCreateOfSize(BddInstance **bddis, extractType et)
{
  SEExtractor* result;
  result = (SEExtractor*)comAlloc(_fbddHeap, sizeof(SEExtractor));
  listConstruct0(&result->lBddis, _fbddHeap);

  result->bddis = bddis;
  result->et = et;
  result->lc = NULL;

  return result;
}

static SEExtractor *
seeCopyConstruct(SEExtractor *see, int size)
{
  int i;
  SEExtractor *result = 
    (SEExtractor*)comAlloc(_fbddHeap, sizeof(SEExtractor));
  
  result->bddis = (BddInstance**)comAlloc(_fbddHeap, 
					  sizeof(BddInstance*)*size);
  for (i=0; i<size; i++)
    result->bddis[i] = see->bddis[i];
  result->et = see->et;
  result->lc = see->lc;

  listConstruct0(&result->lBddis, _fbddHeap);
  listClone(&see->lBddis, &result->lBddis);

  return result;
}

static void
seeDestruct(SEExtractor* see)
{
  listDelAll(&see->lBddis);
  comFree(_fbddHeap, see->bddis);
  comFree(_fbddHeap, see->lc);
  comFree(_fbddHeap, see);
} 

/*********************************/

static DdNode*
getLocalNodeFromBddi(DdManager *ddManager, 
		     BddInstance *bddiParent, BddInstance *bddiChild)
{
  int pos = bddiGetPosChildInParent(bddiParent, bddiChild);
  return Cudd_bddIthVar(ddManager, pos);
}


// findSeeList, if it does not exist, construct new list and add to heap.
static List 
seeGetList(SeManager sem, SEExtractor *see, int size)
{
  int cost = getSeeCost(see, size);
  int key = 2-cost;
  List l = hashGet(&sem->htSEELists[size-2], (void*)key);
  if (l != NULL)
    return l;
  else{
    HeapCell hc;
    l = (List)comAlloc(_fbddHeap, sizeof(struct _List));
    listConstruct0(l, _fbddHeap);
    /*
    hc = heapNewCell(&sem->heapSEEs[size-2], 0, l);
    hc->key = cost;
    */
    hc = heapNewCell(&sem->heapSEEs[size-2], cost, l);

    heapInsertElement(&sem->heapSEEs[size-2], hc);

    hashPut(&sem->htSEELists[size-2], (void*)key, l);
  }
  return l;
}

void
seCBAssign(ISeCallBack cb,
	   void *objCallee,
	   void (*onNewBddc)( void *obj, BddClass *bddc ),
	   void (*onDelBddc)( void *obj, BddClass *bddc ))
{
  cb->objCallee = objCallee;
  cb->onNewBddc = onNewBddc;
  cb->onDelBddc = onDelBddc;
}

/*
  function: seeAddBddi()
  desc: Add a BddInstance 'bddi' to SEExtractor 'see' of size 'size'.
*/
static void
seeAddBddi(SeManager sem, int size, SEExtractor *see, BddInstance *bddi)
{
  SEExtractorListElement *seele;
  ListCell lc;
  clock_t t=0;

  if (see->lc != NULL){
    List l = findSeeList(sem, see, size);
    listDelElement(l, see->lc);
    see->lc = NULL;
  }

  listNewCell(_fbddHeap, lc, bddi);

  if (bddi->bddClass->numVars == size)
    listAddElement(&see->lBddis, lc);
  else
    listAppendElement(&see->lBddis, lc);

  seele = seeleConstruct(see, lc);
  listAppend(&bddi->lSees[size-2], seele);

  t += util_cpu_time();

  // update heap position
  if (listGetSize(&see->lBddis)>1 ||
      (listGetSize(&see->lBddis)==1 
       && listData(BddInstance*, 
		   listGetHead(&see->lBddis))->bddClass->numVars != size)){
    List l;
    if (see->lc == NULL)
      listNewCell(_fbddHeap, see->lc, see);
    
    l = seeGetList(sem, see, size);
    listAppendElement(l, see->lc);
  }
  else{
    assert(see->lc == NULL);
  }
  
  _tHeap += util_cpu_time()-t;
}

static void
seAddExtractor(SeManager sem, BddInstance *bddi1, BddInstance *bddi2, 
	       BddInstance *bddi, extractType et)
{
  SEExtractor *see;
  SEExtractor seeFind;

  seeFind.bddis = (BddInstance**)comAlloc(_fbddHeap, sizeof(BddInstance*)*2);
  seeFind.bddis[0] = bddi1;
  seeFind.bddis[1] = bddi2;
  seeFind.et = et;

  see = (SEExtractor*)hashGet(&sem->htSEEs[0], &seeFind);
  if (see == NULL){
    see = seeCreate(bddi1, bddi2, et);
    hashPut(&sem->htSEEs[0], see, see);

    see->lc = NULL;
    see->et = et;
  }
  seeAddBddi(sem, 2, see, bddi);
  comFree(_fbddHeap, seeFind.bddis);
}

static void
seAddExtractorOfSize(SeManager sem, int size, BddInstance **bddis, 
		     BddInstance *bddi, extractType et)
{
  SEExtractor *see;
  SEExtractor seeFind;

  seeFind.bddis = bddis;
  seeFind.et = et;

  see = (SEExtractor*)hashGet(&sem->htSEEs[size-2], &seeFind);
  if (see == NULL){
    see = seeCreateOfSize(bddis, et);
    hashPut(&sem->htSEEs[size-2], see, see);

    see->lc = NULL;
    assert(et != 0);
    see->et = et;
  }
  else{
    comFree(_fbddHeap, bddis);
  }
  seeAddBddi(sem, size, see, bddi);
}

static void
seeAlign(BddInstance **bddi1, BddInstance **bddi2, extractType *et)
{
  if (*bddi2 < *bddi1){
    BddInstance *bddiTmp;
    bddiTmp = *bddi1;
    *bddi1 = *bddi2;
    *bddi2 = bddiTmp;

    if (*et == ET_AND01)
      *et = ET_AND10;
    else if (*et == ET_AND10)
      *et = ET_AND01;
  }
}

static int
bddiCmp(const void *ptr1, const void *ptr2)
{
  BddInstance *bddi1 = (BddInstance*)ptr1;
  BddInstance *bddi2 = (BddInstance*)ptr2;
  return (bddi1 - bddi2);
}

static void
seeAlignOfSize(int size, BddInstance **bddis, extractType *et)
{
  BddInstance **oldBddis = 
    (BddInstance**)comAlloc(_fbddHeap, sizeof(BddInstance*)*size);
  int *newIndexToOld = (int*)comAlloc(_fbddHeap, sizeof(int)*size);
  int numCubes = (1 << size);
  int i,j;
  unsigned int newCube, oldCube;
  extractType etNew=0;

  // copy bddis to oldBddis
  for (i=0; i<size; i++)
    oldBddis[i] = bddis[i];

  qsort(bddis, size, sizeof(BddInstance*), bddiCmp);

  // construct newIndexToOld
  for (i=0; i<size; i++){
    for (j=0; j<size; j++){
      if (bddis[i] == oldBddis[j]){
	newIndexToOld[i] = j;
      }
    }
  }

  // construct etNew
  for (newCube=0; newCube<numCubes; newCube++){ // for all cubes
    // find the corresponding oldCube.
    oldCube = 0;
    for (j=0; j<size; j++){
      if ((newCube & (1<<j)) != 0)
	oldCube += (1 << newIndexToOld[j]);
    }

    if (((*et) & (1<<oldCube)) != 0) // old cube has a 1 there
      etNew += (1<<newCube);
  }

  *et = etNew;

  comFree(_fbddHeap, newIndexToOld);
  comFree(_fbddHeap, oldBddis);
}

// set val so that bit at index 0 is 0.
static void
normalizeFunc(unsigned int *val, int numVars)
{
  assert (numVars <= 5);

  if (bittest(*val,0)){
    unsigned int mask;
    // complement the function
    if (numVars == 2)
      mask = 0xf;
    else if (numVars == 3)
      mask = 0xff;
    else if (numVars == 4)
      mask = 0xffff;
    else if (numVars == 5);
      mask = 0xffffffff;

    *val = ((*val)^mask); // flip bits
  }
  assert(bittest(*val,0)==FALSE);
}

static DdNode *
bddIntToCube(DdManager* ddManager, int cube, int numVars, int *varMap)
{
  int k;
  DdNode *bddCube;
  DdNode *bddTmp;

  bddCube = Cudd_ReadOne(ddManager);
  Cudd_Ref(bddCube);
  for (k=0; k<numVars; k++){
    DdNode *bddK;
    if (varMap==NULL)
      bddK = Cudd_bddIthVar(ddManager, k);
    else 
      bddK = Cudd_bddIthVar(ddManager, varMap[k]);
    Cudd_Ref(bddK);
    if (bittest(cube, k) == FALSE)
      bddK = Cudd_Complement(bddK);
    
    bddTmp = Cudd_bddAnd(ddManager, bddCube, bddK);
    Cudd_Ref(bddTmp);
    Cudd_RecursiveDeref(ddManager, bddCube);
    Cudd_RecursiveDeref(ddManager, bddK);
    bddCube = bddTmp;
  }
  return bddCube;
}

static DdNode *
bddIntToBdd(DdManager* ddManager, unsigned int func, int numVars)
{
  int numCubes = (1 << numVars);
  DdNode *bddF;
  int cube;

  bddF = Cudd_ReadLogicZero(ddManager);
  Cudd_Ref(bddF);
  for (cube=0; cube<numCubes; cube++){
    // for all minterms.
    if (bittest(func,cube)){
      DdNode *bddTmp;
      DdNode *bddCube;
      bddCube = bddIntToCube(ddManager, cube, numVars, NULL);
      
      bddTmp = Cudd_bddOr(ddManager, bddF, bddCube);
      Cudd_Ref(bddTmp);
      Cudd_RecursiveDeref(ddManager, bddF);
      Cudd_RecursiveDeref(ddManager, bddCube);
      bddF = bddTmp;
    }
  }
  return bddF;
}


/*************************
    Variable Combination
 *************************/
/* vcombInit() and vcombIncr() are functions for enumerating 
   valid variable combinations.
   comb: integer array initialized to the first valid variable combination.
   activeVars: integer array of valid variables indexes.
   Returns FALSE if there is no valid combination
 */
static int 
vcombInit(int *comb, int combSize, 
	  int *activeVars, int activeVarsSize)
{
  int i;

  if (activeVars == NULL){
    // all variables active
    for (i=0; i<combSize; i++) 
      comb[i]=i;
  }
  else{
    // not all variables are active
    if (activeVarsSize < combSize)
      return FALSE;

    for (i=0; i<combSize; i++)
      comb[i] = activeVars[i];
  }
  return TRUE;
}

// returns FALSE if no more combinations.
// activeVars are the variables that are used in generating combinations
static int 
vcombIncr(int *comb, int combSize,
	  int *activeVars, int activeVarsSize)
{
  int i;
  int j=0;
  int *activeCombination;

  if (activeVars != NULL){
    // compute activeCombination
    activeCombination = (int*)comAlloc(_fbddHeap, sizeof(int)*combSize);
    for (i=0; i<combSize; i++){
      // find position of comb[i] in activeVars
      while(!(activeVars[j] == comb[i]))
	j++;
      
      activeCombination[i] = j;
      j++;
    }	  
  }
  else{
    activeCombination = comb;
    //    activeVarsSize = numVars;
  }

  for (i=combSize-1; i>=0; i--){
    if (activeCombination[i]==activeVarsSize-combSize+i){
      // continue to increment more significant digit.
      if (i==0){
	if (activeVars != NULL)
	  comFree(_fbddHeap, activeCombination);
	return FALSE;
      }
    }
    else{
      activeCombination[i]++;
      if (activeCombination[i]<activeVarsSize)
	break;
    }
  }
    
  for (j=i+1; j<combSize; j++)
    activeCombination[j]=activeCombination[j-1]+1;
    
  if (activeVars != NULL){
    // assign comb
    for (i=0; i<combSize; i++)
      comb[i] = activeVars[activeCombination[i]];
    comFree(_fbddHeap, activeCombination);
  }
  return TRUE;
}

static SEExtractorLocal *
seelComputeForVarComb(DdManager *ddManager, BddClass *bddc, 
		    int *varCombination, char size)
{
  unsigned int numCubes = 1 << size;
  unsigned int func=0;
  int i;
  DdNode *bddR1 = NULL;
  DdNode *bddR2 = NULL;

  assert(size >= 2); 

  if (size == 2){
    return 
      seelComputeForVarPair(ddManager, bddc, 
			     Cudd_bddIthVar(ddManager,varCombination[0]), 
			     Cudd_bddIthVar(ddManager,varCombination[1]));
  }

  for (i=0; i<numCubes; i++){
    DdNode *bddCube;
    DdNode *bddCF;

    // generate bddCube
    bddCube = bddIntToCube(ddManager, i, size, varCombination);

    // compute cofactor
    bddCF = Cudd_Cofactor(ddManager, bddc->bdd, bddCube);
    Cudd_Ref(bddCF);
    Cudd_RecursiveDeref(ddManager, bddCube);

    if (bddR1 == NULL)
      bddR1 = bddCF;
    else if (bddCF != bddR1 && bddR2==NULL)
      bddR2 = bddCF;

    if (bddCF == bddR1)
      bitset(func, i);
    else if (bddCF == bddR2)
      // leave as 0
      ;
    else{
      // not a good extracto
      Cudd_RecursiveDeref(ddManager, bddCF);
      return NULL;
    }
    Cudd_RecursiveDeref(ddManager, bddCF);
  }

  // determine if func is a good extractor of size 'size'.
  assert(func < (1<<numCubes) || numCubes == 32);
  normalizeFunc(&func, size);

  {
    int *vars = (int*)comAlloc(_fbddHeap, sizeof(int)*size);
    for (i=0; i<size; i++)
      vars[i] = varCombination[i];
    return seelConstructOfSize(vars, func);
  }
}

/* 
   function: seelAddAllCombsInList()
   desc: 
   Requires:
   'lSeels' is a list of 
*/
static void
seelAddAllCombsInList(List lSeels, List lExpandedSeels)
{
  ListCell c1, c2;

  while(listGetSize(lSeels)>0){
    SEExtractorLocal *seel1;
    c2 = listGetHead(lSeels)->next;
    c1 = listPop(lSeels);
    
    listAppendElement(lExpandedSeels, c1);
    
    seel1 = listData(SEExtractorLocal*, c1);
    while(c2 != NULL){
      extractType et;
      SEExtractorLocal *seel2 = listData(SEExtractorLocal*, c2);

      // determine how to form a new seel with var1 of seel1 and 
      // var 2 of seel2.
      // var2 of seel1 and var1 of seel2 must be of the same polarity.
      assert(seel1->vars[1] == seel2->vars[0]);
      if (seel1->et==ET_AND && seel2->et==ET_AND)
	et = ET_AND;
      else if(seel1->et==ET_OR && seel2->et==ET_OR)
	et = ET_OR;
      else if(seel1->et==ET_XOR && seel2->et==ET_XOR)
	et = ET_XOR;
      else if(seel1->et==ET_AND && seel2->et==ET_AND10)
	et = ET_AND10;
      else if(seel1->et==ET_AND01 && seel2->et==ET_AND)
	et = ET_AND01;
      else if(seel1->et==ET_AND01 && seel2->et==ET_AND10)
	et = ET_OR;
      else if(seel1->et==ET_AND10 && seel2->et==ET_AND01)
	et = ET_AND;
      else if(seel1->et==ET_OR && seel2->et==ET_AND01)
	et = ET_AND01;	
      else if(seel1->et==ET_AND10 && seel2->et==ET_OR)
	et = ET_AND10;
      else{
	assert(FALSE);
      }
      
      seel1 = seelConstruct(seel1->vars[0], seel2->vars[1], et);
      listAppend(lExpandedSeels, seel1);

      c2 = c2->next;
    }
  }
}

/*
  function: vCombComputeActiveVars()
  desc: Identifies variables that are active for sharing extraction.
  A variable is active if not used by any extractor of size smaller than 
  extSize.
  'numActiveVars' returns the number of active variables found.
  'activeVars' is an array of size 'numActiveVars', where the values 
  correspond to the indexes of active variables.
  Assumes extractors of size s, 0<s<extSize have already been computed.
*/
void
vcombComputeActiveVars(BddClass *bddc, int extSize, 
		  int **activeVars, int *numActiveVars)
{
  int i,j;
  char *excludeVars;
  int numActiveVars_ = bddc->numVars;
  int *activeVars_;
  int size;

  excludeVars = (char*)comAlloc(_fbddHeap, sizeof(char)*bddc->numVars);

  for (i=0; i<bddc->numVars; i++)
    excludeVars[i] = FALSE;

  for (size=2; size<extSize; size++){
    ListCell cSeels;
    listForeach(bddc->lSeels[size-2], cSeels){
      int j;
      SEExtractorLocal *seel = listData(SEExtractorLocal*, cSeels);
      for (j=0; j<size; j++){
	if (excludeVars[seel->vars[j]]==FALSE){
	  numActiveVars_--;
	  assert(seel->vars[j]<bddc->numVars);
	  excludeVars[seel->vars[j]] = TRUE;
	}
      }
    }
  }

  activeVars_ = (int*)comAlloc(_fbddHeap, sizeof(int)*numActiveVars_);
  j=0;
  for (i=0; i<numActiveVars_; i++){
    assert(j < bddc->numVars);
    while(j<bddc->numVars && excludeVars[j]==TRUE)
      j++;
    assert(j<bddc->numVars);
    activeVars_[i] = j;
    j++;
  }
  
  *numActiveVars = numActiveVars_;
  *activeVars = activeVars_;

  comFree(_fbddHeap, excludeVars);
}

// Compute disjunctive extractors for BddClass 'bddc'.  Can
// handle extractors of size 2 to 5.
static void 
seelEnumForBddcOfSizeDisj(SeManager sem, BddClass *bddc, int extSize)
{
  int numActiveVars=0;
  int *activeVars;
  DdManager *ddManager = sem->fm->model->ddManager;
  List lSeels;

  int *varCombination;
  
  assert(extSize >= 2);

  if (bddc->lSeels[extSize-2] != NULL)
    return;

  lSeels = (List)comAlloc(_fbddHeap, sizeof(struct _List));
  listConstruct0(lSeels, _fbddHeap);
  assert(bddc->lSeels[extSize-2] == NULL);
  bddc->lSeels[extSize-2] = lSeels;

  if (bddc->numVars < extSize)
    return;

  // compute activeVars
  vcombComputeActiveVars(bddc, extSize, &activeVars, &numActiveVars);

  varCombination = (int*)comAlloc(_fbddHeap, sizeof(int)*extSize);
  vcombInit(varCombination, extSize, activeVars, numActiveVars);
  if (numActiveVars >= extSize){
    do {
      SEExtractorLocal *seel;
      seel = seelComputeForVarComb(ddManager, bddc, 
				  varCombination, extSize);
      if (seel != NULL){
	listAppend(lSeels, seel);
      }
    }while(vcombIncr(varCombination, extSize, activeVars, numActiveVars));

    if(sem->fm->options[opt_verb]){
      int j;
      ListCell cSeel;
      listForeach(lSeels, cSeel){
	SEExtractorLocal *seel = listData(SEExtractorLocal*, cSeel);
	printf("seel:");
	for (j=0; j<extSize; j++)
	  printf("%d ", seel->vars[j]);
	printf("%x\n", seel->et);
      }
    }
  }
  comFree(_fbddHeap, varCombination);
  comFree(_fbddHeap, activeVars);
}

// Compute disjunctive extractors for BddClass 'bddc'.
// Handles extractors of size 2.  Extractors of size greater than 2 are
// handled by calling seelEnumForBddcOfSizeDisj().
static void 
seelEnumForBddcDisj(SeManager sem, BddClass *bddc)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  int i,j;
  int extSize;
  DdNode *nI, *nJ;
  int numVars = bddc->numVars;
  int topLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddc->bdd));

  char *isGoodVar;

  List lSeels;
  ListCell cSeel;
  SEExtractorLocal *seel;

  clock_t t;

  List lSeels3;

  int lb;

  // Check if extractors have already been computed.
  if (bddc->lSeels[0] == NULL){

    _nfse1++;
    
    /*********** Find 2 Variable Extractors ***********/
    if (sem->seMode == FBDD_SE_MODE_EXACT){
      isGoodVar= (char*)comAlloc(_fbddHeap, sizeof(char)*numVars);
      for (i=0; i<numVars; i++)
	isGoodVar[i]=FALSE;
    }
    
    lSeels = (List)comAlloc(_fbddHeap, sizeof(struct _List));
    listConstruct0(lSeels, _fbddHeap);
    assert(bddc->lSeels[0] == NULL);
    bddc->lSeels[0] = lSeels;
    
    t = util_cpu_time();
    
    if (sem->seMode == FBDD_SE_MODE_EXACT){
      for (i=0; i<numVars; i++){
	nI = Cudd_ReadVars(ddManager, 
			   Cudd_ReadInvPerm(ddManager, topLevel+numVars-i-1));
	assert(nI != NULL);
	for (j=i+1; j<numVars; j++){
	  nJ = Cudd_ReadVars(ddManager, 
			     Cudd_ReadInvPerm(ddManager, 
					      topLevel+numVars-j-1));
	  assert(nJ != NULL);
	  seel = seelComputeForVarPair(ddManager, bddc, nI, nJ);
	  if (seel != NULL){
	    listAppend(lSeels, seel);
	    isGoodVar[i]=TRUE;
	    isGoodVar[j]=TRUE;
	  }
	}
      }
      
      if(sem->fm->options[opt_verb]){
	printf("lSeels\n");
	listForeach(lSeels, cSeel){
	  SEExtractorLocal *seel = listData(SEExtractorLocal*, cSeel);
	  printf("seel:%d %d %x\n", seel->vars[0], 
		 seel->vars[1], seel->et);
	}
      }
      
      comFree(_fbddHeap, isGoodVar);
    }
    else if (sem->seMode == FBDD_SE_MODE_FAST){
      lSeels3 = (List)comAlloc(_fbddHeap, sizeof(struct _List));
      listConstruct0(lSeels3, _fbddHeap);
      
      lb = 0;
      for (i=0; i<numVars-1; i++){
	nI = Cudd_ReadVars(ddManager, 
			   Cudd_ReadInvPerm(ddManager, topLevel+numVars-i-1));
	nJ = Cudd_ReadVars(ddManager, 
			   Cudd_ReadInvPerm(ddManager, 
					    topLevel+numVars-(i+1)-1));
	seel = seelComputeForVarPair(ddManager, bddc, nI, nJ);
	
	if (seel == NULL){
	  if (i>lb)
	    seelAddAllCombsInList(lSeels3, lSeels);
	  lb = i+1;
	  listConstruct0(lSeels3, _fbddHeap);
	}
	else{
	  listAppend(lSeels3, seel);
	  
	  if (i+1==numVars-1){
	    seelAddAllCombsInList(lSeels3, lSeels);
	    listConstruct0(lSeels3, _fbddHeap);
	  }
	}
      }
      comFree(_fbddHeap, lSeels3);
    }
  }
  _tCofactors += util_cpu_time() - t;
  
  /****** Find extractors of size 3 to MAX_EXTRACTOR_SIZE *******/
  for (extSize=3; extSize<=maxExtSize; extSize++){
    seelEnumForBddcOfSizeDisj(sem, bddc, extSize);
  }
}

// Compute conjunctive extractors for BddClass 'bddc'.
static void
seelEnumForBddcConj(SeManager sem, BddClass *bddc)
{
  int numActiveVars=0;
  int *activeVars;
  DdManager *ddManager = sem->fm->model->ddManager;
  List lSeels;
  int *varCombination;
  
  lSeels = bddc->lSeels[0];

  // compute activeVars
  vcombComputeActiveVars(bddc, maxExtSize, &activeVars, &numActiveVars);

  if (numActiveVars >= 2){
    varCombination = (int*)comAlloc(_fbddHeap, sizeof(int)*2);
    vcombInit(varCombination, 2, activeVars, numActiveVars);
    do {
      SEExtractorLocal *seel;
      seel = 
	seelComputeForVarPairConj(ddManager, bddc, 
				  Cudd_bddIthVar(ddManager, 
						 varCombination[0]), 
				  Cudd_bddIthVar(ddManager, 
						 varCombination[1]));
      
      if (seel != NULL){
	listAppend(lSeels, seel);

	if(sem->fm->options[opt_verb]){
	  printf("seel:%d %d %x conj\n", 
		 seel->vars[0], seel->vars[1], seel->et);
	}
      }
    }while(vcombIncr(varCombination, 2, activeVars, numActiveVars));
    
    comFree(_fbddHeap, varCombination);
  }
  comFree(_fbddHeap, activeVars);
}

/*
  function: seelEnumForBddc()
  desc: Compute extractors for BddClass 'bddc'.
  Extractors are stored in BddClass->lSeels.
*/
static void
seelEnumForBddc(SeManager sem, BddClass *bddc)
{
  seelEnumForBddcDisj(sem, bddc);
  
  if (sem->fm->options[opt_conj_ext])
    seelEnumForBddcConj(sem, bddc);
  
  if(sem->fm->options[opt_verb])
    printf("\n");
}

void
seAddBddiExtractors(SeManager sem, BddInstance *bddi)
{
  clock_t t = util_cpu_time();

  BddClass *bddc = bddi->bddClass;
  DdManager *ddManager = sem->fm->model->ddManager;
  ListCell cSeel;
  int topLevel;
  int extSize;

  if (bddc->type != INTERNAL_VARIABLE)
    return;

  _nrse1++;

  topLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddc->bdd));

  seelEnumForBddc(sem, bddc);

  // 2 variable extractors
  listForeach(bddc->lSeels[0], cSeel){
    SEExtractorLocal *seel = listData(SEExtractorLocal*, cSeel);
    BddInstance *bddi1, *bddi2;
    extractType et = seel->et;
    
    bddi1 = bddi->inputVars[seel->vars[0]];
    bddi2 = bddi->inputVars[seel->vars[1]];
    
    if (sem->bddiShareAddExtractors==NULL ||
	(bddi1==sem->bddiShareAddExtractors 
	 || bddi2==sem->bddiShareAddExtractors)){
      seeAlign(&bddi1, &bddi2, &et);
      seAddExtractor(sem, bddi1, bddi2, bddi, et);
    }
  }
  
  // 3 or more variable extractors
  for (extSize=3; extSize<=comMin(maxExtSize,bddc->numVars); extSize++){
    listForeach(bddc->lSeels[extSize-2], cSeel){
      SEExtractorLocal *seel = listData(SEExtractorLocal*, cSeel);
      int i;
      BddInstance **bddis = 
	(BddInstance**)comAlloc(_fbddHeap, sizeof(BddInstance*)*extSize);
      char dependant = FALSE;

      extractType et = seel->et;
      for (i=0; i<extSize; i++){
	bddis[i] = bddi->inputVars[seel->vars[i]];
	if (bddis[i]==sem->bddiShareAddExtractors)
	  dependant = TRUE;
      }
      if (sem->bddiShareAddExtractors==NULL || dependant==TRUE){
	assert(et != 0);
	seeAlignOfSize(extSize, bddis, &et);
	assert(et != 0);
	seAddExtractorOfSize(sem, extSize, bddis, bddi, et);
      }
    }
  }

  if (!sem->seActive)
    _tSe += util_cpu_time() - t;
}

void
seAddBddExtractorsUsingBddiShare(SeManager sem, BddInstance *bddi, 
				 BddInstance *bddiShare)
{
  sem->bddiShareAddExtractors = bddiShare;
  seAddBddiExtractors(sem, bddi);
  sem->bddiShareAddExtractors = NULL;  
}

/* function: seRemBddiExtractorsKeepSees()
   desc: Remove extractors dependant on bddi.  If lDeadSees is not NULL, then
   extractors that are removed are placed in lDeadSees.
*/
static char _usingSee = FALSE;
void
seRemBddiExtractorsKeepSees(SeManager sem, BddInstance *bddi, List lDeadSees)
{
  clock_t t;
  int extSize;
  struct _List lKeep;

  t=util_cpu_time();

  for (extSize=2; extSize<=maxExtSize; extSize++){
    listConstruct0(&lKeep, _fbddHeap);
    while(listGetSize(&bddi->lSees[extSize-2])>0){
      ListCell cSeele = listPop(&bddi->lSees[extSize-2]);
      SEExtractorListElement *seele = 
	listData(SEExtractorListElement*, cSeele);
      SEExtractor *see = seele->see;
      
      if (_usingSee){
	int i;
	char dependant = FALSE;

	for (i=0; i<extSize; i++)
	  if (bddiGetMark(see->bddis[i], 14)){
	    dependant = TRUE;
	    break;
	  }
	if (dependant)
	  // remove see
	  ;
	else{
	  listAppendElement(&lKeep, cSeele);
	  continue;
	}
      }

      comFree(_fbddHeap, cSeele);
      
      if (see->lc != NULL){
	List l = findSeeList(sem, see, extSize);
	listDelElement(l, see->lc);
	see->lc = NULL;
      }
      
      listDelElement(&seele->see->lBddis, seele->cell);
      
      if (listGetSize(&see->lBddis)>1 ||
	  (listGetSize(&see->lBddis)==1 
	   && listData(BddInstance*, 
		       listGetHead(&see->lBddis))->bddClass->numVars>extSize)
	  ){
	List l;
	if (see->lc == NULL)
	  listNewCell(_fbddHeap, see->lc, see);
	l = seeGetList(sem, see, extSize);
	listAppendElement(l, see->lc);
      }
      else{
	if (see->lc != NULL){
	  comFree(_fbddHeap, see->lc);
	  see->lc = NULL;
	}
      }

      if (listGetSize(&see->lBddis)==0){
	HashCell hashCell;

	// destroy see
	// remove from heap
	hashCell = hashRemove(&sem->htSEEs[extSize-2], see);
	assert(hashCell != NULL);
	comFree(sem->htSEEs[extSize-2].mem, hashCell);
	if (lDeadSees == NULL)
	  seeDestruct(see);
	else
	  listAppend(lDeadSees, see);
      }
      seeleDestruct(seele);
    }

    listConcat(&bddi->lSees[extSize-2], &lKeep);
  }

  _tHeap += util_cpu_time() - t;
  if (!sem->seActive)
    _tSe += util_cpu_time() - t;
}

void
seRemBddiExtractors(SeManager sem, BddInstance *bddi)
{
  seRemBddiExtractorsKeepSees(sem, bddi, NULL);
}

/*
  function: seRemBddiExtractorsUsingBddiShare()
  desc: Remove all extractors in bddi that share inputs with bddiShare.
*/
void
seRemBddiExtractorsUsingBddiShare(SeManager sem, BddInstance *bddi, 
				  List lDeadSees, 
				  BddInstance *bddiShare)
{
  int i;
  int numVars = bddiShare->bddClass->numVars;

  for (i=0; i<numVars; i++)
    bddiSetMark(bddiShare->inputVars[i], 14, 1);

  _usingSee = TRUE;
  seRemBddiExtractorsKeepSees(sem, bddi, lDeadSees);
  _usingSee = FALSE;

  for (i=0; i<numVars; i++)
    bddiSetMark(bddiShare->inputVars[i], 14, 0);
}

static BddInstance *
seeCreateExtractorBddi(SeManager sem, SEExtractor *see, int extSize)
{
  Model *model = sem->fm->model;
  BddInstance *bddiShare; 
  DdNode *n1, *n2;
  DdNode *bddShare;
  int numVarsShare;
  int *oldIndexToNewShare = NULL;
  int arraySizeShare;
  DdManager *ddManager = model->ddManager;

  int i;

  BddClass *bddcShare;

  if (extSize == 2){
    n1 = Cudd_bddIthVar(ddManager, 0);
    n2 = Cudd_bddIthVar(ddManager, 1);
    bddShare = bddComputeOperation(ddManager, see->et, n1, n2);
    Cudd_Ref(bddShare);
  }
  else{
    DdNode *bddTmp;
    bddShare = bddIntToBdd(ddManager, see->et, extSize);
    assert(Cudd_NodeReadIndex(bddShare)==extSize-1);
    // SLOW: sharedBdd not cached (not utilizing regularity).
    bddTmp = foldBddClass(model, bddShare,
			  &numVarsShare, &oldIndexToNewShare, &arraySizeShare,
			  NULL, NULL);
    Cudd_RecursiveDeref(ddManager, bddShare);
    bddShare = bddTmp;
    assert(Cudd_NodeReadIndex(bddShare)==extSize-1);
  }

  bddcShare = 
    bddcCreateForBdd(model, bddShare, extSize, &model->nextBddcId);

  bddiShare = bddcNewInstance(bddcShare);
  for (i=0; i<extSize; i++){
    int j;
    if (oldIndexToNewShare != NULL)
      j = oldIndexToNewShare[i];
    else
      j = i;
    bddiShare->inputVars[j] = see->bddis[i];
    listAppend(&see->bddis[i]->lFanout,
		   bddiShare);
  }
  comFree(_fbddHeap,oldIndexToNewShare);
  return bddiShare;
}
 
/*
  function: seelEnumForRemainder()
  desc: Computes extractors for remainder 'bddiR' using extractors found
  in 'bddiOrig' (the original function before extraction).  If 'bddiOrig' is
  available, this is faster than computing extractors directly using 
  seelEnumForBddc().
  'bddiShare' is the extracted function.
*/
static void
seelEnumForRemainder(SeManager sem,
		     BddInstance *bddiR, 
		     BddInstance *bddiOrig,
		     BddInstance *bddiShare)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  struct _HashTable htBddiToIndex;
  ListCell cSeel;
  int i;
  List lSeels;
  int sharePos=-1;
  DdNode *n1, *n2;
  int numVars;
  int topLevel;
  clock_t t;

  if (bddiR->bddClass->lSeels[0] != NULL)
    return;

  _nfse1++;

  lSeels = (List)comAlloc(_fbddHeap, sizeof(struct _List));
  listConstruct0(lSeels, _fbddHeap);
  assert(bddiR->bddClass->lSeels[0] == NULL);
  bddiR->bddClass->lSeels[0] = lSeels;

  // populate htBddiToIndex
  // Get position of bddiShare in bddiR->inputVars[].
  hashDefaultConstruct(&htBddiToIndex, _fbddHeap);
  for (i=0; i<bddiR->bddClass->numVars; i++){
    hashPut(&htBddiToIndex, bddiR->inputVars[i], (void*)i+1);
    if (bddiR->inputVars[i] == bddiShare)
      sharePos = i;
  }
  assert(sharePos != -1);
  
  listForeach(bddiOrig->bddClass->lSeels[0], cSeel){
    SEExtractorLocal *seel = listData(SEExtractorLocal*, cSeel);

    BddInstance *bddi1, *bddi2;
    int rPos1, rPos2;

    bddi1 = bddiOrig->inputVars[seel->vars[0]];
    bddi2 = bddiOrig->inputVars[seel->vars[1]];
  
    if (bddi1 == bddiShare->inputVars[0] 
	|| bddi1 == bddiShare->inputVars[1]
	|| bddi2 == bddiShare->inputVars[0] 
	|| bddi2 == bddiShare->inputVars[1]){
      continue;
    }
      
    rPos1 = (int)hashGet(&htBddiToIndex, bddi1)-1;
    rPos2 = (int)hashGet(&htBddiToIndex, bddi2)-1;
    assert(rPos1>=0);
    assert(rPos2>=0);

    listAppend(lSeels, seelConstruct(rPos1, rPos2, seel->et));
  }
  hashDestruct(&htBddiToIndex);

  // find seels in bddiR that contained bddiShare as a variable.
  numVars = bddiR->bddClass->numVars;
  topLevel = 
    Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddiR->bddClass->bdd));

  n1 = Cudd_ReadVars(ddManager, 
		     Cudd_ReadInvPerm(ddManager, 
				      topLevel+numVars-sharePos-1));
  
  t = util_cpu_time();

  if (sem->seMode == FBDD_SE_MODE_FAST){
    int i;

    for(i=sharePos+1; i<numVars; i++){
      SEExtractorLocal *seelNew;
      n2 = Cudd_ReadVars(ddManager, 
			 Cudd_ReadInvPerm(ddManager, 
					  topLevel+numVars-i-1));
      seelNew = seelComputeForVarPair(ddManager, bddiR->bddClass, n1, n2);
      if (seelNew != NULL){
	listAppend(lSeels, seelNew);
      }
      else
	break;
    }

    for(i=sharePos-1; i>=0; i--){
      SEExtractorLocal *seelNew;
      n2 = Cudd_ReadVars(ddManager, 
			 Cudd_ReadInvPerm(ddManager, 
					  topLevel+numVars-i-1));
      seelNew = seelComputeForVarPair(ddManager, bddiR->bddClass, n1, n2);
      if (seelNew != NULL){
	listAppend(lSeels, seelNew);
      }
      else
	break;
    }
  }
  else{
    for(i=0; i<numVars; i++){
      SEExtractorLocal *seelNew;
      if (i == sharePos)
	continue;
      n2 = Cudd_ReadVars(ddManager, 
			 Cudd_ReadInvPerm(ddManager, 
					  topLevel+numVars-i-1));
      seelNew = seelComputeForVarPair(ddManager, bddiR->bddClass, n1, n2);
      if (seelNew != NULL)
	listAppend(lSeels, seelNew);
    }
  }
  _tCofactors += util_cpu_time() - t;
}

// Disjunctive Extraction
static DdNode *
seeComputeRemainderBddDisj(DdManager *ddManager, BddInstance *bddi,
	      SEExtractor *see, int extractIndex, int extSize)
{
  DdNode *bddRemainder;
  DdNode *bddD1, *bddD2;
  DdNode *nX;
  DdNode *bddCube1, *bddCube2;

  DdNode *bddOrig = bddi->bddClass->bdd;
  extractType et = see->et;
  DdNode *n1, *n2;
  BddInstance *bddiVar1, *bddiVar2;

  int varMap[16];

  // create new node used to represent bddExtract.
  // Assume bddOrig is in the global manager (the top most node has the 
  // largest index).

  if (extSize == 2){
    bddiVar1 = see->bddis[0];
    bddiVar2 = see->bddis[1];
    
    n1 = getLocalNodeFromBddi(ddManager, bddi, bddiVar1);
    n2 = getLocalNodeFromBddi(ddManager, bddi, bddiVar2);
        
    switch(et){
    case ET_AND:
      bddCube1 = bddComputeOperation(ddManager, ET_AND, n1, n2);
      Cudd_Ref(bddCube1);
      bddCube2 = bddComputeOperation(ddManager, ET_AND01, n1, n2);
      Cudd_Ref(bddCube2);
      break;
    case ET_OR:
      bddCube1 = bddComputeOperation(ddManager, ET_AND, n1, n2);
      Cudd_Ref(bddCube1);
      bddCube2 = bddComputeOperation(ddManager, ET_OR, n1, n2);
      Cudd_Ref(bddCube2);
      bddCube2 = Cudd_Not(bddCube2);
      break;
    case ET_AND01:
      bddCube1 = bddComputeOperation(ddManager, ET_AND01, n1, n2);
      Cudd_Ref(bddCube1);
      bddCube2 = bddComputeOperation(ddManager, ET_AND, n1, n2);
      Cudd_Ref(bddCube2);
      break;
    case ET_AND10:
      bddCube1 = bddComputeOperation(ddManager, ET_AND10, n1, n2);
      Cudd_Ref(bddCube1);
      bddCube2 = bddComputeOperation(ddManager, ET_AND, n1, n2);
      Cudd_Ref(bddCube2);
      break;
    case ET_XOR:
      bddCube1 = bddComputeOperation(ddManager, ET_AND10, n1, n2);
      Cudd_Ref(bddCube1);
      bddCube2 = bddComputeOperation(ddManager, ET_AND, n1, n2);
      Cudd_Ref(bddCube2);
      break;
    default:
      assert(FALSE);
    }
  }
  else{
    int oneCube;
    int zeroCube=0;
    int i;
    int numCubes = (1<<extSize);
    for (i=0; i<numCubes; i++){
      if (bittest(et, i)){
	oneCube = i;
	break;
      }
    }
    for (i=0; i<extSize; i++){
      DdNode *node = getLocalNodeFromBddi(ddManager, bddi, see->bddis[i]);
      varMap[i] = Cudd_NodeReadIndex(node);
    }
    bddCube1 = bddIntToCube(ddManager, oneCube, extSize, varMap);
    bddCube2 = bddIntToCube(ddManager, zeroCube, extSize, varMap);
  }

  nX = Cudd_ReadVars(ddManager, extractIndex);
  assert(nX);
  Cudd_Ref(nX);

  bddD1 = Cudd_Cofactor(ddManager, bddOrig, bddCube1);
  Cudd_Ref(bddD1);
  Cudd_RecursiveDeref(ddManager, bddCube1);
  bddD2 = Cudd_Cofactor(ddManager, bddOrig, bddCube2);
  Cudd_Ref(bddD2);
  Cudd_RecursiveDeref(ddManager, bddCube2);

  bddRemainder = Cudd_bddIte(ddManager, nX, bddD1, bddD2);
  Cudd_Ref(bddRemainder);
  Cudd_RecursiveDeref(ddManager, bddD1);
  Cudd_RecursiveDeref(ddManager, bddD2);
  Cudd_RecursiveDeref(ddManager, nX);

  return bddRemainder;
}

// Conjunctive Extraction
static DdNode *
seeComputeRemainderBddConj(DdManager *ddManager, BddInstance *bddi,
			  SEExtractor *see, 
			  int extractIndex, int extSize)
{
  DdNode *bddRemainder;
  DdNode *bddD1, *bddD2;
  DdNode *nX;
  DdNode *bddCube1, *bddCube2;

  DdNode *bddOrig = bddi->bddClass->bdd;
  extractType et = see->et;
  DdNode *n1, *n2;
  BddInstance *bddiVar1, *bddiVar2;

  DdNode *bddExtractor;

  int varMap[16];

  // create new node used to represent bddExtract.
  // Assume bddOrig is in the global manager (the top most node has the 
  // largest index).

  nX = Cudd_ReadVars(ddManager, extractIndex);
  assert(nX);
  Cudd_Ref(nX);

  if (extSize == 2){

    bddiVar1 = see->bddis[0];
    bddiVar2 = see->bddis[1];
    
    n1 = getLocalNodeFromBddi(ddManager ,bddi, bddiVar1);
    n2 = getLocalNodeFromBddi(ddManager, bddi, bddiVar2);
        
    switch(et){
    case ET_AND:
      bddExtractor = Cudd_bddAnd(ddManager, n1, n2);
      break;
    case ET_OR:
      bddExtractor = Cudd_bddOr(ddManager, n1, n2);
      break;
    case ET_AND01:
      bddExtractor = Cudd_bddAnd(ddManager, Cudd_Not(n1), n2);
      break;
    case ET_AND10:
      bddExtractor = Cudd_bddAnd(ddManager, n1, Cudd_Not(n2));
      break;
    case ET_XOR:
      bddExtractor = Cudd_bddXor(ddManager, n1, n2);
      break;
    default:
      assert(FALSE);
    }
    Cudd_Ref(bddExtractor);

    bddD1 = Cudd_bddRestrict(ddManager, bddOrig, bddExtractor);
    Cudd_Ref(bddD1);
    bddD2 = Cudd_bddRestrict(ddManager, bddOrig, Cudd_Not(bddExtractor));
    Cudd_Ref(bddD2);
  }
  else{
    int oneCube;
    int zeroCube=0;
    int i;
    int numCubes = (1<<extSize);
    for (i=0; i<numCubes; i++){
      if (bittest(et, i)){
	oneCube = i;
	break;
      }
    }
    for (i=0; i<extSize; i++){
      DdNode *node = getLocalNodeFromBddi(ddManager, bddi, see->bddis[i]);
      varMap[i] = Cudd_NodeReadIndex(node);
    }
    bddCube1 = bddIntToCube(ddManager, oneCube, extSize, varMap);
    bddCube2 = bddIntToCube(ddManager, zeroCube, extSize, varMap);

    bddD1 = Cudd_Cofactor(ddManager, bddOrig, bddCube1);
    Cudd_Ref(bddD1);
    Cudd_RecursiveDeref(ddManager, bddCube1);
    bddD2 = Cudd_Cofactor(ddManager, bddOrig, bddCube2);
    Cudd_Ref(bddD2);
    Cudd_RecursiveDeref(ddManager, bddCube2);
  }

  bddRemainder = Cudd_bddIte(ddManager, nX, bddD1, bddD2);
  Cudd_Ref(bddRemainder);
  Cudd_RecursiveDeref(ddManager, bddD1);
  Cudd_RecursiveDeref(ddManager, bddD2);
  Cudd_RecursiveDeref(ddManager, nX);

  return bddRemainder;
}

// Compute BDD of bddi extracted by see.
static DdNode *
seeComputeRemainderBdd(SeManager sem, BddInstance *bddi,
		      SEExtractor *see, 
		      int extractIndex, int extSize)
{
  DdManager *ddManager = sem->fm->model->ddManager;

  if (sem->fm->options[opt_conj_ext])
    return seeComputeRemainderBddConj(ddManager, bddi,
				     see, extractIndex, extSize);
  else
    return seeComputeRemainderBddDisj(ddManager, bddi,
				     see, extractIndex, extSize);
}

static void
seeCreateRemainderBddi(SeManager sem,
		  SEExtractor *see, 
		  int extSize,
		  BddInstance *bddi, 
		  BddInstance *bddiShared)
     
{
  Model *model = sem->fm->model;
  DdManager *ddManager = model->ddManager;
  int extractNodeIndex;

  DdNode *bddR, *bddTmp;
  int numVars, *oldIndexToNew, arraySize;

  int topLevel;
  BddInstance **oldInputVars;
  BddClass *bddc;

  SERemainder *remainder;

  // Disable Remainder Cache.  It's pulling incorrect data.
  /*
  bddSupport = sercGetSupport(sem, see, bddi, extSize);
  remainder = sercFind(sem, bddSupport, bddi->bddClass);
  */
  remainder = NULL;
  
  topLevel =
    Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddi->bddClass->bdd));
  extractNodeIndex = Cudd_ReadInvPerm(ddManager, topLevel-1);
  
  if (remainder == NULL){
    bddR = seeComputeRemainderBdd(sem, bddi, see, 
		       extractNodeIndex, extSize);
    bddTmp = foldBddClass(model, bddR,
			  &numVars, &oldIndexToNew, &arraySize,
			  NULL,NULL);
    Cudd_RecursiveDeref(ddManager, bddR);
    bddR = bddTmp;
    remainder = serCreate(sem, bddR, numVars, oldIndexToNew, arraySize);
    //    sercPut(sem, bddSupport, bddi->bddClass, remainder);

    _nfse2++;
  }
  else{
    // use cached remainder.
    bddR = remainder->bdd;
    Cudd_Ref(bddR);
    oldIndexToNew = remainder->oldIndexToNew;
    arraySize = remainder->arraySize;
    numVars = remainder->numVars;
  }

  //  sercReleaseSupport(sem, bddSupport);
  _nrse2++;

  bddc = bddcCreateForBdd(model, bddR, numVars, &model->nextBddcId);

  assert(bddc->numVars < bddi->bddClass->numVars);
  
  oldInputVars = bddi->inputVars;
  bddi->inputVars = NULL;
  updateBddisForRoot(model, bddi, bddc,
		     model->oldIndexToNewSF, bddi->bddClass->numVars,
		     oldInputVars,
		     NULL, 0,
		     0,
		     oldIndexToNew, arraySize,
		     extractNodeIndex, bddiShared);
  comFree(_fbddHeap, oldInputVars);

  listAppend(&bddiShared->lFanout, bddi);
}

/*
  function: seeExecuteExtraction()
  desc: Commit extractor 'see'.  Computes shared extractor.  Removes 
  dependant extractors and adds new extractors form in the process.
  'extSize' is the size of extractor see.
*/
void
seeExecuteExtraction(SeManager sem, SEExtractor *see, int extSize)
{
  Model *model = sem->fm->model;
  BddInstance *bddiShare;
  ListCell cBddi;
  int numRemainders=0;
  BddInstance *bddiHead;
  struct _List lSweepBddis;
  int i,j;
  struct _List lDeadSees;
  ListCell cSee;
  char changed;
  SEExtractor *seeWork;
  HashCell hc;
  List lBddis = &see->lBddis;

  if (!(listGetSize(&see->lBddis)>0))
    return;

  listConstruct0(&lDeadSees, _fbddHeap);

  _numSharingsOfSize[extSize-2] += listGetSize(&see->lBddis)-1;

  listConstruct0(&lSweepBddis, _fbddHeap);

  bddiHead = listData(BddInstance*,listGetHead(&see->lBddis));
  assert(!(bddiHead->bddClass->numVars==2 
	   && listGetSize(&see->lBddis)<extSize));

  // unalign see->bddis, to maintain regularity
  // order see->bddis in the order that they are found in the first
  // element of see->lBddis.
  /*
  seeWork = see;
  lBddis = &see->lBddis;
  */

  if (extSize == 2){
    hc = hashRemove(&sem->htSEEs[extSize-2], see);
    comFree(sem->htSEEs[extSize-2].mem, hc);
    
    for (i=0; i<extSize; i++){
      bddiSetMark(see->bddis[i],13,1);
    }

    j=0;
    changed = FALSE;
    for (i=0; i<bddiHead->bddClass->numVars; i++){
      if (bddiGetMark(bddiHead->inputVars[i],13)){
	assert(j<extSize);
	if (see->bddis[j] != bddiHead->inputVars[i])
	  changed = TRUE;
	see->bddis[j] = bddiHead->inputVars[i];
	j++; 
      }
    }
    for (i=0; i<extSize; i++){
      bddiSetMark(see->bddis[i],13,0);
    }

    if (extSize == 2){
      if (changed){
	if (see->et == ET_AND10)
	  see->et = ET_AND01;
	else if (see->et == ET_AND01)
	  see->et = ET_AND10;
      }
    }
    else{
      assert(FALSE);
    }
    hashPut(&sem->htSEEs[extSize-2], see, see);
  }
  
  // Create shared BddInstance
  bddiShare = seeCreateExtractorBddi(sem, see, extSize);
  
  assert(listGetSize(lBddis)>0);
  while(listGetSize(lBddis)>0){
    BddInstance *bddi;
    BddInstance *bddiOld;
    int last = (listGetSize(lBddis)==1);

    cBddi = listGetHead(&see->lBddis);
    bddi = listData(BddInstance*, cBddi);

    if (bddi->bddClass->numVars != extSize){
      _numExtractionsOfSize[extSize-2]++;
    }

    bddiOld = bddiCopyConstruct(bddi);

    if (extSize==2){
      seeWork = seeCopyConstruct(see, extSize);
      seRemBddiExtractorsUsingBddiShare(sem, bddi, &lDeadSees, bddiShare);
      seeCreateRemainderBddi(sem, seeWork, extSize, bddi, bddiShare);
      seelEnumForRemainder(sem, bddi, bddiOld, bddiShare);
      seAddBddExtractorsUsingBddiShare(sem, bddi, bddiShare);

      listDelAll(&seeWork->lBddis);
      comFree(_fbddHeap, seeWork->bddis);
      comFree(_fbddHeap, seeWork);
    }
    else{
      seeWork = seeCopyConstruct(see, extSize);
      seRemBddiExtractors(sem, bddi);
      seeCreateRemainderBddi(sem, seeWork, extSize, bddi, bddiShare);
      seAddBddiExtractors(sem, bddi);

      listDelAll(&seeWork->lBddis);
      comFree(_fbddHeap, seeWork->bddis);
      comFree(_fbddHeap, seeWork);
    }
    if (bddi->bddClass->numVars == 1)
      listAppend(&lSweepBddis, bddi);
    numRemainders++;

    // currently incremental solution not available for 
    // extractors larger than size 2.

    bclRemoveBddi(model->bclm, model, bddiOld);
    if (bddcNumInsts(bddiOld->bddClass)==0 
	&& isBasicGate(model, bddiOld->bddClass)==FALSE){
      cbRun(ISeCallBack, sem, onDelBddc, bddiOld->bddClass);
      bddcDestruct(bddiOld->bddClass);
    }
    bddiDestruct(bddiOld);

    if (last)
      break;
  }
  seAddBddiExtractors(sem, bddiShare);

  listForeach(&lDeadSees, cSee){
    SEExtractor *see = listData(SEExtractor*, cSee);
    if (listGetSize(&see->lBddis)==0){
      seeDestruct(see);
    }
  }
  listDelAll(&lDeadSees);

  sweepIncr(sem->fm->sm, _fbddHeap, &lSweepBddis);

}

void 
bddcFreeSeelList(SeManager sem, BddClass *bddc)
{
  int i;
  if (bddc->lSeels[0] == NULL)
    return;

  assert(_fbddHeap != NULL);

  for (i=0; i<=maxExtSize-2; i++){
    while(listGetSize(bddc->lSeels[i])>0){
      ListCell cSeel = listPop(bddc->lSeels[i]);
      DdNode *bddSupport;

      SEExtractorLocal *seel = listData(SEExtractorLocal*, cSeel);
      bddSupport = sercGetSupportFromVars(sem, seel->vars, i+2);
      sercRemove(sem, bddSupport, bddc);
      assert(sercFind(sem, bddSupport, bddc)==NULL);
      sercReleaseSupport(sem, bddSupport);

      seelDestruct(seel);
      comFree(_fbddHeap, cSeel);
    }
    comFree(_fbddHeap, bddc->lSeels[i]);
    bddc->lSeels[i] = NULL;
  }
}

void
bddiFreeSeeList(BddInstance *bddi)
{
  int i;
  for (i=0; i<=maxExtSize-2; i++){
    while(listGetSize(&bddi->lSees[i])>0){
      ListCell cSeele = listPop(&bddi->lSees[i]);
      seeleDestruct(listData(SEExtractorListElement*, cSeele));
      comFree(_fbddHeap, cSeele);
    }
    listDelAll(&bddi->lSees[i]);
    listConstruct0(&bddi->lSees[i], _fbddHeap);
  }
}

clock_t
seGetTime()
{
  return _tSe;
}

SeManager
seNewManager(FbddManager fm, SeMode seMode)
{
  SeManager sem = 
    (SeManager)comAlloc(_fbddHeap, sizeof(struct _SeManager));
  Model *model = fm->model;
  clock_t t = util_cpu_time();
  int i;

  sem->fm = fm;
  fm->sem = sem;
  sem->seMode = seMode;
  sem->seActive = FALSE;
  sem->bddiShareAddExtractors = NULL;
  listConstruct0(&sem->lCBs, _fbddHeap);

  sercConstruct(sem);

  // Add callbacks to sweep and decompose
  if (sem->fm->sm != NULL){
    sweepCBAssign(&sem->cbSweep, sem, 
	  (void (*)( void *obj, BddInstance *bddi ))seRemBddiExtractors, 
	  (void (*)( void *obj, BddInstance *bddi ))seAddBddiExtractors);
    cbAdd(sem->fm->sm, sem->cSweepCallBack, &sem->cbSweep);
  }
  fbddCBAssign(&sem->cbFbdd, sem, 
	       NULL,
	       (void (*)( void *obj, BddClass *bddc ))bddcFreeSeelList);
  cbAdd(sem->fm, sem->cFbddCallBack, &sem->cbFbdd);
  //  bddcFreeSeelList(_fm->sem, bddc);

  for (i=0; i<=maxExtSize-2; i++){
    hashDefaultConstruct(&sem->htSEEs[i], _fbddHeap);
    sem->htSEEs[i].scheme = hashSeeScheme[i];
  }

  // ERROR
  for (i=0; i<=maxExtSize-2; i++){
    heapConstruct(&sem->heapSEEs[i], _fbddHeap, 1000);
    hashDefaultConstruct(&sem->htSEELists[i], _fbddHeap);
  }

  {
    struct _List lBddis;
    ListCell cBddi;
    
    // find good extractors for all bddis
    listConstruct0(&lBddis, _fbddHeap);
    modelBddiList(_fbddHeap, model, &lBddis);
    listForeach(&lBddis, cBddi){
      BddInstance *bddi = listData(BddInstance*, cBddi);

      if (bddi->bddClass->type == INTERNAL_VARIABLE)
	seAddBddiExtractors(sem, bddi);
    }
    listDelAll(&lBddis);
  }
  _tSe += util_cpu_time() - t;
  return sem;
}

void
seDelManager(SeManager sem)
{
  Model *model = sem->fm->model;
  int i;
  clock_t t = util_cpu_time();
  HashCell cell;

  for (i=0; i<=maxExtSize-2; i++){
    int j,k;
    hashForeach( &sem->htSEEs[i], k, j, cell ){
      SEExtractor *see = hashKey(SEExtractor*, cell);
      assert(hashKey(SEExtractor*, cell) == hashData(SEExtractor*, cell));
      seeDestruct(see);
    }
    hashDestruct(&sem->htSEEs[i]);
    
    hashForeach( &sem->htSEELists[i], k, j, cell ){
      List l = hashData(List, cell);
      comFree(_fbddHeap, l);
    }
    hashDestruct(&sem->htSEELists[i]);
  }

  for (i=0; i<=maxExtSize-2; i++){
    heapDelAll(&sem->heapSEEs[i]);
  }

  {
    struct _List lBddcs;
    ListCell cBddc;
    
    listConstruct0(&lBddcs, _fbddHeap);
    modelBddcList(_fbddHeap, model, &lBddcs);
    listForeach(&lBddcs, cBddc){
      BddClass *bddc = listData(BddClass*, cBddc);
      ListCell cBddi;
      
      listForeach(&bddc->lInsts, cBddi){
	BddInstance *bddi = listData(BddInstance*, cBddi);
	bddiFreeSeeList(bddi);
      }

      bddcFreeSeelList(sem, bddc);
    }
    listDelAll(&lBddcs);

    if (modelHasBackup(model)){
      modelSwapBackup(model);

     listConstruct0(&lBddcs, _fbddHeap);
     modelBddcList(_fbddHeap, model, &lBddcs);
     listForeach(&lBddcs, cBddc){
       BddClass *bddc = listData(BddClass*, cBddc);
       ListCell cBddi;
       
       listForeach(&bddc->lInsts, cBddi){
	 BddInstance *bddi = listData(BddInstance*, cBddi);
	 bddiFreeSeeList(bddi);
       }
       
       bddcFreeSeelList(sem, bddc);
     }
     listDelAll(&lBddcs);
     modelSwapBackup(model);
    }
  }

  bddcFreeSeelList(sem, model->bddcSF);
  bddcFreeSeelList(sem, model->bddcMux);
  bddcFreeSeelList(sem, model->bddcMux2);
  bddcFreeSeelList(sem, model->bddcAnd);
  bddcFreeSeelList(sem, model->bddcOr);
  bddcFreeSeelList(sem, model->bddcXnor);
  bddcFreeSeelList(sem, model->bddcNot);
  bddcFreeSeelList(sem, model->bddcNand);
  bddcFreeSeelList(sem, model->bddcNor);
  bddcFreeSeelList(sem, model->bddcXor);
  bddcFreeSeelList(sem, model->bddcLogicZero);
  bddcFreeSeelList(sem, model->bddcLogicOne);

  sercDestruct(sem);

  // Remove Call backs to sweep and decompose
  if (sem->fm->sm != NULL)
    cbRemove(sem->fm->sm, sem->cSweepCallBack);
  cbRemove(sem->fm, sem->cFbddCallBack);  

  sem->fm->sem = NULL;
  comFree(_fbddHeap, sem);

  _tSe += util_cpu_time() - t;
}

/*
  function: seRun()
  desc: Performs sharing extraction on functions in 'model'.  
  Sharing extraction will only be performed on extractors that have 'lbMatch'
  or more matches.  An optional function parameter seOnNewBddcEvent() is
  called when a new BddClass is created during sharing extraction.
*/
void 
seRun(SeManager sem, int lbMatch)
{
  clock_t t = util_cpu_time();

  sem->seActive = TRUE;
  sem->lbMatch = lbMatch;
  


  while(TRUE){
    HeapCell heapCell;
    SEExtractor *see;
    List lSees;
    ListCell lc;
    int i;
    int bestHeap = -1;
    int bestHeapValue;
    
    // find heap with best extraction.
    for (i=0; i<=maxExtSize-2; i++){
      // remove all empty lists at head of heap.
      while(!heapIsEmpty(&sem->heapSEEs[i])){
	HeapCell hc = heapFindMin( &sem->heapSEEs[i] );
	lSees = heapData(List, hc);
	if (listGetSize(lSees)>0)
	  break;
	else{
	  HeapCell hc;
	  HashCell hashCell;

	  // remove from heap
	  hc = heapDelMin( &sem->heapSEEs[i]);

	  // remove from hash table
	  hashCell = 
	    hashRemove(&sem->htSEELists[i], (void*)(2-hashKey(int, hc)));
	  assert(hashCell != NULL);
	  comFree(sem->htSEELists[i].mem, hashCell);

	  comFree(_fbddHeap, lSees);
	  comFree(heapMem(&sem->heapSEEs[i]), hc);
	}
      }
    
      if (!heapIsEmpty(&sem->heapSEEs[i])){
	HeapCell hc = heapFindMin( &sem->heapSEEs[i] );
	List lSees = heapData(List, hc);
	SEExtractor *see = listData(SEExtractor*, listGetHead(lSees));
	int heapValue = getSeeCost(see, i+2);
	
	if (bestHeap == -1 || heapValue<bestHeapValue){
	  bestHeap = i;
	  bestHeapValue = heapValue;
	}
      }
    }
    
    // break if no heap contains a valid extractor.
    if (bestHeap == -1)
      break;
    
    // commit Best
    heapCell = heapFindMin(&sem->heapSEEs[bestHeap]);
    lSees = heapData(List, heapCell);
    see = listData(SEExtractor*, listGetHead(lSees));

    if (listGetSize(&see->lBddis)<sem->lbMatch){
      break;
    }
    lc = see->lc;

    listRemoveElement(lSees, see->lc);
    see->lc = NULL; // prevent heapRemove (already removed).

    if (listGetSize(&see->lBddis)>0){
      // Commit the extraction
      seeExecuteExtraction(sem, see, bestHeap+2);
    }
    
    comFree(_fbddHeap, lc);
  }
  
  _tSe += util_cpu_time() - t;
  sem->seActive = FALSE;
}


/*
  function: seDump()
  desc: Print statistics collected during sharing extraction.
*/
void
seDump()
{
  printf("Sharing Extraction\n");

  /*
  printf("Extractor Size  %5d %5d %5d %5d\n", 2, 3, 4, 5);
  printf("Sharing Found   %5d %5d %5d %5d\n", 
	 _numExtractionsOfSize[0], 
	 _numExtractionsOfSize[1], 
	 _numExtractionsOfSize[2], 
	 _numExtractionsOfSize[3]);
  */

  printf("------------------------------------------------------------\n");
  printf("Extractor Size  %5d %5d %5d %5d\n", 2, 3, 4, 5);
  printf("Sharing Found   %5d %5d %5d %5d\n", 
	 _numSharingsOfSize[0], 
	 _numSharingsOfSize[1], 
	 _numSharingsOfSize[2], 
	 _numSharingsOfSize[3]);
  printf("------------------------------------------------------------\n");

  printf("number folded extractor enumerations:  %5d\n", _nfse1);
  printf("number regular extractor enumerations: %5d\n", _nrse1);
  printf("number folded remainder computation:   %5d\n", _nfse2);
  printf("number regular remainder computation:  %5d\n", _nrse2);
  printf("\n");
}


/****************************************
             Debug Functions
*****************************************/
/*
  function: checkSeeles()
  desc: checks correctness of two variable extractors.
*/
void
checkSeeles(Model *model)
{
  struct _List lBddis;
  ListCell cBddi;

  listConstruct0(&lBddis, _fbddHeap);
  modelBddiList(_fbddHeap, model, &lBddis);

  listForeach(&lBddis,cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    ListCell cSeele;

    listForeach(&bddi->lSees[0], cSeele){
      SEExtractorListElement *seele =
	listData(SEExtractorListElement*, cSeele);
      
      // check that seele->cell is found in seele->see->lBddis
      {
	ListCell cell;
	char found = FALSE;
	listForeach(&seele->see->lBddis, cell){
	  if (cell == seele->cell)
	    found = TRUE;
	}
	assert(found);
      }
    }
  }

  listDelAll(&lBddis);
}

/****************************************
        SEExtractor Hash Scheme 
*****************************************/
static Juint COMCALL 
_seeHashCode( IHashScheme si, HashTable h, void* key)
{
  SEExtractor *see = key;

  return (((unsigned int)(see->bddis[0])>>2)*PRIME1 
	  + ((unsigned int)(see->bddis[1])>>2)*PRIME2 
	  + ((unsigned int)(see->et))*PRIME3 
	  );
}

static Jboolean COMCALL 
_seeEquals(IHashScheme si, HashTable h, void* key, void *another)
{
  SEExtractor *see1 = (SEExtractor*)key;
  SEExtractor *see2 = (SEExtractor*)another;

  if (see1->bddis[0] == see2->bddis[0]
      && see1->bddis[1] == see2->bddis[1]
      && see1->et == see2->et)
    return TRUE;
  else
    return FALSE;
}

static struct  _DHashScheme	_hashSeeSchemeVtbl = {
  _hashQuery, _hashAddref, _hashRelease,
  _seeHashCode, _seeEquals
};

static struct _IHashScheme	_hashSeeScheme = { &_hashSeeSchemeVtbl };
static const IHashScheme hashSeeScheme2 = &_hashSeeScheme;

static Juint COMCALL
_seeHashCode3( IHashScheme si, HashTable h, void* key)
{
  SEExtractor *see = key;

  return (((unsigned int)(see->bddis[0])>>2)*PRIME1
          + ((unsigned int)(see->bddis[1])>>2)*PRIME2
          + ((unsigned int)(see->bddis[2])>>2)*PRIME3
          + ((unsigned int)(see->et))*PRIME4
          );
}

static Jboolean COMCALL
_seeEquals3(IHashScheme si, HashTable h, void* key, void *another)
{
  SEExtractor *see1 = (SEExtractor*)key;
  SEExtractor *see2 = (SEExtractor*)another;

  if (see1->bddis[0] == see2->bddis[0]
      && see1->bddis[1] == see2->bddis[1]
      && see1->bddis[2] == see2->bddis[2]
      && see1->et == see2->et)
    return TRUE;
  else
    return FALSE;
}

static struct  _DHashScheme     _hashSeeScheme3Vtbl = {
  _hashQuery, _hashAddref, _hashRelease,
  _seeHashCode3, _seeEquals3
};

static struct _IHashScheme      _hashSeeScheme3 = { &_hashSeeScheme3Vtbl };
static const IHashScheme       hashSeeScheme3 = &_hashSeeScheme3;

static Juint COMCALL
_seeHashCode4( IHashScheme si, HashTable h, void* key)
{
  SEExtractor *see = key;

  return (((unsigned int)(see->bddis[0])>>2)*PRIME1
          + ((unsigned int)(see->bddis[1])>>2)*PRIME2
          + ((unsigned int)(see->bddis[2])>>2)*PRIME3
          + ((unsigned int)(see->bddis[3])>>2)*PRIME4
          + ((unsigned int)(see->et))*PRIME5
          );
}

static Jboolean COMCALL
_seeEquals4(IHashScheme si, HashTable h, void* key, void *another)
{
  SEExtractor *see1 = (SEExtractor*)key;
  SEExtractor *see2 = (SEExtractor*)another;

  if (see1->bddis[0] == see2->bddis[0]
      && see1->bddis[1] == see2->bddis[1]
      && see1->bddis[2] == see2->bddis[2]
      && see1->bddis[3] == see2->bddis[3]
      && see1->et == see2->et)
    return TRUE;
  else
    return FALSE;
}

static struct  _DHashScheme     _hashSeeScheme4Vtbl = {
  _hashQuery, _hashAddref, _hashRelease,
  _seeHashCode4, _seeEquals4
};

static struct _IHashScheme      _hashSeeScheme4 = { &_hashSeeScheme4Vtbl };
static const IHashScheme         hashSeeScheme4 = &_hashSeeScheme4;

static Juint COMCALL
_seeHashCode5( IHashScheme si, HashTable h, void* key)
{
  SEExtractor *see = key;

  return (((unsigned int)(see->bddis[0])>>2)*PRIME1
          + ((unsigned int)(see->bddis[1])>>2)*PRIME2
          + ((unsigned int)(see->bddis[2])>>2)*PRIME3
          + ((unsigned int)(see->bddis[3])>>2)*PRIME4
          + ((unsigned int)(see->bddis[4])>>2)*PRIME5
          + ((unsigned int)(see->et))*PRIME6
          );
}

static Jboolean COMCALL
_seeEquals5(IHashScheme si, HashTable h, void* key, void *another)
{
  SEExtractor *see1 = (SEExtractor*)key;
  SEExtractor *see2 = (SEExtractor*)another;

  if (see1->bddis[0] == see2->bddis[0]
      && see1->bddis[1] == see2->bddis[1]
      && see1->bddis[2] == see2->bddis[2]
      && see1->bddis[3] == see2->bddis[3]
      && see1->bddis[4] == see2->bddis[4]
      && see1->et == see2->et)
    return TRUE;
  else
    return FALSE;
}

static struct  _DHashScheme     _hashSeeScheme5Vtbl = {
  _hashQuery, _hashAddref, _hashRelease,
  _seeHashCode5, _seeEquals5
};

static struct _IHashScheme      _hashSeeScheme5 = { &_hashSeeScheme5Vtbl };
static const IHashScheme hashSeeScheme5 = &_hashSeeScheme5;

const IHashScheme hashSeeScheme[] 
  = {&_hashSeeScheme, &_hashSeeScheme3, 
     &_hashSeeScheme4, &_hashSeeScheme5};
