/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <assert.h>

#include "fbddutil.h"
#include "cuddInt.h"
#include "cudd.h"
#include "se.h"

extern IHashScheme hashSECacheScheme;

typedef struct _SECacheKey
{
  DdNode *bddSupport;
  BddClass *bddc;
} SECacheKey;

void
sercConstruct(SeManager sem)
{
  hashDefaultConstruct(&sem->htRemainderCache, _fbddHeap);
  sem->htRemainderCache.scheme = hashSECacheScheme;
}

void
sercDestruct(SeManager sem)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  int j, k;
  HashCell hc;

  hashForeach( &sem->htRemainderCache, k, j, hc ){
    SECacheKey *seckFree;
    SERemainder *remainder;

    seckFree = hashKey(SECacheKey*, hc);
    remainder = hashData(SERemainder*, hc);

    Cudd_RecursiveDeref(ddManager, seckFree->bddSupport);
    comFree(_fbddHeap, seckFree);

    serDestruct(sem, remainder);
  }

  hashDestruct(&sem->htRemainderCache);
}

void
sercReleaseSupport(SeManager sem, DdNode *bddSupport)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  Cudd_RecursiveDeref(ddManager, bddSupport);
}

DdNode *
sercGetSupportFromVars(SeManager sem, int *vars, int extSize)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  int i;
  DdNode *bddResult;

  bddResult = Cudd_ReadOne(ddManager);
  Cudd_Ref(bddResult);

  for (i=0; i<extSize; i++){
    DdNode *n = Cudd_bddIthVar(ddManager, vars[i]);
    DdNode *bddTmp;
    Cudd_Ref(n);

    bddTmp = Cudd_bddAnd(ddManager, n, bddResult);
    Cudd_Ref(bddTmp);
    Cudd_RecursiveDeref(ddManager, n);
    Cudd_RecursiveDeref(ddManager, bddResult);
    bddResult = bddTmp;
  }

  return bddResult;
}

DdNode *
sercGetSupport(SeManager sem, SEExtractor *see, BddInstance *bddi, int extSize)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  DdNode *bddResult;
  int i;
  int numVars = bddi->bddClass->numVars;
  int foundCount=0;

  bddResult = Cudd_ReadOne(ddManager);
  Cudd_Ref(bddResult);

  // mark bddis used by see.
  for(i=0; i<extSize; i++){
    BddInstance *seeBddi = see->bddis[i];
    bddiSetMark(seeBddi, 13, TRUE);
  }
  for (i=0; i<numVars; i++){
    if (bddiGetMark(bddi->inputVars[i], 13)){
      DdNode *n = Cudd_bddIthVar(ddManager, i);
      DdNode *bddTmp;
      Cudd_Ref(n);

      bddTmp = Cudd_bddAnd(ddManager, n, bddResult);
      Cudd_Ref(bddTmp);
      Cudd_RecursiveDeref(ddManager, n);
      Cudd_RecursiveDeref(ddManager, bddResult);
      bddResult = bddTmp;
      foundCount++;
    }
  }
  assert(foundCount == extSize);
  for(i=0; i<extSize; i++){
    BddInstance *seeBddi = see->bddis[i];
    bddiSetMark(seeBddi, 13, FALSE);
  }

  return bddResult;
}

SERemainder *
sercFind(SeManager sem, DdNode *bddSupport, BddClass *bddc)
{
  SECacheKey seck;

  seck.bddSupport = bddSupport;
  seck.bddc = bddc;

  return (SERemainder*)hashGet(&sem->htRemainderCache, &seck);
}

void
sercPut(SeManager sem, DdNode *bddSupport, BddClass *bddc, SERemainder *ser)
{
  SECacheKey *seck;

  seck = (SECacheKey*)comAlloc(_fbddHeap, sizeof(SECacheKey));
  seck->bddSupport = bddSupport;
  seck->bddc = bddc;
  Cudd_Ref(seck->bddSupport);
  hashPut(&sem->htRemainderCache, seck, ser);
}

void
sercRemove(SeManager sem, DdNode *bddSupport, BddClass *bddc)
{
  DdManager *ddManager = sem->fm->model->ddManager;
  HashCell hc;
  SECacheKey seck;
  SERemainder *remainder;

  seck.bddSupport = bddSupport;
  seck.bddc = bddc;

  // ERROR: remove hashGet
  if (hashGet(&sem->htRemainderCache, &seck) != NULL){
    hc = hashRemove(&sem->htRemainderCache, &seck);
    if (hc != NULL){
      SECacheKey *seckFree = hashKey(SECacheKey*, hc);
      remainder = hashData(SERemainder*, hc);
      
      Cudd_RecursiveDeref(ddManager, seckFree->bddSupport);
      comFree(_fbddHeap, seckFree);
      serDestruct(sem, remainder);
      comFree(sem->htRemainderCache.mem, hc);
    }
  }
}

/****************************************
        SECache Hash Scheme 
*****************************************/
static Juint COMCALL 
_seCacheHashCode( IHashScheme si, HashTable h, void* key)
{
  SECacheKey *rck = (SECacheKey *)key;

  return (((unsigned int)(rck->bddSupport))*PRIME1
	  + ((unsigned int)(rck->bddc)>>2)*PRIME2 
	  );
}

static Jboolean COMCALL 
_seCacheEquals(IHashScheme si, HashTable h, void* key, void *another)
{
  SECacheKey *rck1 = (SECacheKey *)key;
  SECacheKey *rck2 = (SECacheKey *)another;

  if (rck1->bddSupport == rck2->bddSupport
      && rck1->bddc == rck2->bddc)
    return TRUE;
  else
    return FALSE;
}

static struct  _DHashScheme	_hashSECacheSchemeVtbl = {
  _hashQuery, _hashAddref, _hashRelease,
  _seCacheHashCode, _seCacheEquals
};

static struct _IHashScheme	_hashSECacheScheme = { &_hashSECacheSchemeVtbl };
IHashScheme hashSECacheScheme = &_hashSECacheScheme;
