/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "cuddInt.h"
#include "cudd.h"

DdNode *
buildLogicZero(DdManager *manager)
{
  DdNode *f;
  f = Cudd_ReadLogicZero(manager);
  return f;
}

DdNode *
buildLogicOne(DdManager *manager)
{
  DdNode *f;
  f = Cudd_ReadOne(manager);
  return f;
}

// Build Source Follower
DdNode *
buildSF(DdManager *manager)
{
  DdNode *f;
  f = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(f);
  return f;
}

// Build Inverter
DdNode *
buildNOT(DdManager *manager)
{
  DdNode *f, *tmp;

  f = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(f);

  tmp = Cudd_Not(f);
  Cudd_Ref(tmp);

  Cudd_RecursiveDeref(manager, f);
  
  return tmp;
}

DdNode *
buildMUX(DdManager *manager)
{
  DdNode *f, *a, *b, *s;
  
  // index 2: select signal
  // index 1: signal 'a'
  // index 0: signal 'b'
  // f = sa + s'b

  b = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(b);

  a = Cudd_bddIthVar(manager, 1);
  Cudd_Ref(a);

  s = Cudd_bddIthVar(manager, 2);
  Cudd_Ref(s);

  f = Cudd_bddIte(manager, s, a, b);
  Cudd_Ref(f);
  Cudd_RecursiveDeref(manager, a);
  Cudd_RecursiveDeref(manager, b);
  Cudd_RecursiveDeref(manager, s);

  return f;
}

DdNode *
buildMUX2(DdManager *manager)
{
  DdNode *f, *a, *b, *s;

  // index 2: select signal
  // index 0: signal 'a'
  // index 1: signal 'b'
  // f = sa + s'b

  b = Cudd_bddIthVar(manager, 1);
  Cudd_Ref(b);

  a = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(a);

  s = Cudd_bddIthVar(manager, 2);
  Cudd_Ref(s);

  f = Cudd_bddIte(manager, s, a, b);
  Cudd_Ref(f);
  Cudd_RecursiveDeref(manager, a);
  Cudd_RecursiveDeref(manager, b);
  Cudd_RecursiveDeref(manager, s);

  return f;
}

DdNode *
buildOR(DdManager *manager)
{
  DdNode *f, *a, *b;

  // index 1: signal 'a'
  // index 0: signal 'b'
  // f = a + b

  b = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(b);

  a = Cudd_bddIthVar(manager, 1);
  Cudd_Ref(a);

  f = Cudd_bddOr(manager, a, b);
  Cudd_Ref(f);
  Cudd_RecursiveDeref(manager, a);
  Cudd_RecursiveDeref(manager, b);

  return f;
}

DdNode *
buildAND(DdManager *manager)
{
  DdNode *f, *a, *b;

  // index 1: signal 'a'
  // index 0: signal 'b'
  // f = ab

  b = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(b);

  a = Cudd_bddIthVar(manager, 1);
  Cudd_Ref(a);

  f = Cudd_bddAnd(manager, a, b);
  Cudd_Ref(f);
  Cudd_RecursiveDeref(manager, a);
  Cudd_RecursiveDeref(manager, b);

  return f;
}

DdNode *
buildAND01(DdManager *manager)
{
  DdNode *f, *a, *b;

  // index 1: signal 'a'
  // index 0: signal 'b'
  // f = b'a

  b = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(b);
  b = Cudd_Not(b);

  a = Cudd_bddIthVar(manager, 1);
  Cudd_Ref(a);

  f = Cudd_bddAnd(manager, a, b);
  Cudd_Ref(f);
  Cudd_RecursiveDeref(manager, a);
  Cudd_RecursiveDeref(manager, b);

  return f;
}

DdNode *
buildAND10(DdManager *manager)
{
  DdNode *f, *a, *b;

  // index 1: signal 'a'
  // index 0: signal 'b'
  // f = ba'

  b = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(b);

  a = Cudd_bddIthVar(manager, 1);
  Cudd_Ref(a);
  a = Cudd_Not(a);

  f = Cudd_bddAnd(manager, a, b);
  Cudd_Ref(f);
  Cudd_RecursiveDeref(manager, a);
  Cudd_RecursiveDeref(manager, b);

  return f;
}


DdNode *
buildXNOR(DdManager *manager)
{
  DdNode *f, *a, *b;
  
  b = Cudd_bddIthVar(manager, 0);
  Cudd_Ref(b);

  a = Cudd_bddIthVar(manager, 1);
  Cudd_Ref(a);
  
  f = Cudd_bddXnor(manager, a, b);
  Cudd_Ref(f);
  Cudd_RecursiveDeref(manager, a);
  Cudd_RecursiveDeref(manager, b);
  
  return f;
}

DdNode*
buildOROfSize(DdManager *ddManager, int size)
{
  DdNode *nOr;
  DdNode *nTmp, *nTmp2;
  int i;

  nOr = Cudd_ReadLogicZero(ddManager);
  Cudd_Ref(nOr);
  for (i=0; i<size; i++){
    nTmp = Cudd_bddIthVar(ddManager, i);
    Cudd_Ref(nTmp);
    
    nTmp2 = Cudd_bddOr(ddManager, nOr, nTmp);
    Cudd_Ref(nTmp2);
    Cudd_RecursiveDeref(ddManager, nOr);
    Cudd_RecursiveDeref(ddManager, nTmp);
    nOr = nTmp2;
  }
  
  return nOr;
}

