/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <assert.h>

#include "cudd.h"
#include "metautil.h"
#include "fbddutil.h"
#include "fbddopt.h"

void
bddcSimplify(FbddManager fm, BddClass *bddc)
{
  Model *model = fm->model;
  DdNode *node;
  BddInstance **newInputVars;
  int i;    
  ListCell bddiCell;
  BddClass *bddcB;
  int *oldIndexToNew;
  int arraySize;
  int numVars;
  
  int oldNumVars = bddc->numVars;

  // copy bdd to another manager, reduce the bdd, and copy it back.
  if (isBasicGate(model, bddc)==TRUE)
    return;

  node = foldBddClass(model, bddc->bdd,
		      &numVars, &oldIndexToNew, &arraySize, 
		      NULL, NULL);

  bddcB = NULL;
  if (node != bddc->bdd){
    ListCell cBddi;

    // no instances of type bddc->bdd will exist
    assert(isBasicGateBdd(model, bddc->bdd) == FALSE);
    bclRemove(model->bclm, bddc->bdd);

    // bdd has changed, update the bddClass and it's bddInstances.
    bddcB = bddcCreateForBdd(model, node, 
			      numVars, &model->nextBddcId);

    listForeach(&bddc->lInsts, cBddi){
      ListCell cNew;
      BddInstance *bddi = listData(BddInstance*, cBddi);
      
      bddi->bddClass = bddcB;
      
      listNewCell(_fbddHeap, cNew, bddi);
      listAppendElement(&bddcB->lInsts, cNew);
      bddi->cBddcInst = cNew;
    }

    // Reordering may have resulted in the removal of some variables and   
    // the variable ordering maybe be different.  This information is stored
    // in numVars and oldIndexToNew.  Update BddInstances.
    listForeach(&bddc->lInsts, bddiCell){
      BddInstance *bddiPtr = listData(BddInstance*, bddiCell);
      
      newInputVars = comAlloc(_fbddHeap, sizeof(BddInstance*)*numVars);

      assert(oldNumVars >= numVars);
      
      for (i=0; i<numVars; i++)
	newInputVars[i] = NULL;

      for (i=0; i<oldNumVars; i++){
	if (oldIndexToNew[i] != -1){
	  assert(newInputVars[oldIndexToNew[i]] == NULL);
	  newInputVars[oldIndexToNew[i]] = bddiPtr->inputVars[i];
	}
      }
      
      // check every input has been assigned
      for (i=0; i<numVars; i++)
	assert(newInputVars[i]);
      
      comFree(_fbddHeap, bddiPtr->inputVars);
      bddiPtr->inputVars = newInputVars;
    }      
    
    // instances from bddc were moved to bddcB
    Cudd_RecursiveDeref(model->ddManager, bddc->bdd);
    bddcDestroyKeepInstances(bddc);
  }
  else
    // destroy bddc
    Cudd_RecursiveDeref(model->ddManager, bddc->bdd);

  comFree(_fbddHeap, oldIndexToNew);
}

void
fbddSimplify(FbddManager fm)
{
  struct _List lBddcs;
  ListCell cell;
  IMem tmpArena = _fbddHeap;

  listConstruct0(&lBddcs, tmpArena);
  modelBddcList(tmpArena, fm->model, &lBddcs);
  listForeach(&lBddcs, cell){
    BddClass *bddc = listData(BddClass*, cell);
    bddcSimplify(fm, bddc);
  }
  listDelAll(&lBddcs);
}
