/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
  DESC: The sweep routines perform simple redundany removal through
  constant propagation, merging repeating support variables and 
  removing dangling logic networks.
*/

#include <assert.h>
#include <time.h>

#include "metautil.h"
#include "fbdd.h"
#include "fbddopt.h"
#include "fbddutil.h"

void
sweepCBAssign(ISweepCallBack cb,
	      void *objCallee,
	      void (*beforeSweepBddi)( void *obj, BddInstance *bddi ),
	      void (*afterSweepBddi)( void *obj, BddInstance *bddi ))
{
  cb->objCallee = objCallee;
  cb->beforeSweepBddi = beforeSweepBddi;
  cb->afterSweepBddi = afterSweepBddi;
}

/*********************
     Input Config
**********************/
// Returns TRUE if input can be simplified.
// Returns FALSE otherwise.
// index of inputSimplify = -1 if const zero
// -2 if const one
// otherwise the value represents the first index where
// this inputvar was encountered.
InputConfig
icCompute(IMem _mem, int numVars, 
	  char findConsts, BddInstance **inputVars)
{
  // bddi to first variable index with this bddi
  int i;
  InputConfig inputSimplify;

  inputSimplify = (int*)comAlloc(_mem, sizeof(int)*(numVars+1));

  // first index represents whether the inputs can be simplified.
  inputSimplify[0] = 0; // set to FALSE;

  for (i=0; i<numVars; i++){
    BddInstance *bddiInput = inputVars[i];
    BddClass *bddcInput = bddiInput->bddClass;
    if (findConsts 
	&& bddcInput->numVars==0 && bddcInput->type==INTERNAL_VARIABLE
	&& !Cudd_IsComplement(bddcInput->bdd)
	){
      // found constant zero
      inputSimplify[i+1] = CONST_ONE; // const zero
      // INCOMPLETE: if lFanout is 0, can remove subtree
      inputSimplify[0]++;
    }
    else if (findConsts 
	     && bddcInput->numVars==0 && bddcInput->type==INTERNAL_VARIABLE
	     && Cudd_IsComplement(bddcInput->bdd)
	     ){
      // found constant one
      inputSimplify[i+1] = CONST_ZERO; // const one
      inputSimplify[0]++;
    }
    else{
      int varIndex;
      int j;

      // check for repeating inputs
      if (bddiGetMark(bddiInput, 10) == FALSE){
	assert(i<MAX_BDD_VARS);
	varIndex = i;
	bddiSetMark(bddiInput, 10, TRUE);
      }
      else{
	// bddi has been found before
	for (j=0; j<numVars; j++)
	  if (inputVars[j] == bddiInput){
	    varIndex = j;
	    break;
	  }
	inputSimplify[0]++;
      }
      inputSimplify[i+1] = varIndex;
    }
  }

  for (i=0; i<numVars; i++){
    bddiSetMark(inputVars[i],10,FALSE);
  }

  return inputSimplify;
}

int
icNumSS(InputConfig ic)
{
  return ic[0];
}

int *
icComputeStringKey(IMem mem, int numVars, InputConfig inputConfig)
{
  int *result;
  int i;

  result = (int*)comAlloc(mem, sizeof(int)*(numVars+1));
  for(i=0; i<numVars; i++){
    result[i] = icReadVar(inputConfig,i)+1;
  }
  result[numVars] = 0;

  return result;
}

char
icCompare(int numVars, InputConfig ic1, InputConfig ic2)
{
  int i;

  if (icCanSimplify(ic1) != icCanSimplify(ic2))
    return FALSE;

  for (i=0; i<numVars; i++){
    if (icReadVar(ic1,i) != icReadVar(ic2,i))
      return FALSE;
  }
  return TRUE;
}

/*********************
        Sweep
**********************/
// Requires:
// reorder: will always reorder if TRUE, otherwise, reorders only if 
// canSimplify.
BddClass *
sweepBddc(SweepManager sm,
	  BddClass *bddc, 
	  InputConfig inputSimplify, char reorder,
	  int *oldIndexToNew)
{
  Model *model = sm->fm->model;
  DdManager *ddmSweep = sm->ddm;
  int i;
  DdNode *nTmp;
  BddClass *bddcSimplified;
  clock_t t;
  int numVars;
  int arraySize;
  int *oldIndexToNew3;
  int canSimplify = FALSE;

  DdNode *bddSupport;
  DdNode *nVar;

  assert(bddc->type != LATCH);

  if (bddc->numVars == 0){
    return bddc;
  }

  t = util_cpu_time();

  // if can't simplify and don't reorder, then bddcSimplified = bddc.
  canSimplify = icCanSimplify(inputSimplify);
  if (!canSimplify && !reorder){
    for (i=0; i<bddc->numVars; i++)
      oldIndexToNew[i] = i;
    tSimplify += util_cpu_time() - t;
    return bddc;
  }

  for (i=0; i<bddc->numVars+1; i++){
    Cudd_bddIthVar(ddmSweep, i);
  }

  // copy to ddmSweep
  nTmp = bddCopyAndReIndex(model->ddManager, bddc->bdd, ddmSweep,
			 &numVars, &oldIndexToNew3, &arraySize,
			 NULL, NULL);

  bddSupport = Cudd_Support(ddmSweep, nTmp);
  Cudd_Ref(bddSupport);

  for (i=0; i<bddc->numVars; i++){
    DdNode *nChild=NULL;

    // check var i still exists in nTmp
    {
      DdNode *bddAnd;
      nVar = Cudd_bddIthVar(ddmSweep,  oldIndexToNew3[i]);
      Cudd_Ref(nVar);
      bddAnd = Cudd_bddAnd(ddmSweep, bddSupport, nVar);
      Cudd_Ref(bddAnd);
      if (bddAnd != bddSupport){
	Cudd_RecursiveDeref(ddmSweep, nVar);
	Cudd_RecursiveDeref(ddmSweep, bddAnd);
	continue;
      }
      Cudd_RecursiveDeref(ddmSweep, nVar);
      Cudd_RecursiveDeref(ddmSweep, bddAnd);
    }
    
    if (icReadVar(inputSimplify,i) == CONST_ZERO){
      // zero constant
      nChild = Cudd_ReadLogicZero(ddmSweep);
      Cudd_Ref(nChild);
    }
    else if (icReadVar(inputSimplify,i) == CONST_ONE){
      // one constant
      nChild = Cudd_ReadOne(ddmSweep);
      Cudd_Ref(nChild);
    }
    else if (icReadVar(inputSimplify,i) != i){
      // this input is representated by a variable found before.
      assert(icReadVar(inputSimplify,i)>=0 
	     && icReadVar(inputSimplify,i)<arraySize);
      nChild =
	Cudd_bddIthVar(ddmSweep, oldIndexToNew3[icReadVar(inputSimplify,i)]);
      Cudd_Ref(nChild);
    }
    
    if (icReadVar(inputSimplify,i) != i){
      DdNode *nCompose;
      assert(nChild);
      assert(i>=0 && i<arraySize);

      nCompose = Cudd_bddCompose(ddmSweep, nTmp, nChild, oldIndexToNew3[i]);
      Cudd_Ref(nCompose);
      Cudd_RecursiveDeref(ddmSweep, nTmp);
      Cudd_RecursiveDeref(ddmSweep, nChild);
      nTmp = nCompose;

      Cudd_RecursiveDeref(ddmSweep, bddSupport);
      bddSupport = Cudd_Support(ddmSweep, nTmp);
      Cudd_Ref(bddSupport);
    }
    else
      assert(!nChild);
  }

  Cudd_RecursiveDeref(ddmSweep, bddSupport);
  
  // final bdd is in nTmp.
  {
    DdNode *nSimplified;
    int *oldIndexToNew1;
    int numVars1, arraySize1;
    clock_t t2;
    
    // Variable reorder
    if (reorder==TRUE || canSimplify){
      t2 = util_cpu_time();
      Cudd_ReduceHeap(ddmSweep, CUDD_REORDER_TYPE, 1);
      tReduceHeap += util_cpu_time() - t2;
    }
    
    nSimplified = bddCopyAndReIndex(ddmSweep, nTmp, model->ddManager,
				    &numVars1, &oldIndexToNew1, &arraySize1,
				    NULL, NULL);

    Cudd_RecursiveDeref(ddmSweep, nTmp);

    // create a new BddClass if one does not already exist for nSimplified.
    bddcSimplified =
      bddcCreateForBdd(model, nSimplified, numVars1,
		       &model->nextBddcId);

    // populate oldIndexToNew
    for (i=0; i<bddc->numVars; i++){
      if (oldIndexToNew3[i]>=arraySize1 
	  || oldIndexToNew1[oldIndexToNew3[i]] == -1)
	oldIndexToNew[i] = -1;
      else
	oldIndexToNew[i] = oldIndexToNew1[oldIndexToNew3[i]];
    }
    comFree(_fbddHeap,oldIndexToNew1);
  }
  comFree(_fbddHeap,oldIndexToNew3);
  
  tSimplify += util_cpu_time() - t;
  return bddcSimplified;
}

static void
sweepDanglingNetworkHelper(SweepManager sm, Nml *nml, BddInstance* bddi)
{
  int i;
  int numVars;
  Model *model = sm->fm->model;

  if ( bddiNumFanout(bddi) > 0 
       || bddi->bddClass->type == PRIMARY_INPUT
       || nmlGetBddiMark(nml, bddi)
      )
    return;

  nmlPutBddiMark(nml, bddi, 1);

  //  cbRun(ISweepCallBack, sm, beforeDestroyBddi, bddi);
  cbRun(ISweepCallBack, sm, beforeSweepBddi, bddi);

  numVars = bddi->bddClass->numVars;

  bclRemoveBddi(model->bclm, model, bddi);
  if (bddcNumInsts(bddi->bddClass)==0 && !isBasicGate(model, bddi->bddClass))
    bddcDestruct(bddi->bddClass);

  for (i=0; i<numVars; i++){
    int nDel;
    
    nDel = listDelData(&bddi->inputVars[i]->lFanout, 
		       bddi);
    sweepDanglingNetworkHelper(sm, nml, bddi->inputVars[i]);
  }

  bddiDestruct(bddi);
}

// Will not free empty bddcs.
// But will free bddis.  Take care not to depend on static lists of bddis
// when using this function.
void 
sweepDanglingNetwork(SweepManager sm, BddInstance *bddi)
{
  Nml nml;
  nmlConstruct(&nml);
  sweepDanglingNetworkHelper(sm, &nml, bddi);
  nmlDestruct(&nml);
}

void
sweepDanglingLatchNetworks(SweepManager sm)
{
  Model *model = sm->fm->model;
  IMem arena;
  ListCell cBddi;
  struct _List lTemp;
  List lCurr;
  int i;
  struct _List lBddis;

  arena = _fbddHeap;

  // mark reachable nodes
  listConstruct0(&lBddis, arena);
  modelBddiList(arena, model, &lBddis);  
  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    bddiSetMark(bddi, 15, TRUE);
  }
  listDelAll(&lBddis);

  for (i=0; i<5; i++){ // cycle through all latch types
    switch(i){
    case 0: lCurr = &model->bddcLatchFe->lInsts; break;
    case 1: lCurr = &model->bddcLatchRe->lInsts; break;
    case 2: lCurr = &model->bddcLatchAh->lInsts; break;
    case 3: lCurr = &model->bddcLatchAl->lInsts; break;
    case 4: lCurr = &model->bddcLatchAs->lInsts; break;
    }

    listConstruct0(&lTemp, _fbddHeap);
    while(listGetSize(lCurr)>0){
      BddInstance *bddi;
      cBddi = listGetHead(lCurr);
      bddi = listData(BddInstance*, cBddi);
      
      // if bddi not reachable
      if (!bddiGetMark(bddi, 15)){
	ListCell cFanout;
	// break the cycles.
	listForeach(&bddi->lFanout, cFanout){
	  BddInstance *bddiFanout = listData(BddInstance*, cFanout);
	  int i;
	  for (i=0; i<bddiFanout->bddClass->numVars; i++){
	    if (bddiFanout->inputVars[i] == bddi){
	      assert(i != bddiFanout->bddClass->numVars); // check found it
	      bddiFanout->inputVars[i] = 
		model->bddiDummy; // redirect to dummy bddi
	      listAppend(&model->bddiDummy->lFanout, bddiFanout);
	    }
	  }
	}
	listDelAll(&bddi->lFanout);
	
	// destruct network
	sweepDanglingNetwork(sm, bddi);
      }
      else{
	cBddi = listPop(lCurr);
	listAppendElement(&lTemp, cBddi);
      }
    }

    switch(i){
    case 0: model->bddcLatchFe->lInsts = lTemp; break;
    case 1: model->bddcLatchRe->lInsts = lTemp; break;
    case 2: model->bddcLatchAh->lInsts = lTemp; break;
    case 3: model->bddcLatchAl->lInsts = lTemp; break;
    case 4: model->bddcLatchAs->lInsts = lTemp; break;
    }
  }

  // unmark reachable nodes
  listConstruct0(&lBddis, arena);
  modelBddiList(arena, model, &lBddis);  
  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    bddiSetMark(bddi, 15, FALSE);
  }
  listDelAll(&lBddis);
}

// Constant Fold and merge equivalent input vars into one.
// Does not update fanout.
// Returns 2, if bddi was destroyed.
// Returns 1, if changes were made.
// Returns 0, if no changes were made.
// Requires: updateMPs: TRUE if MPs to be updated
int
sweepBddi(SweepManager sm, BddInstance *bddi, List lDanglingOutputs)
{
  Model *mdl = sm->fm->model;
  BddInstance **oldInputVars;
  int numVars;
  int i;
  char canSimplify = FALSE;
  int result = 0;
  InputConfig inputSimplify;
  IMem _arena;

  if (bddi->bddClass->type != INTERNAL_VARIABLE
      || bddi->bddClass == mdl->bddcLogicOne
      || bddi->bddClass == mdl->bddcLogicZero
      )
    return FALSE;

  _arena = _fbddHeap;

  cbRun(ISweepCallBack, sm, beforeSweepBddi, bddi);

  oldInputVars = bddi->inputVars;
  numVars = bddi->bddClass->numVars;

  // Remove SF or NOT inputs.
  for(i=0; i<numVars; i++){
    BddInstance *bddiInput = bddi->inputVars[i];
    if (bddiInput->bddClass == mdl->bddcSF
	|| bddiInput->bddClass == mdl->bddcNot
	){
      cbRun(ISweepCallBack, sm, beforeSweepBddi, bddiInput);
      if (bddiInput->bddClass == mdl->bddcNot){
	DdNode *bdd;
	DdNode *nComplement;
	BddClass *bddcNew;
	BddInstance **inputVars;

	bdd = bddi->bddClass->bdd;

	nComplement = Cudd_Not(Cudd_bddIthVar(mdl->ddManager, i));
	Cudd_Ref(nComplement);

	bdd = Cudd_bddCompose(mdl->ddManager, bdd, nComplement, i);
	Cudd_Ref(bdd);
	Cudd_RecursiveDeref(mdl->ddManager, nComplement);
	
	bddcNew = bddcCreateForBdd(mdl, bdd, numVars,
				   &mdl->nextBddcId);
	
	// Add bddi to bddcNew
	listDelElement(&bddi->bddClass->lInsts, 
		       bddi->cBddcInst);
	if (bddcNumInsts(bddi->bddClass) == 0
	    && !isBasicGate(mdl, bddi->bddClass)){
	  bclRemove(mdl->bclm, bddi->bddClass->bdd);
	  bddcDestruct(bddi->bddClass);
	}

	inputVars = bddi->inputVars;
	bddi->inputVars = NULL;
	bddcAddInstance(bddcNew, bddi);
	comFree(_fbddHeap, bddi->inputVars);
	bddi->inputVars = inputVars;
      }

      listDelDataAll(&bddiInput->lFanout, bddi);
      
      // attach bddi to bddiInput->inputVars[0]
      bddi->inputVars[i] = bddiInput->inputVars[0];
      listAppend(&bddiInput->inputVars[0]->lFanout, bddi);
      
      // If signal bddiInput not used, destroy it.
      if (listGetSize(&bddiInput->lFanout)==0){
	listDelDataAll(&bddiInput->inputVars[0]->lFanout, 
			  bddiInput);
	listDelElement(&bddiInput->bddClass->lInsts, 
		       bddiInput->cBddcInst);
	bddiDestruct(bddiInput);
      }
      else{
	cbRun(ISweepCallBack, sm, afterSweepBddi, bddiInput);	
	cbRun(ISweepCallBack, sm, beforeSweepBddi, bddi);  
      }
    }
  }
  //  cbRun(ISweepCallBack, sm, afterSweepBddi, bddi);	

  inputSimplify = icCompute(_arena, numVars, TRUE, oldInputVars);
  canSimplify = icCanSimplify(inputSimplify);

  if (canSimplify == FALSE){
    result = 0;
  }
  else{
    BddClass *bddcSimplified;
    int *oldIndexToNew;
    
    oldIndexToNew = (int*)comAlloc(_arena, sizeof(int)*numVars); 
    bddcSimplified = sweepBddc(sm, bddi->bddClass, 
			       inputSimplify, FALSE, oldIndexToNew);
    assert(bddcSimplified != bddi->bddClass);

    listDelElement(&bddi->bddClass->lInsts, bddi->cBddcInst);
    if (bddcSimplified != bddi->bddClass 
	&& bddcNumInsts(bddi->bddClass) == 0
	&& !isBasicGate(mdl, bddi->bddClass)){
      bclRemove(mdl->bclm, bddi->bddClass->bdd);
      bddcDestruct(bddi->bddClass);
    }

    bddi->inputVars = NULL;
    bddcAddInstance(bddcSimplified, bddi);    
    
    // update inputVars
    for (i=0; i<numVars; i++){
      if (oldIndexToNew[i] != -1){
	bddi->inputVars[oldIndexToNew[i]] = oldInputVars[i];
      }
    }
      
    // update fanout of inputVars
    for (i=0; i<numVars; i++){
      if (icReadVar(inputSimplify,i) != i || oldIndexToNew[i] == -1){
	int found;
	
	// input was removed because it was a constant or because
	// it was shared at this gate at a different position 
	// or was eliminated during simplification.
	found = listDelData(&oldInputVars[i]->lFanout, bddi);
    	assert(found);
	if (bddiNumFanout(oldInputVars[i])==0){
	  ListCell cNew;
	  listNewCell(_fbddHeap, cNew, oldInputVars[i]);
	  listAppendElement(lDanglingOutputs, cNew);
	}
      }
    }
    
    comFree(_arena, oldIndexToNew);
    comFree(_fbddHeap, oldInputVars);
    result = 1;
  }
  
  comFree(_arena, inputSimplify);

  cbRun(ISweepCallBack, sm, afterSweepBddi, bddi);
  
  return result;
}


static void
sweepModelHelper(SweepManager sm, BddInstance *bddi)
{
  int numVars = bddi->bddClass->numVars;
  int i;
  int sweepResult;

  struct _List lDanglingOutputs;
  
  if (bddiGetMark(bddi, 11))
    return;

  bddiSetMark(bddi, 11, TRUE);

  switch(bddi->bddClass->type){
  case PRIMARY_INPUT:
    return;
  case PRIMARY_OUTPUT: 
    if (bddiGetMark(bddi->inputVars[0], 11) == FALSE){
      sweepModelHelper(sm, bddi->inputVars[0]);
    }
    return;
  case LATCH:
    bddiSetMark(bddi, 11, TRUE);
    if (bddiGetMark(bddi->inputVars[0], 11) == FALSE)
      sweepModelHelper(sm, bddi->inputVars[0]);
    return;
  default:
    break;
  }

  // sweep children
  for (i=0; i<numVars; i++){
    BddInstance *bddiChild = bddi->inputVars[i];
    assert(i < numVars);
    assert(i < bddi->bddClass->numVars);
    sweepModelHelper(sm, bddiChild);
  }

  // sweep self
  listConstruct0(&lDanglingOutputs, _fbddHeap);
  sweepResult = sweepBddi(sm, bddi, &lDanglingOutputs);
  if (sweepResult>0){
    ListCell cOutput;
    listForeach(&lDanglingOutputs, cOutput){
      sweepDanglingNetwork(sm, listData(BddInstance*, cOutput));
    }
  }
  listDelAll(&lDanglingOutputs);
}

static void
remBddiFromLSweepBddis(SweepManager sm, BddInstance *bddi)
{
  listDelData(sm->lSweepIncBddis, bddi);
}

// lBddis, is the list of bddis that need to be sweeped.
// Assumes all other bddis in the network are sweeped.
void
sweepIncr(SweepManager sm, IMem mem, List lBddis)
{
  clock_t t = util_cpu_time();
  Model *model = sm->fm->model;
  ListCell cCB;
  struct _ISweepCallBack cbSweep;

  sweepCBAssign(&cbSweep, sm, 
	(void (*)( void *obj, BddInstance *bddi))remBddiFromLSweepBddis,
		NULL);
  cbAdd(sm, cCB, &cbSweep);

  sm->lSweepIncBddis = lBddis;
  sm->memSweepInc = mem;
  
  while(listGetSize(lBddis)>0){
    ListCell cBddi = listPop(lBddis);
    BddInstance *bddi = listData(BddInstance*, cBddi);
    int sweepResult;
    struct _List lDanglingOutputs;
    struct _List lFanouts;

    listConstruct0(&lDanglingOutputs, _fbddHeap);
    listConstruct0(&lFanouts, _fbddHeap);
    listClone(&bddi->lFanout, &lFanouts);
    
    sweepResult = sweepBddi(sm, bddi, &lDanglingOutputs);
    if (sweepResult==2 
	|| bddi->bddClass==model->bddcSF || bddi->bddClass==model->bddcNot
	|| bddi->bddClass==model->bddcLogicOne 
	|| bddi->bddClass==model->bddcLogicZero){
      ListCell cFanout;
      // Parent need to be sweeped.
      // Add parents to lBddis
      listForeach(&lFanouts, cFanout){
	BddInstance *bddiFanout = listData(BddInstance*, cFanout);
	if (bddiFanout->bddClass->type == INTERNAL_VARIABLE
	    && listSearch(lBddis, bddiFanout)==NULL)
	  listAppend(lBddis, bddiFanout);
      }
    }
    listDelAll(&lFanouts);
    if (sweepResult>0){
      ListCell cOutput;

      listForeach(&lDanglingOutputs, cOutput){
	sweepDanglingNetwork(sm, listData(BddInstance*, cOutput));
      }
    }
    listDelAll(&lDanglingOutputs);

    comFree(mem, cBddi);
  }

  cbRemove(sm, cCB);

  tSweep += util_cpu_time() - t;
}

SweepManager 
sweepNewManager(FbddManager fm)
{
  SweepManager sm = (SweepManager)comAlloc(_fbddHeap, 
					   sizeof(struct _SweepManager));

  sm->fm = fm;
  fm->sm = sm;
  sm->ddm = newDdManager();
  listConstruct0(&sm->lCBs, _fbddHeap);

  return sm;
}

void
sweepDelManager(SweepManager sm)
{
  Cudd_Quit(sm->ddm);
  sm->fm->sm = NULL;
  comFree(_fbddHeap, sm);
}

void 
sweepRun(SweepManager sm)
{
  ListCell cOutput, cBddi;
  IMem mem;
  struct _List lBddis;
  Model *model = sm->fm->model;
  
  listForeach(&model->lOutputs, cOutput){
    BddInstance *bddiOutput = listData(BddInstance*, cOutput);
    sweepModelHelper(sm, bddiOutput);
  }

  // clear bddi marks
  mem = _fbddHeap;
  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);
  listForeach(&lBddis, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    bddiSetMark(bddi, 11, FALSE);
  }
  listDelAll(&lBddis);
}
