/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <assert.h>

#include "metautil.h"

#include "fbddutil.h"
#include "fbddopt.h"


typedef struct _WriteDotData
{
  Nml nmlBddi;
  FILE *dotFile;
  struct _List listBddc;
} WriteDotData;

static void
bddiWriteDotHelper(WriteDotData *wdd, BddInstance *bddi)
{
  ListCell cell;
  int bddcInList = FALSE;

  if (bddi == NULL)
    return;

  // do not reprocess bddi if already processed
  if (nmlGetBddiMark(&wdd->nmlBddi, bddi))
    return;
  
  nmlPutBddiMark(&wdd->nmlBddi, bddi, 1);

  // if corresponding bddc not in list, insert it.
  // SLOW.
  listForeach(&wdd->listBddc, cell){
    if (listData(BddClass*, cell) == bddi->bddClass){
      bddcInList = TRUE;
      break;
    }
  }

  if (bddcInList == FALSE && bddi->bddClass->type != PRIMARY_INPUT
      && bddi->bddClass->type != PRIMARY_OUTPUT){
    listNewCell(_fbddHeap, cell, bddi->bddClass);
    listAppendElement(&wdd->listBddc, cell);
  }

  // Write Edges and Nodes
  if (bddi->bddClass->type == PRIMARY_INPUT){
    // Primary Input
    fprintf(wdd->dotFile, 
	    "inst_%d [shape=record,label=\"<o> PI_%s\"]\n", 
	    (int)bddi, bddi->id);
  }
  else if(bddi->bddClass->type == PRIMARY_OUTPUT){
    int i;
    // Primary Input
    fprintf(wdd->dotFile, 
	    "inst_%d [shape=record,label=\"<i> PO_%s\"]\n", 
	    (int)bddi, bddi->id);

    // write edges to children
    for (i=0; i<bddi->bddClass->numVars; i++){
      fprintf(wdd->dotFile, "inst_%d:i -> inst_%d:o;\n", 
	      (int)bddi, (int)(bddi->inputVars[i]));
    }

    for (i=0; i<bddi->bddClass->numVars; i++)
      bddiWriteDotHelper(wdd, bddi->inputVars[i]);
  }
  else{
    // Internal Node
    int i;

    // write node
    fprintf(wdd->dotFile,
	    "inst_%d [shape=record,label=\"{<o> %s|{", 
	    (int)bddi, bddi->bddClass->name);
    for (i=0; i<bddi->bddClass->numVars; i++){
      if (i!=0)
	fprintf(wdd->dotFile, "|");
      fprintf(wdd->dotFile, "<i%d> %d", i, i);
    }
    fprintf(wdd->dotFile, "}}\"]\n");
    
    // write edges to children
    for (i=0; i<bddi->bddClass->numVars; i++){
      fprintf(wdd->dotFile, "inst_%d:i%d -> inst_%d:o;\n", 
	      (int)bddi, i, (int)(bddi->inputVars[i]));
    }

    for (i=0; i<bddi->bddClass->numVars; i++)
      bddiWriteDotHelper(wdd, bddi->inputVars[i]);
  }
}

void
modelWriteDot(Model *model, char *instFile, char *bddcFile)
{
  FILE *dotFile;
  DdNode **outNodes;
  char **outNames;
  ListCell cell;
  int i=0;
  DdManager *manager=model->ddManager;
  WriteDotData wdd;

  listConstruct0(&wdd.listBddc, _fbddHeap); // list of unique bddc's
  
  // write bddi to dot
  dotFile = fopen(instFile, "w");
  fprintf(dotFile, "digraph structs {\n");
  fprintf(dotFile, "size = \"10,8\"; orientation=landscape; ");
  fprintf(dotFile, "edge [dir=none];\n");
  fprintf(dotFile, "node [shape=record];\n");
  
  wdd.dotFile = dotFile;
  nmlConstruct(&wdd.nmlBddi);
  listForeach(&(model->lOutputs), cell){
    bddiWriteDotHelper(&wdd, listData(BddInstance*, cell));
  }
  nmlDestruct(&wdd.nmlBddi);
    
  fprintf(dotFile, "}\n");
  fclose(dotFile);

  // write bddc's to dot
  outNodes = 
    (DdNode**)comAlloc(_fbddHeap,sizeof(DdNode*)*listGetSize(&wdd.listBddc));
  outNames = 
    (char**)comAlloc(_fbddHeap,sizeof(char*)*listGetSize(&wdd.listBddc));
    
  // build outNodes and outNames
  listForeach(&wdd.listBddc, cell){
    BddClass *bddc;
    
    bddc = listData(BddClass*, cell);
    outNodes[i] = bddc->bdd; 
    outNames[i] = bddc->name;
    
    i++;
  }
  
  dotFile = fopen(bddcFile, "w");
  
  Cudd_DumpDot(manager, listGetSize(&wdd.listBddc), 
	       outNodes, NULL, outNames, dotFile);
  
  comFree(_fbddHeap,outNodes);
  comFree(_fbddHeap,outNames);
  fclose(dotFile);
  
  listDelAll(&wdd.listBddc );
}


void
fPrintSimpleGate(FILE *file, BddClass *bddc, int numInputs, char *sCubes)
{
  int i;

  fprintf(file, ".model m%d\n.inputs ", (int)bddc);
  for (i=0; i<numInputs; i++)
    fprintf(file, "N%d ", i);
  fprintf(file, "\n");  
  fprintf(file, ".outputs Nout\n.names ");
  for (i=0; i<numInputs; i++)
    fprintf(file, "N%d ", i);
  fprintf(file, "Nout\n");  
  fprintf(file, "%s", sCubes);
  fprintf(file, ".end\n\n");
  
}

void bddWriteCubeSet(FILE *file, DdManager *ddManager, BddClass *bddc);
void bddWriteVCubeSet(FILE *file, DdManager *ddManager, BddClass *bddc);

void
fPrintComplexGate(FILE *file, BddClass *bddc, DdManager *ddManager)
{
  int i;

  fprintf(file, ".model m%d\n.inputs ", (int)bddc);
  for (i=0; i<bddc->numVars; i++)
    fprintf(file, "N%d ", i);
  fprintf(file, "\n");  
  fprintf(file, ".outputs Nout\n.names ");
  for (i=0; i<bddc->numVars; i++)
    fprintf(file, "N%d ", i);
  fprintf(file, "Nout\n");

  bddWriteCubeSet(file, ddManager, bddc);

  fprintf(file, ".end\n\n");
}

void
fPrintVComplexGate(FILE *file, BddClass *bddc, DdManager *ddManager)
{
  int i;

  fprintf(file, "primitive pg%d(o, ", (int)bddc);
  for (i=0; i<bddc->numVars; i++)
    if (i==bddc->numVars-1)
      fprintf(file, "i%d);\n", i);
    else
      fprintf(file, "i%d, ", i);
  fprintf(file, "output o; input ");
  for (i=0; i<bddc->numVars; i++)
    if (i==bddc->numVars-1)
      fprintf(file, "i%d;\n", i);
    else
      fprintf(file, "i%d, ", i);
  fprintf(file, "table\n");

  bddWriteVCubeSet(file, ddManager, bddc);

  fprintf(file, "endtable\n");
  fprintf(file, "endprimitive\n");
}

void
fbddWriteBlif(FbddManager fm, char *blifFile)
{
  FILE *file;
  ListCell cell;
  int i=0;
  Model *model = fm->model;
  struct _List lBddis;
  struct _List _lBddcs;
  List lBddcs = &_lBddcs;
  char bddcInList;
  char *oNames[1];
  IMem mem;

  file = fopen(blifFile, "w");

  fprintf(file, ".model %s\n", model->name);

  fprintf(file, ".inputs");
  listForeach( &model->lInputs, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    fprintf(file, " %s", bddi->id);
  }
  fprintf(file, "\n");

  fprintf(file, ".outputs");
  listForeach(&model->lOutputs, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    fprintf(file, " %s", bddi->id);
  }
  fprintf(file, "\n");

  if (listGetSize(&model->bddcClock->lInsts)>1){
    fprintf(file, ".clock");
    listForeach(&model->bddcClock->lInsts, cell){
      BddInstance *bddi = listData(BddInstance*, cell);
      if (bddi == model->bddiGenericClk)
	continue;
      fprintf(file, " %s", bddi->id);
    }
    fprintf(file, "\n");
  }

  // Print Latches
  listForeach(&model->bddcLatchFe->lInsts, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    int val = (bddi->mark & 0x0300) >> 8;
    if (bddi->inputVars[1] == model->bddiGenericClk)
      fprintf(file, ".latch n%d n%d %d\n", 
	      (int)bddi->inputVars[0], (int)bddi, val);
    
    else
    fprintf(file, ".latch n%d n%d fe n%d %d\n", 
	    (int)bddi->inputVars[0], (int)bddi,
	    (int)bddi->inputVars[1], val);
  }
  listForeach(&model->bddcLatchRe->lInsts, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    int val = (bddi->mark & 0x0300) >> 8;
    fprintf(file, ".latch n%d n%d re n%d %d\n", 
	    (int)bddi->inputVars[0], (int)bddi,
	    (int)bddi->inputVars[1], val);
  }
  listForeach(&model->bddcLatchAh->lInsts, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    int val = (bddi->mark & 0x0300) >> 8;
    fprintf(file, ".latch n%d n%d ah n%d %d\n", 
	    (int)bddi->inputVars[0], (int)bddi,
	    (int)bddi->inputVars[1], val);
  }
  listForeach(&model->bddcLatchAl->lInsts, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    int val = (bddi->mark & 0x0300) >> 8;
    fprintf(file, ".latch n%d n%d al n%d %d\n", 
	    (int)bddi->inputVars[0], (int)bddi,
	    (int)bddi->inputVars[1], val);
  }
  listForeach(&model->bddcLatchAs->lInsts, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    int val = (bddi->mark & 0x0300) >> 8;
    fprintf(file, ".latch n%d n%d as n%d %d\n", 
	    (int)bddi->inputVars[0], (int)bddi,
	    (int)bddi->inputVars[1], val);
  }
  fprintf(file, "\n");

  // print out logic blocks
  mem = _fbddHeap;
  listConstruct0(&lBddis, mem);
  modelBddiList(mem, model, &lBddis);
  listConstruct0(lBddcs, _fbddHeap);

  listForeach (&lBddis, cell){
    ListCell bddcCell;

    BddInstance *bddi = listData(BddInstance*, cell);

    if (bddi == model->bddiGenericClk)
      continue;

    if (bddi->bddClass->type == PRIMARY_OUTPUT){
      fprintf(file, ".names n%d %s\n", (int)bddi->inputVars[0], bddi->id);
      fprintf(file, "1 1\n");
    }
    else if(bddi->bddClass->type == PRIMARY_INPUT){
      fprintf(file, ".names %s n%d\n", bddi->id, (int)bddi);
      fprintf(file, "1 1\n");
    }
    else if(bddi->bddClass->type == LATCH){
      // do nothing, handled earlier
    }
    else{
      fprintf(file, ".subckt m%d", (int)bddi->bddClass);
      
      // print input
      for(i=0; i<bddi->bddClass->numVars; i++){
	fprintf(file, " N%d=n%d", i, (int)bddi->inputVars[i]);
      }
      // print output
      fprintf(file, " Nout=n%d", (int)bddi);

      fprintf(file, "\n");

      // update list of bddc's
      bddcInList = FALSE;
      listForeach(lBddcs, bddcCell){
	if (listData(BddClass*, bddcCell) == bddi->bddClass){
	  bddcInList = TRUE;
	  break;
	}
      }

      if (bddcInList == FALSE){
	listNewCell(_fbddHeap, bddcCell, bddi->bddClass);
	listAppendElement(lBddcs, bddcCell);
      }
    }
  }
  fprintf(file, "\n");
  listDelAll(&lBddis);

  oNames[0] = (char*)comAlloc(_fbddHeap,sizeof(char)*5);
  strcpy(oNames[0], "Nout");

  listForeach(lBddcs, cell){
    BddClass *bddc;

    bddc = listData(BddClass*, cell);

    if (bddc == model->bddcAnd){
      fPrintSimpleGate(file, bddc, 2, "11 1\n");
      continue;
    }
    else if (bddc == model->bddcOr){
      fPrintSimpleGate(file, bddc, 2, "1- 1\n-1 1\n");
      continue;
    }
    else if (bddc == model->bddcXor){
      fPrintSimpleGate(file, bddc, 2, "10 1\n01 1\n");
      continue;
    }
    else if (bddc == model->bddcNand){
      fPrintSimpleGate(file, bddc, 2, "0- 1\n-0 1\n");
      continue;
    }
    else if (bddc == model->bddcNor){
      fPrintSimpleGate(file, bddc, 2, "00 1\n");
      continue;
    }
    else if (bddc == model->bddcXnor){
      fPrintSimpleGate(file, bddc, 2, "11 1\n00 1\n");
      continue;
    }
    else if (bddc->bdd == model->bddcNot->bdd){
      fPrintSimpleGate(file, bddc, 1, "0 1\n");
      continue;
    }
    else if (bddc->bdd == model->bddcSF->bdd){
      fPrintSimpleGate(file, bddc, 1, "1 1\n");
      continue;
    }
    else if(bddc == model->bddcMux){
      fPrintSimpleGate(file, bddc, 3, "-11 1\n1-0 1\n");
      continue;
    }
    else if(bddc == model->bddcMux2){
      fPrintSimpleGate(file, bddc, 3, "1-1 1\n-10 1\n");
      continue;
    }
    else if(bddc == model->bddcLogicOne){
      fPrintSimpleGate(file, bddc, 0, "1\n");
      continue;
    }
    else if(bddc == model->bddcLogicZero){
      fPrintSimpleGate(file, bddc, 0, "0\n");
      continue;
    }
    else if(bddc->type == LATCH){
      continue;
    }
    else{
      fPrintComplexGate(file, bddc, model->ddManager);

    }
  }
  listDelAll(lBddcs);
  comFree(_fbddHeap,oNames[0]);

  fclose(file);

}

static void
computeVerilogName(char *strVerilogName, char *strOrigName)
{
  char c;
  while((c=*strOrigName++) != 0){
    if ((c>='a' && c<='z') || (c>='A' && c<='Z') || (c>='0' && c<='9')){
      *strVerilogName = c;
    }
    else{
      *strVerilogName = '_';
    }
    strVerilogName++;
  }
  *strVerilogName=0;
}

static void
writeSignalList(FILE *file, List list)
{
  ListCell cBddi;
  char strVerilogName[100];

  listForeach(list, cBddi){
    BddInstance *bddi = listData(BddInstance*, cBddi);
    computeVerilogName(strVerilogName, bddi->id);
    if (cBddi == listGetTail(list)){
      fprintf(file, "p%s ", strVerilogName);
    }
    else{
      fprintf(file, "p%s, \n", strVerilogName);
    }
  }
}

static void
getSignalStr(char str[32], BddInstance *bddi)
{
  char strTmp[32];
  if (bddi->bddClass->type == PRIMARY_OUTPUT ||
      bddi->bddClass->type == PRIMARY_INPUT){
    computeVerilogName(strTmp, bddi->id);
    sprintf(str, "p%s", strTmp); 
  }
  else if (bddi->bddClass->numVars == 0 
	   && bddi->bddClass->type==INTERNAL_VARIABLE
	   && !Cudd_IsComplement(bddi->bddClass->bdd))
    // == _model->bddcLogicOne->bdd)
    sprintf(str, "1");
  else if (bddi->bddClass->numVars == 0
	   && bddi->bddClass->type==INTERNAL_VARIABLE
	   && Cudd_IsComplement(bddi->bddClass->bdd))
    // bddi->bddClass->bdd == _model->bddcLogicZero->bdd)
    sprintf(str, "0");
  else
    sprintf(str, "n%d", (int)bddi);
};

void
fbddWriteVerilog(FbddManager fm, char *verilogFile)
{
  FILE *file;
  ListCell cell;
  Model *model = fm->model;
  struct _List lBddis;
  struct _List _lBddcs;
  List lBddcs = &_lBddcs;
  ListCell cBddc;

  IMem mem;
  char first;
  char strVerilogName[50];
  struct _List lRegs;

  struct _List lPriInputs;
  ListCell cPriInput;
  
  file = fopen(verilogFile, "w");

  // print out logic blocks
  mem = _fbddHeap;
  listConstruct0(&lBddis, _fbddHeap);
  modelBddiList(mem, model, &lBddis);
  
  computeVerilogName(strVerilogName, model->name);
  fprintf(file, "module %s (", strVerilogName);
  //  fprintf(file, "module mmm (");
  // write input signals

  listConstruct0(&lPriInputs, _fbddHeap);
  listForeach(&model->bddcPrimaryIn->lInsts, cPriInput){
    BddInstance *bddi = listData(BddInstance*, cPriInput);
    listAppend(&lPriInputs, bddi);
  }
  listForeach(&model->bddcClock->lInsts, cPriInput){
    BddInstance *bddi = listData(BddInstance*, cPriInput);
    listAppend(&lPriInputs, bddi);    
  }

  writeSignalList(file, &lPriInputs);
  fprintf(file, ", ");

  // write output signals
  writeSignalList(file, &model->lOutputs);
  fprintf(file, ");\n");
  
  fprintf(file, "input ");
  writeSignalList(file, &lPriInputs);
  fprintf(file, ";\n");

  listDelAll(&lPriInputs);
  
  fprintf(file, "output ");
  writeSignalList(file, &model->lOutputs);
  fprintf(file, ";\n");
  
  fprintf(file, "wire ");
  first = TRUE;
  listForeach (&lBddis, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    if (bddi->bddClass->type != PRIMARY_OUTPUT && 
	bddi->bddClass->type != LATCH){
      if (first == TRUE){
	fprintf(file, "n%d", (int)bddi);
	first = FALSE;
      }
      else
	fprintf(file, ",\n  n%d", (int)bddi);
    }
  }
  fprintf(file, ";\n\n");

  listConstruct0(&lRegs, _fbddHeap);
  listForeach(&model->bddcLatchFe->lInsts, cell)
    listAppend(&lRegs, listData(BddInstance*, cell));
  listForeach(&model->bddcLatchRe->lInsts, cell)
    listAppend(&lRegs, listData(BddInstance*, cell));
  listForeach(&model->bddcLatchAh->lInsts, cell)
    listAppend(&lRegs, listData(BddInstance*, cell));
  listForeach(&model->bddcLatchAl->lInsts, cell)
    listAppend(&lRegs, listData(BddInstance*, cell));
  listForeach(&model->bddcLatchAs->lInsts, cell)
    listAppend(&lRegs, listData(BddInstance*, cell));

  if (listGetSize(&lRegs)>0){
    fprintf(file, "reg ");
    first = TRUE;
    listForeach (&lRegs, cell){
      BddInstance *bddi = listData(BddInstance*, cell);
      if (first == TRUE){
	fprintf(file, "n%d", (int)bddi);
	first = FALSE;
      }
      else
	fprintf(file, ", n%d\n", (int)bddi);
    }
    fprintf(file, ";\n\n");
  }
  listDelAll(&lRegs);

  listForeach (&lBddis, cell){
    BddInstance *bddi = listData(BddInstance*, cell);
    BddClass *bddc = bddi->bddClass;
    char str[3][32];
    int i;
    int m;

    m = comMin(bddc->numVars, 3);
    for (i=0; i<m; i++)
      getSignalStr(str[i], bddi->inputVars[i]);

    if (bddi->bddClass->type == PRIMARY_OUTPUT){
      computeVerilogName(strVerilogName, bddi->id);
      fprintf(file, "assign p%s = n%d;\n", strVerilogName, 
	      (int)bddi->inputVars[0]);
    }
    else if(bddi->bddClass->type == PRIMARY_INPUT){
      /*
      computeVerilogName(strVerilogName, bddi->id);
      fprintf(file, "assign n%d = p%s;\n", (int)bddi, strVerilogName);
      */
    } 
    else{
      if (bddc->bdd == model->bddcLogicZero->bdd){
	/*
	fprintf(file, "assign n%d = 0;\n", 
		(int)bddi);
	*/
      }
      else if (bddc->bdd == model->bddcLogicOne->bdd){
	/*
	fprintf(file, "assign n%d = 1;\n", 
		(int)bddi);
	*/
      }
      else if (bddc->bdd == model->bddcAnd->bdd){
	fprintf(file, "assign n%d = %s & %s;\n", 
		(int)bddi, str[0], str[1]);
      }
      else if (bddc->bdd == model->bddcOr->bdd){
	fprintf(file, "assign n%d = %s | %s;\n", 
		(int)bddi, str[0], str[1]);
      }
      else if (bddc->bdd == model->bddcXor->bdd){
	fprintf(file, "assign n%d = %s ^ %s;\n", 
		(int)bddi,
		str[0], 
		str[1]);
      }
      else if (bddc->bdd == model->bddcNand->bdd){
	fprintf(file, "nand (n%d, %s, %s);\n", 
		(int)bddi,
		str[0], 
		str[1]);
      }
      else if (bddc->bdd == model->bddcNor->bdd){
	fprintf(file, "nor (n%d, %s, %s);\n", 
		(int)bddi,
		str[0], 
		str[1]);
      }
      else if (bddc->bdd == model->bddcXnor->bdd){
	fprintf(file, "assign n%d = %s ~^ %s;\n", 
		(int)bddi,
		str[0], 
		str[1]);
      }
      else if (bddc->bdd == model->bddcNot->bdd){
	fprintf(file, "assign n%d = ~%s;\n", 
		(int)bddi,
		str[0]);
      }
      else if (bddc->bdd == model->bddcSF->bdd){
	fprintf(file, "assign n%d = %s;\n", 
		(int)bddi,
		str[0]);
      }
      else if(bddc->bdd == model->bddcMux->bdd){
	fprintf(file, "assign n%d = (%s == 1'b1) ? %s : %s;\n",
		(int)bddi,
		str[2],
		str[1],
		str[0]);
      }
      else if(bddc->bdd == model->bddcMux2->bdd){
	fprintf(file, "assign n%d = (%s == 1'b1) ? %s : %s;\n",
		(int)bddi,
		str[2],
		str[0],
		str[1]);
      }
      else if(bddc->type == LATCH){
	if (bddc == model->bddcLatchFe)
	  fprintf(file, "always @(negedge %s) n%d = %s;\n",
		  str[1],
		  (int)bddi,
		  str[0]);
	else if (bddc == model->bddcLatchRe)
	  fprintf(file, "always @(posedge %s) n%d = %s;\n",
		  str[1],
		  (int)bddi,
		  str[0]);
	else if (bddc == model->bddcLatchAh){
	  fprintf(file, "always @ (%s, %s)\n", 
		  str[0], str[1]);
	  fprintf(file, "begin:\n");
	  fprintf(file, "if (%s == 1)\n", str[1]);
	  fprintf(file, "n%d = %s;\n", (int)bddi, str[0]);
	  fprintf(file, "end\n");
	}
	else if (bddc == model->bddcLatchAl){
	  fprintf(file, "always @ (%s, %s)\n", 
		  str[0], str[1]);
	  fprintf(file, "begin:\n");
	  fprintf(file, "if (%s == 0)\n", str[1]);
	  fprintf(file, "n%d = %s;\n", (int)bddi, str[0]);
	  fprintf(file, "end\n");
	}
	else if (bddc == model->bddcLatchAs){
	  exit(1);
	  assert(FALSE);
	}
      }
      else{
	// print complex gate
	fprintf(file, "pg%d(n%d, ", 
		(int)bddc,
		(int)bddi);
	for (i=0; i<bddc->numVars; i++){
	  getSignalStr(str[0], bddi->inputVars[i]);
	  if (i!=bddc->numVars-1)
	    fprintf(file, "%s, ", str[0]);
	  else
	    fprintf(file, "%s);\n", str[0]);
	  //	assert(FALSE);
	}
      }
    }
  }
  fprintf(file, "\n");
  fprintf(file, "endmodule\n");

  listDelAll(&lBddis);

  // print out all complex gates
  listConstruct0(lBddcs, _fbddHeap);
  modelBddcList(mem, model, lBddcs);
  listForeach(lBddcs, cBddc){
    BddClass *bddc = listData(BddClass*, cBddc);
    if (bddc == model->bddcAnd)
      continue;
    else if (bddc == model->bddcOr)
      continue;
    else if (bddc == model->bddcXor)
      continue;
    else if (bddc == model->bddcNand)
      continue;
    else if (bddc == model->bddcNor)
      continue;
    else if (bddc == model->bddcXnor)
      continue;
    else if (bddc->bdd == model->bddcNot->bdd)
      continue;
    else if (bddc->bdd == model->bddcSF->bdd)
      continue;
    else if(bddc == model->bddcMux)
      continue;
    else if(bddc == model->bddcMux2)
      continue;
    else if(bddc == model->bddcLogicOne)
      continue;
    else if(bddc == model->bddcLogicZero)
      continue;
    else if(bddc->type == LATCH)
      continue;
    else{
      fPrintVComplexGate(file, bddc, model->ddManager);      
    }
  }
  listDelAll(lBddcs);
  
  fclose(file);
}

void
bddWriteDot(char *filename, DdManager *manager, DdNode *bdd)
{
  FILE *f = fopen(filename, "w");
  DdNode *nodes[1];

  nodes[0] = bdd;

  Cudd_DumpDot( manager, 1, nodes, NULL, NULL, f);

  fclose(f);
}

void
bddWriteDDcal(char *filename, DdManager *manager, DdNode *bdd)
{
  FILE *f = fopen(filename, "w");
  DdNode *nodes[1];

  nodes[0] = bdd;

  Cudd_DumpDDcal( manager, 1, nodes, NULL, NULL, f);

  fclose(f);
}

typedef struct _Literal
{
  int level;
  char isTrue;
} Literal;

typedef struct _CubeSetData
{
  DdNode *bddLogicOne;
  DdNode *bddLogicZero;
  struct _List lCube; // list of levels that are in the cube.
  int topLevel;
  int numVars;
  FILE *fOut;
  DdManager *ddManager;
} CubeSetData;

void
printCube(CubeSetData *csd)
{
  int i,j;
  int lastLevel;
  ListCell cell;
  char *strCube = (char*)comAlloc(_fbddHeap, sizeof(char)*(csd->numVars+1));
  lastLevel = csd->topLevel;

  strCube[csd->numVars]=0;
  j=csd->numVars-1;

  listForeach(&csd->lCube, cell){
    Literal *literal = listData(Literal*, cell);
    for (i=lastLevel; i<literal->level-1; i++){
      strCube[j]='-';
      j--;
    }
    if (literal->isTrue){
      strCube[j]='1';
      j--;
    }
    else{
      strCube[j]='0';
      j--;
    }
    lastLevel = literal->level;
  }
  for (i=lastLevel; i<csd->topLevel+csd->numVars-1; i++){
    strCube[j]='-';
    j--;
  }
  fprintf(csd->fOut, "%s", strCube);
  fprintf(csd->fOut, " 1\n");

  comFree(_fbddHeap, strCube);
}

void
printVCubeOne(CubeSetData *csd)
{
  int i,j;
  int lastLevel;
  ListCell cell;
  char *strCube = (char*)comAlloc(_fbddHeap, sizeof(char)*(csd->numVars+1));
  lastLevel = csd->topLevel;

  strCube[csd->numVars]=0;
  j=csd->numVars-1;

  listForeach(&csd->lCube, cell){
    Literal *literal = listData(Literal*, cell);
    for (i=lastLevel; i<literal->level-1; i++){
      strCube[j]='?';
      j--;
    }
    if (literal->isTrue){
      strCube[j]='1';
      j--;
    }
    else{
      strCube[j]='0';
      j--;
    }
    lastLevel = literal->level;
  }
  for (i=lastLevel; i<csd->topLevel+csd->numVars-1; i++){
    strCube[j]='?';
    j--;
  }
  fprintf(csd->fOut, "%s", strCube);
  fprintf(csd->fOut, " : 1;\n");

  comFree(_fbddHeap, strCube);
}

void
printVCubeZero(CubeSetData *csd)
{
  int i,j;
  int lastLevel;
  ListCell cell;
  char *strCube = (char*)comAlloc(_fbddHeap, sizeof(char)*(csd->numVars+1));
  lastLevel = csd->topLevel;

  strCube[csd->numVars]=0;
  j=csd->numVars-1;

  listForeach(&csd->lCube, cell){
    Literal *literal = listData(Literal*, cell);
    for (i=lastLevel; i<literal->level-1; i++){
      strCube[j]='?';
      j--;
    }
    if (literal->isTrue){
      strCube[j]='1';
      j--;
    }
    else{
      strCube[j]='0';
      j--;
    }
    lastLevel = literal->level;
  }
  for (i=lastLevel; i<csd->topLevel+csd->numVars-1; i++){
    strCube[j]='?';
    j--;
  }
  fprintf(csd->fOut, "%s", strCube);
  fprintf(csd->fOut, " : 0;\n");

  comFree(_fbddHeap, strCube);
}

DdNode *
normalizeChild(DdNode *nChild, char parentIsComplement)
{
  if ((Cudd_IsComplement(nChild) 
       && parentIsComplement)
      || (!Cudd_IsComplement(nChild) 
	  && !parentIsComplement)){
    nChild = Cudd_Regular(nChild);
  }
  else{
    nChild = Cudd_Complement(nChild);
  }
  
  return nChild;
}

void
bddWriteCubeSetHelper(CubeSetData *csd, DdNode *node)
{
  DdNode *nChild;
  Literal literal;
  ListCell cell;
  
  if (Cudd_IsConstant(node)){
    if (node == csd->bddLogicOne){
      printCube(csd);
    }
  }
  else{
    literal.level = Cudd_ReadPerm(csd->ddManager, Cudd_NodeReadIndex(node));

    // traverse then edge
    nChild = Cudd_T(node);
    nChild = normalizeChild(nChild, Cudd_IsComplement(node));
    assert(nChild != NULL);
    literal.isTrue = TRUE;
    listNewCell(_fbddHeap, cell, &literal);
    listAppendElement(&csd->lCube, cell);
    bddWriteCubeSetHelper(csd, nChild);
    listRemoveElement(&csd->lCube, cell);
    comFree(_fbddHeap, cell);

    // traverse else edge
    nChild = Cudd_E(node);
    nChild = normalizeChild(nChild, Cudd_IsComplement(node));
    assert(nChild != NULL);
    literal.isTrue = FALSE;
    listNewCell(_fbddHeap, cell, &literal);
    listAppendElement(&csd->lCube, cell);
    bddWriteCubeSetHelper(csd, nChild);
    listRemoveElement(&csd->lCube, cell);
    comFree(_fbddHeap, cell);
  }
}

void
bddWriteCubeSet(FILE *file, DdManager *ddManager, BddClass *bddc)
{
  CubeSetData csd;

  csd.ddManager = ddManager;
  csd.bddLogicOne = Cudd_ReadOne(ddManager);
  listConstruct0(&csd.lCube, _fbddHeap);

  csd.fOut = file;
  csd.numVars = bddc->numVars;
  csd.topLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddc->bdd));

  assert(!Cudd_IsConstant(bddc->bdd));

  bddWriteCubeSetHelper(&csd, bddc->bdd);

  listDestruct(&csd.lCube);
}

void
bddWriteVCubeSetHelper(CubeSetData *csd, DdNode *node)
{
  DdNode *nChild;
  Literal literal;
  ListCell cell;
 
  if (Cudd_IsConstant(node)){
    if (node == csd->bddLogicOne){
      printVCubeOne(csd);
    }
    else if (node == csd->bddLogicZero){
      printVCubeZero(csd);
    }
  }
  else{
    literal.level = Cudd_ReadPerm(csd->ddManager, Cudd_NodeReadIndex(node));

    // traverse then edge
    nChild = Cudd_T(node);
    nChild = normalizeChild(nChild, Cudd_IsComplement(node));
    assert(nChild != NULL);
    literal.isTrue = TRUE;
    listNewCell(_fbddHeap, cell, &literal);
    listAppendElement(&csd->lCube, cell);
    bddWriteVCubeSetHelper(csd, nChild);
    listRemoveElement(&csd->lCube, cell);
    comFree(_fbddHeap, cell);

    // traverse else edge
    nChild = Cudd_E(node);
    nChild = normalizeChild(nChild, Cudd_IsComplement(node));
    assert(nChild != NULL);
    literal.isTrue = FALSE;
    listNewCell(_fbddHeap, cell, &literal);
    listAppendElement(&csd->lCube, cell);
    bddWriteVCubeSetHelper(csd, nChild);
    listRemoveElement(&csd->lCube, cell);
    comFree(_fbddHeap, cell);
  }
}

void
bddWriteVCubeSet(FILE *file, DdManager *ddManager, BddClass *bddc)
{
  CubeSetData csd;

  csd.ddManager = ddManager;
  csd.bddLogicOne = Cudd_ReadOne(ddManager);
  csd.bddLogicZero = Cudd_ReadLogicZero(ddManager);
  listConstruct0(&csd.lCube, _fbddHeap);

  csd.fOut = file;
  csd.numVars = bddc->numVars;
  csd.topLevel = Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(bddc->bdd));

  assert(!Cudd_IsConstant(bddc->bdd));

  bddWriteVCubeSetHelper(&csd, bddc->bdd);

  listDestruct(&csd.lCube);
}

void
printMatchVersusSize(char *strFile, Model *model)
{
  IMem mem;
  struct _List lBddcs;
  ListCell cBddc;
  int *matchVersusSize;
  int i;
  FILE *file;
  int numBddis = 0;
  int numBddcs = 0;

  mem = _fbddHeap;
  
  matchVersusSize = (int*)comAlloc(mem, sizeof(int)*200);
  for(i=0; i<200; i++)
    matchVersusSize[i]=0;

  file = fopen(strFile, "w");
  fprintf(file, "numvars numinsts\n");
  listConstruct0(&lBddcs, _fbddHeap);
  modelBddcList(mem, model, &lBddcs);
  listForeach(&lBddcs, cBddc){
    BddClass *bddc = listData(BddClass*, cBddc);
    matchVersusSize[bddc->numVars] += bddcNumInsts(bddc);
    fprintf(file, "%d %d\n", bddc->numVars, bddcNumInsts(bddc));
    numBddcs++;
    numBddis+=bddcNumInsts(bddc);
  }
  fprintf(file, "\n");
  fprintf(file, "%d %d\n", numBddcs, numBddis);

  listDelAll(&lBddcs);

  fclose(file);
  comFree(mem, matchVersusSize);
  
}
